@@ -1,6 +1,8 @@
 /* libsh/ -prototypes */
 
 /* libsh/builtin.c */
+extern conscell *sh_car	    __((conscell *, conscell *));
+extern conscell *sh_glob    __((conscell *, conscell *));
 extern int       sh_builtin __((int argc, const char *argv[]));
 extern int       sh_include __((int argc, const char *argv[]));
 extern conscell *sh_return  __((conscell *, conscell *, int *));
@@ -22,8 +24,8 @@ extern int        glob_match __((int *pattern, int *eopattern, const char *s));
 extern char       globchars[];
 extern void       glob_init __((void));
 extern int        pathcmp __((const void *ap, const void *bp));
-extern int        squish  __((conscell *d, char **bufp, int **ibufp));
-extern conscell * expand  __((conscell *d));
+extern int        squish  __((conscell *d, char **bufp, int **ibufp, int doglob));
+extern conscell * expand  __((conscell *d, int variant));
 
 /* libsh/interpret.c */
 extern int  magic_number;
@@ -35,7 +37,7 @@ extern int setfreefd __((void));
 extern void assign   __((conscell *, conscell *, struct osCmd *));
 #endif
 #ifdef SPTREE_H
-extern int xundefun __((struct spblk *));
+extern int xundefun __((void *, struct spblk *));
 #endif
 #ifdef TOKEN_NARGS /* Must have include "libsh/sh.h" for this */
 extern void functype __((const char *, struct shCmd **, struct sslfuncdef **));
@@ -56,7 +58,6 @@ extern conscell * s_last     __((conscell *));
 extern int        s_equal1   __((conscell *, conscell *));
 extern int        s_equal    __((conscell *, conscell *));
 extern conscell * s_nth      __((conscell *list, int n));
-extern void       s_set_prev __((conscell *prev, conscell *list));
 extern void       s_grind    __((conscell *, FILE *));
 extern void       _grind     __((conscell *));
 extern conscell * s_catstring __((conscell *));
@@ -105,11 +106,12 @@ extern int sh_test __((int argc, const char *argv[]));
 /* libsh/trap.c */
 extern int  sprung;
 extern int  interrupted;
-extern const char * traps[];
-extern RETSIGTYPE (*orig_handler[])(int);
+/* extern const char * traps[]; */
+extern RETSIGTYPE (*orig_handler[]) __((int));
+extern void traphup  __((int));
 extern void trapsnap __((void));
 extern void trap_handler __((int));
-extern int eval __((const char *script, const char *scriptname, const char *savefil));
+extern int eval __((const char *script, const char *scriptname, const char *savefil, const struct stat *));
 #ifdef S_IFMT
 extern int loadeval __((int fcfd, const char *path, struct stat *srcstbufp));
 extern int  leaux    __((int, const char*, struct stat *));
@@ -164,7 +166,6 @@ extern int funclevel;
 /* Globals .... */
 
 extern char *getenv __((const char *));
-extern char *optarg;
 extern char *strerror __((int));
 extern char *strsignal __((int));
 extern int D_assign;
@@ -172,8 +173,6 @@ extern int D_compare;
 extern int D_functions;
 extern int D_matched;
 extern int errno;
-extern int opterr;
-extern int optind;
 extern void *tmalloc __((const size_t n));
 #ifdef SPTREE_H
 extern void sp_null __((struct sptree *));
