@@ -3,6 +3,8 @@
  *	This will be free software, but only when it is finished.
  */
 
+#define _SYS_STREAM_H /* Block inclusion of Solaris 10 <sys/stream.h> */
+
 #include "hostenv.h"
 
 #include <stdio.h>
@@ -122,6 +124,7 @@ struct notary {
 	const char	*dsn;
 };
 
+/* NOTE: This structure is constructed at couple of places! */
 struct address {
 	const char	*a_pname;		/* printable representation */
 	struct addr	*a_tokens;		/* RFC822 token list */
@@ -166,6 +169,8 @@ typedef enum {	normal,		/* nil */
 		Resent,		/* this is a resent- header field */
 		eIdentinfo,	/* identification info		  */
 		eChannel,	/* received by this transport channel (mailer)*/
+		eErrorMsg,	/* Flag-like header telling this is created
+				   by the SCHEDULER for err msgs! */
 		eExternal,	/* external source of info, untrustable */
 		eEnvid,		/* SMTP-DSN ENVID string */
 		eFrom,		/* envelope sender address */
@@ -234,12 +239,16 @@ struct envelope {
 	const char	*e_file;	/* message file name		*/
 	FILE		*e_fp;		/* message file pointer		*/
 	const char	*e_messageid;	/* message id string		*/
+	char		e_spoolid[32];  /* spoolid of the message file;
+					   64-bit decimal integer: 20 ch +
+					   compact encoded timestamp: 9 ch +
+					   string end NUL: 1  */
 };
 
 #ifdef	RFC976
-#define HDR_SCANNER(x)	scan822(&(x), strlen((const char *)x), '!', '%', 0, (token822 **)0)
+#define HDR_SCANNER(x)	scan822(&(x), strlen((const char *)x), '!', '%', (token822 **)0)
 #else	/* !RFC976 */
-#define HDR_SCANNER(x)	scan822(&(x), strlen((const char *)x), 0, 0, 0, (token822 **)0)
+#define HDR_SCANNER(x)	scan822(&(x), strlen((const char *)x), 0, 0, (token822 **)0)
 #endif	/* RFC976 */
 
 struct Zpasswd {
@@ -254,6 +263,8 @@ struct Zpasswd {
   /* Our ``Z'' extensions */
   long quota1, quota2;
   const char *mail_forward;
+  const char *mailspooldir;
+  const char *mailfolderdir;
 };
 
 struct Zgroup {
