@@ -10,10 +10,10 @@
 #include <signal.h>
 
 #ifndef RETSIGTYPE
-#define RETSIGTYPE void	/* Propably not globally portable.. */
+#define RETSIGTYPE void	/* Probably not globally portable.. */
 #endif
 
-#ifdef	SV_INTERRUPT	/* BSDism */
+#if defined(SV_INTERRUPT) && !defined(HAVE_SIGPROCMASK)	/* BSDism */
 	/* ================ BSD 4.3+ (?) STUFF ================ */
 #define	SIGNAL_HANDLE(X,Y)	\
 	{	struct sigvec sv;			\
@@ -33,7 +33,7 @@
 #define SIGNAL_HOLD(SIG) \
 	sigblock(sigmask(SIG))
 #define SIGNAL_RELEASE(SIG) \
-	sigsetmask(~sigmask(SIG) & sigblock(SIGNAL_HOLD(sig)))
+	sigsetmask(~sigmask(SIG) & sigblock(SIGNAL_HOLD(SIG)))
 #define SIGNAL_IGNORE(SIG) \
 	SIGNAL_HANDLE(SIG,SIG_IGN)
 
@@ -52,6 +52,13 @@
 		sigemptyset(&act.sa_mask);	\
 		sigaction (X, &act, NULL);	\
 	}
+#define SIGNAL_ACTION(X, Y) \
+	{	struct sigaction act;		\
+		act.sa_sigaction = Y;		\
+		act.sa_flags   = SA_SIGINFO|SA_INTERRUPT|SA_NODEFER; \
+		sigemptyset(&act.sa_mask);	\
+		sigaction (X, &act, NULL);	\
+	}
 #define SIGNAL_HANDLESAVE(X,Y,OLD) \
 	{	struct sigaction act, oact;	\
 		act.sa_handler = Y;		\
@@ -62,16 +69,16 @@
 		OLD = oact.sa_handler;		\
 	}
 #define SIGNAL_HOLD(SIG) \
-	{	sigset_t	sigmask;	\
-		sigemptyset(&sigmask);		\
-		sigaddset(&sigmask,SIG);	\
-		sigprocmask(SIG_BLOCK,&sigmask,NULL);	\
+	{	sigset_t	sigmsk;		\
+		sigemptyset(&sigmsk);		\
+		sigaddset(&sigmsk,SIG);		\
+		sigprocmask(SIG_BLOCK,&sigmsk,NULL);	\
 	}
 #define SIGNAL_RELEASE(SIG) \
-	{	sigset_t	sigmask;	\
-		sigemptyset(&sigmask);		\
-		sigaddset(&sigmask,SIG);	\
-		sigprocmask(SIG_UNBLOCK,&sigmask,NULL);	\
+	{	sigset_t	sigmsk;		\
+		sigemptyset(&sigmsk);		\
+		sigaddset(&sigmsk,SIG);		\
+		sigprocmask(SIG_UNBLOCK,&sigmsk,NULL);	\
 	}
 #define SIGNAL_IGNORE(SIG) \
 	SIGNAL_HANDLE(SIG,SIG_IGN)
@@ -95,16 +102,16 @@
 		OLD = oact.sa_handler;		\
 	}
 #define SIGNAL_HOLD(SIG) \
-	{	sigset_t	sigmask;	\
-		sigemptyset(&sigmask);		\
-		sigaddset(&sigmask,SIG);	\
-		sigprocmask(SIG_BLOCK,&sigmask,NULL);	\
+	{	sigset_t	sigmsk;		\
+		sigemptyset(&sigmsk);		\
+		sigaddset(&sigmsk,SIG);		\
+		sigprocmask(SIG_BLOCK,&sigmsk,NULL);	\
 	}
 #define SIGNAL_RELEASE(SIG) \
-	{	sigset_t	sigmask;	\
-		sigemptyset(&sigmask);		\
-		sigaddset(&sigmask,SIG);	\
-		sigprocmask(SIG_UNBLOCK,&sigmask,NULL);	\
+	{	sigset_t	sigmsk;		\
+		sigemptyset(&sigmsk);		\
+		sigaddset(&sigmsk,SIG);		\
+		sigprocmask(SIG_UNBLOCK,&sigmsk,NULL);	\
 	}
 #define SIGNAL_IGNORE(SIG) \
 	SIGNAL_HANDLE(SIG,SIG_IGN)
@@ -131,3 +138,24 @@
 #ifndef SIGCHLD
 #define SIGCHLD SIGCLD
 #endif
+
+#ifdef  HAVE_WAITPID
+# include <sys/wait.h>
+#else
+# ifdef HAVE_WAIT3
+#  include <sys/wait.h> /* Has BSD wait3() */
+# else
+#  ifdef HAVE_SYS_WAIT_H /* POSIX.1 compatible */
+#   include <sys/wait.h>
+#  else /* Not POSIX.1 compatible, lets fake it.. */
+extern int wait();
+#  endif
+# endif
+#endif
+
+#ifndef WEXITSTATUS
+# define WEXITSTATUS(s) (((s) >> 8) & 0377)
+#endif
+#ifndef WSIGNALSTATUS
+# define WSIGNALSTATUS(s) ((s) & 0177)
+#endif
