@@ -8,6 +8,7 @@
 #include "zmsignal.h" /* used for kill(2) */
 #include "libz.h"
 #include "libc.h"
+#include "mail.h"
 
 #ifndef	_IOFBF
 #define	_IOFBF	0
@@ -46,7 +47,7 @@ killprevious(sig, pidfile)
 	if ((sig == 0 || sig == SIGTERM || -sig == SIGTERM)
 	    && (fp = fopen(path, "w+")) != NULL) {
 		setvbuf(fp, buf, _IOFBF, sizeof buf);
-		fprintf(fp, "%d\n", getpid());
+		fprintf(fp, "%d\n", (int)getpid());
 		rc = fflush(fp);
 		if (fclose(fp) != 0)
 		  rc |= 1; /* indicate error */
@@ -56,3 +57,19 @@ killprevious(sig, pidfile)
 	free(path);
 	return rc;
 }
+
+int
+killpidfile(pidfile)
+	const char *pidfile;
+{
+	int rc;
+	char *path;
+
+	path = emalloc((unsigned)(strlen(postoffice) + strlen(pidfile) + 2));
+	sprintf(path, "%s/%s", postoffice, pidfile);
+
+	rc = unlink(path);
+
+	free(path);
+	return rc;
+}
