@@ -59,14 +59,6 @@
 #include <string.h>
 #include <errno.h>
 
-union sockaddr_uni {
-    struct sockaddr     sa;
-    struct sockaddr_in  v4;
-#if defined(AF_INET6) && defined(INET6)
-    struct sockaddr_in6 v6;
-#endif
-};
-
 #include "libc.h"
 #include "zmalloc.h"
 #include "libz.h"
@@ -101,20 +93,34 @@ struct mbuf;
 #endif
 
 int
-loadifaddresses(sockaddrp)
-struct sockaddr ***sockaddrp;
+loadifaddresses(sockaddrp
+#ifdef TESTMODE
+		, ifsp
+#endif
+		)
+     Usockaddr ***sockaddrp;
+#ifdef TESTMODE
+     void ***ifsp;
+#endif
 {
 	int i;
 	int sapcount = -1;
 	int sapspace = 2;
-	union sockaddr_uni **sap;
+	Usockaddr **sap;
+#ifdef TESTMODE
+	void **ifs;
+#endif
 
-	sap = (void*)malloc(sizeof(union sockaddr_uni*) * (sapspace + 2));
+	sap = (void*)malloc(sizeof(Usockaddr*) * (sapspace + 2));
+#ifdef TESTMODE
+	ifs = (void*)malloc(sizeof(void*) * (sapspace + 2));
+#endif
 
 	if (! sap)
 	  return -3; /* UAARGH! */
 
 #ifdef HAVE_GETIFADDRS
+	/* #warning "loadifaddresses(): HAVE_GETIFADDRS" */
 	{
 	  struct ifaddrs *ifar = NULL, *ifa;
 
@@ -132,8 +138,11 @@ struct sockaddr ***sockaddrp;
 
 	      if (sapcount +2 >= sapspace) {
 		sapspace <<= 1;
-		sap = (void*)realloc(sap, (sizeof(union sockaddr_uni*) *
+		sap = (void*)realloc(sap, (sizeof(Usockaddr*) *
 					   (sapspace + 2)));
+#ifdef TESTMODE
+		ifs = (void*)realloc(ifs, (sizeof(void*) * (sapspace +2)));
+#endif
 	      }
 
 	      if (! sap) {
@@ -147,7 +156,10 @@ struct sockaddr ***sockaddrp;
 
 		/* pick the whole sockaddr package! */
 		memcpy(si4, sa, sizeof(*si4));
-		sap[++sapcount] = (union sockaddr_uni *)si4;
+		sap[++sapcount] = (Usockaddr *)si4;
+#ifdef TESTMODE
+		ifs[sapcount] = strdup(ifa->ifa_name);
+#endif
 	      }
 
 #if defined(AF_INET6) && defined(INET6)
@@ -158,7 +170,10 @@ struct sockaddr ***sockaddrp;
 
 		/* pick the whole sockaddr package! */
 		memcpy(si6, sa, sizeof(*si6));
-		sap[++sapcount] = (union sockaddr_uni *)si6;
+		sap[++sapcount] = (Usockaddr *)si6;
+#ifdef TESTMODE
+		ifs[sapcount] = strdup(ifa->ifa_name);
+#endif
 	      }
 #endif
 	    }
@@ -174,6 +189,7 @@ struct sockaddr ***sockaddrp;
 #else /* not HAVE_GETIFADDRS */
 
 #if defined(SIOCGLIFCONF) && defined(AF_INET6) && defined(INET6)
+	/* #warning "loadifaddresses(): SIOCGLIFCONF && INET6" */
 	/*#warning "experimental INET6 related SIOCGLIFCONF code activated!"*/
 	{
 	  struct lifconf lifc;
@@ -223,7 +239,7 @@ struct sockaddr ***sockaddrp;
 	  for (i = 0; i < lifc.lifc_len; ) {
 
 	    struct lifreq *lifr = (struct lifreq *) &lifc.lifc_buf[i];
-	    union sockaddr_uni *sa = (union sockaddr_uni *) &lifr->lifr_addr;
+	    Usockaddr *sa = (Usockaddr *) &lifr->lifr_addr;
 #ifdef SIOCGLIFFLAGS
 	    struct lifreq lifrf;
 #endif
@@ -275,8 +291,11 @@ struct sockaddr ***sockaddrp;
 
 	    if (sapcount +2 >= sapspace) {
 	      sapspace <<= 1;
-	      sap = (void*)realloc(sap, (sizeof(union sockaddr_uni*) *
+	      sap = (void*)realloc(sap, (sizeof(Usockaddr*) *
 					 (sapspace + 2)));
+#ifdef TESTMODE
+	      ifs = (void*)realloc(ifs, (sizeof(void*) * (sapspace +2)));
+#endif
 	    }
 
 	    if (! sap) {
@@ -292,7 +311,7 @@ struct sockaddr ***sockaddrp;
 
 	      /* pick the whole sockaddr package! */
 	      memcpy(si4, sa, sizeof(*si4));
-	      sap[++sapcount] = (union sockaddr_uni *)si4;
+	      sap[++sapcount] = (Usockaddr *)si4;
 	    } else if (sa->sa.sa_family == AF_INET6) {
 	      struct sockaddr_in6 *si6 = (void*)malloc(sizeof(*si6));
 	      if (si6 == NULL)
@@ -300,7 +319,7 @@ struct sockaddr ***sockaddrp;
 
 	      /* pick the whole sockaddr package! */
 	      memcpy(si6, sa, sizeof(*si6));
-	      sap[++sapcount] = (union sockaddr_uni *)si6;
+	      sap[++sapcount] = (Usockaddr *)si6;
 	    }
 	  }
 
@@ -313,6 +332,7 @@ done_this_ipv6:
 #else /* SIOCGLIFCONF && IPv6 */
 
 #ifdef SIOCGIFCONF
+	/* #warning "loadifaddresses(): SIOCGIFCONF && INET(4)" */
 	{
 	  struct ifconf ifc;
 	  int ifbufsize = 4 * sizeof(struct ifreq) + 4;
@@ -363,7 +383,7 @@ done_this_ipv6:
 	  for (i = 0; i < ifc.ifc_len; ) {
 
 	    struct ifreq *ifr = (struct ifreq *) &ifc.ifc_buf[i];
-	    union sockaddr_uni *sa = (union sockaddr_uni *) &ifr->ifr_addr;
+	    Usockaddr *sa = (Usockaddr *) &ifr->ifr_addr;
 #ifdef SIOCGIFFLAGS
 	    struct ifreq ifrf;
 #endif
@@ -413,8 +433,11 @@ done_this_ipv6:
 
 	    if (sapcount +2 >= sapspace) {
 	      sapspace <<= 1;
-	      sap = (void*)realloc(sap, (sizeof(union sockaddr_uni*) *
+	      sap = (void*)realloc(sap, (sizeof(Usockaddr*) *
 					 (sapspace + 2)));
+#ifdef TESTMODE
+	      ifs = (void*)realloc(ifs, (sizeof(void*) * (sapspace +2)));
+#endif
 	    }
 	    if (! sap) {
 	      close(s);
@@ -422,13 +445,13 @@ done_this_ipv6:
 	      return -4; /* UAARGH! */
 	    }
 
-	    if (sa->sa.sa_family == AF_INET) {
+	    if (sa->v4.sin_family == AF_INET) {
 	      struct sockaddr_in *si4 = (void*)malloc(sizeof(*si4));
 	      if (si4 == NULL)
 		break;
 	      /* pick the whole sockaddr package! */
 	      memcpy(si4, &ifr->ifr_addr, sizeof(*si4));
-	      sap[++sapcount] = (union sockaddr_uni *)si4;
+	      sap[++sapcount] = (Usockaddr *)si4;
 	    }
 	  }
 	  close(s);
@@ -443,17 +466,22 @@ done_this_ipv6:
 
 #endif /* not HAVE_GETIFADDRS */
 
-	  *sockaddrp = (struct sockaddr **)sap;
+	  *sockaddrp = sap;
 
 	  sap[++sapcount] = NULL;
 
+#ifdef TESTMODE
+	  *ifsp = ifs;
+	  ifs[sapcount] = NULL;
+#endif
+
 	  return sapcount;
 	}
 
 #ifndef TESTMODE /* We test ONLY of  loadifaddresses() routine! */
 
 static             int    nmyaddrs = 0;
-static union sockaddr_uni  ** myaddrs = NULL;
+static Usockaddr  ** myaddrs = NULL;
 
 static void stashmyaddress __((const char *));
 static void
@@ -462,14 +490,9 @@ stashmyaddress(host)
 {
 	int naddrs;
 	struct hostent *hp, hent;
-	union {
-	  struct in_addr ia4;
-#if defined(AF_INET6) && defined(INET6)
-	  struct in6_addr ia6;
-#endif
-	} au;
 	int addrsiz, af, rc;
 	void *addrs[2];
+	Usockaddr au;
 
 	if (host == NULL || *host == 0) return;
 
@@ -490,13 +513,13 @@ stashmyaddress(host)
 	  if (strncasecmp(host,"[IPv6:",6)==0) {
 	    af = AF_INET6;
 	    addrsiz = IN6ADDRSZ;
-	    rc = inet_pton(af, host+6, &au.ia6);
+	    rc = inet_pton(af, host+6, &au.v6);
 	  } else
 #endif
 	    if (*host == '[') {
 	      af = AF_INET;
 	      addrsiz = INADDRSZ;
-	      rc = inet_pton(af, host+1, &au.ia4);
+	      rc = inet_pton(af, host+1, &au.v4);
 	    } else
 	      return;
 
@@ -530,33 +553,31 @@ stashmyaddress(host)
 	/* malloc(size) == realloc(NULL, size) */
 	myaddrs = (void*)realloc((void*)myaddrs,
 				 (nmyaddrs + naddrs +1) *
-				 sizeof(union sockaddr_uni*));
+				 sizeof(Usockaddr*));
 
 	if (!myaddrs) return; /* Uurgh.... */
 
 	for (hp_init(hp); *hp_getaddr() != NULL; hp_nextaddr()) {
 	  if (hp->h_addrtype == AF_INET) {
-	    struct sockaddr_in *si;
+	    Usockaddr *si;
 	    si = (void*)malloc(sizeof(*si));
-	    if (!si) {
-	      return;
-	    }
-	    myaddrs[nmyaddrs++] = (union sockaddr_uni *) si;
+	    if (!si) return;
 	    memset(si,0,sizeof(*si));
-	    si->sin_family = AF_INET;
-	    memcpy(&si->sin_addr.s_addr, *hp_getaddr(), hp->h_length);
+
+	    myaddrs[nmyaddrs++] = si;
+	    si->v4.sin_family = AF_INET;
+	    memcpy(&si->v4.sin_addr.s_addr, *hp_getaddr(), hp->h_length);
 	  }
 #if defined(AF_INET6) && defined(INET6)
 	  if (hp->h_addrtype == AF_INET6) {
-	    struct sockaddr_in6 *si6;
+	    Usockaddr *si6;
 	    si6 = (void*)malloc(sizeof(*si6));
-	    if (!si6) {
-	      return;
-	    }
-	    myaddrs[nmyaddrs++] = (union sockaddr_uni *) si6;
+	    if (!si6) return;
 	    memset(si6,0,sizeof(*si6));
-	    si6->sin6_family = AF_INET;
-	    memcpy(&si6->sin6_addr.s6_addr, *hp_getaddr(), hp->h_length);
+
+	    myaddrs[nmyaddrs++] = si6;
+	    si6->v6.sin6_family = AF_INET6;
+	    memcpy(&si6->v6.sin6_addr.s6_addr, *hp_getaddr(), hp->h_length);
 	  }
 #endif
 	}
@@ -568,7 +589,7 @@ stashmyaddresses(host)
 const char *host;
 {
 	const char *s1, *s2, *zenv;
-	union sockaddr_uni **sa;
+	Usockaddr **sa;
 	int sacnt;
 
 	/* Clear them all away */
@@ -584,7 +605,7 @@ const char *host;
 
 	/* Now start fillig -- interface addresses, if you can get them */
 
-	sacnt = loadifaddresses((struct sockaddr ***)&sa);
+	sacnt = loadifaddresses( &sa );
 
 	if (sacnt > 0) {
 	  /* Okay, we GOT some addresses, I bet we got them all!
@@ -622,17 +643,17 @@ const char *host;
 
 int
 matchmyaddress(_sa)
-	struct sockaddr *_sa;
+	Usockaddr *_sa;
 {
 	int i;
-	union sockaddr_uni *sau = (union sockaddr_uni *) _sa;
+	Usockaddr *sau = _sa;
 
 	if (!myaddrs)
 		stashmyaddresses(NULL);
 	if (!myaddrs) return 0; /* Don't know my addresses ! */
 	
 	/* Match loopback net.. */
-	if (sau->sa.sa_family == AF_INET) {
+	if (sau->v4.sin_family == AF_INET) {
 	  int net;
 
 	  net = (ntohl(sau->v4.sin_addr.s_addr) >> 24) & 0xFF;
@@ -646,12 +667,12 @@ matchmyaddress(_sa)
 
 	for (i = 0; i < nmyaddrs; ++i) {
 	  /* if this is myself, skip to next MX host */
-	  if (sau->sa.sa_family == myaddrs[i]->sa.sa_family) {
-	    if (sau->sa.sa_family == AF_INET &&
+	  if (sau->v4.sin_family == myaddrs[i]->v4.sin_family) {
+	    if (sau->v4.sin_family == AF_INET &&
 		memcmp(&sau->v4.sin_addr, &myaddrs[i]->v4.sin_addr, 4) == 0)
 	      return 1;
 #if defined(AF_INET6) && defined(INET6)
-	    if (sau->sa.sa_family == AF_INET6 &&
+	    if (sau->v4.sin_family == AF_INET6 &&
 		memcmp(&sau->v6.sin6_addr, &myaddrs[i]->v6.sin6_addr, 16) == 0)
 	      return 1;
 #endif
@@ -668,7 +689,7 @@ matchmyaddresses(ai)
 	for ( ; ai ; ai = ai->ai_next ) {
 	  i = 0;
 	  if (ai->ai_addr)
-	    i = matchmyaddress(ai->ai_addr);
+	    i = matchmyaddress((Usockaddr*)ai->ai_addr);
 	  if (i) return i;
 	}
 	return 0;
@@ -684,27 +705,31 @@ int argc;
 char *argv[];
 {
   int cnt, i;
-  struct sockaddr **sa;
+  Usockaddr **sa;
+  void **ifs;
   char buf[80];
+  char *ifname;
 
-  cnt = loadifaddresses(&sa);
+  cnt = loadifaddresses(&sa, &ifs);
 
   printf("loadifaddresses rc=%d\n", cnt);
 
   for (i = 0; i < cnt; ++i) {
-    switch(sa[i]->sa_family) {
+    ifname = ifs ? ifs[i] : NULL;
+    if (!ifname) ifname = "?";
+    switch(sa[i]->v4.sin_family) {
     case AF_INET:
-      inet_ntop(AF_INET, &((struct sockaddr_in **) sa)[i]->sin_addr, buf, sizeof(buf));
-      printf("IPv4: [%s]\n", buf);
+      inet_ntop(AF_INET, &sa[i]->v4.sin_addr, buf, sizeof(buf));
+      printf("IPv4: %8s [%s]\n", ifname, buf);
       break;
 #if defined(AF_INET6) && defined(INET6)
     case AF_INET6:
-      inet_ntop(AF_INET6, &((struct sockaddr_in6 **) sa)[i]->sin6_addr, buf, sizeof(buf));
-      printf("IPv6: [IPv6:%s]\n",buf);
+      inet_ntop(AF_INET6, &sa[i]->v6.sin6_addr, buf, sizeof(buf));
+      printf("IPv6: %8s [IPv6:%s]\n", ifname, buf);
       break;
 #endif
     default:
-      printf("Unknown socket address family: %d\n", sa[i]->sa_family);
+      printf("Unknown socket address family: %d\n", sa[i]->v4.sin_family);
       break;
     }
   }
