@@ -2,6 +2,10 @@
  *
  * Minimum spoolid space is about 17 chars!  Have at least 32 !
  * (8 for a timestamp, 8+1 for the i-node number and terminating null..)
+ *      64-bit decimal integer:    20 ch +
+ *      compact encoded timestamp:  9 ch +
+ *      string end NUL:             1 ch
+ *  Total: 30 chars.
  */
 
 #include <sys/types.h>
@@ -12,13 +16,13 @@
 
 /* These ARE NOT MIME BASE64 characters, but something by which it is
    fairly easily to MANUALLY decode the following result.. */
-char taspid_encodechars[] =
+const char taspid_encodechars[] =
 	"ABCDEFGHIJKLMNOPQRSTUVWXYZ0123abcdefghijklmnopqrstuvwxyz4567890-=";
 
-void taspoolid(buf,mtime,inodenum)
-char *buf;
-time_t mtime;
-long inodenum;
+void taspoolid(buf,inodenum,mtime,mtimens)
+     char *buf;
+     time_t mtime;
+     long inodenum, mtimens;
 {
   char *s = buf;
   struct tm *tt;
@@ -49,6 +53,15 @@ long inodenum;
   /* Seconds */
   *s++ = taspid_encodechars[tt->tm_sec];
 
+  if (mtimens != 0) { /* Add nanoseconds to the spoolid ONLY IF it differs from zero! */
+    /* Nanoseconds */
+    *s++ = taspid_encodechars[ (mtimens >> 24) & 63 ]; /*   1.1 s  */
+    *s++ = taspid_encodechars[ (mtimens >> 18) & 63 ]; /*  16.8 ms */
+    *s++ = taspid_encodechars[ (mtimens >> 12) & 63 ]; /* 262.1 s */
+    *s++ = taspid_encodechars[ (mtimens >>  6) & 63 ]; /*   4.1 s */
+    *s++ = taspid_encodechars[ (mtimens      ) & 63 ]; /*    64 ns */
+  }
+
   *s = 0; /* terminate zero */
 
   /* .. and finally attach the inode-number part of the spool file name. */
