@@ -29,6 +29,8 @@
  * 
  */
 
+#if 0
+
 /* The Inner Net License, Version 2.00
 
   The author(s) grant permission for redistribution and use in source and
@@ -87,7 +89,9 @@ SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 #include <arpa/inet.h>
 #include <arpa/nameser.h> /* Sol 2.6 barfs without this.. */
 #include <resolv.h>
+#ifdef HAVE_SYS_UN_H
 #include <sys/un.h>
+#endif
 #include <sys/utsname.h>
 #include <netdb.h>
 #if !defined(EAI_AGAIN) || !defined(AI_NUMERICHOST)
@@ -101,8 +105,10 @@ SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 #endif /* min */
 
 #ifndef AF_LOCAL
-# define AF_LOCAL AF_UNIX
-# define PF_LOCAL PF_UNIX
+# ifdef AF_UNIX
+#  define AF_LOCAL AF_UNIX
+#  define PF_LOCAL PF_UNIX
+# endif
 #endif
 
 static char * nrl_domainname __((void));
@@ -156,10 +162,28 @@ nrl_domainname ()
   return domain;
 }
 
+/* This is NASTY, GLIBC has changed the type after instroducing
+   this function, Sol (2.)8 has 'int', of upcoming POSIX standard
+   revision I don't know.. */
+
+#ifndef GETNAMEINFOFLAGTYPE
+# if defined(__GLIBC__) && defined(__GLIBC_MINOR__)
+#  if __GLIBC__ >= 2 && __GLIBC_MINOR__ >= 2
+	/* I am not sure that it was already 2.2(.0) that had
+	   this change, but 2.2.2 has it... */
+#   define GETNAMEINFOFLAGTYPE unsigned int
+#  else
+#   define GETNAMEINFOFLAGTYPE int
+#  endif
+# else
+#  define GETNAMEINFOFLAGTYPE int
+# endif
+#endif
 
 int
 getnameinfo __((const struct sockaddr *sa, size_t addrlen, char *host,
-		size_t hostlen, char *serv, size_t servlen, int flags));
+		size_t hostlen, char *serv, size_t servlen,
+		GETNAMEINFOFLAGTYPE flags));
 
 int
 getnameinfo (sa, addrlen, host, hostlen, serv, servlen, flags)
@@ -169,7 +193,7 @@ getnameinfo (sa, addrlen, host, hostlen, serv, servlen, flags)
      size_t hostlen;
      char *serv;
      size_t servlen;
-     int flags;
+     GETNAMEINFOFLAGTYPE flags;
 {
   int serrno = errno;
   int ok = 0;
@@ -178,10 +202,12 @@ getnameinfo (sa, addrlen, host, hostlen, serv, servlen, flags)
     return -1;
 
   switch (sa->sa_family) {
+#ifdef AF_LOCAL
   case AF_LOCAL:
     if (addrlen < (size_t) (((struct sockaddr_un *) NULL)->sun_path))
       return -1;
     break;
+#endif
   case AF_INET:
     if (addrlen < sizeof (struct sockaddr_in))
       return -1;
@@ -234,6 +260,7 @@ getnameinfo (sa, addrlen, host, hostlen, serv, servlen, flags)
       }
       break;
 
+#ifdef AF_LOCAL
     case AF_LOCAL:
       if (!(flags & NI_NUMERICHOST)) {
 	struct utsname utsname;
@@ -250,6 +277,7 @@ getnameinfo (sa, addrlen, host, hostlen, serv, servlen, flags)
 
       strncpy (host, "localhost", hostlen);
       break;
+#endif
 
     default:
       return -1;
@@ -275,9 +303,11 @@ getnameinfo (sa, addrlen, host, hostlen, serv, servlen, flags)
 	}
 	break;
 
+#ifdef AF_LOCAL
       case AF_LOCAL:
 	strncpy (serv, ((struct sockaddr_un *) sa)->sun_path, servlen);
 	break;
+#endif
     }
 
   if (host && (hostlen > 0))
@@ -287,3 +317,5 @@ getnameinfo (sa, addrlen, host, hostlen, serv, servlen, flags)
   errno = serrno;
   return 0;
 }
+
+#endif
