@@ -4,6 +4,11 @@
  */
 
 #include "hostenv.h"
+#include "zmsignal.h"
+
+#ifndef HAVE_STRSIGNAL
+
+#if !HAVE_DECL_SYS_SIGLIST
 #ifndef	HAVE_SYS_SIGLIST /* No  sys_siglist[]  at the libc ? */
 const char *sys_siglist[] = {
 	"Signal 0",
@@ -40,9 +45,9 @@ const char *sys_siglist[] = {
 	"Signal 31"
 };
 #endif	/* HAVE_SYS_SIGLIST */
+#endif
 
-#ifndef HAVE_STRSIGNAL
-#ifndef SYS_SIGLIST_DECLARED /* Not declared anywhere ? */
+#if !HAVE_DECL_SYS_SIGLIST /* Not declared anywhere ? */
 extern const char *sys_siglist[];
 #endif
 
