@@ -34,18 +34,18 @@ STATIC u_int	outlen, outsize;
 
 struct smallSymbol shsymbol[CHARSETSIZE];
 
-struct kwdefn {
+static struct kwdefn {
 	const char *name;
 	Keyword	    kw;
 } kwlist[] = {	/* most used first order */
 {	"in",		kIn		},
 #ifdef	MAILER
-{	"ssift",	kSift		}, /* String SIFT */
-{	"tfiss",	kTfis		},
+{	"ssift",	kSSift		}, /* String SIFT */
+{	"tfiss",	kTfisS		},
 {	"tsift",	kTSift		}, /* Token SIFT */
-{	"tfist",	kTTfis		},
-{	"sift",		kTSift		}, /* Token SIFT -- compab.alias */
-{	"tfis",		kTTfis		},
+{	"tfist",	kTfisT		},
+{	"sift",		kSift		}, /* Obsolete Token SIFT */
+{	"tfis",		kTfis		},
 #endif	/* MAILER */
 {	"case",		kCase		},
 {	"esac",		kEsac		},
@@ -166,7 +166,7 @@ ShInitIFS(uIFS)
 void
 ShInit()
 {
-	register int i;
+	register u_int i;
 	register const u_char *cp;
 
 	for (i = 0; i < (sizeof shsymbol/sizeof shsymbol[0]); ++i) {
@@ -249,6 +249,10 @@ STATIC struct errmsgs {
 { eIllegalLeftParen,		"illegal position for unquoted left parenthesis"},
 { eIllegalConnector,		"illegal connector after backgrounding"	},
 { eIllegalTokenUnbalancedList,	"illegal token within list"		},
+{ eObsoleteSift,		"'sift' is obsolete form of 'tsift'" },
+{ eObsoleteTfis,		"'tfis' is obsolete form of 'tfist'" },
+{ eTfissMisparity,		"had 'tfiss', expected 'tfist'" },
+{ eTfistMisparity,		"had 'tfist', expected 'tfiss'" },
 };
 
 STATIC const char *errfilename;   /* set from inputname in SslWalker */
@@ -258,7 +262,7 @@ STATIC void
 sslw_error(errCode)
 	ErrorCodes errCode;
 {
-	int i;
+	u_int i;
 	char *cp, *bos;
 	const char *msg = "unknown error";
 
@@ -737,7 +741,7 @@ SslWalker(inputname, tfp, eotp)
 	register int	i, bufferIndex;
 	TableOperation	opcode;	/* the current operation */
 	int		branchOriginTop, branchOrigin[127];
-	char		*cp, *herebuf, *buffer;
+	unsigned char		*cp, *herebuf, *buffer;
 	int		bufferSize;
 
 	/*
@@ -915,8 +919,7 @@ SslWalker(inputname, tfp, eotp)
 			if (isset('S'))
 				printf("identifying '%s'\n", buffer);
 			for (kwp = &kwlist[0]; kwp->name != NULL; ++kwp) {
-				if (*(kwp->name) == buffer[0]
-				    && strcmp(kwp->name, buffer) == 0) {
+				if (strcmp(kwp->name, buffer) == 0) {
 					resultValue = (int)(kwp->kw);
 					break;
 				}
