@@ -25,7 +25,7 @@ int v_record = 0;	/* record variable accesses */
 int v_changed = 0;	/* a variable in v_accessed was changed */
 struct vaccess *v_accessed = NULL;
 
-extern char **environ; /* Should be declated in  <stdlib.h>, but isn't.. */
+extern char **environ; /* Should be declared in  <stdlib.h>, but isn't.. */
 
 void
 v_written(l)
@@ -89,6 +89,9 @@ STATIC struct vsync {
 };
 
 
+FILE *vfindlog;
+int vfindtest;
+
 /*
  * The ifs_flush() routine is defined below, it maintains this global value:
  */
@@ -99,21 +102,50 @@ char *ifs;		/* input file separator characters */
  * Find the value of a named variable by looking through the scoped p-lists.
  */
 
+#include <fcntl.h>
+
 conscell *
 v_find(name)
 	const char * name;
 {
 	register conscell *l, *pl, *scope;
 	int nlen = strlen(name);
+	int varscopedepth = 0;
+	int varposcount = 0;
+	int varsearchcount = 0;
+	
+	if (!vfindtest) {
+	  char *s = getenv("VFINDLOGFILE");
+	  if (s)  {
+	    int fd = open(s, O_WRONLY|O_APPEND|O_CREAT, 0644);
+	    if (fd >= 0) {
+	      if (dup2(fd, 99) >  0) {
+		vfindlog = fdopen(99, "a");
+	      }
+	      close(fd);
+	    }
+	  }
+	  vfindtest = 1;
+	}
+
 
 	if (name == NULL) return NULL; /* No input name, no output var.. */
 
 	/* if (fvcache.namesymbol > 0 && fvcache.namesymbol == symbol(name))
 		return fvcache.location; */
 
+	if (vfindlog)
+	  fprintf(vfindlog, "vfind{%d:%s}\t", nlen, name);
+
 	pl = NULL;
 	for (scope = car(envarlist); scope != NULL; scope = cdr(scope)) {
+		++ varscopedepth;
+		varposcount = 0;
+
 		for (l = car(scope); l != NULL; pl = cdr(l), l = cddr(l)) {
+			++varsearchcount;
+			++varposcount;
+
 			if (l->slen == nlen &&
 			    memcmp(name, l->cstring, nlen) == 0){
 				if (l != car(scope)) {
@@ -135,10 +167,18 @@ v_find(name)
 					v_accessed = v;
 				}
 #endif	/* MAILER */
+				if (vfindlog)
+				  fprintf(vfindlog, "%d\t%d\t%d\n",
+					  varscopedepth, varposcount, varsearchcount);
+
 				return l;
 			}
 		}
 	}
+	if (vfindlog)
+	  fprintf(vfindlog, "%d\t%d\t%d\n",
+		  0, 0, varsearchcount);
+
 	return NULL;
 }
 
@@ -150,6 +190,9 @@ v_find(name)
  */
 
 
+#define ALWAYS_QUOTE_EXPANDS
+
+
 conscell *
 v_expand(s, caller, retcode)
 	const char *s;			/* variable name */
@@ -160,6 +203,7 @@ v_expand(s, caller, retcode)
 	register int n;
 	register char *cp;
 	char np[CHARSETSIZE+1]; /* each possible option, plus last NUL */
+	int noquote = 0;
 
 	GCVARS3;
 	GCPRO3(d,l,tmp);
@@ -172,6 +216,7 @@ v_expand(s, caller, retcode)
 	 */
 	switch (*s) {
 	case '@':
+		noquote = NOQUOTEIFQUOTED;
 	case '*':
 		if (caller == NULL || cdar(caller->argv) == NULL) {
 			goto end_v_expand;
@@ -180,9 +225,13 @@ v_expand(s, caller, retcode)
 		for (l = d; l != NULL && cdr(l) != NULL ; l = cdr(l)) {
 			tmp = conststring(" ",1);
 			cdr(tmp) = cdr(l);
-			l = cdr(l) = tmp;
-			if (*s == '@')
-				tmp->flags |= NOQUOTEIFQUOTED;
+			cdr(l)   = tmp;
+			l        = cdr(l);
+			tmp->flags |= noquote;
+#ifdef ALWAYS_QUOTE_EXPANDS
+			if (!noquote)
+			  tmp->flags |= QUOTEDSTRING;
+#endif
 		}
 		/* grindef("ARGW = ", ncons(d)); */
 		goto end_v_expand;
@@ -198,6 +247,9 @@ v_expand(s, caller, retcode)
 		}
 		d = copycell(d);
 		cdr(d) = NULL;
+#ifdef ALWAYS_QUOTE_EXPANDS
+		d->flags |= QUOTEDSTRING;
+#endif
 		/* d = s_copy_tree(d); */ /* XXX: Needed ? */
 		goto end_v_expand;
 
@@ -260,6 +312,10 @@ v_expand(s, caller, retcode)
 	if ((d = v_find(s)) != NULL) {
 		d = copycell(cdr(d));
 		cdr(d) = NULL;
+#ifdef ALWAYS_QUOTE_EXPANDS
+		if (STRING(d))
+		  d->flags |= QUOTEDSTRING;
+#endif
 	}
  end_v_expand:
 
@@ -404,9 +460,9 @@ v_envinit()
 		*--cp = '=';
 	}
 	if (!gotpath) {
-		s = conststring(PATH,strlen(PATH));
+		s = conststring(PATH,sizeof(PATH)-1);
 		nconc(envarlist, s);
-		s = conststring(DEFAULT_PATH,strlen(DEFAULT_PATH));
+		s = conststring(DEFAULT_PATH,sizeof(DEFAULT_PATH)-1);
 		nconc(envarlist, s);
 	}
 	/* now make it into a list of plists */
@@ -414,7 +470,7 @@ v_envinit()
 	envarlist = ncons(envarlist);	/* ((env)) */
 	/* ... and prepend the normal scope */
 	/* ... and put it in the list of pre-defined non-env. variables */
-	s = conststring(ENVIRONMENT,strlen(ENVIRONMENT));
+	s = conststring(ENVIRONMENT,sizeof(ENVIRONMENT)-1);
 	cdr(s) = envarlist;	/* must only use s_push with envarlist now */
 	cdr(s_last(e)) = s;	/* (envcopy ENVIRONMENT (env))		*/
 	s = ncons(e);		/* s = (envcopy ENVIRONMENT (env))	*/
