@@ -17,8 +17,8 @@ export LINES
 
 echo "<HTML><HEAD><TITLE>"
 basename "$1"
-echo "</TITLE></HEAD><BODY BGCOLOR=white><PRE>"
-groff  -t -man -Tascii "$1" | \
+echo '</TITLE></HEAD><BODY BGCOLOR=white><script src="http://www.google-analytics.com/urchin.js" type="text/javascript"> </script> <script type="text/javascript"> _uacct = "UA-194078-3"; urchinTracker(); </script> <PRE>'
+groff -t -man -Tascii "$1" | \
     perl -e '
         #select STDIN; $| = 1;
         select STDERR; $| = 1;
@@ -85,18 +85,29 @@ groff  -t -man -Tascii "$1" | \
         s{</I><B>_</B><I>}{_}og;
 	# Ordinary man-pages
         s{<I>([-.0-9a-zA-Z_]+)</I>\((\dzm)\)}{<A HREF="\1.\2.html"><I>\1</I>(\2)</A>}og;
+
 	# Ordinary PERL PODs
         s{<I>([-.0-9a-zA-Z_]+::[-.0-9a-zA-Z_]+)</I>\((\d\w+)\)}{<A HREF="\1.\2.html"><I>\1</I>(\2)</A>}og;
 	print;' | \
-    perl -ne '
-	if (m{^<B>(.*)</B>$}o) {
+    perl -e '
+	@labels=();
+	while (<STDIN>) {
+	  if (m{^<B>(.*)</B>$}o) {
 	    my $n = $1; $n =~ s/ /_/g;
 	    printf "<A NAME=\"%s\"></A>",$n;
-	}
-	if (m{^   <B>(.*)</B>$}o) {
+	    push @labels, $n;
+	  }
+	  if (m{^   <B>(.*)</B>$}o) {
 	    my $n = $1; $n =~ s/ /_/g;
 	    printf "<A NAME=\"%s\"></A>",$n;
+	    push @labels, $n;
+	  }
+	  print;
 	}
-	print;
+	printf "<p><p>\n<ul></n";
+	foreach $n (@labels) {
+	    printf "<li> <A HREF=\"#%s\">%s</A>\n",$n,$n;
+	}
+	printf "</ul>\n";
 	'
 echo "</PRE></BODY></HTML>"
