@@ -1,7 +1,7 @@
 /*
  *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
- *	Some functions Copyright 1991-2002 Matti Aarnio.
+ *	Some functions Copyright 1991-2006 Matti Aarnio.
  */
 
 /*
@@ -9,28 +9,14 @@
  *  script, along with its support facilities.
  */
 
-#include "mailer.h"
-#include <stdio.h>
-#include <sys/types.h>
-#ifdef HAVE_UNISTD_H
-# include <unistd.h>
-#endif
-#include <ctype.h>
-#include <fcntl.h>
-#include <sys/file.h>			/* O_RDONLY for run_praliases() */
-#include <pwd.h>			/* for run_homedir() */
-#include <grp.h>			/* for run_grpmems() */
-#include <errno.h>
+#include "router.h"
 #ifdef HAVE_SYS_UN_H
 #include <sys/socket.h>
 #include <sys/un.h>
 #endif
 
-#include "zsyslog.h"
-#include "shmmib.h"
-#include "sysexits.h"
-
-extern struct MIB_MtaEntry *MIBMtaEntry;
+#include "zmpoll.h"  /* We have ZMPOLL -wrapper for preferrably real poll(2),
+			or as a backup, for select(2).  */
 
 #ifdef HAVE_DIRENT_H
 # include <dirent.h>
@@ -48,13 +34,15 @@ extern struct MIB_MtaEntry *MIBMtaEntry;
 #endif /* HAVE_DIRENT_H */
 
 #ifdef HAVE_DIRFD
-#   define Z_DIRFD(dir) dirfd(dir)
+#   define Z_DIRFD(dirp) dirfd(dirp)
 #else
 # ifdef USE_DIRFD_DD_FD
-#   define Z_DIRFD(dir) ((dir).dd_fd)
+#   define Z_DIRFD(dirp) ((dirp)->dd_fd)
 # else
 #  ifdef USE_DIRFD_D_FD
-#   define Z_DIRFD(dir) ((dir).dd_fd)
+#   define Z_DIRFD(dirp) ((dirp)->d_fd)
+#  else
+#   error "ERROR:ERROR:ERROR: bad dirfd() autoconf"
 #  endif
 # endif
 #endif
@@ -67,16 +55,8 @@ extern struct MIB_MtaEntry *MIBMtaEntry;
 #include <sys/resource.h>
 #endif
 
-#include "zmsignal.h"
-
-#include "zsyslog.h"
-#include "mail.h"
 #include "interpret.h"
 #include "io.h"
-#include "libz.h"
-#include "libc.h"
-
-#include "prototypes.h"
 
 #ifdef	HAVE_LOCKF
 #ifdef	F_OK
@@ -98,21 +78,12 @@ extern struct MIB_MtaEntry *MIBMtaEntry;
 #define	_IOFBF	0
 #endif	/* !_IOFBF */
 
-
-extern const char *traps[];
-extern time_t time __((time_t *));
-extern int routerdirloops;
-
-#ifndef strchr
-extern char *strchr(), *strrchr();
-#endif
-
-extern void free_gensym __((void));
-
 static int gothup = 0;
 
-static RETSIGTYPE sig_exit __((int));
-static RETSIGTYPE
+extern struct itimerval profiler_itimer_at_start;
+
+
+RETSIGTYPE
 sig_exit(sig)
 int sig;
 {
@@ -136,16 +107,13 @@ int sigarg;
 }
 
 static void dohup __((int));
-static void
-dohup(sig)
-int sig;
+static void dohup(sig)
+     int sig;
 {
 	gothup = 0;
-	if (traps[SIGHUP] != NULL)
-		eval(traps[SIGHUP], "trap", NULL, NULL);
+	traphup(sig);
 }
 
-
 /*
  * Run the Router in Daemon mode.
  */
@@ -167,6 +135,8 @@ struct router_child {
   int   fromchild;
   int   childpid;
   int   hungry;
+  struct zmpollfd *fdpto;
+  struct zmpollfd *fdpfrom;
 
   char *linebuf;
   int   linespace;
@@ -189,21 +159,6 @@ struct router_child {
 
 struct router_child routerchilds[MAXROUTERCHILDS];
 
-extern int nrouters;
-extern const char *logfn;
-
-/* ../scheduler/pipes.c */
-extern int  pipes_create         __((int *tochild, int *fromchild));
-extern void pipes_close_parent   __((int *tochild, int *fromchild));
-extern void pipes_to_child_fds   __((int *tochild, int *fromchild));
-extern void pipes_shutdown_child __((int tochild)); /* At parent, shutdown channel towards child */
-
-/* ../scheduler/resources.c */
-extern int  resources_query_nofiles  __((void));
-extern void resources_maximize_nofiles __((void));
-extern void resources_limit_nofiles __((int nfiles));
-extern int  resources_query_pipesize __((int fildes));
-
 static void child_server __((int tofd, int frmfd));
 static int  rd_doit __((const char *filename, const char *dirs));
 static int  parent_reader __((int waittime));
@@ -290,10 +245,26 @@ static int  start_child (i)
 
 	  int idx;
 
-	  pipes_to_child_fds(tofd,frmfd);
-	  for (idx = resources_query_nofiles(); idx >= 3; --idx)
-	    close(idx);
+#if 0
+#if	defined(SA_NOCLDSTOP)||defined(SA_ONSTACK)||defined(SA_RESTART)
+	/* ================ POSIX.1 STUFF ================ */
+	  extern void sigact_segv __((int, siginfo_t *, void *));
+	  extern void sigact_bus __((int, siginfo_t *, void *));
 
+	  SIGNAL_ACTION(SIGSEGV, sigact_segv);
+	  SIGNAL_ACTION(SIGBUS,  sigact_bus);
+#endif
+#endif
+
+	  /* Run the profiler itimer, if it was on before a fork... */
+	  setitimer(ITIMER_PROF, &  profiler_itimer_at_start, NULL);
+
+
+	  pipes_to_child_fds(tofd,frmfd);
+	  for (idx = resources_query_nofiles(); idx >= 3; --idx) {
+	    if (idx == 99) continue; /* Special debug handle, if ever used */
+	    close(idx);
+	  }
 #if 0
 	  resources_maximize_nofiles();
 #endif
@@ -353,10 +324,11 @@ int signum;
 
 #ifdef  HAVE_WAIT4
 #ifdef HAVE_SYS_RESOURCE_H
-	  pid = wait4(-1, &statloc, WNOHANG, &r);
+	  pid = wait4(0, &statloc, WNOHANG, &r);
 #else
-	  pid = wait4(-1, &statloc, WNOHANG, NULL);
+	  pid = wait4(0, &statloc, WNOHANG, NULL);
 #endif
+	  if (pid <= 0) break; /* No processes */
 #else
 #ifdef  HAVE_WAIT3
 #ifdef HAVE_SYS_RESOURCE_H
@@ -364,11 +336,13 @@ int signum;
 #else
 	  pid = wait3(&statloc, WNOHANG, NULL);
 #endif
+	  if (pid <= 0) break; /* No processes */
 #else
 #ifdef	HAVE_WAITPID
 	  pid = waitpid(-1, &statloc, WNOHANG);
 #else
-	  pid = wait(&statloc);
+	  pid = wait(&statloc); :: ERROR :: ERROR :: ERROR ::
+	    /* THIS IS EXTREMELY BAD CASE! */
 #endif
 #endif
 #endif
@@ -568,109 +542,58 @@ static void _parent_writer(rc)
 }
 
 
-#ifdef	HAVE_SELECT
-
-#ifdef _AIX /* The select.h  defines NFDBITS, etc.. */
-# include <sys/types.h>
-# include <sys/select.h>
-#endif
-
-
-#if	defined(BSD4_3) || defined(sun)
-#include <sys/file.h>
-#endif
-#include <errno.h>
-#include <netdb.h>
-#include <netinet/in.h>
-#include <sys/socket.h>
-#include <sys/time.h>
-
-#ifndef	NFDBITS
-/*
- * This stuff taken from the 4.3bsd /usr/include/sys/types.h, but on the
- * assumption we are dealing with pre-4.3bsd select().
- */
-
-typedef long	fd_mask;
-
-#ifndef	NBBY
-#define	NBBY	8
-#endif	/* NBBY */
-#define	NFDBITS		((sizeof fd_mask) * NBBY)
-
-/* SunOS 3.x and 4.x>2 BSD already defines this in /usr/include/sys/types.h */
-#ifdef	notdef
-typedef	struct fd_set { fd_mask	fds_bits[1]; } fd_set;
-#endif	/* notdef */
-
-#ifndef	_Z_FD_SET
-#define	_Z_FD_SET(n, p)   ((p)->fds_bits[0] |= (1 << (n)))
-#define	_Z_FD_CLR(n, p)   ((p)->fds_bits[0] &= ~(1 << (n)))
-#define	_Z_FD_ISSET(n, p) ((p)->fds_bits[0] & (1 << (n)))
-#define _Z_FD_ZERO(p)	  memset((char *)(p), 0, sizeof(*(p)))
-#endif	/* !FD_SET */
-#endif	/* !NFDBITS */
-
-#ifdef FD_SET
-#define _Z_FD_SET(sock,var) FD_SET(sock,&var)
-#define _Z_FD_CLR(sock,var) FD_CLR(sock,&var)
-#define _Z_FD_ZERO(var) FD_ZERO(&var)
-#define _Z_FD_ISSET(i,var) FD_ISSET(i,&var)
-#else
-#define _Z_FD_SET(sock,var) var |= (1 << sock)
-#define _Z_FD_CLR(sock,var) var &= ~(1 << sock)
-#define _Z_FD_ZERO(var) var = 0
-#define _Z_FD_ISSET(i,var) ((var & (1 << i)) != 0)
-#endif
-
 /* Return info re has some waiting been done */
 
 static int parent_reader(waittime)
 	int waittime;
 {
-  fd_set rdset, wrset;
-  int i, highfd, fd, rc;
-  struct timeval tv;
+  int i, fd, rc, highfd;
+  struct zmpollfd *fds = NULL;
+  struct zmpollfd *notifyfdp = NULL;
 
   if (notifysocket_reinit && notifysocket_reinit < now)
     notifysock_init();
 
  redo_again:;
 
-  _Z_FD_ZERO(rdset);
-  _Z_FD_ZERO(wrset);
+  highfd = 0; /* actually counts polled fd's, does not track
+		 highest fd like with select(2) ! */
 
-  highfd = notifysocket;
   if (notifysocket >= 0)
-    _Z_FD_SET(notifysocket, rdset);
+    zmpoll_addfd(&fds, &highfd, notifysocket, -1, &notifyfdp);
 
   for (i = 0; i < MAXROUTERCHILDS; ++i) {
     /* Places to read from ?? */
     fd = routerchilds[i].fromchild;
-    if (fd >= 0) {
-      _Z_FD_SET(fd, rdset);
-      if (highfd < fd)
-	highfd = fd;
-    }
+    if (fd >= 0)
+      zmpoll_addfd(&fds, &highfd, fd, -1, &routerchilds[i].fdpfrom);
+    else
+      routerchilds[i].fdpfrom = NULL; /* Mark that we do NOT have dataset! */
+
     /* Something wanting to write ?? */
     fd = routerchilds[i].tochild;
     if (fd >= 0 &&
-	routerchilds[i].childout < routerchilds[i].childsize) {
-      _Z_FD_SET(fd, wrset);
-      if (highfd < fd)
-	highfd = fd;
-    }
+	routerchilds[i].childout < routerchilds[i].childsize)
+      /* Have fd, but nothing to write.. */
+      zmpoll_addfd(&fds, &highfd, -1, fd, &routerchilds[i].fdpto);
+    else
+      routerchilds[i].fdpto = NULL; /* Mark that we do NOT have dataset! */
   }
+
 #if 0
-  if (highfd < 0) return 0; /* Nothing to do! */
+  if (highfd < 0) {
+    if (fds) free(fds);
+    return 0; /* Nothing to do! */
+  }
 #endif
 
-  tv.tv_sec = waittime;
-  tv.tv_usec = 0;
-  rc = select(highfd+1, &rdset, &wrset, NULL, &tv);
+  rc = zmpoll( fds, highfd, waittime*1000 /* millisecs */ );
 
-  if (rc == 0) return 1; /* Nothing to do, leave..
-			    Did sleep for a second! */
+  if (rc == 0) {
+    if (fds) free(fds);
+    return 1; /* Nothing to do, leave..
+		 Did sleep for a waittime! */
+  }
 
   if (rc < 0) {
     /* Drat, some error.. */
@@ -681,37 +604,27 @@ static int parent_reader(waittime)
       _parent_writer(&routerchilds[i]);
       _parent_reader(&routerchilds[i]);
     }
+    if (fds) free(fds);
     return 0;  /* Urgh, an error.. */
   }
 
-  /* Ok, select gave indication of *something* being ready for read */
+  /* Ok, zmpoll() gave indication of *something* being ready for read */
 
-  if (notifysocket >= 0 && _Z_FD_ISSET(notifysocket, rdset))
+  if (notifyfdp  &&  (notifyfdp->revents & (ZM_POLLIN|ZM_POLLERR|ZM_POLLHUP)))
     notify_reader(notifysocket);
 
   for (i = 0; i < MAXROUTERCHILDS; ++i) {
-    fd = routerchilds[i].tochild;
-    if (fd >= 0 && _Z_FD_ISSET(fd, wrset))
+    struct router_child *r = &routerchilds[i];
+
+    if (r->fdpto && r->fdpto->revents & ZM_POLLOUT)
       _parent_writer(&routerchilds[i]);
-    fd = routerchilds[i].fromchild;
-    if (fd >= 0 && _Z_FD_ISSET(fd, rdset))
+    if (r->fdpfrom && r->fdpfrom->revents & (ZM_POLLIN|ZM_POLLERR|ZM_POLLHUP))
       _parent_reader(&routerchilds[i]);
   }
+
   return 1; /* Did some productive job, don't sleep at the caller.. */
 }
 
-#else /* NO HAVE_SELECT */
-static int parent_reader(waittime)
-	int waittime;
-{
-  int i;
-  /* No select, but can do non-blocking -- we hope.. */
-  for (i = 0; i < MAXROUTERCHILDS; ++i)
-    _parent_reader(&routerchilds[i], waittime);
-  return 0;
-}
-#endif
-
 
 
 /*
@@ -904,6 +817,7 @@ static void child_server(tofd,frmfd)
 struct dirstatname {
 	struct stat st;
 	long ino;
+	int  sizekb;
 	char *dir; /* Points to stable strings.. */
 	char name[1]; /* Allocate enough size */
 };
@@ -1004,12 +918,14 @@ dq_insert(DQ, ino, file, dir)
 	*/
 
 	MIBMtaEntry->rt.ReceivedMessages += 1;
-	MIBMtaEntry->rt.StoredMessages   += 1;
+	MIBMtaEntry->rt.StoredMessages    = dq->wrkcount;
 
 	i = (stbuf.st_size + stbuf.st_blksize -1)/1024;
 	MIBMtaEntry->rt.ReceivedVolume += i;
 	MIBMtaEntry->rt.StoredVolume   += i;
 
+	dsn->sizekb = i;
+
 	return 0;
 }
 
@@ -1142,7 +1058,11 @@ int syncweb(rc)
 	int wrkidx = dq->wrkcount -1;
 
 	/* Any work to do ? */
-	if (dq->wrksum == 0) return 0;
+	if (dq->wrksum == 0) {
+	  MIBMtaEntry->rt.StoredMessages   = 0;
+	  MIBMtaEntry->rt.StoredVolume     = 0;
+	  return 0;
+	}
 
 	time(&now);
 
@@ -1612,9 +1532,9 @@ run_daemon(argc, argv)
 
 	/* instantiate the signal handler.. */
 #ifdef SIGCLD
-	SIGNAL_HANDLE(SIGCLD,  sig_chld);
+	sig_chld(SIGCLD);
 #else
-	SIGNAL_HANDLE(SIGCHLD, sig_chld);
+	sig_chld(SIGCHLD);
 #endif
 
 	SIGNAL_HANDLE(SIGTERM, sig_exit);	/* mustexit = 1 */
@@ -1687,7 +1607,7 @@ run_daemon(argc, argv)
 	      break; /* fork failed.. */
 	}
 
-	sleep(5); /* Wait a bit before continuting so that
+	sleep(5); /* Wait a bit before continuing so that
 		     child processes have a change to start */
 
 	/* Collect start reports (initial "#hungry\n" lines) */
@@ -1697,19 +1617,40 @@ run_daemon(argc, argv)
 
 	    time(&now);
 	    if (now > nextdirscan) {
+	      long sizekbsum = 0;
+	      long msgcountsum = 0;
+
 	      nextdirscan = now + 10;
 
 	      for (i = 0; i < ROUTERDIR_CNT; ++i) {
-		if (routerdirs2[i])
+		if (routerdirs2[i]) {
 		  dirqueuescan(routerdirs2[i], dirq[i], 1);
+		  msgcountsum += dirq[i]->wrkcount;
+		  for (ii = 0; ii < dirq[i]->wrkcount; ++i) {
+		    sizekbsum += dirq[i]->stats[ii]->sizekb;
+		  }
+		}
 	      }
+
+	      /* Resync the queue size and count every now and then...
+		 so that even if the sequencer() leaks them, they get
+		 back into sanity quick enough.. */
+	      MIBMtaEntry->rt.StoredMessages   = msgcountsum;
+	      MIBMtaEntry->rt.StoredVolume     = sizekbsum;
+
+#ifdef SIGCLD /* re-instantiate the child processor */
+	      sig_chld(SIGCLD);
+#else
+	      sig_chld(SIGCHLD);
+#endif
 	    }
 
 	    for (i = 0; i < MAXROUTERCHILDS; ++i) {
 	      if ( routerchilds[i].tochild < 0 &&
 		   routerchilds[i].dq &&
-		   routerchilds[i].dq->wrkcount )
+		   routerchilds[i].dq->wrkcount ) {
 		start_child(i);
+	      }
 
 	      if ( routerchilds[i].tochild >= 0 &&
 		   routerchilds[i].childsize == 0 &&
