@@ -2,9 +2,10 @@
  *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
  *
- *	Modifications/maintance, Matti Aarnio, over years 1990-2003
+ *	Modifications/maintance, Matti Aarnio, over years 1990-2004
  *
  *	'longestmatch' driver kissg@sztaki.hu 970209
+ *	'pathalias.at' driver added by Daniel Kiper <dkiper@netspace.com.pl>.
  */
 
 /*
@@ -17,26 +18,10 @@
  * needs or capabilities of each host.
  */
 
-#include "mailer.h"
-#ifdef HAVE_UNISTD_H
-#include <unistd.h>
-#endif
-#include <ctype.h>
-#include <sys/file.h>
-#ifdef HAVE_FCNTL_H
-# include <fcntl.h>
-#endif
+#include "router.h"
 #include "libdb/search.h"
-#include "splay.h"
 #include "sleepycatdb.h"
 
-#include "prototypes.h"
-
-extern long crc32  __((const void *));
-extern long crc32n __((const void *, int));
-
-extern struct sptree *spt_databases; /* At conf.c */
-
 /*
  * The following tables describes the kinds of lookups we support.
  */
@@ -171,17 +156,18 @@ struct db_kind {
 
 /* drivers */
 static conscell	*find_domain       __((conscell *DBFUNC(lookupfn), search_info *sip));
+static conscell	*find_at_domain    __((conscell *DBFUNC(lookupfn), search_info *sip));
 static conscell	*find_nodot_domain __((conscell *DBFUNC(lookupfn), search_info *sip));
 static conscell *find_longest_match __((conscell *DBFUNC(lookupfn), search_info *sip));
 /* others.. */
 static void      cacheflush __((struct db_info *dbip));
-extern conscell	*readchunk  __((const char *file, long offset));
-static int	 iclistdbs  __((struct spblk *spl));
+static int	 iclistdbs  __((void *, struct spblk *spl));
 
 
 static void (*cachemarkupfunc) __((conscell*));
 static int
-iccachemarkup(spl)
+iccachemarkup(p, spl)
+	void *p;
 	struct spblk *spl;
 {
 	struct db_info *dbip = (struct db_info *)spl->data;
@@ -205,7 +191,7 @@ cache_gc_markup_iterator(mrkupfunc)
      void (*mrkupfunc)__((conscell*));
 {
   cachemarkupfunc = mrkupfunc;
-  sp_scan(iccachemarkup, (struct spblk *)NULL, spt_databases);
+  sp_scan(iccachemarkup, NULL, (struct spblk *)NULL, spt_databases);
 }
 
 static void
@@ -264,6 +250,8 @@ run_relation(argc, argv)
 		case 'd':	/* driver routine */
 			if (STREQ(zoptarg, "pathalias.nodot"))
 				proto_config.driver = find_nodot_domain;
+			else if (STREQ(zoptarg, "pathalias.at"))
+				proto_config.driver = find_at_domain;
 			else if (STREQ(zoptarg, "pathalias"))
 				proto_config.driver = find_domain;
 			else if (STREQ(zoptarg, "longestmatch"))
@@ -493,7 +481,8 @@ icdbspltree(name)
  */
 
 static int
-iclistdbs(spl)
+iclistdbs(p, spl)
+	void *p;
 	struct spblk *spl;
 {
 	struct db_kind *dbkp;
@@ -557,6 +546,8 @@ iclistdbs(spl)
 	  printf("{pa}");  i += 4;
 	} else if (dbip->driver == find_domain) {
 	  printf("{pa.}"); i += 5;
+	} else if (dbip->driver == find_at_domain) {
+	  printf("{pa.at}"); i += 7;
 	} else if (dbip->driver == find_longest_match) {
 	  printf("{lm}");  i += 4;
 	}
@@ -586,7 +577,8 @@ iclistdbs(spl)
 static int  dbs_atexit_set;
 
 static int
-dbs_atexit_close(spl)
+dbs_atexit_close(p, spl)
+	void *p;
 	struct spblk *spl;
 {
 	struct db_info *dbip;
@@ -610,7 +602,7 @@ dbs_atexit_close(spl)
 
 static void dbs_atexit()
 {
-  sp_scan(dbs_atexit_close, (struct spblk *)NULL, spt_databases);
+  sp_scan(dbs_atexit_close, NULL, (struct spblk *)NULL, spt_databases);
 }
 
 
@@ -634,7 +626,7 @@ run_db(argc, argv)
 
 		printf("#DBname Type{lookup,sub} cache{inuse/max} ttl Flgs File/param\n");
 
-		sp_scan(iclistdbs, (struct spblk *)NULL, spt_databases);
+		sp_scan(iclistdbs, NULL, (struct spblk *)NULL, spt_databases);
 		return 0;
 	}
 
@@ -739,6 +731,7 @@ run_db(argc, argv)
 		if (dbip->postproc == Indirect) {
 			si.file = dbip->subtype;
 			(*dbip->close)(&si,"db flush indirect");
+			si.file = dbip->file;
 		}
 		break;
 	case 'o':	/* owner */
@@ -841,6 +834,9 @@ dblookup(dbname, argc, argv30)
 	} else
 	  khash = crc32(key);
 
+	/* if (!dbip->file || !*dbip->file)
+	   fprintf(stderr,"relation: '%s' has void file\n",dbname); */
+
 	memset(&si, 0, sizeof(si));
 	si.file      =  dbip->file;
 	si.cfgfile   =  dbip->cfgfile;
@@ -1012,9 +1008,7 @@ dblookup(dbname, argc, argv30)
 		if (D_db)
 			fprintf(stderr, "NIL\n");
 		if (!(DB_NEG_CACHE & dbip->flags)) {
-			if (dbip->cache_size > 0) {
-				free(realkey);
-			}
+			if (realkey) free(realkey);
 			UNGCPRO3;
 			if (si.argv0) free((void*)si.argv0);
 			si.argv0 = NULL;
@@ -1056,6 +1050,9 @@ dblookup(dbname, argc, argv30)
 		cache->value   = l;
 		if (D_db)
 			fprintf(stderr, "... added '%s' to cache", realkey);
+		/* The "realkey" went into cache. */
+		realkey = NULL;
+
 		if (si.ttl > 0) {
 			cache->expiry = now + si.ttl;
 			if (D_db)
@@ -1066,15 +1063,15 @@ dblookup(dbname, argc, argv30)
 				fprintf(stderr, "\n");
 		}
 		ll = l;
-		/* The "realkey" went into cache. */
 
 	} else {
 
 		ll = l;
-		if (realkey) free(realkey);
-
 	}
 
+	if (realkey) free(realkey);
+	realkey = NULL;
+
 	UNGCPRO3;
  post_subst:
 
@@ -1169,18 +1166,20 @@ dblookup(dbname, argc, argv30)
 
 	  malloc_failure:
 
-	    if (si.argv0) free((void*)si.argv0);
-	    si.argv0 = NULL;
-
 	    ll = l;
 
 	    UNGCPRO3;
 	  }
 
 	}
+
+	if (si.argv0) free((void*)si.argv0);
+	si.argv0 = NULL;
+
 	return ll;
 }
 
+#if 0
 const char *
 dbfile(dbname)
 	const char *dbname;
@@ -1196,6 +1195,7 @@ dbfile(dbname)
 	}
 	return dbip->file;
 }
+#endif
 
 /*
  * Flush all cache entries from a database definition.
@@ -1233,9 +1233,10 @@ cacheflush(dbip)
  * Flush everything; back to original state
  */
 
-static void	_sptdbreset __((struct spblk *spl));
+static void	_sptdbreset __((void *, struct spblk *spl));
 static void
-_sptdbreset(spl)
+_sptdbreset(p, spl)
+	void *p;
 	struct spblk *spl;
 {
 	char *av[4];
@@ -1251,7 +1252,7 @@ _sptdbreset(spl)
 void
 dbfree()
 {
-	sp_scan(_sptdbreset, (struct spblk *)NULL, spt_databases);
+	sp_scan(_sptdbreset, NULL, (struct spblk *)NULL, spt_databases);
 }
 #endif	/* MALLOC_TRACE */
 
@@ -1358,6 +1359,10 @@ find_domain(lookupfn, sip)
 		buf[0] = '.';
 		memcpy(buf + 1, realkey, keylen+1);
 		sip->key = buf;
+
+		if (sip->argv0) free((void*)sip->argv0);
+		sip->argv0 = dupnstr(realkey, keylen);
+
 		l = (*lookupfn)(sip);
 #ifndef HAVE_ALLOCA
 		free(buf);
@@ -1381,12 +1386,59 @@ find_domain(lookupfn, sip)
 	}
 
 	if (l) {
-	    if (sip->argv0) free((void*)sip->argv0);
-	    sip->argv0 = dupnstr(realkey, keylen);
-	    return l;
+	  if (sip->argv0) free((void*)sip->argv0);
+	  sip->argv0 = dupnstr(realkey, keylen);
 	}
 
-	return NULL;
+	return l;
+}
+
+/*
+ * The lookup sequence for user@ is:
+ *
+ *	user@
+ *	@LOCAL@
+ *
+ * The lookup sequence for user@foo.bar.edu is:
+ *
+ *	user@foo.bar.edu
+ *	foo.bar.edu
+ *	.foo.bar.edu
+ *	.bar.edu
+ *	.edu
+ *	.
+ *
+ * The lookup sequence for foo.bar.edu is:
+ *
+ *	foo.bar.edu
+ *	.foo.bar.edu
+ *	.bar.edu
+ *	.edu
+ *	.
+ */
+
+static conscell *
+find_at_domain(lookupfn, sip)
+	conscell *DBFUNC(lookupfn);
+	search_info *sip;
+{
+	register const char *cp;
+	conscell *l;
+
+	if ((cp = strchr(sip->key, '@')) == NULL)
+		return find_domain(lookupfn, sip);
+
+	if ((l = (*lookupfn)(sip)) != NULL)
+		return l;
+
+	if (*(cp + 1) == '\0') {
+		sip->key = sip->flags & DB_MAPTOLOWER ? "@local@" : "@LOCAL@";
+		return (*lookupfn)(sip);
+	}
+
+	sip->key = cp + 1;
+
+	return find_domain(lookupfn, sip);
 }
 
 /*
