@@ -1,7 +1,10 @@
 /*
  *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
- *	Some functions Copyright 1991-2001 Matti Aarnio.
+ *	Some functions Copyright 1991-2006 Matti Aarnio.
+ *
+ *
+ *	SRS support added by Daniel Kiper <dkiper@netspace.com.pl>.
  */
 
 /*
@@ -9,36 +12,52 @@
  * are callable from the configuration file.
  */
 
-#include "mailer.h"
-#include <stdio.h>
-#include <sys/types.h>
-#ifdef HAVE_UNISTD_H
-# include <unistd.h>
+#include "router.h"
+
+#ifndef LONG_MAX
+#if SIZEOF_LONG == 4
+# define LONG_MAX 2147483647L /* For 32 bit machine! */
+#else
+# define LONG_MAX 9223372036854775807L /* For 64 bit machine! */
+#endif
 #endif
-#include <ctype.h>
-#include <fcntl.h>
-#include <sys/file.h>			/* O_RDONLY for run_praliases() */
-#include <pwd.h>			/* for run_homedir() */
-#include <grp.h>			/* for run_grpmems() */
-#include <errno.h>
-#include <sysexits.h>
 
-#include "zmsignal.h"
-#include "zsyslog.h"
-#include "mail.h"
-#include "interpret.h"
-#include "io.h"
-#include "libz.h"
-#include "libc.h"
+#ifndef LONG_MIN
+#if SIZEOF_LONG == 4
+# define LONG_MIN -2147483648L /* For 32 bit machine! */
+#else
+# define LONG_MIN -9223372036854775808  /* For 64 bit machine! */
+#endif
+#endif
 
-#include "prototypes.h"
 
-#ifndef	_IOFBF
-#define	_IOFBF	0
-#endif	/* !_IOFBF */
+#ifdef HAVE_SRS2_H
 
-extern conscell *s_value;
+#include <srs2.h>
 
+#define VAR_SRS_ALWAYSREWRITE	"srs_alwaysrewrite"
+#define VAR_SRS_DOMAIN		"srs_domain"
+#define VAR_SRS_HASHLENGTH	"srs_hashlength"
+#define VAR_SRS_HASHMIN		"srs_hashmin"
+#define VAR_SRS_MAXAGE		"srs_maxage"
+#define VAR_SRS_NOFORWARD	"srs_noforward"
+#define VAR_SRS_NOREVERSE	"srs_noreverse"
+#define VAR_SRS_SECRETS		"srs_secrets"
+#define VAR_SRS_SEPARATOR	"srs_separator"
+
+#define MIN_SRS_HASHLENGTH	1
+#define MAX_SRS_HASHLENGTH	4096
+
+#define MIN_SRS_HASHMIN		1
+
+#define MIN_SRS_MAXAGE		1
+#define MAX_SRS_MAXAGE		4096
+
+#define LIMIT_SIZE		2
+#define LIMIT_MIN		0
+#define LIMIT_MAX		1
+
+#endif
 
 /* The builtin functions are declared and initialized here.  */
 
@@ -46,22 +65,22 @@ extern conscell *s_value;
 static int run_hostname    ARGCV;
 static int run_whataddress ARGCV;
 static int run_iserrormsg  ARGCV;
+static int run_isinteractive ARGCV;
 static int run_erraddrlog  ARGCV;
 
 static conscell *run_dblookup __((conscell *avl, conscell *il));
-extern conscell *sh_car       __((conscell *avl, conscell *il));
 static conscell *run_cadr     __((conscell *avl, conscell *il));
 static conscell *run_caddr    __((conscell *avl, conscell *il));
 static conscell *run_cadddr   __((conscell *avl, conscell *il));
 static conscell *run_listexpand   __((conscell *avl, conscell *il));
+#ifdef HAVE_SRS2_H
+static conscell *run_srs_forward   __((conscell *avl, conscell *il));
+static conscell *run_srs_reverse   __((conscell *avl, conscell *il));
+#endif
 #if 0
 static conscell *run_newattribute __((conscell *avl, conscell *il));
 #endif
 
-extern int run_doit      ARGCV;
-extern int run_stability ARGCV;
-extern int run_process   ARGCV;
-
 static int run_grpmems   ARGCV;
 #if 0 /* dead code */
 static int run_praliases ARGCV;
@@ -88,15 +107,6 @@ static int run_syslog    ARGCV;
 static int run_malcontents ARGCV;
 #endif	/* CSRIMALLOC */
 
-extern const char *traps[];
-extern int nobody;
-extern time_t time __((time_t *));
-extern int routerdirloops;
-
-#ifndef strchr
-extern char *strchr(), *strrchr();
-#endif
-
 struct shCmd fnctns[] = {
 {	"relation",	run_relation,	NULL,	NULL,	0	},
 {	DBLOOKUPNAME,	NULL,	run_dblookup,	NULL,	SH_ARGV	},
@@ -106,6 +116,7 @@ struct shCmd fnctns[] = {
 {	"untrace",	run_trace,	NULL,	NULL,	0	},
 {	"hostname",	run_hostname,	NULL,	NULL,	0	},
 {	"iserrormsg",	run_iserrormsg,	NULL,	NULL,	0	},
+{	"isinteractive",	run_isinteractive,	NULL,	NULL,	0	},
 {	"sender",	run_whataddress,NULL,	NULL,	0	},
 {	"recipient",	run_whataddress,NULL,	NULL,	0	},
 {	"erraddron",	run_erraddrlog,	NULL,	NULL,	0	},
@@ -147,6 +158,10 @@ struct shCmd fnctns[] = {
 #if	defined(XMEM) && defined(CSRIMALLOC)
 {	"malcontents",	run_malcontents,NULL,	NULL,	0	},
 #endif	/* CSRIMALLOC */
+#ifdef HAVE_SRS2_H
+{	"srs_forward",	NULL,	run_srs_forward, NULL,	SH_ARGV	},
+{	"srs_reverse",	NULL,	run_srs_reverse, NULL,	SH_ARGV	},
+#endif
 /* The rest have been added locally */
 { NULL, NULL, NULL, NULL, 0 }
 };
@@ -192,6 +207,17 @@ static struct debugind {
 	{	NULL,			0		}
 };
 
+/* add builtin router functions to list of builtin shell functions */
+void router_functions_init()
+{
+	register struct shCmd *shcmdp;
+
+	for (shcmdp = &fnctns[0]; shcmdp->name != NULL; ++shcmdp)
+		sp_install(symbol(shcmdp->name),
+			   (void*)shcmdp, 0, spt_builtins);
+}
+
+
 /* The builtin trace function. This is also used by command line debug specs */
 
 int
@@ -254,7 +280,6 @@ run_gensym(argc, argv)
 	return 0;
 }
 
-extern void free_gensym __((void));
 void
 free_gensym()
 {
@@ -327,6 +352,18 @@ run_iserrormsg(argc, argv)
 	return !isErrorMsg;
 }
 
+/*
+ * $(isinteractive dummyargs)  returns a flag 
+ */
+
+static int
+run_isinteractive(argc, argv)
+	int argc;
+	const char *argv[];
+{
+	return !isInteractive;
+}
+
 /*
  * this is like accton(), but for logging errors in addresses.
  */
@@ -562,7 +599,10 @@ run_praliases(argc, argv)
 
 	for (h = e->e_headers; h != NULL; h = h->h_next) {
 		h->h_descriptor = &aliashdr;
-		h->h_contents = hdr_scanparse(e, h, 1, 0);
+		/* NOTE: (although this is dead code)
+		   the '#' commenting in input file must be handled
+		   in THIS code, not delayed into  hdr_scanparse() ! */
+		h->h_contents = hdr_scanparse(e, h, 0);
 		h->h_stamp = hdr_type(h);
 		if (h->h_stamp == BadHeader) {
 			hdr_errprint(e, h, stderr, "alias expansion");
@@ -662,7 +702,6 @@ FILE *fp;
 	int lastfd = getdtablesize();
 	fd_set fdset;
 	struct timeval tv;
-	extern int errno;
 
 	fprintf(fp,"File handles: %d/",lastfd);
 
@@ -1186,6 +1225,13 @@ run_listexpand(avl, il)
 			++n;
 		}
 
+		/* NOTE:
+		   the '#' commenting in input file must be handled
+		   in THIS code, not delayed into  hdr_scanparse() ! */
+
+		if (zlinebuf[0] == '#')
+		  continue;
+
 /* #ifdef SENDMAIL_COMPABILITY_KLUDGE */
 		/**
 		 * Additional sendmail compability kludge ++haa
@@ -1215,7 +1261,7 @@ run_listexpand(avl, il)
 		if (c && (*s == ',' || *s == '\n'))
 		  *s = '\0';
 
-		hs.h_contents = hdr_scanparse(e, &hs, 1, 1);
+		hs.h_contents = hdr_scanparse(e, &hs, 1);
 		hs.h_stamp = hdr_type(&hs);
 
 		if (hs.h_stamp == BadHeader || hs.h_contents.a == NULL) {
@@ -1301,7 +1347,7 @@ run_listexpand(avl, il)
 		if (c && (*s == ',' || *s == '\n'))
 		  *s = '\0';
 
-		hs.h_contents = hdr_scanparse(e, &hs, 1, 1);
+		hs.h_contents = hdr_scanparse(e, &hs, 1);
 		hs.h_stamp = hdr_type(&hs);
 
 		if (hs.h_stamp == BadHeader || hs.h_contents.a == NULL) {
@@ -2020,7 +2066,9 @@ optsave(type, e)
 				sprintf(name,"_%s", fyitable[i].fyiname);
 				squirrel(e, name, fyitable[i].fyitext);
 			}
-			fprintf(stderr, "*** %s\n", fyitable[i].fyitext);
+			fprintf(stderr, "%s: *** %s\n",
+				e->e_spoolid,
+				fyitable[i].fyitext);
 			return;
 		}
 	}
@@ -2088,13 +2136,14 @@ run_822syntax(argc, argv)
 	hs.h_lines = makeToken(argv[1], strlen(argv[1]));
 	hs.h_lines->t_type = Line;
 	hs.h_lines->t_next = NULL;
-	hs.h_contents = hdr_scanparse(&es, &hs, 1, 0);
+	hs.h_contents = hdr_scanparse(&es, &hs, 0);
 	hs.h_stamp = hdr_type(&hs);
 	if (hs.h_stamp == BadHeader) {
+	    if (z_isterminal(0))
 		hdr_errprint(&es, &hs, stderr, "RFC822/976/2822");
-		return 1;
+	    return 1;
 	} else if (hs.h_contents.a == NULL)
-		return 1;
+	    return 1;
 	return 0;
 }
 
@@ -2222,3 +2271,194 @@ run_dequote(argc, argv)
 {
   return run_condquote_(argc,argv,0);
 }
+
+#ifdef HAVE_SRS2_H
+
+static void
+zmailer_srs_set_bool(srs, var_srs_name, srs_set_func)
+	srs_t *srs;
+	char *var_srs_name;
+	int (*srs_set_func)(srs_t *, srs_bool);
+{
+	char *endptr;
+	long int tmp;
+	conscell *var_srs;
+	srs_bool invalid_var_srs = TRUE;
+
+	if ((var_srs = v_find(var_srs_name)) != NULL) {
+		if (cdr(var_srs) != NULL && !LIST(cdr(var_srs))) {
+			tmp = strtol(cdr(var_srs)->string, &endptr, 10);
+			if (cdr(var_srs)->string != '\0' && *endptr == '\0' &&
+			    (tmp != LONG_MAX || errno != ERANGE) && tmp >= 0) {
+				invalid_var_srs = FALSE;
+				srs_set_func(srs, tmp ? TRUE : FALSE);
+			}
+		}
+		if (invalid_var_srs == TRUE)
+			fprintf(stderr, "SRS: invalid %s - ignored\n", var_srs_name);
+	}
+}
+
+static void
+zmailer_srs_set_int(srs, var_srs_name, limit, srs_set_func)
+	srs_t *srs;
+	char *var_srs_name;
+	long int limit[LIMIT_SIZE];
+	int (*srs_set_func)(srs_t *, int);
+{
+	char *endptr;
+	long int tmp;
+	conscell *var_srs;
+	srs_bool invalid_var_srs = TRUE;
+
+	if ((var_srs = v_find(var_srs_name)) != NULL) {
+		if (cdr(var_srs) != NULL && !LIST(cdr(var_srs))) {
+			tmp = strtol(cdr(var_srs)->string, &endptr, 10);
+			if (cdr(var_srs)->string != '\0' && *endptr == '\0' &&
+			    (tmp != LONG_MIN && tmp != LONG_MAX || errno != ERANGE) &&
+			    tmp >= limit[LIMIT_MIN] && tmp <= limit[LIMIT_MAX]) {
+				invalid_var_srs = FALSE;
+				srs_set_func(srs, (int)tmp);
+			}
+		}
+		if (invalid_var_srs == TRUE)
+			fprintf(stderr, "SRS: invalid %s - ignored\n", var_srs_name);
+	}
+}
+
+static srs_t *
+zmailer_srs_init(void)
+{
+	int ret_srs;
+	long int limit[LIMIT_SIZE] = {MIN_SRS_HASHLENGTH, MAX_SRS_HASHLENGTH};
+	conscell *var_srs;
+	srs_t *srs;
+
+	if ((srs = srs_new()) == NULL)
+		return NULL;
+
+	if ((var_srs = v_find(VAR_SRS_SECRETS)) == NULL || cdr(var_srs) == NULL
+	    || !LIST(cdr(var_srs)) || (var_srs = cadr(var_srs)) == NULL) {
+		fprintf(stderr, "SRS: invalid " VAR_SRS_SECRETS "\nSRS: SRS disabled\n");
+		srs_free(srs);
+		return NULL;
+	}
+
+	while (var_srs != NULL) {
+		if (LIST(var_srs)) {
+			fprintf(stderr, "SRS: invalid " VAR_SRS_SECRETS "\nSRS: SRS disabled\n");
+			srs_free(srs);
+			return NULL;
+		}
+		if ((ret_srs = srs_add_secret(srs, var_srs->string)) != SRS_SUCCESS) {
+			fprintf(stderr, "SRS: %s\nSRS: SRS disabled\n", srs_strerror(ret_srs));
+			srs_free(srs);
+			return NULL;
+		}
+		var_srs = cdr(var_srs);
+	}
+
+	zmailer_srs_set_int(srs, VAR_SRS_HASHLENGTH, limit, srs_set_hashlength);
+
+	return srs;
+}
+
+static conscell *
+run_srs_forward(avl, il)
+	conscell *avl, *il;
+{
+	void *tmp;
+	int ret_srs;
+	conscell *var_srs, *var_srs_domain;
+	size_t tmp_len;
+	srs_t *srs;
+
+	if ((il = cdar(avl)) == NULL)
+		return NULL;
+
+	if (LIST(il) || (srs = zmailer_srs_init()) == NULL)
+		return copycell(il);
+
+	if ((var_srs_domain = v_find(VAR_SRS_DOMAIN)) == NULL || cdr(var_srs_domain) == NULL
+	    || LIST(cdr(var_srs_domain))) {
+		fprintf(stderr, "SRS: invalid " VAR_SRS_DOMAIN "\nSRS: SRS disabled\n");
+		srs_free(srs);
+		return copycell(il);
+	}
+
+	if ((var_srs = v_find(VAR_SRS_SEPARATOR)) != NULL)
+		if (cdr(var_srs) == NULL || LIST(cdr(var_srs))
+		    || strlen(cdr(var_srs)->string) > 1
+		    || srs_set_separator(srs, (char)*cdr(var_srs)->string) != SRS_SUCCESS)
+			fprintf(stderr, "SRS: invalid " VAR_SRS_SEPARATOR " - ignored\n");
+
+	zmailer_srs_set_bool(srs, VAR_SRS_ALWAYSREWRITE, srs_set_alwaysrewrite);
+	zmailer_srs_set_bool(srs, VAR_SRS_NOFORWARD, srs_set_noforward);
+
+	tmp_len = strlen(il->string) + strlen(cdr(var_srs_domain)->string) + srs_get_hashlength(srs) + 66;
+	tmp = tmalloc(tmp_len);
+
+	ret_srs = srs_forward(srs, tmp, tmp_len, il->string, cdr(var_srs_domain)->string);
+
+	switch (ret_srs) {
+		case SRS_SUCCESS:
+			srs_free(srs);
+			tmp_len = strlen(tmp);
+			return newstring(dupnstr(tmp, tmp_len), tmp_len);
+		case SRS_ENOTREWRITTEN:
+			if (srs_get_noforward(srs) == TRUE)
+				break;
+		default:
+			fprintf(stderr, "SRS: %s\n", srs_strerror(ret_srs));
+	}
+	srs_free(srs);
+	return copycell(il);
+}
+
+static conscell *
+run_srs_reverse(avl, il)
+	conscell *avl, *il;
+{
+	void *tmp;
+	int ret_srs;
+	long int limit[LIMIT_SIZE];
+	size_t tmp_len;
+	srs_t *srs;
+
+	if ((il = cdar(avl)) == NULL)
+		return NULL;
+
+	if (LIST(il) || !SRS_IS_SRS_ADDRESS(il->string)
+	    || (srs = zmailer_srs_init()) == NULL)
+		return copycell(il);
+
+	limit[LIMIT_MIN] = MIN_SRS_HASHMIN;
+	limit[LIMIT_MAX] = srs_get_hashlength(srs);
+	zmailer_srs_set_int(srs, VAR_SRS_HASHMIN, limit, srs_set_hashmin);
+
+	limit[LIMIT_MIN] = MIN_SRS_MAXAGE;
+	limit[LIMIT_MAX] = MAX_SRS_MAXAGE;
+	zmailer_srs_set_int(srs, VAR_SRS_MAXAGE, limit, srs_set_maxage);
+
+	zmailer_srs_set_bool(srs, VAR_SRS_NOREVERSE, srs_set_noreverse);
+
+	tmp = tmalloc(tmp_len = strlen(il->string) + 1);
+
+	ret_srs = srs_reverse(srs, tmp, tmp_len, il->string);
+
+	switch (ret_srs) {
+		case SRS_SUCCESS:
+			srs_free(srs);
+			tmp_len = strlen(tmp);
+			return newstring(dupnstr(tmp, tmp_len), tmp_len);
+		case SRS_ENOTREWRITTEN:
+			if (srs_get_noreverse(srs) == TRUE)
+				break;
+		default:
+			fprintf(stderr, "SRS: %s\n", srs_strerror(ret_srs));
+	}
+	srs_free(srs);
+	return copycell(il);
+}
+
+#endif
