@@ -1,62 +1,33 @@
 /*
  *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
+ *
+ *	Header rewriting extended by Daniel Kiper <dkiper@netspace.com.pl>.
  */
 /*
  *	Some modifications  by
- *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2003
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2004
  */
 
 /*
  * This file contains most of the RFC822-specific message manipulation.
  */
 
-#include "mailer.h"
-#include <stdio.h>
-#include <sys/types.h>
-#include <sys/param.h>
-#include <fcntl.h>
-#include <sys/file.h>
-#include <errno.h>
-#include "mail.h"
-#include "libz.h"
-#ifdef HAVE_UNISTD_H
-#include <unistd.h>
-#endif
-#include "zsyslog.h"
+#include "router.h"
 
 #ifdef HAVE_SYS_UN_H
 #include <sys/socket.h>
 #include <sys/un.h>
 #endif
-
-#include "prototypes.h"
-#include "libsh.h"
-#include "zmsignal.h"
-#include "shmmib.h"
-
-#ifndef _IOFBF
-#define _IOFBF  0
-#endif  /* !_IOFBF */
-#ifndef _IOLBF
-#define _IOLBF  0200
-#endif  /* !_IOFBF */
-
-#ifdef	HAVE_LOCKF
-#ifdef	F_OK
-#undef	F_OK
-#endif	/* F_OK */
-#ifdef	X_OK
-#undef	X_OK
-#endif	/* X_OK */
-#ifdef	W_OK
-#undef	W_OK
-#endif	/* W_OK */
-#ifdef	R_OK
-#undef	R_OK
-#endif	/* R_OK */
-
-#endif	/* HAVE_LOCKF */
+#ifdef HAVE_SYS_RESOURCE_H
+#ifdef HAVE_GETRUSAGE
+#include <sys/time.h>
+#include <sys/resource.h>
+#ifdef HAVE_UNISTD_H
+#include <unistd.h>
+#endif
+#endif
+#endif
 
 static void	reject __((struct envelope *e, const char *msgfile));
 
@@ -65,6 +36,7 @@ static void	prdsndata __((conscell *info, FILE *fp, const char *comment));
 static conscell	*find_errto __((conscell *list));
 
 #define dprintf	if (D_sequencer) printf
+#define Vprintf	if (vfp) fprintf
 
 #define	QCHANNEL(x)	(x)
 #define	QHOST(x)	cdr(x)
@@ -128,6 +100,8 @@ run_rfc822(argc, argv)
 	int status = PERR_OK, errflg;
 	memtypes oval;
 	GCVARS2;
+	struct stat stbuf;
+	long infilesize_kb = 0;
 
 	errflg = 0;
 #if 0
@@ -175,16 +149,26 @@ run_rfc822(argc, argv)
 	    fflush(stderr);
 	    /* abort(); */
 	  }
+	  fstat(FILENO(e->e_fp), &stbuf);
 	  status = PERR_OK;
 	  setvbuf(e->e_fp, buf, _IOFBF, sizeof buf);
 	  e->e_file = file;
 	  status = makeLetter(e);
+
+	  if (stbuf.st_blksize == 0) stbuf.st_blksize = 1024;
+	  infilesize_kb = (stbuf.st_size + stbuf.st_blksize -1) / 1024;
+
 	}
 
 	if (status == PERR_OK)
 	  /* passing the file is redundant but nice for backtraces */
 	  status = sequencer(e, file);
 
+	if (MIBMtaEntry->rt.StoredMessages > 0)
+	  MIBMtaEntry->rt.StoredMessages        -= 1;
+	if (MIBMtaEntry->rt.StoredVolume > 0)
+	  MIBMtaEntry->rt.StoredVolume          -= infilesize_kb;
+
 	switch (status) {
 	case PERR_OK:		/* 0 */
 		break;
@@ -281,7 +265,22 @@ makeLetter(e)
 #endif	/* !HAVE_STRUCT_STAT_ST_BLKSIZE */
 
 
-	taspoolid(e->e_spoolid, e->e_statbuf.st_mtime, (long)e->e_statbuf.st_ino);
+	taspoolid(e->e_spoolid, (long)e->e_statbuf.st_ino,
+		  e->e_statbuf.st_mtime,
+#ifdef HAVE_STRUCT_STAT_ST_ATIM_TV_NSEC
+		  e->e_statbuf.st_mtim.tv_nsec
+#else
+#ifdef HAVE_STRUCT_STAT_ST_ATIM___TV_NSEC
+		  e->e_statbuf.st_mtim.__tv_nsec
+#else
+#ifdef HAVE_STRUCT_STAT_ST_ATIMENSEC
+		  e->e_statbuf.st_mtimensec
+#else
+		  0
+#endif
+#endif
+#endif
+		  );
 
 
 	inheader = 0; /* 0: envelope, 1: RFC-822 */
@@ -372,14 +371,14 @@ makeLetter(e)
 					progname, h->h_descriptor->class);
 				fwrite((char *)zlinebuf, sizeof (char), -i, stderr);
 				putc('\n', stderr);
-				h->h_contents = hdr_scanparse(e, h, 0, 0);
+				h->h_contents = hdr_scanparse(e, h, 0);
 				h->h_stamp = hdr_type(h);
 				return PERR_BADSUBMIT;
 				/* break; */
 			case eFrom:		/* pity to break the elegance */
 				osem = h->h_descriptor->semantics;
 				h->h_descriptor->semantics = Mailbox;
-				h->h_contents = hdr_scanparse(e, h, 0, 0);
+				h->h_contents = hdr_scanparse(e, h, 0);
 				h->h_stamp = hdr_type(h);
 				h->h_descriptor->semantics = osem;
 				break;
@@ -390,7 +389,7 @@ makeLetter(e)
 				e->e_eHeaders = h->h_next;
 				break;
 			default:		/* a real envelope header */
-				h->h_contents = hdr_scanparse(e, h, 0, 0);
+				h->h_contents = hdr_scanparse(e, h, 0);
 				h->h_stamp = hdr_type(h);
 				break;
 			}
@@ -417,7 +416,7 @@ makeLetter(e)
 		nh = h->h_next;
 		h->h_next = ph;
 		if (h->h_descriptor) {
-			h->h_contents = hdr_scanparse(e, h, 0, 0);
+			h->h_contents = hdr_scanparse(e, h, 0);
 			h->h_stamp = hdr_type(h);
 			if (!hdr_nilp(h))	/* excise null-valued headers */
 				ph = h;
@@ -1021,7 +1020,7 @@ mkSender(e, name, flag)
 	/* X: sprintf(buf, "%s <%s>", fullname, name); more or less */
 	sh->h_lines = makeToken(name, strlen((char *)name));
 	sh->h_lines->t_type = Line;
-	sh->h_contents = hdr_scanparse(e, sh, 0, 0);
+	sh->h_contents = hdr_scanparse(e, sh, 0);
 #endif
 	/* X: optimize: save this in a hashtable somewhere */
 	UNGCPRO1;
@@ -1234,7 +1233,7 @@ sequencer(e, file)
 	struct envelope *e;
 	const char *file;
 {
-	struct header  *h, *ph, *nh = NULL, *oh, *msgidh, **hp;
+	struct header  *h, *ph, *nh = NULL, *oh, *rh, *msgidh, **hp;
 	struct header  *rcvdhdr;
 	struct address *a, *ap;
 	struct addr    *p = NULL;
@@ -1263,9 +1262,57 @@ sequencer(e, file)
 	struct stat stbuf;
 	long infilesize_kb, taskfilesize_kb;
 	GCVARS5;
+	double worktimeu, worktimes;
+
+
+	worktimeu = worktimes = 0.0l;
+#if defined(RUSAGE_SELF) && defined(HAVE_GETRUSAGE)
+	{
+	  struct rusage rr;
+	  if (getrusage(RUSAGE_SELF, &rr) == 0) {
+	    worktimeu = ( ((double)rr.ru_utime.tv_sec) +
+			  ((double)rr.ru_utime.tv_usec) * 0.000001l );
+	    worktimes = ( ((double)rr.ru_stime.tv_sec) +
+			  ((double)rr.ru_stime.tv_usec) * 0.000001l );
+	  }
+	}
+#endif
+
 
 	time(&start_now);
 
+	if (e == NULL) {
+	  /* No envelope ??? */
+	  return PERR_OK;
+	}
+
+	FindEnvelope(eVerbose);
+	if (h  &&  h->h_contents.a  &&  h->h_contents.a->a_tokens) {
+	  if (h->h_contents.a->a_tokens->p_tokens    &&
+	      h->h_contents.a->a_tokens->p_tokens->t_type == String)
+	    h->h_contents.a->a_tokens->p_tokens->t_type = Atom;
+	  printToken(verbosefile, verbosefile + sizeof verbosefile,
+		     h->h_contents.a->a_tokens->p_tokens,
+		     (token822 *)NULL, 0);
+	  /*
+	   * We have to be careful how we open this file, since one
+	   * might imagine someone trying to append to /etc/passwd
+	   * using this stuff.  The only safe way is to open it
+	   * with the permissions of the owner of the message file.
+	   */
+	  SETEUID(e->e_statbuf.st_uid);
+
+	  vfp = fopen(verbosefile, "a");
+	  if (vfp) {
+	    fseek(vfp, (off_t)0, 2);
+	    setvbuf(vfp, NULL, _IOLBF, 0);
+	    fprintf(vfp, "router processing message %s\n", file);
+	  }
+	  SETEUID(0);
+	} else
+	  vfp = NULL;
+
+
 	if (schedulersubdirhash < 0) {
 	  const char *s = getzenv("SCHEDULERDIRHASH");
 	  if (s && ((s[0] == '1' || s[0] == '2') && s[1] == 0))
@@ -1299,17 +1346,17 @@ sequencer(e, file)
 
 	errors_to = NULL; /* to be gotten rid off.. */
 
-	if (e == NULL) {
-		/* No envelope ??? */
-		return PERR_OK;
-	}
-
 	dprintf("Sender authentication\n");
+	Vprintf(vfp,"Sender authentication\n");
 	e->e_trusted = isgoodguy(e->e_statbuf.st_uid);
 	dprintf("Sender is%s trusted\n", e->e_trusted ? "" : "n't");
+	Vprintf(vfp,"Sender is%s trusted\n", e->e_trusted ? "" : "n't");
 
-	if (deferuid)
+	if (deferuid) {
+	  Vprintf(vfp,"DEFERRING DUE TO FILE OWNER PROBLEMS\n");
+	  if (vfp) fclose(vfp);
 	  return PERR_DEFERRED;
+	}
 
 	FindEnvelopeLast(eRcvdFrom);
 	rcvdhdr = h;
@@ -1333,6 +1380,7 @@ sequencer(e, file)
 	       that the message writer coded in... */
 
 	    dprintf("Message has 'rcvdfrom' envelope header, but we don't trust it!\n");
+	    Vprintf(vfp, "Message has 'rcvdfrom' envelope header, but we don't trust it!\n");
 	    s = uidpwnam(e->e_statbuf.st_uid);
 	    ps = "";
 	    totlen = 10 + strlen(s) + 60;
@@ -1365,14 +1413,17 @@ sequencer(e, file)
 
 	  rcvdhdr->h_lines    = makeToken(ts+9, strlen(ts+9));
 	  rcvdhdr->h_lines->t_type = Line;
-	  rcvdhdr->h_contents = hdr_scanparse(e, rcvdhdr, 0, 0);
+	  rcvdhdr->h_contents = hdr_scanparse(e, rcvdhdr, 0);
 	  rcvdhdr->h_stamp    = hdr_type(rcvdhdr);
 
 	  smtprelay     = ts+9;
 	}
 
-	if (deferuid)
+	if (deferuid) {
+	  Vprintf(vfp,"DEFERRING due to unspecified reason\n");
+	  if (vfp) fclose(vfp);
 	  return PERR_DEFERRED;
+	}
 
 	dprintf("Parse envelope and message header\n");
 	e->e_messageid = NULL;
@@ -1401,8 +1452,11 @@ sequencer(e, file)
 		}
 	}
 
-	if (deferuid)
+	if (deferuid) {
+	  Vprintf(vfp,"DEFERRING due to unspecified reason\n");
+	  if (vfp) fclose(vfp);
 	  return PERR_DEFERRED;
+	}
 
 	dprintf("Determine if message is a Resent-* type thing\n");
 	e->e_resent = 0;
@@ -1450,8 +1504,12 @@ sequencer(e, file)
 		  e->e_messageid = saveAddress(msgidh->h_contents.a->a_tokens);
 	}
 
-	if (perr)
-		return perr;
+	if (perr) {
+	  Vprintf(vfp,"ERRORING OUT with code %s\n",
+		  (perr == PERR_LOOP ? "PERR_LOOP" : "PERR_ENVELOPE"));
+	  if (vfp) fclose(vfp);
+	  return perr;
+	}
 
 	/* put pertinent message information in the zsh environment */
 	setenvinfo(e);
@@ -1497,6 +1555,8 @@ sequencer(e, file)
 
 	if (deferuid) {
 	  UNGCPRO5;
+	  Vprintf(vfp,"DEFERRING due to unspecified reason\n");
+	  if (vfp) fclose(vfp);
 	  return PERR_DEFERRED;
 	}
 
@@ -1645,7 +1705,7 @@ sequencer(e, file)
 			    h = makeHeader(spt_eheaders, ss, 4);
 			    h->h_lines = makeToken(ss+5,2);
 			    h->h_lines->t_type = Line;
-			    h->h_contents = hdr_scanparse(e, h, 0, 0);
+			    h->h_contents = hdr_scanparse(e, h, 0);
 			    /* And now we bend the result beyond any
 			       recognition.. */
 			    h->h_contents.a->a_tokens->p_type     =  anAddress;
@@ -1753,6 +1813,8 @@ sequencer(e, file)
 
 	if (deferuid) {
 	  UNGCPRO5;
+	  Vprintf(vfp,"DEFERRING due to unspecified reason\n");
+	  if (vfp) fclose(vfp);
 	  return PERR_DEFERRED;
 	}
 
@@ -1760,9 +1822,12 @@ sequencer(e, file)
 	FindEnvelope(eTo);
 	if (h == NULL) {
 		dprintf("No recipient(s) specified in the envelope\n");
+		Vprintf(vfp,"No recipient(s) specified in the envelope\n");
 		if (header_error) {
 			dprintf("Due to header error, we ignore message\n");
+			Vprintf(vfp,"Due to header error, we ignore message\n");
 			UNGCPRO5;
+			if (vfp) fclose(vfp);
 			return PERR_HEADER;
 		}
 		ph = e->e_eHeaders;
@@ -1786,6 +1851,8 @@ sequencer(e, file)
 		dprintf("Are we supposed to be psychic?\n");
 		if (ph == e->e_eHeaders) {
 		  UNGCPRO5;
+		  Vprintf(vfp,"Bailing out - no recipients\n");
+		  if (vfp) fclose(vfp);
 		  return PERR_NORECIPIENTS;
 		}
 	}
@@ -1809,10 +1876,10 @@ sequencer(e, file)
 	}
 
 	/*
-	 * Log the message after we find the envelope From address, otherwise
-	 * log entries might have empty sender fields.
+	 * Log the message after we find the envelope From address,
+	 * otherwise log entries might have empty sender fields.
 	 */
-	if (e->e_messageid)
+	if (e->e_messageid && !no_logmessage)
 		logmessage(e);
 
 	if (header_error) {
@@ -1848,7 +1915,7 @@ sequencer(e, file)
 			    nh = makeHeader(spt_headers, ss, 4);
 			    nh->h_lines = makeToken(ss+5,3);
 			    nh->h_lines->t_type = Line;
-			    nh->h_contents = hdr_scanparse(e, nh, 0, 0);
+			    nh->h_contents = hdr_scanparse(e, nh, 0);
 			    nh->h_stamp = hdr_type(nh);
 			    FindHeader("to",e->e_resent);
 			    if (h == NULL) {
@@ -1982,6 +2049,8 @@ sequencer(e, file)
 		}
 		if (deferuid) {
 		  UNGCPRO5;
+		  Vprintf(vfp,"DEFERRING due to unspecified reason\n");
+		  if (vfp) fclose(vfp);
 		  return PERR_DEFERRED;
 		}
 	}
@@ -1997,7 +2066,9 @@ sequencer(e, file)
 
 	if (routed_addresses == NULL)	/* they were probably all deferred */ {
 		printf("No routed addresses -> deferred\n");
+		Vprintf(vfp,"No routed addresses -> deferred\n");
 		UNGCPRO5;
+		if (vfp) fclose(vfp);
 		return PERR_DEFERRED;
 	}
 
@@ -2125,6 +2196,7 @@ sequencer(e, file)
 
 	if (deferuid) {
 	  UNGCPRO5;
+	  if (vfp) fclose(vfp);
 	  return PERR_DEFERRED;
 	}
 
@@ -2142,8 +2214,11 @@ sequencer(e, file)
 			isSenderAddr = h->h_descriptor->user_type == Sender;
 			isRecpntAddr = h->h_descriptor->user_type == Recipient;
 			if (isSenderAddr || isRecpntAddr) {
-				if (!STRING(rwp->info))	/* just addresses */
+				if (LIST(rwp->info)) {
+					if (car(rwp->info) == NULL)
 					continue;
+					*hp = hdr_rewrite(car(rwp->info)->string, h);
+				} else
 				*hp = hdr_rewrite(rwp->info->string, h);
 				hp = &(*hp)->h_next;
 			}
@@ -2153,22 +2228,27 @@ sequencer(e, file)
  
 	if (deferuid) {
 	  UNGCPRO5;
+	  if (vfp) fclose(vfp);
 	  return PERR_DEFERRED;
 	}
 
 	if (rwhead == NULL) {
 	  UNGCPRO5;
+	  if (vfp) fclose(vfp);
 	  return PERR_NORECIPIENTS;
 	}
 
 	dprintf("Emit specification to the transport system\n");
-#ifdef	USE_ALLOCA
-	ofpname = (char *)alloca(30);
+
+#ifndef USE_ALLOCA
+	ofpname = (char*)emalloc(16+strlen(TRANSPORTDIR)+strlen(pfile));
 #else
-	ofpname = (char *)emalloc(30);
+	/* This actually reallocs more space from stack, but then it
+	   is just stack space and will disappear.. */
+	ofpname = (char*)alloca(16+strlen(TRANSPORTDIR)+strlen(pfile));
 #endif
-	/* Guaranteed unique within this machine */
-	sprintf(ofpname,".%ld.%d", (long)time(NULL), (int)getpid());
+	sprintf(ofpname, "../%s/%s.%s", TRANSPORTDIR, subdirhash, pfile);
+
 	ofp = fopen(ofpname, "w+");
 	if (ofp == NULL) {
 #ifndef USE_ALLOCA
@@ -2176,6 +2256,7 @@ sequencer(e, file)
 #endif
 		printf("Creation of control file failed\n");
 		UNGCPRO5;
+		if (vfp) fclose(vfp);
 		return PERR_CTRLFILE;
 	}
 	setvbuf(ofp, vbuf, _IOFBF, sizeof vbuf);
@@ -2187,33 +2268,9 @@ sequencer(e, file)
 	fprintf(ofp, "%c%c0x%08lx\n",
 		_CF_FORMAT, _CFTAG_NORMAL, (long)_CF_FORMAT_KNOWN_SET);
 
-	FindEnvelope(eVerbose);
-	if (h  &&  h->h_contents.a  &&  h->h_contents.a->a_tokens) {
-		if (h->h_contents.a->a_tokens->p_tokens    &&
-		    h->h_contents.a->a_tokens->p_tokens->t_type == String)
-			h->h_contents.a->a_tokens->p_tokens->t_type = Atom;
-		printToken(verbosefile, verbosefile + sizeof verbosefile,
-			   h->h_contents.a->a_tokens->p_tokens,
-			   (token822 *)NULL, 0);
-		/*
-		 * We have to be careful how we open this file, since one
-		 * might imagine someone trying to append to /etc/passwd
-		 * using this stuff.  The only safe way is to open it
-		 * with the permissions of the owner of the message file.
-		 */
-		SETEUID(e->e_statbuf.st_uid);
-
-		vfp = fopen(verbosefile, "a");
-		if (vfp) {
-			fseek(vfp, (off_t)0, 2);
-			setvbuf(vfp, NULL, _IOLBF, 0);
-			fprintf(vfp, "router processed message %s\n", file);
-			fprintf(ofp, "%c%c%s\n", _CF_VERBOSE, _CFTAG_NORMAL,
-				verbosefile);
-		}
-		SETEUID(0);
-	} else
-		vfp = NULL;
+	if (vfp)
+	  fprintf(ofp, "%c%c%s\n", _CF_VERBOSE, _CFTAG_NORMAL,
+		  verbosefile);
 
 	fprintf(ofp, "%c%c%s\n",
 		_CF_MESSAGEID, _CFTAG_NORMAL, pfile);
@@ -2404,9 +2461,27 @@ sequencer(e, file)
 		putc(_CF_MSGHEADERS, ofp);
 		putc('\n', ofp);
 		if (vfp)
+			if (LIST(rwp->info)) {
+				if (car(rwp->info) != NULL)
+					if (cdar(rwp->info) != NULL)
+						fprintf(vfp, "headers rewritten using '%s' and '%s' function:\n",
+							car(rwp->info)->string, cdar(rwp->info)->string);
+					else
+						fprintf(vfp, "headers rewritten using '%s' function:\n",
+							car(rwp->info)->string);
+			} else
 		  fprintf(vfp, "headers rewritten using '%s' function:\n",
 			  rwp->info->string);
 
+		if (LIST(rwp->info) && car(rwp->info) != NULL && cdar(rwp->info) != NULL) {
+			rh = header_rewrite(cdar(rwp->info)->string, NULL, NULL, HRR_AB);
+			for (; rh != NULL; rh = rh->h_next) {
+				hdr_print(rh, ofp);
+				if (vfp)
+					hdr_print(rh, vfp);
+			}
+		}
+
 		/* print the header, replacing all To:, Cc:, fields with
 		   the corresponding fields as stored with the rewrite set. */
 		nh = rwp->urw.h;
@@ -2423,9 +2498,26 @@ sequencer(e, file)
 				  hdr_print(nh, vfp);
 				nh = nh->h_next;
 			} else {
-				hdr_print(h, ofp);
+				if (LIST(rwp->info) && car(rwp->info) != NULL && cdar(rwp->info) != NULL) {
+					if ((rh = header_rewrite(cdar(rwp->info)->string, h, ofp, HRR_RR)) == NULL)
+						continue;
+				} else
+					rh = h;
+				for (; rh != NULL; rh = rh->h_next) {
+					hdr_print(rh, ofp);
+					if (vfp)
+						hdr_print(rh, vfp);
+					if (rh == h)
+						break;
+				}
+			}
+		}
+		if (LIST(rwp->info) && car(rwp->info) != NULL && cdar(rwp->info) != NULL) {
+			rh = header_rewrite(cdar(rwp->info)->string, NULL, NULL, HRR_AE);
+			for (; rh != NULL; rh = rh->h_next) {
+				hdr_print(rh, ofp);
 				if (vfp)
-				  hdr_print(h, vfp);
+					hdr_print(rh, vfp);
 			}
 		}
 		putc('\n', ofp);
@@ -2477,6 +2569,18 @@ sequencer(e, file)
 	}
 
 
+	fflush(ofp);
+	{
+	  long l1, l2, l3, l4;
+	  fd_statfs(FILENO(ofp), &l1, &l2, &l3, &l4);
+
+	  MIBMtaEntry->sys.TportSpoolFreeSpace = l1;
+	  MIBMtaEntry->sys.TportSpoolUsedSpace = l2;
+	  MIBMtaEntry->sys.TportSpoolFreeFiles = l3;
+	  MIBMtaEntry->sys.TportSpoolUsedFiles = l4;
+	}
+	
+
 	if ((fclose(ofp) != 0) || ofperrors ||
 	    (erename(file, qpath) != 0)) { /* ORIGINAL FILE PATH! */
 	  zunlink(qpath);
@@ -2485,6 +2589,7 @@ sequencer(e, file)
 	  free(ofpname);
 	  free(qpath);
 #endif
+	  if (vfp) fclose(vfp);
 	  return PERR_CTRLFILE;
 	}
 
@@ -2507,6 +2612,7 @@ sequencer(e, file)
 	  free(qpath);
 	  free(path);
 #endif
+
 	  return PERR_CTRLFILE;
 	}
 
@@ -2521,11 +2627,29 @@ sequencer(e, file)
 		}
 #endif
 		fclose(vfp);
+		vfp = NULL;
 	}
 
+#if defined(RUSAGE_SELF) && defined(HAVE_GETRUSAGE)
+	{
+	  struct rusage rr;
+	  if (worktimeu != 0.0l && worktimes != 0.0l &&
+	      getrusage(RUSAGE_SELF, &rr) == 0) {
+	    worktimeu = ( ((double)rr.ru_utime.tv_sec) +
+			  ((double)rr.ru_utime.tv_usec) * 0.000001l
+			  - worktimeu );
+	    worktimes = ( ((double)rr.ru_stime.tv_sec) +
+			  ((double)rr.ru_stime.tv_usec) * 0.000001l
+			  - worktimes );
+	  } else {
+	    worktimeu = worktimes = 0.0l;
+	  }
+	}
+#endif
+
 	rtsyslog(e->e_spoolid, e->e_statbuf.st_mtime,
 		 fromaddr, smtprelay, (int) e->e_statbuf.st_size,
-		 onrcpts, msgidstr, start_now);
+		 onrcpts, msgidstr, start_now, worktimeu, worktimes);
 
 	MIBMtaEntry->rt.TransmittedMessages   += 1;
 	MIBMtaEntry->rt.ReceivedRecipients    += inrcpts;
@@ -2534,10 +2658,6 @@ sequencer(e, file)
 	MIBMtaEntry->rt.TransmittedVolume     += infilesize_kb;
 	MIBMtaEntry->rt.TransmittedVolume2    += taskfilesize_kb;
 
-	MIBMtaEntry->rt.StoredMessages        -= 1;
-	MIBMtaEntry->rt.StoredVolume          -= infilesize_kb;
-
-
 #ifdef AF_UNIX
 	do {
 
