@@ -7,27 +7,11 @@
  * This file implements most of an RFC822/976 scanner/parser.
  */
 
-#include "mailer.h"
-#include <stdio.h>
-#include <sys/types.h>
-#include <ctype.h>
-#ifdef HAVE_UNISTD_H
-#include <unistd.h>
-#endif
-#include "io.h"
-#include "prototypes.h"
-
-#ifndef _IOFBF
-#define _IOFBF  0
-#endif  /* !_IOFBF */
-
+#include "router.h"
 
 struct headerinfo nullhdr = { 0, nilHeaderSemantics, nilUserType, normal };
 int D_rfc822 = 0; /* Debug traceing */
 
-extern int do_hdr_warning; /* If set, headers with errors in them are
-			      printed "as is" -- h_line */
-
 /*
  * Apply RFC822 scanner and parser to the list of Line tokens passed as args.
  *
@@ -36,10 +20,10 @@ extern int do_hdr_warning; /* If set, headers with errors in them are
  */
 
 union misc
-hdr_scanparse(e, h, commentflag, no_line_crossing)
+hdr_scanparse(e, h, no_line_crossing)
 	struct envelope *e;
 	struct header *h;
-	int commentflag, no_line_crossing;
+	int no_line_crossing;
 {
 	register token822 *t;
 	long		 len;
@@ -136,7 +120,7 @@ hdr_scanparse(e, h, commentflag, no_line_crossing)
 		while (len > 0) {
 			ocp = cp;
 			nt = t;
-			scan_t = scan822(&cp, len, c1, c2, commentflag, &nt);
+			scan_t = scan822(&cp, len, c1, c2, &nt);
 			if (nt != t && !no_line_crossing) {
 				/* compound token across line */
 
@@ -377,7 +361,7 @@ mkMessageId(e, unixtime)
 #endif
 	h->h_lines = makeToken(buf, strlen(buf));
 	h->h_lines->t_type = Line;
-	h->h_contents = hdr_scanparse(e, h, 0, 0);
+	h->h_contents = hdr_scanparse(e, h, 0);
 	/* OPTIMIZE: save this in a hashtable somewhere */
 	return h;
 }
@@ -404,7 +388,7 @@ mkToHeader(e, buf)
 	h->h_next = 0;
 	h->h_lines = makeToken(buf, strlen(buf));
 	h->h_lines->t_type = Line;
-	h->h_contents = hdr_scanparse(e, h, 0, 0);
+	h->h_contents = hdr_scanparse(e, h, 0);
 	/* OPTIMIZE: save this in a hashtable somewhere */
 	return h;
 }
