@@ -4,7 +4,7 @@
  */
 /*
  *	Lots of modifications (new guts, more or less..) by
- *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2003
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2004
  */
 
 
@@ -12,54 +12,81 @@
  * ZMailer router, main and miscellany routines.
  */
 
-#include "mailer.h"
-#include <grp.h>
-#include <pwd.h>
-#include <signal.h>
-#include <ctype.h>
-#include <fcntl.h>
-#include <sys/file.h>
-#include "mail.h"
-#include "zsyslog.h"
-#include "zmsignal.h"
-#include "interpret.h"
-#include "splay.h"
-
-#include "prototypes.h"
-
-extern const char *postoffice; /* At libzmailer.a: mail.c */
-
-extern struct shCmd fnctns[];
-extern time_t time __((time_t *));
+#define NO_IO_H
+#include "router.h"
 
 static void initialize __((const char *configfile, int argc, const char *argv[]));
-static void logit __((const char *file, const char *id, const char *from, const char *to));
+static void logit __((const char *id, const char *from, const char *to));
+static void Imode_smtpserver __((void));
 
 
-extern const char *progname; /* At zmsh_init() et.friends */
 const char * mailshare;
-const char * myhostname = 0;
+const char * myhostname;
 const char * pidfile = PID_ROUTER;
 const char * logfn;
 time_t	now;
 
-extern memtypes stickymem;
-extern conscell *s_value;
-
-#include "shmmib.h"
-
-int	mustexit = 0;
-int	canexit = 0;
-int	router_id = 0;
+int	mustexit;
+int	canexit;
+int	router_id;
 int	deferit;
 int	deferuid;
-int	savefile = 0;
+int	savefile;
 const char * zshopts = "-O";
 int	nosyslog = 1;
-int	routerdirloops = 0;
-int	do_hdr_warning = 0;
-int	workermode = 0;
+int	routerdirloops;
+int	do_hdr_warning;
+int	workermode;
 int	nrouters = 1;
+int	isInteractive;
+int	no_logmessage;
+
+#define IMODE_NONE 0
+#define IMODE_SMTPSERVER 1
+
+int	I_mode = IMODE_NONE;
+
+
+struct itimerval profiler_itimer_at_start;
+
+
+#if	defined(SA_NOCLDSTOP)||defined(SA_ONSTACK)||defined(SA_RESTART)
+	/* ================ POSIX.1 STUFF ================ */
+
+#if 0
+void sigact_segv(sig, info, data)
+     int sig;
+     siginfo_t *info;
+     void *data;
+{
+  printf("SEGV: signo=%d errno=%d code=%d pid=%d uid=%d status=%x utime=%ld stime=%ld value=%p (int=%d ptr=%p) addr=%p band=%ld fd=%d\n",
+	 info->si_signo, info->si_errno, info->si_code,
+	 info->si_pid,   info->si_uid,   info->si_status,
+	 info->si_utime, info->si_stime, info->si_value.sival_ptr,
+	 info->si_int,   info->si_ptr,   info->si_addr,
+	 info->si_band,  info->si_fd);
+  fflush(stdout);
+  exit(SIGSEGV);
+}
+
+
+void sigact_bus(sig, info, data)
+     int sig;
+     siginfo_t *info;
+     void *data;
+{
+  printf("BUS: signo=%d errno=%d code=%d pid=%d uid=%d status=%x utime=%ld stime=%ld value=%p (int=%d ptr=%p) addr=%p band=%ld fd=%d\n",
+	 info->si_signo, info->si_errno, info->si_code,
+	 info->si_pid,   info->si_uid,   info->si_status,
+	 info->si_utime, info->si_stime, info->si_value.sival_ptr,
+	 info->si_int,   info->si_ptr,   info->si_addr,
+	 info->si_band,  info->si_fd);
+  fflush(stdout);
+  exit(SIGBUS);
+}
+#endif
+#endif
+
 
 int
 main(argc, argv)
@@ -75,6 +102,8 @@ main(argc, argv)
 	FILE *fp;
 #endif	/* XMEM */
 
+	getitimer(ITIMER_PROF, & profiler_itimer_at_start);
+
 #ifdef HAVE_SETGROUPS
 	/* We null supplementary groups list entirely */
 	setgroups(0, NULL);
@@ -105,7 +134,7 @@ main(argc, argv)
 
 
 	while (1) {
-		c = zgetopt(argc, (char*const*)argv, "m:n:dikf:o:t:L:P:r:sSVwWZ:");
+		c = zgetopt(argc, (char*const*)argv, "m:n:diI:kf:o:t:lL:P:r:sSVwWZ:");
 		if (c == EOF)
 			break;
 	  
@@ -153,6 +182,12 @@ main(argc, argv)
 		case 'i':	/* first read config file, then read from tty */
 			interactiveflg = 1;
 			break;
+		case 'I':
+			if (strcmp(zoptarg,"smtpserver") == 0)
+			  I_mode = IMODE_SMTPSERVER;
+			else
+			  ++errflg;
+			break;
 		case 'k':	/* kill the previous daemon upon startup */
 			killflg = 1;
 			break;
@@ -166,6 +201,9 @@ main(argc, argv)
 		case 'L':	/* override default log file */
 			logfn = zoptarg;
 			break;
+		case 'l':
+			no_logmessage = 1;
+			break;
 		case 'P':	/* override default postoffice */
 			postoffice = zoptarg;
 			break;
@@ -196,7 +234,7 @@ main(argc, argv)
 
 	if (errflg || (interactiveflg && daemonflg)) {
 		fprintf(stderr,
-			"Usage: %s [ -dikV -n #routers -t traceflag -f configfile -L logfile -P postoffice -Z zenvfile]\n",
+			"Usage: %s [ -dikVl -n #routers -t traceflag -f configfile -L logfile -P postoffice -Z zenvfile]\n",
 			progname);
 		exit(128+errflg);
 	}
@@ -299,7 +337,10 @@ main(argc, argv)
 
 	/* We (and our children) run with SIGPIPE ignored.. */
 	SIGNAL_HANDLE(SIGPIPE, SIG_IGN);
-
+#if 0
+	SIGNAL_ACTION(SIGSEGV, sigact_segv);
+	SIGNAL_ACTION(SIGBUS,  sigact_bus);
+#endif
 
 	initialize(config, argc - c, &argv[c]);
 
@@ -311,8 +352,6 @@ main(argc, argv)
 	mal_leaktrace(1);
 #endif /* MALLOC_TRACE */
 
-
-
 	if (daemonflg) {
 		if (chdir(postoffice) < 0 || chdir(ROUTERDIR) < 0)
 		  fprintf(stderr, "%s: cannot chdir.\n", progname);
@@ -320,13 +359,16 @@ main(argc, argv)
 		if (offout < ftell(stdout) || offerr < ftell(stderr)) {
 		  fprintf(stderr, "%d %d %d %d\n",
 			  (int) offout, (int) ftell(stdout),
-			  (int) offerr, (int) ftell(stderr));
+			  (int) offerr, (int) ftell(stderr) );
 		  fprintf(stderr, "%s: daemon not started.\n", progname);
 		  die(1, "errors during startup");
 		}
 
-		if (tac == 0)		/* leave worldy matters behind */
+		if (tac == 0) {		/* leave worldy matters behind */
 		  detach();
+		  close(0); open("/dev/null", O_RDONLY, 0);
+		  setitimer(ITIMER_PROF, &  profiler_itimer_at_start, NULL);
+		}
 		printf("%s: router daemon (%s)\n\tstarted at %s\n",
 		       progname, Version, rfc822date(&now));
 		if (killflg)
@@ -377,12 +419,17 @@ write(30, "\n", 1);
 	  run_daemon(1, &av[0]);
 	  /* NOTREACHED */
 	} else if (interactiveflg) {
+
+	  if (I_mode == IMODE_SMTPSERVER) {
+	    Imode_smtpserver();
+	  } else {
 #ifdef	MALLOC_TRACE
-	  zshtoplevel(NULL);
+	    zshtoplevel(NULL);
 #else	/* !MALLOC_TRACE */
-	  trapexit(zshtoplevel(NULL));
+	    trapexit(zshtoplevel(NULL));
 #endif	/* MALLOC_TRACE */
-	  /* NOTREACHED */
+	    /* NOTREACHED */
+	  }
 	}
 #ifdef	MALLOC_TRACE
 	dbfree();
@@ -427,13 +474,7 @@ initialize(configfile, argc, argv)
 	zshinit(ac, av);
 
 	/* add builtin router functions to list of builtin shell functions */
-	{
-		register struct shCmd *shcmdp;
-
-		for (shcmdp = &fnctns[0]; shcmdp->name != NULL; ++shcmdp)
-			sp_install(symbol(shcmdp->name),
-				   (void*)shcmdp, 0, spt_builtins);
-	}
+	router_functions_init();
 
 	/* initialize splay trees in router */
 	av[0] = "relation";
@@ -616,8 +657,7 @@ logmessage(e)
 					/* print what we've got so far */
 					if (to != buf) {
 						*to = '\0';
-						logit(e->e_file,
-						      e->e_messageid,
+						logit(e->e_spoolid,
 						      from, buf);
 						to = buf;
 					}
@@ -637,7 +677,7 @@ logmessage(e)
 	}
 	if (to != buf) {
 		*to = '\0';
-		logit(e->e_file, e->e_messageid, from, buf);
+		logit(e->e_spoolid, from, buf);
 	}
 }
 
@@ -649,25 +689,25 @@ logmessage(e)
  */
 
 static void
-logit(file, id, from, to)
-	const char *file, *id, *from, *to;
+logit(id, from, to)
+	const char *id, *from, *to;
 {
 	int flen, baselen;
 
 	if (id == NULL)
-		id = file;
-	baselen = strlen(file) + strlen(id) + 4;
+		id = "<>";
+	baselen = strlen(id) + 4;
 	flen = strlen(from);
 	if (flen > 0)
-	  printf("%.200s: file: %s %s => %s\n", id, file, from, to);
+	  printf("%.30s: fromto: %s => %s\n", id, from, to);
 	else
-	  printf("%.200s: file: %s %s\n",       id, file, to);
+	  printf("%.30s: fromto: <> => %s\n",       id, to);
 	if (!nosyslog) {
 	  if (flen > 0)
-	    zsyslog((LOG_INFO, "%.200s: file: %.150s %.200s => %.200s",
-		     id, file, from, to));
+	    zsyslog((LOG_INFO, "%.30s: fromto: %.200s => %.200s",
+		     id, from, to));
 	  else
-	    zsyslog((LOG_INFO, "%.200s: file: %.150s %.200s", id, file, to));
+	    zsyslog((LOG_INFO, "%.30s: fromto: <> => %.200s", id, to));
 	}
 }
 
@@ -703,3 +743,83 @@ printfds()
 	fprintf(stderr,"\n");
 	fflush(stderr);
 }
+
+
+void
+Imode_smtpserver __((void))
+{
+	int n, c;
+	const char *key;
+	const char *data;
+	memtypes oval;
+	const char *av[4];
+	char *linebuf = malloc(20000);
+	int linespc = 19500;
+	int cmdcount = 0;
+	char *sh_memlevel = getlevel(MEM_SHCMD);
+
+	oval = stickymem;
+	stickymem = MEM_TEMP;	/* per-message space */
+
+	setlinebuf(stdin);
+
+	SIGNAL_HANDLE(SIGTERM, sig_exit);	/* mustexit = 1 */
+
+	isInteractive = 1;
+
+	av[0] = "server";
+	av[1] = "init";
+	av[2] = NULL;
+	s_apply(2, &av[0]);
+
+	while ( !mustexit ) {
+
+	  ++cmdcount;
+	  if (cmdcount > 1000) {
+	    av[0] = "server";
+	    av[1] = "purge";
+	    av[2] = NULL;
+	    s_apply(2, &av[0]);
+	    cmdcount = 0;
+	  }
+
+	  fprintf(stdout, "#hungry\n");
+	  fflush(stdout);
+
+	  n = 0;
+	  while (( c = fgetc(stdin) ) != EOF) {
+	    if (n + 10  > linespc) {
+	      linespc += 2000;
+	      linebuf = realloc(linebuf, linespc);
+	    }
+	    linebuf[n++] = c;
+	    if (c == '\n') break;
+	  }
+	  linebuf[n] = 0;
+
+	  if (n <= 0) break;
+
+	  /* process!   linebuf[]  has input data.. */
+
+	  key  = strtok(linebuf, "\t");
+	  data = strtok(NULL, "\n");
+
+	  deferit = 0;
+	  v_set(DEFER, "");
+
+	  gensym = 1;
+
+	  av[0] = "server";
+	  av[1] = key;
+	  av[2] = data;
+	  av[3] = NULL;
+
+	  s_apply(3, &av[0]); /* "server" key argument */
+
+	  free_gensym();
+	  setlevel(MEM_SHCMD,sh_memlevel);
+	  
+	  fflush(stdout);
+	}
+	stickymem = oval;
+}
