@@ -1,15 +1,35 @@
 /*
  *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
+ *
+ *	Header rewriting extended by Daniel Kiper <dkiper@netspace.com.pl>.
  */
 
-#include "hostenv.h"
-#include "mailer.h"
+#include "router.h"
+#include "vis.h"
 
-#include "prototypes.h"
+#ifndef LONG_MAX
+#if SIZEOF_LONG == 4
+# define LONG_MAX 2147483647L /* For 32 bit machine! */
+#else
+# define LONG_MAX 9223372036854775807L /* For 64 bit machine! */
+#endif
+#endif
 
-conscell **return_valuep = NULL;
-conscell *s_value        = NULL;
+#ifndef LONG_MIN
+#if SIZEOF_LONG == 4
+# define LONG_MIN -2147483648L /* For 32 bit machine! */
+#else
+# define LONG_MIN -9223372036854775808  /* For 64 bit machine! */
+#endif
+#endif
+
+
+#define HRR_VIS_DISABLE		"hrr_vis_disable"
+#define HRR_UNVIS_DISABLE	"hrr_unvis_disable"
+
+conscell **return_valuep;
+conscell *s_value;
 
 int
 l_apply(fname, l)
@@ -69,6 +89,28 @@ n_apply(cpp, argc, argv)
 	return retval;
 }
 
+static int
+get_hrr_vis_flag(var_hrr_vis_name, name, stage_name)
+	const char *var_hrr_vis_name;
+	const char *name;
+	const char *stage_name;
+{
+	char *endptr;
+	long int tmp;
+	conscell *var_hrr_vis;
+
+	if ((var_hrr_vis = v_find(var_hrr_vis_name)) != NULL) {
+		if (cdr(var_hrr_vis) != NULL && !LIST(cdr(var_hrr_vis))) {
+			tmp = strtol(cdr(var_hrr_vis)->string, &endptr, 10);
+			if (cdr(var_hrr_vis)->string != '\0' && *endptr == '\0' &&
+			    (tmp != LONG_MAX || errno != ERANGE) && tmp >= 0)
+				return tmp ? 1 : 0;
+		}
+		fprintf(stderr, "%s(%s): invalid %s - ignored\n", name, stage_name, var_hrr_vis_name);
+	}
+	return 0;
+}
+
 /*
  * Call func with tokenlist t as the argument.
  * This is used to rewrite an address.
@@ -274,7 +316,7 @@ hdr_rewrite(name, h)
 			strcpy(s, cs);
 			/* t = HDR_SCANNER(cs); */
 			t = scan822((const char**)&s, strlen(s),
-				    '!', '%', 0, &ap->a_tokens->p_tokens);
+				    '!', '%', &ap->a_tokens->p_tokens);
 
 			stickymem = osticky;
 
@@ -295,6 +337,118 @@ hdr_rewrite(name, h)
 	return nh;
 }
 
+struct header *
+header_rewrite(name, h, fp, stage)
+	const char *name;
+	struct header *h;
+	FILE *fp;
+	int stage;
+{
+	const char *av[5], *stage_name[] = {"hrr_ab", "hrr_rr", "hrr_ae"};
+	const char *tmp, *token;
+	token822 *t;
+	struct header *nh = NULL, *th = NULL;
+
+	av[0] = name;
+	av[1] = stage_name[stage];
+	av[4] = NULL;
+
+	if (stage == HRR_RR) {
+		sb_external(FILENO(fp));
+		hdr_print(h, fp);
+		if ((tmp = sb_retrieve(FILENO(fp))) == NULL
+		    || (tmp = strchr(tmp, ':')) == NULL)
+			return h;
+		++tmp;
+		*(char *)(tmp + strlen(tmp) - 1) = '\0';
+		av[2] = h->h_pname;
+		if (get_hrr_vis_flag(HRR_VIS_DISABLE, name, stage_name[stage]))
+			av[3] = tmp;
+		else {
+			av[3] = tmalloc(strlen(tmp) * 4 + 1);
+			strvis((char *)av[3], tmp, VIS_OCTAL | VIS_GLOB | VIS_WHITE);
+		}
+	} else
+		av[2] = av[3] = "";
+
+	if (s_apply(4, av) == -1 || s_value == NULL)
+		return h;
+
+	if (!LIST(s_value) || car(s_value) == NULL || !LIST(car(s_value))) {
+		fprintf(stderr, "%s(%s): returned value is invalid\n", name, stage_name[stage]);
+		return h;
+	}
+
+	if (stage == HRR_RR && caar(s_value) == NULL)
+		return NULL;
+
+	for (s_value = car(s_value); s_value != NULL; s_value = cdr(s_value)) {
+
+		if (!LIST(s_value) || car(s_value) == NULL || LIST(car(s_value))
+		    || cdar(s_value) == NULL || LIST(cdar(s_value))) {
+			fprintf(stderr, "%s(%s): returned value is invalid\n", name, stage_name[stage]);
+			return h;
+		}
+
+		if (get_hrr_vis_flag(HRR_UNVIS_DISABLE, name, stage_name[stage]))
+			tmp = cdar(s_value)->string;
+		else {
+			tmp = tmalloc(strlen(cdar(s_value)->string) + 1);
+			if (strunvis((char *)tmp, cdar(s_value)->string) == -1) {
+				fprintf(stderr, "%s(%s): invalid escape sequence\n", name, stage_name[stage]);
+				return h;
+			}
+		}
+
+		if (nh == NULL)
+			nh = th = makeHeader(spt_headers, car(s_value)->string, strlen(car(s_value)->string));
+		else {
+			th->h_next = makeHeader(spt_headers, car(s_value)->string, strlen(car(s_value)->string));
+			th = th->h_next;
+		}
+
+		th->h_descriptor = &nullhdr;
+		th->h_lines = NULL;
+
+		t = NULL;
+
+		while (1) {
+			while ((*tmp == '\n') || ((*tmp == '\r') && (*(tmp + 1) == '\n'))) {
+				if (*tmp == '\r')
+					++tmp;
+				++tmp;
+			}
+			if (*tmp == '\0')
+				break;
+			if ((tmp = strpbrk(token = tmp, "\n")) == NULL) {
+				if (t == NULL) {
+					th->h_lines = makeToken(token, strlen(token));
+					th->h_lines->t_type = Line;
+				} else {
+					t->t_next = makeToken(token, strlen(token));
+					t->t_next->t_type = Line;
+				}
+				break;
+			}
+			if (t == NULL) {
+				th->h_lines = t = makeToken(token, *(tmp - 1) == '\r' ? tmp - token - 1 : tmp - token);
+				th->h_lines->t_type = Line;
+			} else {
+				t->t_next = makeToken(token, *(tmp - 1) == '\r' ? tmp - token - 1 : tmp - token);
+				t->t_next->t_type = Line;
+				t = t->t_next;
+			}
+		}
+
+		if (th->h_lines == NULL) {
+			fprintf(stderr, "%s(%s): returned value is invalid\n", name, stage_name[stage]);
+			return h;
+		}
+	}
+
+	return nh;
+}
+
 void
 setenvinfo(e)
 	struct envelope *e;
@@ -325,6 +479,9 @@ setenvinfo(e)
 		CONSTSTR(e->e_messageid);
 	}
 
+	CONSTSTR("spoolid");
+	CONSTSTR(e->e_spoolid);
+
 	CONSTSTR("uid");
 	NEWSTRD(e->e_statbuf.st_uid);
 
@@ -399,7 +556,7 @@ char *newattribute_2(onam,nam,val)
 	    UNGCPRO4;
 	    return NULL; /* ?? */
 	  }
-	  if (strcmp(nam,lc->string)==0) {
+	  if (STREQ(nam,lc->string)) {
 	    if (!cdr(lc)) {
 	      UNGCPRO4;
 	      return NULL;
