@@ -4,7 +4,7 @@
  */
 /*
  *	Lots of modifications (new guts, more or less..) by
- *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-1995
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2003
  */
 
 #include "hostenv.h"
@@ -16,44 +16,89 @@ int	/* return non-zero when there are some childs
 	   available for start				*/
 doagenda()
 {
-	struct thread *ncuritem, *nncuritem;
+	struct thread *thr, *nthr;
 	int didsomething = 0;
 
-	mytime(&now);
+	timed_log_reinit(); /* 	internal: mytime(&now); */
 
-	ncuritem = thread_head;
+
+	thr = thread_head;
 
 	if (verbose)
 	  sfprintf(sfstdout,"curitem %p curitem->wakeup %lu now %d\n",
-		   ncuritem, ncuritem ? ncuritem->wakeup : 0, (int)now);
+		   thr, thr ? thr->wakeup : 0, (int)now);
 
-	/* thread_head -chain should be in time order, thus the while-loop
-	   should be traversed only once, because  thread_start kicks
-	   right away (or reschedules) */
+	/* thread_head -chain should be in time order, thus
+	   the while-loop should be traversed only once, because
+	   thread_start kicks  right away (or reschedules) */
 
-	while (ncuritem  && ncuritem->wakeup <= now  ) {
-	  /* Object pointed by  ncuritem  may disappear due to
+	for ( ; thr  && thr->wakeup <= now ; thr = nthr ) {
+	  /* Object pointed by  thr  may disappear due to
 	     expiration at thread_start() .. */
-	  nncuritem = ncuritem->nexttr;
+	  nthr = thr->nexttr;
+
 	  /* Not running, and wakeup in past ? */
-	  if (ncuritem->proc == NULL && ncuritem->wakeup <= now)
+
+	  if (thr->proc == NULL && thr->wakeup <= now) {
 	    /* Try to start it! */
-	    while (thread_start(ncuritem, 0))
+
+	    while (thread_start(thr, 0))
 	      ++ didsomething;
 
-	  ncuritem = nncuritem;
+	  }
 
-	  queryipccheck();
+	  queryipccheck(); /* updates the 'now' variable too... */
 	}
 
-	mytime(&now);
-
 	/* if (verbose)
 	   printf("alarmed %d\n", now);  */
 
 	return (didsomething);
 }
 
+
+int	/* Return the number of messages expired this time around. */
+doexpiry2()
+{
+	struct thread *thr, *nthr;
+	int didsomething = 0, rc;
+	time_t timelimit;
+
+	timed_log_reinit(); /* 	internal: mytime(&now); */
+
+	timelimit = now + expiry2_timelimit;
+
+	thr = thread_head;
+
+	if (verbose)
+	  sfprintf(sfstdout,"curitem %p curitem->wakeup %lu now %d\n",
+		   thr, thr ? thr->wakeup : 0, (int)now);
+
+	/* thread_head -chain should be in time order, thus
+	   the while-loop should be traversed only once, because
+	   thread_start kicks right away (or reschedules) */
+
+	for ( ; thr  && thr->wakeup <= now ; thr = nthr ) {
+	  /* Object pointed by  thr  may disappear due to
+	     expiration at thread_start() .. */
+	  nthr = thr->nexttr;
+
+	  rc = thread_expire2(thr, timelimit, 0, NULL);
+	  if (rc)
+	    didsomething += rc;
+
+	  queryipccheck(); /* updates the 'now' variable too... */
+
+	  if (now > timelimit) break;
+	}
+
+	/* if (verbose)
+	   printf("alarmed %d\n", now);  */
+
+	return (didsomething);
+}
+
+
 /* Do immediate scheduling of given host-indentifier. */
 int
 turnme(turnarg)
@@ -62,7 +107,7 @@ const char *turnarg;
 	struct spblk *spl;
 	struct web *wp;
 	char *cp = strchr(turnarg,' ');
-	struct thread *ncuritem, *nncuritem;
+	struct thread *thr, *nthr;
 	spkey_t spk;
 	int rc = 0;
 
@@ -79,18 +124,18 @@ const char *turnarg;
 	}
 	wp = (struct web *)spl->data;
 	
-	ncuritem = thread_head;
-	while (ncuritem != NULL) {
-	  /* Object pointed by  ncuritem  may disappear due to
+	thr = thread_head;
+	while (thr != NULL) {
+	  /* Object pointed by  thr  may disappear due to
 	     expiration at thread_start()	 */
-	  nncuritem = ncuritem->nexttr;
+	  nthr = thr->nexttr;
 
-	  if (wp == ncuritem->whost && ncuritem->proc == NULL) {
-	    ncuritem->wakeup = 0; /* Force its starttime! */
-	    rc += thread_start(ncuritem, 1);
+	  if (wp == thr->whost && thr->proc == NULL) {
+	    thr->wakeup = 0; /* Force its starttime! */
+	    rc += thread_start(thr, 1);
 	    /* We MAY get multiple matches, though it is unlikely.. */
 	  }
-	  ncuritem = nncuritem;
+	  thr = nthr;
 	}
 	return rc;
 }   
