@@ -1,7 +1,7 @@
 /*
  *	ZMailer 2.99.53+ Scheduler "mailq2" routines
  *
- *	Copyright Matti Aarnio <mea@nic.funet.fi> 1999-2003
+ *	Copyright Matti Aarnio <mea@nic.funet.fi> 1999-2006
  *
  */
 
@@ -17,11 +17,6 @@
 
 #include "prototypes.h"
 
-#ifdef _AIX /* The select.h  defines NFDBITS, etc.. */
-# include <sys/types.h>
-# include <sys/select.h>
-#endif
-
 #ifdef HAVE_SYS_LOADAVG_H
 #include <sys/loadavg.h>
 #endif
@@ -42,45 +37,7 @@
 #endif
 #include "libc.h"
 
-
-#ifndef	NFDBITS
-/*
- * This stuff taken from the 4.3bsd /usr/include/sys/types.h, but on the
- * assumption we are dealing with pre-4.3bsd select().
- */
-
-typedef long	fd_mask;
-
-#ifndef	NBBY
-#define	NBBY	8
-#endif	/* NBBY */
-#define	NFDBITS		((sizeof fd_mask) * NBBY)
-
-/* SunOS 3.x and 4.x>2 BSD already defines this in /usr/include/sys/types.h */
-#ifdef	notdef
-typedef	struct fd_set { fd_mask	fds_bits[1]; } fd_set;
-#endif	/* notdef */
-
-#ifndef	_Z_FD_SET
-#define	_Z_FD_SET(n, p)   ((p)->fds_bits[0] |= (1 << (n)))
-#define	_Z_FD_CLR(n, p)   ((p)->fds_bits[0] &= ~(1 << (n)))
-#define	_Z_FD_ISSET(n, p) ((p)->fds_bits[0] & (1 << (n)))
-#define _Z_FD_ZERO(p)	  memset((char *)(p), 0, sizeof(*(p)))
-#endif	/* !FD_SET */
-#endif	/* !NFDBITS */
-
-#ifdef FD_SET
-#define _Z_FD_SET(sock,var) FD_SET(sock,&var)
-#define _Z_FD_CLR(sock,var) FD_CLR(sock,&var)
-#define _Z_FD_ZERO(var) FD_ZERO(&var)
-#define _Z_FD_ISSET(i,var) FD_ISSET(i,&var)
-#else
-#define _Z_FD_SET(sock,var) var |= (1 << sock)
-#define _Z_FD_CLR(sock,var) var &= ~(1 << sock)
-#define _Z_FD_ZERO(var) var = 0
-#define _Z_FD_ISSET(i,var) ((var & (1 << i)) != 0)
-#endif
-
+#include "zmpoll.h"
 
 
 static void mq2interpret __((struct mailq *, char *));
@@ -415,7 +372,7 @@ void mq2_register(fd, addr)
 
   mq2_puts(mq,"version zmailer 2.0\n");
 
-  mq2test = mq2_authuser(mq, NULL);
+  mq2test = mq2_authuser(mq, mq2authfile, NULL);
 
   if (! mq2test) {
 
@@ -448,31 +405,34 @@ void mq2_register(fd, addr)
 }
 
 /* EXTERNAL */
-int mq2add_to_mask(rdmaskp, wrmaskp, maxfd)
-     fd_set *rdmaskp, *wrmaskp;
-     int maxfd;
+void mq2add_to_poll(fds, fdscountp)
+     struct zmpollfd **fds;
+     int *fdscountp;
 {
   struct mailq *mq = mq2root;
 
   for ( ; mq ; mq = mq->nextmailq ) {
+    mq->fds = NULL;
+
     if (mq->fd < 0)
       continue;
 
-    if (mq->fd > maxfd)
-      maxfd = mq->fd;
+    /* _Z_FD_SET(mq->fd, *rdmaskp);
+       if (mq->outbufcount < mq->outbufsize)
+       _Z_FD_SET(mq->fd, *wrmaskp);
+    */
 
-    _Z_FD_SET(mq->fd, *rdmaskp);
+    zmpoll_addfd(fds, fdscountp, mq->fd,
+		 (mq->outbufcount < mq->outbufsize ? mq->fd : -1),
+		 &mq->fds);
 
-    if (mq->outbufcount < mq->outbufsize)
-      _Z_FD_SET(mq->fd, *wrmaskp);
   }
-
-  return maxfd;
 }
 
+
 /* EXTERNAL */
-void mq2_areinsets(rdmaskp, wrmaskp)
-     fd_set *rdmaskp, *wrmaskp;
+void mq2_areinsets(fds)
+     struct zmpollfd *fds;
 {
     struct mailq *mq;
 
@@ -486,7 +446,7 @@ void mq2_areinsets(rdmaskp, wrmaskp)
     mq = mq2root;
     while ( mq ) {
       struct mailq *mq2 = mq->nextmailq;
-      if (mq->fd >= 0 && _Z_FD_ISSET(mq->fd, *wrmaskp)) {
+      if ( mq->fds && (mq->fds->revents & ZM_POLLOUT) ) {
 	mq2_wflush(mq);
       }
 
@@ -502,7 +462,9 @@ void mq2_areinsets(rdmaskp, wrmaskp)
     mq = mq2root;
     while ( mq ) {
       struct mailq *mq2 = mq->nextmailq;
-      if (mq->fd < 0 || _Z_FD_ISSET(mq->fd, *rdmaskp)) {
+      if ( mq->fds &&
+	   (mq->fds->revents & (ZM_POLLIN|ZM_POLLERR|ZM_POLLHUP)) ) {
+	mq->fds = NULL;
 	mq2_read(mq);
       }
       mq = mq2;
@@ -609,6 +571,68 @@ static int mq2cmd_etrn(mq,s)
   return 0;
 }
 
+static void
+scheduler_mailq_extra(fp)
+     Sfio_t *fp;
+{
+  int pid;
+  int pipe_fd[2];
+  const char *scheduler_M_extra_cmd = getzenv("SCHEDULER_M_EXTRA");
+
+  if (!scheduler_M_extra_cmd) return; /* Nothing to do.. */
+
+  if (pipe(pipe_fd) != 0)
+    return; /* Sigh.. We don't worry about e.g. EINTR here. */
+
+  pid = fork();
+
+  if (pid > 0) {
+    /* Parent! */
+
+    /* We are ourselves going to die rather quickly... */
+# ifdef SIGCLD
+    SIGNAL_HANDLE(SIGCLD,SIG_IGN);		/* Auto-reap the kids.. */
+# else
+    SIGNAL_HANDLE(SIGCHLD,SIG_IGN);
+# endif
+
+    close(pipe_fd[1]); /* Parent does not write */
+
+    for (;;) {
+      char c;
+      int rc = read(pipe_fd[0], &c, 1);
+      if (rc == 0) break; /* EOF */
+      if (rc == 1) {
+	sfputc(fp, c);
+	continue;
+      }
+      if (rc < 0 && errno == EINTR)
+	continue;
+      break; /* ANYTHING AT ALL: break out! */
+    }
+    close(pipe_fd[0]);
+
+    return;
+  }
+  if (pid == 0) {
+    int i;
+    /* Child! */
+    close(pipe_fd[0]); /* Child does not read.. */
+
+    if (pipe_fd[1] != 1)
+      dup2(pipe_fd[1],1); /* STDOUT */
+    if (pipe_fd[1] != 2)
+      dup2(pipe_fd[1],2); /* STDERR */
+
+    close(0);
+    i = open("/dev/null",O_RDONLY,0);
+    if (i > 0) dup2(i, 0);
+
+    execl(scheduler_M_extra_cmd, scheduler_M_extra_cmd, NULL);
+    _exit(255);
+ }
+}
+
 
 /* INTERNAL */
 static void mq2_show_snmp(mq)
@@ -634,6 +658,8 @@ static void mq2_show_snmp(mq)
   r = (Z_SHM_MIB_is_attached() > 0); /* Attached and WRITABLE ? */
 
 
+#define SCHEDULER_MAILQ_EXTRA scheduler_mailq_extra(fp);
+
 #include "mailq.inc" /* Shared stuff with  mailq.c  program */
 
 
@@ -828,7 +854,7 @@ static void mq2interpret(mq,s)
 
   if (mq->auth == 0 && strcmp(s,"AUTH") == 0) {
     MIBMtaEntry->sc.MQ2sockCommandAUTH ++;
-    mq2auth(mq,t);
+    mq2auth(mq, mq2authfile, t);
     if (! mq->auth)
       MIBMtaEntry->sc.MQ2sockAuthRej ++;
     return;
