@@ -4,7 +4,7 @@
  */
 /*
  *	Lots of modifications (new guts, more or less..) by
- *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2001
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2003
  *
  *	We produce RFC 1894 format reports with everything it contains.
  */
@@ -30,6 +30,8 @@ extern char *strnsave();
 #endif
 extern int errno;
 
+int store_error_on_error;
+
 extern int default_full_content; /* at conf.c */
 
 /*
@@ -476,7 +478,7 @@ reporterrs(cfpi, delayreports)
 	char boundarystr[400];
 	char rptspoolid[30];
 	time_t mtime;
-	long ino;
+	long ino, mtimens;
 	int actionsets[5]; /* 0:DELIVERED, 1:FAILED, 2:RELAYED, 3:DELAYED,
 			      4:EXPANDED */
 	ACTSETENUM thisaction;
@@ -810,7 +812,22 @@ reporterrs(cfpi, delayreports)
 
 	  fstat(sffileno(errfp),&stbuf); /* doesn't matter exactly what,
 					  as long as unique */
-	  taspoolid(boundarystr, stbuf.st_ctime, (long)stbuf.st_ino);
+
+	  taspoolid(boundarystr, (long)stbuf.st_ino, stbuf.st_mtime,
+#ifdef HAVE_STRUCT_STAT_ST_ATIM_TV_NSEC
+		    stbuf.st_mtim.tv_nsec
+#else
+#ifdef HAVE_STRUCT_STAT_ST_ATIM___TV_NSEC
+		    stbuf.st_mtim.__tv_nsec
+#else
+#ifdef HAVE_STRUCT_STAT_ST_ATIMENSEC
+		    stbuf.st_mtimensec
+#else
+		    0
+#endif
+#endif
+#endif
+		    );
 
 	  strcat(boundarystr, "=_/");
 	  strcat(boundarystr, dom);
@@ -825,6 +842,7 @@ reporterrs(cfpi, delayreports)
 	} else {
 	  sfprintf(errfp, "Reporting-MTA: x-local-hostname; -unknown-\n");
 	}
+	sfprintf(errfp, "Return-Path: <%s>\n", eaddr);
 	if (envid != NULL) {
 	  sfprintf(errfp, "Original-Envelope-Id: ");
 	  decodeXtext(errfp,envid);
@@ -1044,16 +1062,19 @@ be in subsequent parts of this MESSAGE/DELIVERY-STATUS structure.\n\n");
 	/* And cap the tail with paired MIME boundary.. */
 	sfprintf(errfp, "--%s--\n", boundarystr);
 
-	ino = 0; mtime = 0;
+	ino = 0; mtime = mtimens = 0;
 	close(cfp->fd);
 	free_cfp_memory(cfp);
 
 	if (no_error_report > 0) {
-	  sfmail_close_alternate_async(errfp,POSTMANDIR,":error-on-error", msgwriteasync);
+	  if (store_error_on_error)
+	    sfmail_close_alternate_async(errfp,POSTMANDIR,":error-on-error", msgwriteasync);
+	  else
+	    sfmail_abort(errfp);
 	  sprintf(rptspoolid, "POSTMAN :error-on-error"); /* < 30 chr ! */
 	} else {
-	  _sfmail_close_async(errfp, &ino, &mtime, msgwriteasync);	/* XX: check for error */
-	  taspoolid(rptspoolid, mtime, ino);
+	  _sfmail_close_async(errfp, &ino, &mtime, &mtimens, msgwriteasync);	/* XX: check for error */
+	  taspoolid(rptspoolid, ino, mtime, mtimens);
 	}
 
 	if (do_syslog)
@@ -1198,9 +1219,10 @@ delayreport(vp)
 /* FIXME: reporting subsystem needs tuning! */
 /*        ... implementing/tuning DELAYED reports! */
 
-static int ctl_report_1 __((struct spblk *spl));
-static int ctl_report_1(spl)
-struct spblk *spl;
+static int ctl_report_1 __((void *, struct spblk *spl));
+static int ctl_report_1(p, spl)
+	void *p;
+	struct spblk *spl;
 {
 	struct vertex *vp, *nvp;
 	struct ctlfile * cfp = (struct ctlfile *)spl->data;
@@ -1236,7 +1258,7 @@ struct spblk *spl;
 void
 interim_report_run __((void))
 {
-	sp_scan(ctl_report_1, NULL, spt_mesh[L_CTLFILE]);
+	sp_scan(ctl_report_1, NULL, NULL, spt_mesh[L_CTLFILE]);
 }
 
 
