@@ -4,7 +4,7 @@
  */
 /*
  *	Lots of modifications (new guts, more or less..) by
- *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2003
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2004
  */
 
 /*
@@ -75,10 +75,6 @@ extern char *optarg;
 
 struct sptree *spt_mesh[SIZE_L];
 
-#ifdef	MALLOC_TRACE
-struct conshell *envarlist = NULL;
-#endif	/* MALLOC_TRACE */
-
 #define TRANSPORTMAXNOFILES 32 /* Number of files a transporter may
 				  need open -- or any of its children.. */
 int	transportmaxnofiles = TRANSPORTMAXNOFILES; /* Default value */
@@ -103,6 +99,7 @@ time_t	sched_starttime;
 int	do_syslog = 0;
 int	verbose = 0;
 int	querysocket = -1;	/* fd of TCP socket to listen for queries */
+int	querysocket6 = -1;	/* fd of TCP socket to listen for queries */
 int	msgwriteasync = 0;
 int	D_alloc = 0;
 int	hungry_childs = 0;
@@ -127,6 +124,8 @@ extern int forkrate_limit;	/* How many forks per second ? */
 int	mailqmode = 1;		/* ZMailer v1.0 mode on mailq */
 char *  mailqsock;
 char *  notifysock;
+int     QSpoolFd = -1;
+int     TSpoolFd = -1;
 
 int    interim_report_interval = 5*60; /* 5 minutes */
 
@@ -174,6 +173,8 @@ static void   init_timeserver __((void));
 extern time_t time __((time_t *));
 #endif
 
+struct itimerval profiler_itimer_at_start;
+
 static int loginitsched __((int));
 static int loginitsched(sig)
 int sig;
@@ -242,11 +243,13 @@ void timed_log_reinit()
 	lstat(logfn,&stbuf) == 0 &&
 	S_ISREG(stbuf.st_mode)) {
 
-      long fsspace;
-      fsspace = free_fd_statfs(sffileno(sfstdout));
-      MIBMtaEntry->sys.LogFreeSpace = fsspace;
-      fsspace = used_fd_statfs(sffileno(sfstdout));
-      MIBMtaEntry->sys.LogUsedSpace = fsspace;
+      long bavail, bused, iavail, iused;
+      if (! fd_statfs(sffileno(sfstdout), &bavail, &bused, &iavail, &iused)) {
+	MIBMtaEntry->sys.LogUsedSpace = bused;
+	MIBMtaEntry->sys.LogFreeSpace = bavail;
+	MIBMtaEntry->sys.LogUsedFiles = iused;
+	MIBMtaEntry->sys.LogFreeFiles = iavail;
+      }
     }
   }
 }
@@ -277,7 +280,7 @@ int global_maxkids = 1000;
 time_t now;
 
 Sfio_t * vfp_open(cfp)
-struct ctlfile *cfp;
+	struct ctlfile *cfp;
 {
 	Sfio_t *vfp;
 	int fd;
@@ -304,7 +307,7 @@ static void cfp_free __((struct ctlfile *cfp, struct spblk *spl));
 static void cfp_free0 __((struct ctlfile *cfp));
 
 static void cfp_free0(cfp)
-struct ctlfile *cfp;
+	struct ctlfile *cfp;
 {
 
 	struct vertex *vp, *nvp;
@@ -325,7 +328,7 @@ struct ctlfile *cfp;
 	if (!cfp->head) {
 	  /* This should *not* be happening.. */
 	  sfprintf(sfstderr,
-		   "%s: SHOULD NOT HAPPEN: cfp->head == NULL; spoolid: %s\n",
+		   "%s: SHOULD NOT HAPPEN: (except for ETRN files) cfp->head == NULL; spoolid: %s\n",
 		   progname, cfp->spoolid);
 	  unctlfile(cfp, 1);
 	  return;
@@ -344,8 +347,8 @@ struct ctlfile *cfp;
 }
 
 static void cfp_free(cfp, spl)
-struct ctlfile *cfp;
-struct spblk   *spl;
+	struct ctlfile *cfp;
+	struct spblk   *spl;
 {
 	/* Delete from the  spt_mesh[]  */
 
@@ -360,9 +363,10 @@ struct spblk   *spl;
 }
 
 
-static int ctl_free __((struct spblk *spl));
-static int ctl_free(spl)
-struct spblk *spl;
+static int ctl_free __((void *, struct spblk *spl));
+static int ctl_free(p, spl)
+	void *p;
+	struct spblk *spl;
 {
 	cfp_free((struct ctlfile *)spl->data, NULL);
 	return 0;
@@ -375,7 +379,7 @@ struct spblk *spl;
  *  previously released..  (cfp->contents, for example)
  */
 void free_cfp_memory(cfp)
-struct ctlfile *cfp;
+	struct ctlfile *cfp;
 {
 	if (cfp->contents)	free(cfp->contents);
 	if (cfp->vfpfn)		free(cfp->vfpfn);
@@ -521,6 +525,9 @@ main(argc, argv)
 
 	const char *t, *syslogflg;
 
+	getitimer(ITIMER_PROF, & profiler_itimer_at_start);
+
+
 #ifdef GLIBC_MALLOC_DEBUG__ /* memory allocation debugging with GLIBC */
 	old_malloc_hook = __malloc_hook;
 	__malloc_hook = my_malloc_hook;
@@ -543,7 +550,7 @@ main(argc, argv)
 	ArgvSave   = (char *) argv[0];
 	EOArgvSave = argv[argc-1] + strlen(argv[argc-1]) + 1;
 
-	mytime(&sched_starttime);
+	time(&sched_starttime);
 
 	memset(&dirqb,0,sizeof(dirqb));
 	dirscan_mesh = sp_init();
@@ -771,16 +778,6 @@ main(argc, argv)
 	}
 
 
-	if (logfn != NULL) {
-	  /* loginit is a signal handler, so can't pass log */
-	  if (loginitsched(SIGHUP) < 0) /* do setlinebuf() there */
-	    die(1, "log initialization failure");
-	  /* close and reopen log files */
-	  SIGNAL_HANDLE(SIGHUP, (RETSIGTYPE(*)__((int))) loginitsched);
-	} else {
-	  SIGNAL_IGNORE(SIGHUP); /* no surprises please */
-	  sfset(sfstdout, SF_LINE, 1);
-	}
 #ifdef USE_SIGREAPER
 # ifdef SIGCLD
 	SIGNAL_HANDLE(SIGCLD,  sig_chld);
@@ -820,8 +817,6 @@ main(argc, argv)
 	  /* NOTREACHED */
 	}
 
-
-
 	if (postoffice == NULL && (postoffice = getzenv("POSTOFFICE")) == NULL)
 	  postoffice = POSTOFFICE;
 
@@ -829,6 +824,19 @@ main(argc, argv)
 	  sfprintf(sfstderr, "%s: cannot chdir to %s/%s.\n",
 		   progname, postoffice, TRANSPORTDIR);
 
+
+	if (logfn != NULL) {
+	  /* loginit is a signal handler, so can't pass log */
+	  if (loginitsched(SIGHUP) < 0) /* do setlinebuf() there */
+	    die(1, "log initialization failure");
+	  /* close and reopen log files */
+	  SIGNAL_HANDLE(SIGHUP, (RETSIGTYPE(*)__((int))) loginitsched);
+	} else {
+	  SIGNAL_IGNORE(SIGHUP); /* no surprises please */
+	  sfset(sfstdout, SF_LINE, 1);
+	}
+
+
 	if (rendezvous == NULL && (rendezvous = getzenv("RENDEZVOUS")) == NULL)
 	  rendezvous = qoutputfile;
 	if (daemonflg) {
@@ -842,12 +850,23 @@ main(argc, argv)
 	    /* NOTREACHED */
 	  }
 	  detach();		/* leave worldy matters behind */
-	  mytime(&now);
+
+	  setitimer(ITIMER_PROF, &  profiler_itimer_at_start, NULL);
+
+	  close(0);
+	  open("/dev/null",O_RDONLY,0); /* This is good for fd0 reading.. */
+
+
+	  time(&now);
 	  sfprintf(sfstdout, "%s: scheduler daemon (%s)\n\tpid %d started at %s\n",
 		   progname, Version, (int)getpid(), (char *)rfc822date(&now));
 	}
 
 
+	/* call this to create the timeserver -- if possible */
+	init_timeserver(); /* Will take around 3-4 secs.. */
+
+
 	/* Now we are either interactive, or daemon, lets attach monitoring
 	   memory block.. and fill it in.  */
 
@@ -857,14 +876,22 @@ main(argc, argv)
 
 	/* Zero the gauges at our startup.. */
 	MIBMtaEntry->sc.StoredMessagesSc		= 0;
-	MIBMtaEntry->sc.StoredThreadsSc		= 0;
+	MIBMtaEntry->sc.StoredThreadsSc			= 0;
 	MIBMtaEntry->sc.StoredVerticesSc		= 0;
 	MIBMtaEntry->sc.StoredRecipientsSc		= 0;
-	MIBMtaEntry->sc.StoredVolumeSc		= 0;
+	MIBMtaEntry->sc.StoredVolumeSc			= 0;
 	MIBMtaEntry->sc.TransportAgentProcessesSc	= 0;
-	MIBMtaEntry->sc.TransportAgentsActiveSc	= 0;
+	MIBMtaEntry->sc.TransportAgentsActiveSc		= 0;
 	MIBMtaEntry->sc.TransportAgentsIdleSc		= 0;
 
+	/* Zero also these transport agent gauges at our startup.. */
+	MIBMtaEntry->tasmtp.TaProcCountG		= 0;
+	MIBMtaEntry->tasmcm.TaProcCountG		= 0;
+	MIBMtaEntry->tambox.TaProcCountG		= 0;
+	MIBMtaEntry->tahold.TaProcCountG		= 0;
+	MIBMtaEntry->taerrm.TaProcCountG		= 0;
+	MIBMtaEntry->taexpi.TaProcCountG		= 0;
+	MIBMtaEntry->tarert.TaProcCountG		= 0;
 
 	/* Actually we want this to act as daemon,
 	   even when not in daemon mode.. */
@@ -889,10 +916,6 @@ main(argc, argv)
 	SIGNAL_HANDLE(SIGTERM, sig_exit);	/* split */
 	SIGNAL_HANDLE(SIGQUIT, sig_quit);	/* Slow shutdown */
 
-	/* call it to create the timeserver -- if possible */
-	init_timeserver(); /* Will take around 3-4 secs.. */
-
-
 	if (optind < argc) {
 	  /* process the specified control files only */
 	  for (; optind < argc; ++optind) {
@@ -1056,7 +1079,7 @@ main(argc, argv)
 	/* Doing nicely we would kill the childs here, but we are not
 	   such a nice people -- we just discard incore data, and crash out.. */
 
-	sp_scan(ctl_free, NULL, spt_mesh[L_CTLFILE]);
+	sp_scan(ctl_free, NULL, NULL, spt_mesh[L_CTLFILE]);
 
 #ifdef	MALLOC_TRACE
 	mal_dumpleaktrace(stderr);
@@ -1077,6 +1100,10 @@ int sig;
 		close(querysocket);
 		querysocket = -1;
 	}
+	if (querysocket6 >= 0) {		/* give up mailq socket asap */
+		close(querysocket6);
+		querysocket6 = -1;
+	}
 	if (canexit)
 		die(0, "signal");
 	mustexit = 1;
@@ -1903,7 +1930,23 @@ slurp(fd, ino)
 	cfp->fd       = fd;
 	cfp->dirind   = -1; /* Not known -- or top-level */
 	cfp->uid      = stbuf.st_uid;
+
 	cfp->envctime = stbuf.st_ctime;
+#ifdef HAVE_STRUCT_STAT_ST_ATIM_TV_NSEC
+	cfp->envctimens = stbuf.st_ctim.tv_nsec;
+#else
+#ifdef HAVE_STRUCT_STAT_ST_ATIM___TV_NSEC
+	cfp->envctimens = stbuf.st_ctim.__tv_nsec;
+#else
+#ifdef HAVE_STRUCT_STAT_ST_ATIMENSEC
+	cfp->envctimens = stbuf.st_ctimensec;
+#else
+	cfp->envctimens = 0;
+#endif
+#endif
+#endif
+
+
 	cfp->contents = contents;
 
 	/* 
@@ -2448,11 +2491,26 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 					another machine, and run things
 					in there, and still have same
 					expiration times.. */
+#ifdef HAVE_STRUCT_STAT_ST_ATIM_TV_NSEC
+	cfp->mtimens = stbuf.st_mtim.tv_nsec;
+#else
+#ifdef HAVE_STRUCT_STAT_ST_ATIM___TV_NSEC
+	cfp->mtimens = stbuf.st_mtim.__tv_nsec;
+#else
+#ifdef HAVE_STRUCT_STAT_ST_ATIMENSEC
+	cfp->mtimens = stbuf.st_mtimensec;
+#else
+	cfp->mtimens = 0;
+#endif
+#endif
+#endif
 
 	/* Reuse the buffer ...
 	   Generate same spoolid string that all other subsystems also
 	   report to syslog. */
-	taspoolid(path2, cfp->mtime, cfp->id);
+
+	taspoolid(path2, cfp->id, cfp->mtime, cfp->mtimens);
+
 	cfp->spoolid = strsave(path2);
 	if (!cfp->spoolid) {
 	  /* malloc() failure.. */
@@ -2813,7 +2871,8 @@ static void vtxdo(vp, cehdr, path)
 }
 
 
-int vtxredo(spl)
+int vtxredo(p, spl)
+	void *p;
         struct spblk *spl;
 {
         struct ctlfile *cfp = (struct ctlfile *)spl->data;
@@ -2821,7 +2880,7 @@ int vtxredo(spl)
 
         /* assert cfp != NULL */
 	for (vp = cfp->head ; vp != NULL ; vp = vp->next[L_CTLFILE]) {
-	  vtxdo(vp, rrcf_head, NULL);
+	  vtxdo(vp, p /* -> rrcf_head */, NULL);
 	}
         return 0;
 }
@@ -3080,7 +3139,7 @@ static void init_timeserver()
 	  timeserver_segment->pid = ppid;
 
 	  time(&now);
-	  sleep(3);
+	  sleep(2);
 
 #ifdef HAVE_SELECT
 	  now2 = timeserver_segment->tv.tv_sec;
@@ -3088,7 +3147,7 @@ static void init_timeserver()
 	  now2 = timeserver_segment->time_sec;
 #endif
 	  if (now2 == 0 || now2 == now) {
-	    /* HUH ?! No incrementation in 3 seconds ?? */
+	    /* HUH ?! No incrementation in 2 seconds ?? */
 	    /* Start ticking in alarm() mode! */
 	    kill (ppid, SIGTERM);
 	    timeserver_pid = 0;
@@ -3108,11 +3167,32 @@ static void init_timeserver()
 	{
 
 	  int space_check_count = 0;
+	  char *p;
 
 	  MIBMtaEntry->sc.schedulerTimeserverStartTime = time(NULL);
 	  MIBMtaEntry->sc.schedulerTimeserverStarts ++;
 
 
+	  close(0); close(1); close(2);
+	  if (statuslog) sfclose(statuslog);
+
+	  /* Create file handles (and temp files) for studying
+	     file system (free) space state. */
+
+	  p = malloc(strlen(TRANSPORTDIR)+50);
+	  if (p) {
+	    sprintf(p, "../%s/.%d.spacetest", QUEUEDIR, getpid());
+	    QSpoolFd = open( p, O_RDWR|O_CREAT, 0600 );
+	    unlink(p);unlink(p);
+	    
+	    sprintf(p, "../%s/.%d.spacetest", TRANSPORTDIR, getpid());
+	    TSpoolFd = open( p, O_RDWR|O_CREAT, 0600 );
+	    unlink(p);unlink(p);
+
+	    free(p);
+	  }
+
+
 #ifdef HAVE_SETPROCTITLE
 	  setproctitle("[TimeServer]");
 #else
@@ -3125,9 +3205,9 @@ static void init_timeserver()
 	  ppid = getppid(); /* who is our parent ? */
 
 #ifdef HAVE_SELECT
-	  gettimeofday(&timeserver_segment->tv, NULL);
+	  gettimeofday((struct timeval *)&timeserver_segment->tv, NULL);
 #else
-	  time(&timeserver_segment->time_sec);
+	  time((time_t *)&timeserver_segment->time_sec);
 #endif
 
 	  for(;;) {
@@ -3138,12 +3218,12 @@ static void init_timeserver()
 	    tv.tv_sec = 0;
 	    tv.tv_usec = 300000;
 
-	    gettimeofday(&timeserver_segment->tv, NULL);
+	    gettimeofday((struct timeval *)&timeserver_segment->tv, NULL);
 
 	    rc = select(0,NULL,NULL,NULL,&tv);
 #else
-	    time(&timeserver_segment->time_sec);
-	    sleep(1);
+	    time((time_t *)&timeserver_segment->time_sec);
+	    sleep(1); /* We don't have select() to use.. */
 #endif
 	    /* Is the parent still alive ?? */
 	    if (kill(ppid, 0) != 0)
@@ -3152,9 +3232,23 @@ static void init_timeserver()
 	    /* Now check and fill in the filesystem free space
 	       gauges */
 	    if (--space_check_count  < 0) {
-	      long spc = Z_SHM_FileSysFreeSpace();
 
-	      MIBMtaEntry->sys.SpoolFreeSpace = spc;
+	      long bavail, bused, iavail, iused;
+	      if (QSpoolFd >= 0 &&
+		  ! fd_statfs(QSpoolFd, &bavail, &bused, &iavail, &iused)) {
+		MIBMtaEntry->sys.SpoolUsedSpace = bused;
+		MIBMtaEntry->sys.SpoolFreeSpace = bavail;
+		MIBMtaEntry->sys.SpoolUsedFiles = iused;
+		MIBMtaEntry->sys.SpoolFreeFiles = iavail;
+	      }
+	      if (TSpoolFd >= 0 &&
+		  ! fd_statfs(TSpoolFd, &bavail, &bused, &iavail, &iused)) {
+		MIBMtaEntry->sys.TportSpoolUsedSpace = bused;
+		MIBMtaEntry->sys.TportSpoolFreeSpace = bavail;
+		MIBMtaEntry->sys.TportSpoolUsedFiles = iused;
+		MIBMtaEntry->sys.TportSpoolFreeFiles = iavail;
+	      }
+
 	      /* Log FS space is monitored by the self-timed
 		 log reinit code. */
 
