@@ -4,7 +4,7 @@
  */
 /*
  *	Lots of modifications (new guts, more or less..) by
- *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2003
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2006
  */
 
 
@@ -25,7 +25,7 @@
 #include "zsyslog.h"
 #include <sysexits.h>
 #ifdef HAVE_SYS_RESOURCE_H
-#include <sys/resource.h>
+#include <sys/resource.h> /* for setpriority */
 #endif
 #ifdef HAVE_SYS_UN_H
 #include <sys/socket.h>
@@ -39,6 +39,7 @@
 #include "libz.h"
 
 #include "libc.h"
+#include "zmpoll.h"
 
 extern int forkrate_limit;
 extern int freeze;
@@ -1039,128 +1040,90 @@ static void waitandclose(fd)
 	reclaim(fd, cpids[fd].tofd);
 }
 
-#ifdef	HAVE_SELECT
 
-#ifdef _AIX /* The select.h  defines NFDBITS, etc.. */
-# include <sys/types.h>
-# include <sys/select.h>
-#endif
-
-#ifndef	NFDBITS
-/*
- * This stuff taken from the 4.3bsd /usr/include/sys/types.h, but on the
- * assumption we are dealing with pre-4.3bsd select().
- */
-
-typedef long	fd_mask;
-
-#ifndef	NBBY
-#define	NBBY	8
-#endif	/* NBBY */
-#define	NFDBITS		((sizeof fd_mask) * NBBY)
-
-/* SunOS 3.x and 4.x>2 BSD already defines this in /usr/include/sys/types.h */
-#ifdef	notdef
-typedef	struct fd_set { fd_mask	fds_bits[1]; } fd_set;
-#endif	/* notdef */
-
-#ifndef	_Z_FD_SET
-#define	_Z_FD_SET(n, p)   ((p)->fds_bits[0] |= (1 << (n)))
-#define	_Z_FD_CLR(n, p)   ((p)->fds_bits[0] &= ~(1 << (n)))
-#define	_Z_FD_ISSET(n, p) ((p)->fds_bits[0] & (1 << (n)))
-#define _Z_FD_ZERO(p)	  memset((char *)(p), 0, sizeof(*(p)))
-#endif	/* !FD_SET */
-#endif	/* !NFDBITS */
-
-#ifdef FD_SET
-#define _Z_FD_SET(sock,var) FD_SET(sock,&var)
-#define _Z_FD_CLR(sock,var) FD_CLR(sock,&var)
-#define _Z_FD_ZERO(var) FD_ZERO(&var)
-#define _Z_FD_ISSET(i,var) FD_ISSET(i,&var)
-#else
-#define _Z_FD_SET(sock,var) var |= (1 << sock)
-#define _Z_FD_CLR(sock,var) var &= ~(1 << sock)
-#define _Z_FD_ZERO(var) var = 0
-#define _Z_FD_ISSET(i,var) ((var & (1 << i)) != 0)
-#endif
-
-
-int in_select = 0;
+int in_poll = 0;
 
 int
 mux(timeout)
 time_t timeout;
 {
-	int	i, n, maxf;
-	fd_set	rdmask;
-	fd_set	wrmask;
-	struct timeval tv;
+	int	i, n;
+	int wait_secs;
 	struct procinfo *proc = cpids;
 
+	int fdscount = 0;
+	static struct zmpollfd *fds = NULL;
+
+	struct zmpollfd *queryfds  = NULL;
+	struct zmpollfd *query6fds = NULL;
+	struct zmpollfd *notifyfds = NULL;
+
 	timed_log_reinit();
 
-	if (in_select) {
+	if (in_poll) {
 	  sfprintf(sfstderr,"**** recursed into mux()! ***\n");
 	  return 0;
 	}
 
 	queryipccheck();
 
-	tv.tv_sec = timeout - now; /* Timeout in seconds */
+	wait_secs = timeout - now; /* Timeout in seconds */
 	if (timeout < now)
-	  tv.tv_sec = 0;
-	tv.tv_usec = 0;
+	  wait_secs = 0;
 
-	maxf = -1;
-	_Z_FD_ZERO(rdmask);
-	_Z_FD_ZERO(wrmask);
+	fdscount = 0;
 	readsockcnt = 0;
 	if (cpids != NULL)
-	  for (proc = cpids,i = 0; i < scheduler_nofiles ; ++i,++proc)
+	  for (proc = cpids,i = 0; i < scheduler_nofiles ; ++i,++proc) {
+	    proc->fdpfrom = proc->fdpto = NULL;
+
 	    if (proc->pid != 0) {
 	      /* Something can be read ? */
-	      _Z_FD_SET(i, rdmask);
-	      if (maxf < i)
-		maxf = i;
+	      zmpoll_addfd(&fds, &fdscount, i, -1, &(proc->fdpfrom));
+	      /* sfprintf(sfstderr,"**** QX zmpoll_addfd(fds, %d, %d, %d, fdpfrom) QX ***\n", fdscount, i, -1); */
 
 	      ++readsockcnt;
 	      /* Something to write ? */
 	      if (proc->cmdlen > 0 && proc->tofd >= 0) {
-		_Z_FD_SET(proc->tofd, wrmask);
-		if (maxf < proc->tofd)
-		  maxf = proc->tofd;
+	        zmpoll_addfd(&fds, &fdscount, -1,  proc->tofd, &(proc->fdpto));
+	        /* sfprintf(sfstderr,"**** QX zmpoll_addfd(fds, %d, %d, %d, fdpto) QX ***\n", fdscount, -1, proc->tofd); */
+	      } else {
+		proc->fdpto = NULL;
 	      }
 
 	    }
+	  }
 
-	if (querysocket >= 0) {
-	  _Z_FD_SET(querysocket, rdmask);
-	  if (maxf < querysocket)
-	    maxf = querysocket;
-	}
-	if (notifysocket >= 0) {
-	  _Z_FD_SET(notifysocket, rdmask);
-	  if (maxf < notifysocket)
-	    maxf = notifysocket;
-	}
+	if (querysocket >= 0)
+	  zmpoll_addfd(&fds, &fdscount, querysocket,  -1, &queryfds);
 
-	/* Although we don't react on the results of these MQ2 fd's,
-	   getting them to break timeouts is important for MAILQv2
-	   responsiveness. ! */
+	if (querysocket6 >= 0)
+	  zmpoll_addfd(&fds, &fdscount, querysocket6, -1, &query6fds);
+
+	if (notifysocket >= 0)
+	  zmpoll_addfd(&fds, &fdscount, notifysocket, -1, &notifyfds);
+
+	/* Although we don't react on the results of these MQ2 fd's
+	   here in main loop, getting them to break timeouts is
+	   important for MAILQv2 responsiveness. ! */
 
 	if (mailqmode == 2)
-	  maxf = mq2add_to_mask(&rdmask, &wrmask, maxf);
+	  mq2add_to_poll(&fds, &fdscount);
 
-	if (maxf < 0)
+	if (fdscount == 0) {
 	  return -1;
+	}
 
-	++maxf;
 	/* sfprintf(sfstderr, "about to select on %x [%d]\n",
 	   mask.fds_bits[0], maxf); */
 
-	in_select = 1;
+	in_poll = 1;
+
+	/* n = select(maxf, &rdmask, &wrmask, NULL, &tv); */
+	n = zmpoll(fds, fdscount, wait_secs * 1000);
+	if (verbose)
+	  sfprintf(sfstderr,"**** QX zmpoll(fds, %d, %d * 1000) = %d; QX ***\n", fdscount, wait_secs, n);
 
-	n = select(maxf, &rdmask, &wrmask, NULL, &tv);
 
 	if (n < 0) {
 	  int err = errno;
@@ -1168,51 +1131,60 @@ time_t timeout;
 	  timed_log_reinit();
 
 	  /* sfprintf(sfstderr, "got an interrupt (%d)\n", errno); */
-	  in_select = 0;
-	  if (errno == EINTR || errno == EAGAIN)
+	  in_poll = 0;
+	  if (err == EINTR || err == EAGAIN) {
 	    return 0;
-	  if (errno == EINVAL || errno == EBADF) {
+	  }
+	  if (err == EINVAL || err == EBADF) {
 	    sfprintf(sfstderr, "** select() returned errno=%d\n", err);
-	    for (i = 0; i < maxf; ++i) {
-	      if (_Z_FD_ISSET(i,rdmask)  &&  fcntl(i,F_GETFL,0) < 0)
-		sfprintf(sfstderr,"** Invalid fd on a select() rdmask: %d\n",i);
-	      if (_Z_FD_ISSET(i,wrmask)  &&  fcntl(i,F_GETFL,0) < 0)
-		sfprintf(sfstderr,"** Invalid fd on a select() wrmask: %d\n",i);
+	    for (i = 0; i < fdscount; ++i) {
+	      if (fds[i].events & ZM_POLLIN && fcntl(fds[i].fd,F_GETFL,0) < 0)
+		sfprintf(sfstderr,"** Invalid fd on a zmpoll() ZM_POLLIN fds: %d\n", fds[i].fd);
+	      if (fds[i].events & ZM_POLLOUT && fcntl(fds[i].fd,F_GETFL,0) < 0)
+		sfprintf(sfstderr,"** Invalid fd on a zmpoll() ZM_POLLOUT fds: %d\n", fds[i].fd);
 	    }
 	    sfsync(sfstderr);
 	    abort(); /* mux() select() error EINVAL or EBADF !? */
 	  }
-	  perror("select() returned unknown error ");
+	  perror("poll() returned unknown error ");
 	  fflush(stderr);
 	  sfsync(sfstderr);
+#if 0
 	  abort(); /* Select with unknown error */
+#endif
 	} else if (n == 0) {
 	  /* sfprintf(sfstderr, "abnormal 0 return from select!\n"); */
 	  /* -- just a timeout -- fast or long */
 	  timed_log_reinit();
-	  in_select = 0;
+	  in_poll = 0;
 	  return 1;
 	} else {
 	  /*sfprintf(sfstderr, "got %d ready (%x)\n", n, rdmask.fds_bits[0]);*/
 
 	  /* In case we really should react.. */
-	  if (querysocket >= 0 &&
-	      _Z_FD_ISSET(querysocket, rdmask))
+	  if (querysocket >= 0 && queryfds &&
+	      queryfds->revents & (ZM_POLLIN|ZM_POLLERR|ZM_POLLHUP))
+	    queryipccheck();
+	  if (querysocket6 >= 0 && query6fds &&
+	      query6fds->revents & (ZM_POLLIN|ZM_POLLERR|ZM_POLLHUP))
 	    queryipccheck();
 	  
-	  if (notifysocket >= 0 &&
-	      _Z_FD_ISSET(notifysocket, rdmask))
+	  if (notifysocket >= 0 && notifyfds &&
+	      notifyfds->revents & (ZM_POLLIN|ZM_POLLERR|ZM_POLLHUP))
 	    receive_notify(notifysocket);
 
 	  if (cpids != NULL) {
 
-	    for (proc = cpids, i = 0; i < maxf; ++i, ++proc) {
+	    for (proc = cpids, i = 0; i < scheduler_nofiles; ++i, ++proc) {
+	      /* sfprintf(sfstderr,"**** QX i = %d QX ***\n", i); */
 	      
 	      timed_log_reinit();
 
 	      if (proc->pid < 0 ||
-		  (proc->pid > 0 && _Z_FD_ISSET(i, rdmask))) {
-		_Z_FD_CLR(i, rdmask);
+	          (proc->pid > 0 && proc->fdpfrom &&
+		   proc->fdpfrom->revents & (ZM_POLLIN|ZM_POLLERR|ZM_POLLHUP))) {
+
+		/* _Z_FD_CLR(i, rdmask); */
 		/*sfprintf(sfstderr,"that is fd %d\n",i);*/
 		/* do non-blocking reads from this fd */
 		readfrom(i);
@@ -1220,7 +1192,8 @@ time_t timeout;
 
 	      /* In case we have non-completed 'feeds', try feeding them */
 	      if (proc->pid > 0    &&  proc->tofd >= 0   &&
-		  proc->cmdlen > 0 &&  _Z_FD_ISSET(proc->tofd, wrmask))
+		  proc->cmdlen > 0 &&
+		  proc->fdpto && proc->fdpto->revents & ZM_POLLOUT)
 		flush_child(proc);
 
 	      /* Because this loop might take a while ... */
@@ -1231,12 +1204,14 @@ time_t timeout;
 	    }
 	  }
 
-	  in_select = 0;
+	  in_poll = 0;
 	}
 	/* sfprintf(sfstderr, "return from mux\n"); */
 	return 0;
 }
 
+
+
 /* Call it how often you wish, but act it only once a second, or so.. */
 static time_t lastqueryipccheck;
 static time_t qipcretry;
@@ -1245,10 +1220,6 @@ void
 queryipccheck()
 {
 	int	n;
-	fd_set	rdmask;
-	fd_set	wrmask;
-	struct timeval tv;
-	int maxfd;
 
 	timed_log_reinit(); /* internal: mytime(&now); */
 
@@ -1264,42 +1235,52 @@ queryipccheck()
 	   */
 	}
 
-	if (querysocket >= 0 || notifysocket >= 0) {
+	if ( (querysocket  >= 0) ||
+	     (querysocket6 >= 0) ||
+	     (notifysocket >= 0)    ) {
 
-	  maxfd = querysocket;
+	  int fdscount;
+	  struct zmpollfd *fds = NULL;
 
-	  tv.tv_sec = 0;
-	  tv.tv_usec = 0;
+	  struct zmpollfd *queryfds = NULL;
+	  struct zmpollfd *query6fds = NULL;
+	  struct zmpollfd *notifyfds = NULL;
+
+	  fdscount = 0;
 
-	  _Z_FD_ZERO(rdmask);
-	  _Z_FD_ZERO(wrmask);
 	  if (querysocket >= 0)
-	    _Z_FD_SET(querysocket, rdmask);
-	  if (notifysocket >= 0) {
-	    _Z_FD_SET(notifysocket, rdmask);
-	    if (notifysocket > maxfd)
-	      maxfd = notifysocket;
-	  }
+	    zmpoll_addfd( &fds, &fdscount, querysocket,  -1, &queryfds);
+
+	  if (querysocket6 >= 0)
+	    zmpoll_addfd( &fds, &fdscount, querysocket6, -1, &query6fds);
+
+	  if (notifysocket >= 0)
+	    zmpoll_addfd( &fds, &fdscount, notifysocket, -1, &notifyfds);
+
 
 	  if (mailqmode == 2) {
-	    maxfd = mq2add_to_mask(&rdmask, &wrmask, maxfd);
-	    n = select(maxfd+1, &rdmask, &wrmask, NULL, &tv);
+	    mq2add_to_poll(&fds, &fdscount);
+
+	    n = zmpoll(fds, fdscount, /* wait: */ 0 /* ms */ );
+
 	    if (n > 0)
-	      mq2_areinsets(&rdmask, &wrmask);
+	      mq2_areinsets(fds);
+
 	  } else 
-	    n = select(maxfd+1, &rdmask, &wrmask, NULL, &tv);
+	    n = zmpoll(fds, fdscount, /* wait: */ 0 /* ms */ );
 
-	  if (n > 0 && notifysocket >= 0 &&
-	      _Z_FD_ISSET(notifysocket, rdmask)) {
+	  if (notifyfds &&
+	      (notifyfds->revents & (ZM_POLLIN|ZM_POLLERR|ZM_POLLHUP)))
 	    receive_notify(notifysocket);
-	  }
+	  if (notifyfds) notifyfds->revents = 0;
+
+	  if (query6fds &&
+	      (query6fds->revents & (ZM_POLLIN|ZM_POLLERR|ZM_POLLHUP))) {
 
-	  if (n > 0 && querysocket >= 0 &&
-	      _Z_FD_ISSET(querysocket, rdmask)) {
 	    Usockaddr raddr;
 	    int raddrlen = sizeof(raddr);
 
-	    n = accept(querysocket, (struct sockaddr *)&raddr, &raddrlen);
+	    n = accept(querysocket6, (struct sockaddr *)&raddr, (void*)&raddrlen);
 	    if (n >= 0) {
 	      if (mailqmode == 1) {
 		int pid;
@@ -1358,6 +1339,61 @@ queryipccheck()
 	      }
 	    }
 	  }
+	  if (query6fds) query6fds->revents = 0;
+
+	  if (queryfds &&
+	      (queryfds->revents & (ZM_POLLIN|ZM_POLLERR|ZM_POLLHUP))) {
+
+	    Usockaddr raddr;
+	    int raddrlen = sizeof(raddr);
+
+	    n = accept(querysocket, (struct sockaddr *)&raddr, (void*)&raddrlen);
+	    if (n >= 0) {
+	      if (mailqmode == 1) {
+		int pid;
+
+		MIBMtaEntry->sc.MQ1sockConnects ++;
+		MIBMtaEntry->sc.MQ1sockParallel ++;
+
+		pid = fork();
+		if (pid == 0) {
+#if defined(F_SETFD)
+		  fcntl(n, F_SETFD, 1); /* close-on-exec */
+#endif
+#ifdef USE_TCPWRAPPER
+#ifdef HAVE_TCPD_H /* TCP-Wrapper code */
+		  if (wantconn(n, "mailq") == 0) {
+		    char *msg = "500 TCP-WRAPPER refusing 'mailq' query from your whereabouts\r\n";
+		    int   len = strlen(msg);
+		    write(n,msg,len);
+		    MIBMtaEntry->sc.MQ1sockParallel --;
+		    MIBMtaEntry->sc.MQ1sockTcpWrapRej ++;
+		    _exit(0);
+		  }
+#endif
+#endif
+		  qprint(n);
+		  close(n);
+		  MIBMtaEntry->sc.MQ1sockParallel --;
+		  /* Silence memory debuggers about this child's
+		     activities by doing exec() on the process.. */
+		  /* execl("/bin/false","false",NULL); */
+		  _exit(0); /* _exit() should be silent, too.. */
+		}
+		if (pid < 0)
+		  MIBMtaEntry->sc.MQ1sockParallel --;
+		close(n);
+	      } else {
+		/* mailqmode == 2 */
+
+		MIBMtaEntry->sc.MQ2sockConnects ++;
+
+		mq2_register(n, &raddr);
+	      }
+	    }
+	  }
+	  if (queryfds) queryfds->revents = 0;
+
 	}
 }
 
@@ -1398,6 +1434,7 @@ queryipcinit()
 	  } else
 	    modecode = 0;
 
+	  /* Actually we support only AF_UNIX for the notify... */
 #ifdef  AF_UNIX
 	  if (modecode == 2) {
 	    struct sockaddr_un sad;
@@ -1440,6 +1477,99 @@ queryipcinit()
 	  break;
 	}
 
+#if defined(AF_INET6) && defined(INET6)
+	while (querysocket6 < 0) {
+
+	  modedata = NULL;
+	  modecode = 0;
+	  if (mailqsock) {
+	    if (cistrncmp(mailqsock,"UNIX:",5)==0) {
+	      modedata = (char *)mailqsock+5;
+	      modecode = 2;
+	    } else if (*mailqsock == '/') {
+	      /* If it begins with '/', it is AF_UNIX socket */
+	      modedata = (char *)mailqsock;
+	      modecode = 2;
+	    } else if (cistrncmp(mailqsock,"TCP:",4)==0) {
+	      modedata = (char *)mailqsock+4;
+	      modecode = 1;
+	    } else {
+	      /* The default mode is TCP/IP socket */
+	      modedata = (char *)mailqsock;
+	      modecode = 1;
+	    }
+	  } else {
+	    /* The default mode is TCP/IP socket */
+	    modedata = (char *)mailqsock;
+	    modecode = 1;
+	  }
+
+	  if (modecode == 1) {
+	    struct servent *serv;
+	    Usockaddr ua;
+	    int on = 1;
+	    int port = 174;
+	    char *modedata2 = NULL;
+
+	    if (modedata) {
+	      modedata2 = strchr(modedata,'@');
+	      if (modedata2) *modedata2++ = 0;
+	    }
+
+	    if (!modedata || !*modedata || sscanf(modedata,"%d",&port) != 1) {
+	      serv = getservbyname(modedata ? modedata : "mailq", "tcp");
+	      if (serv == NULL) {
+		sfprintf(sfstderr, "No 'mailq' tcp service defined!\n");
+		port = 174; /* magic knowledge */
+	      } else
+		port = ntohs(serv->s_port);
+	    }
+	    if (modedata2)  modedata2[-1] = '@';
+
+	    querysocket6 = socket(PF_INET6, SOCK_STREAM, 0);
+	    if (querysocket6 < 0) {
+	      perror("querysocket6: socket(PF_INET6) fail");
+	      break;
+	    }
+
+	    memset( & ua, 0, sizeof(ua) );
+#if 0
+	    zgetbindaddr(NULL, AF_INET6, &ua); /* error returns
+						  without setting
+						  anything. */
+#endif
+	    ua.v6.sin6_family   = AF_INET6;
+	    ua.v6.sin6_flowinfo = 0;
+	    ua.v6.sin6_port     = htons(port);
+
+	    setsockopt(querysocket6, SOL_SOCKET, SO_REUSEADDR, (void*)&on, sizeof(on));
+#ifdef SO_REUSEPORT
+	    setsockopt(querysocket6, SOL_SOCKET, SO_REUSEPORT, (void*)&on, sizeof(on));
+#endif
+
+	    if (bind(querysocket6,(struct sockaddr*)&ua,sizeof(ua.v6)) < 0) {
+	      perror("bind:TCP6 mailq socket");
+	      close(querysocket6);
+	      querysocket6 = -1;
+	      break;
+	    }
+
+	    fd_nonblockingmode(querysocket6);
+
+#if defined(F_SETFD)
+	    fcntl(querysocket6, F_SETFD, 1); /* close-on-exec */
+#endif
+	    if (listen(querysocket6, 5) < 0) {
+	      perror("listen:TCP6 mailq socket");
+	      close(querysocket6);
+	      querysocket6 = -1;
+	      break;
+	    }
+	  }
+	  break;
+	}
+#endif /* defined(AF_INET6) && defined(INET6) */
+
 	while (querysocket < 0) {
 
 	  modedata = NULL;
@@ -1510,8 +1640,8 @@ queryipcinit()
 	      break;
 	    }
 	  }
-#endif
-#ifdef	AF_INET
+#endif /* AF_UNIX */
+#ifdef AF_INET
 	  if (modecode == 1) {
 	    struct servent *serv;
 	    Usockaddr ua;
@@ -1539,47 +1669,18 @@ queryipcinit()
 		ua.v4.sin_addr.s_addr = htonl(INADDR_ANY);
 	    ua.v4.sin_port        = htons(port);
 	    ua.v4.sin_family      = AF_INET;
-	    querysocket = -1;
-#ifdef INET6
-	    querysocket = socket(PF_INET6, SOCK_STREAM, 0);
-#endif
-	    if (querysocket < 0)
-	      querysocket = socket(PF_INET, SOCK_STREAM, 0);
-#ifdef INET6
-	    else {
-	      memset(&ua, 0, sizeof(ua));
-	      if (zgetbindaddr(NULL, AF_INET6, &ua))
-		  ua.v4.sin_addr.s_addr = htonl(INADDR_ANY);
-	      ua.v6.sin6_port   = htons(port);
-	      ua.v6.sin6_family = AF_INET6;
-	    }
-#endif
+	    querysocket = socket(PF_INET, SOCK_STREAM, 0);
 	    if (querysocket < 0) {
-#ifdef INET6
-	      perror("querysocket: socket(PF_INET6/PF_INET)");
-#else
 	      perror("querysocket: socket(PF_INET)");
-#endif
 	      break;
 	    }
 	    setsockopt(querysocket, SOL_SOCKET, SO_REUSEADDR, (void*)&on, sizeof(on));
-
-#ifdef INET6
-	    if (ua.v6.sin6_family == AF_INET6) {
-	      if (bind(querysocket, (struct sockaddr *)&ua, sizeof ua.v6) < 0) {
-		perror("bind:TCP6 mailq socket");
-		close(querysocket);
-		querysocket = -1;
-		break;
-	      }
-	    } else
-#endif
-	      if (bind(querysocket, (struct sockaddr *)&ua, sizeof ua.v4) < 0) {
-		perror("bind:TCP4 mailq socket");
-		close(querysocket);
-		querysocket = -1;
-		break;
-	      }
+	    if (bind(querysocket, (struct sockaddr *)&ua, sizeof ua.v4) < 0) {
+	      perror("bind:TCP4 mailq socket");
+	      close(querysocket);
+	      querysocket = -1;
+	      break;
+	    }
 
 	    fcntl(querysocket, F_SETFL,
 		  fcntl(querysocket, F_GETFL, 0)|O_NONBLOCK);
@@ -1600,54 +1701,15 @@ queryipcinit()
 	  break;
 	}
 
-	if (querysocket >= 0 && notifysocket >= 0)
+	if ((querysocket >= 0 || querysocket6 >= 0) && notifysocket >= 0) {
 	  qipcretry = 0; /* Successfull init done. */
-	else {
+	  
+	} else {
 	  mytime(&now); 
 	  qipcretry = now + 5; /* Will do retry soon.. */
-
 	}
 }
 
-#else	/* !HAVE_SELECT */
-
-int
-mux(timeout)
-time_t timeout;
-{
-	int	fd;
-
-	/*
-	 * Nice'n easy and simpleminded: grab a random file descriptor,
-	 * and sit and read off it until something happens.
-	 * Some very complicated mux'ing schemes (with shared pipes'n stuff)
-	 * are possible in the absence of async i/o like select() or the
-	 * simulation USG supplies, but it ain't worth the hassle.
-	 */
-	readsockcnt = 0;
-	if (cpids != NULL)
-	  for (fd = 0; fd < scheduler_nofiles ; ++fd)
-	    if (cpids[fd].pid != 0) {
-	      readfrom(fd);
-	      ++readsockcnt;
-	    }
-
-	mytime(&now);
-	if (timeout > now)
-	  sleep(1);
-	return 1;
-}
-
-void queryipccheck()
-{
-	/* NOTHING AT ALL -- No select(), no querysocket.. */
-}
-
-void
-queryipcinit()
-{
-}
-#endif	/* HAVE_SELECT */
 
 static void readfrom(fd)
 	int fd;
@@ -1752,6 +1814,7 @@ static void readfrom(fd)
 	free(buf);
 }
 
+#if 0 /* System does not use MKDIR/RMDIR anymore.. */
 #if defined(USE_BINMKDIR) || defined(USE_BINRMDIR)
 
 /*
@@ -1815,6 +1878,7 @@ system(name)
 }
 
 #endif
+#endif
 
 
 #ifdef USE_SIGREAPER
@@ -1835,7 +1899,7 @@ int signum;
 	  pid = waitpid(-1, &statloc, WNOHANG);
 #else
 #ifdef  HAVE_WAIT4
-	  pid = wait4(-1, &statloc, WNOHANG, NULL);
+	  pid = wait4(0, &statloc, WNOHANG, NULL);
 #else
 #ifdef  HAVE_WAIT3
 	  pid = wait3(&statloc, WNOHANG, NULL);
