@@ -1,7 +1,7 @@
 /*
  *   mx_client_verify() -- subroutine for ZMailer smtpserver
  *
- *   By Matti Aarnio <mea@nic.funet.fi> 1997-1999,2002-2003
+ *   By Matti Aarnio <mea@nic.funet.fi> 1997-1999,2002-2004,2006
  */
 
 #include "smtpserver.h"
@@ -9,32 +9,133 @@
 
 extern int use_ipv6;
 
-static int dnsmxlookup __((const char*, int, int, int));
+static int dnsmxlookup __((struct policystate *, const char*, int, int, int));
 
 extern int debug;
 static char * txt_buf = NULL;
 
+struct mxset {
+  int pref;
+  char islocal;
+  char type; /* bitset: 1 | 2 */
+  char *mx;
+};
+
+static int validatev4address __(( const Usockaddr *usap ));
+static int validatev4address(usap)
+     const Usockaddr *usap;
+{
+	/* we validate only ipv4 addresses, ipv6 we let pass as "ok" */
+	if (usap->v4.sin_family == AF_INET) {
+	  unsigned long ipv4addr = ntohl(usap->v4.sin_addr.s_addr);
+
+	  /* Things rejected at this test:
+
+	     - 10.0.0.0 to 10.255.255.255
+	     - 172.16.0.0 to 172.31.255.255
+	     - 192.168.0.0 to 192.168.255.255
+	     - Loopback
+	     - Local-link for auto-DHCP
+	     - IPv4 multicast
+	     - Class E ("Don't Use")
+	     - 0.0.0.0 and 255.255.255.255 are bogus
+	  */
+
+#ifndef IN_CLASSA
+#define IN_CLASSA(a)            ((((unsigned long)(a)) & 0x80000000) == 0)
+#define IN_CLASSB(a)            ((((unsigned long)(a)) & 0xc0000000) == 0x80000000)
+#define IN_CLASSC(a)            ((((unsigned long)(a)) & 0xe0000000) == 0xc0000000)
+#endif
+
+
+	  unsigned int octet1, octet2;
+
+	  if ((! IN_CLASSA(ipv4addr)) &&
+	      (! IN_CLASSB(ipv4addr)) &&
+	      (! IN_CLASSC(ipv4addr)))
+	    return 1; /* Not unicast address classes */
+
+	  octet1 = ( ipv4addr >> 24 ) & 255U;
+	  octet2 = ( ipv4addr >> 16 ) & 255U;
+
+
+	  if ((! IN_CLASSA(ipv4addr)) &&
+	      (! IN_CLASSB(ipv4addr)) &&
+	      (! IN_CLASSC(ipv4addr)))
+	    return 1; /* Not unicast address classes */
+
+	  if (octet1 ==   0 ||
+	      octet1 ==  10 ||
+	      octet1 == 127 ||
+	      octet1 == 255   ) return 1; /* Some of "bad ones" */ 
+	  if (octet1 == 172 &&
+	      (octet2 >= 16 && octet2 <= 31))
+	    return 1; /* private internets, class B */
+
+	  if (octet1 == 192 && octet2 == 168)
+	    return 1; /* private internets, class C */
+
+	  if (octet1 == 169 && octet2 == 254)
+	    return 1; /* Automatic private IP address */
+
+	}
+
+#if defined(AF_INET6) && defined(INET6)
+	if (usap->v6.sin6_family == AF_INET6) {
+	  /* loopback is forbidden, also anything but unicast */
+	}
+#endif
+
+	/* No rejects so far, so we did accept it! */
+	return 0;
+}
+
+static int validatev4addresses __(( const struct addrinfo *aip ));
+static int validatev4addresses(aip)
+     const struct addrinfo *aip;
+{
+	int i;
+
+	for (; aip ; aip = aip->ai_next ) {
+	  i = validatev4address( (Usockaddr *)(aip->ai_addr) );
+	  if (i) return i;
+	}
+	return 0;
+}
+
+/*
+ * return values:
+ *   state->islocaldomain
+ *   ret: 0 = not found ( = reject )
+ *   ret: 1 = FOUND MX MATCH ( = ACCEPT )
+ *   ret: 2 = reject by MX rule
+ *   ret: negatives are EX_**** values.
+ */
+
 static int
-dnsmxlookup(host, depth, mxmode, qtype)
+dnsmxlookup(state, host, depth, mxmode, qtype)
+	struct policystate *state;
 	const char *host;
 	int depth;
 	int mxmode;
 	int qtype;
 {
 	HEADER *hp;
-	msgdata *eom, *cp;
-	int qlen, n, i, qdcount, ancount, nscount, arcount, maxpref, class;
+	msgdata *eom, *cp, *cpnext;
+	int qlen, n, j, qdcount, ancount, nscount, arcount, maxpref, class;
 	u_short type;
 	int saw_cname = 0, had_mx_record = 0;
 	int ttl;
 	struct addrinfo req, *ai;
 #define MAXMX 128
-	char *mx[MAXMX];
-	int   mxtype[MAXMX];
-	int mxcount;
+	struct mxset mxs[MAXMX];
+	int mxcount, mxislocal = 0;
+	int mx_ip_is_invalid = 0;
 	querybuf qbuf, answer;
 	msgdata buf[8192], realname[8192];
 
+	memset( mxs, 0, sizeof(mxs) );
+
 	if (depth == 0)
 	  h_errno = 0;
 
@@ -107,7 +208,9 @@ dnsmxlookup(host, depth, mxmode, qtype)
 	realname[0] = '\0';
 	maxpref = 70000;
 	mxcount = 0;
-	while (--ancount >= 0 && cp < eom) {
+
+	for ( ; --ancount >= 0 && cp < eom; cp = cpnext) {
+
 	  n = dn_expand((msgdata *)&answer, eom, cp, (void*)buf, sizeof buf);
 	  if (n < 0)
 	    break;
@@ -118,38 +221,40 @@ dnsmxlookup(host, depth, mxmode, qtype)
 	  NS_GET32(ttl,   cp); /* ttl   */
 	  NS_GET16(n,     cp); /* dlen  */
 
-	  if (cp + n > eom) {
+	  cpnext = cp + n;
+
+	  if (cpnext > eom) {
 	    /* BAD data.. */
 	    break;
 	  }
 
 	  if (type == T_CNAME) {
-	    cp += dn_expand((msgdata *)&answer, eom, cp,
-			    (void*)realname, sizeof realname);
+	    dn_expand((msgdata *)&answer, eom, cp,
+		      (void*)realname, sizeof realname);
 	    saw_cname = 1;
 	    continue;
 	  }
 
 	  if (type != qtype)  {
 	    /* Not looked for .. */
-	    cp += n;
 	    continue;
 	  }
 
 	  if (type == T_MX) {
-	    cp += 2; /* MX preference value */
+	    int pref;
+	    NS_GET16(pref, cp); /* MX preference value */
 	    n = dn_expand((msgdata *)&answer, eom, cp, (void*)buf, sizeof buf);
 	    if (n < 0)
 	      break;
-	    cp += n;
 
 	    if (debug)
 	      printf("000  MX[%d] = '%s'\n", mxcount, buf);
 
 	    if (mxcount < MAXMX) {
-	      mx[mxcount] = strdup((const char *)buf);
-	      if (!mx[mxcount]) break; /* Out of memory ?? */
-	      mxtype[mxcount] = 0;
+	      mxs[mxcount].mx   = strdup((const char *)buf);
+	      mxs[mxcount].pref = pref;
+	      mxs[mxcount].type = 0;
+	      if (!mxs[mxcount].mx) break; /* Out of memory ?? */
 	      ++mxcount;
 	    }
 	    /* If too many MXes, just skip the rest.. */
@@ -159,7 +264,7 @@ dnsmxlookup(host, depth, mxmode, qtype)
 	  } /* ===== END OF MX DATA PROCESING ========= */
 
 	  if (type == T_TXT) {
-	    int len = (*cp) & 0xFF; /* 0..255 chars */
+	    int i, len = (*cp) & 0xFF; /* 0..255 chars */
 
 	    /* Mal-formed inputs are possible overflowing the buffer.. */
 	    if (len > (eom - cp))
@@ -174,32 +279,34 @@ dnsmxlookup(host, depth, mxmode, qtype)
 	    memcpy(txt_buf, cp, len);
 	    txt_buf[len] = '\0';
 	    for (i = 0; i < mxcount; ++i) {
-	      if (mx[i]) free(mx[i]);
-	      mx[i] = NULL;
+	      if (mxs[i].mx) free(mxs[i].mx);
+	      mxs[i].mx = NULL;
 	    }
 	    return 1; /* OK! */
 	  }
 
 	  /* If reached here, skip the data tail */
 	  /* In theory could be an abort even..  */
-	  cp += n;
+
 	} /* ===== END OF DNS ANSWER PROCESSING ======= */
 
 
 	if (ancount > 0) {
 	  /* Sigh, waste of time :-( */
-	  for (i = 0; i < mxcount; ++i) if (mx[i]) free(mx[i]);
+	  int i;
+	  for (i = 0; i < mxcount; ++i) if (mxs[i].mx) free(mxs[i].mx);
 	  return -EX_SOFTWARE;
 	}
 
 
+#if 0
 	if (qtype == T_MX && !mxmode && had_mx_record) {
 	  /* Accept if found ANYTHING! */
 	  if (debug) printf("000-  ... accepted!\n");
-	  for (i = 0; i < mxcount; ++i) if (mx[i]) free(mx[i]);
+	  for (i = 0; i < mxcount; ++i) if (mxs[i].mx) free(mxs[i].mx);
 	  return 1;
 	}
-
+#endif
 
 	/* Now skip the AUTHORITY SECTION data */
 
@@ -225,8 +332,8 @@ dnsmxlookup(host, depth, mxmode, qtype)
 	}
 
 	if (debug)
-	  printf("000-  nscount=%d (== 0 ?)  arcount=%d cp<eom ? (%d)\n",
-		 nscount, arcount, cp < eom);
+	  printf("000-  nscount=%d (== 0 ?)  arcount=%d  eom-cp=%d\n",
+		 nscount, arcount, (int)(eom-cp));
 
 	/* Ok, can continue to pick the ADDITIONAL SECTION data */
 
@@ -234,7 +341,10 @@ dnsmxlookup(host, depth, mxmode, qtype)
 	   look for the 'AA' bit.  If it isn't set, we don't use this
 	   data, but do explicite lookups below. */
 
-	while (hp->aa && nscount == 0 && arcount > 0 && cp < eom) {
+	for ( ;
+	      hp->aa && nscount == 0 && arcount > 0 && cp < eom;
+	      cp = cpnext) {
+
 	  n = dn_expand((msgdata *)&answer, eom, cp, (void*)buf, sizeof buf);
 	  if (n < 0) { cp = eom; break; }
 	  cp += n;
@@ -245,16 +355,17 @@ dnsmxlookup(host, depth, mxmode, qtype)
 	  cp += NS_INT32SZ;    /* ttl   - long  */
 	  NS_GET16(n, cp);     /* dlen  - short */
 
-	  if (cp + n > eom) { cp = eom; break; }
+	  cpnext = cp + n;
+
+	  if (cpnext > eom)    { continue; /* BAD BAD! */ }
 
 	  if (class != C_IN) {
-	    cp += n; --nscount;
+	    --arcount;
 	    continue;
 	  }
 
 	  /* Ok, we have Type IN data in the ADDITIONAL SECTION */
 
-
 	  /* A and AAAA are known here! */
 
 	  if (type == T_A
@@ -264,23 +375,21 @@ dnsmxlookup(host, depth, mxmode, qtype)
 	      ) {
 
 	    Usockaddr usa;
+	    
+	    --arcount;
 
 	    /* Pick the address data */
-	    for (i = 0; i < mxcount; ++i) {
+	    for (n = 0; n < mxcount; ++n) {
 	      /* Is this known (wanted) name ?? */
-	      if (strcasecmp((const char *)buf, mx[i]) == 0) {
+	      if (strcasecmp((const char *)buf, mxs[n].mx) == 0) {
 		/* YES! */
 
-		mxtype[i] |= (type == T_A) ? 1 : 2 ; /* bitflag: 1 or 2 */
+		mxs[n].type |= (type == T_A) ? 1 : 2 ; /* bitflag: 1 or 2 */
 
 		/* We do have a wanted name! */
 
 		/* build addrinfo block, pick addresses */
 
-		/* WARNING! WARNING! WARNING! WARNING! WARNING! WARNING!
-		   This assumes intimate knowledge of the system
-		   implementation of the  ``struct addrinfo'' !  */
-
 		memset(&usa, 0, sizeof(usa));
 
 		switch (type) {
@@ -302,36 +411,52 @@ dnsmxlookup(host, depth, mxmode, qtype)
 		if (debug) {
 		  if (usa.v4.sin_family == AF_INET) {
 		    inet_ntop(AF_INET, (void*) & usa.v4.sin_addr, (char *)buf, sizeof(buf));
-		    printf("000-  matching %s AR address IPv4:[%s]\n", mx[i], buf);
+		    printf("000-  matching %s AR address IPv4:[%s]\n", mxs[n].mx, buf);
 		  }
 #if defined(AF_INET6) && defined(INET6)
 		  else if (usa.v6.sin6_family == AF_INET6) {
-		    inet_ntop(AF_INET6, (void*) & usa.v6.sin6_addr, buf, sizeof(buf));
-		    printf("000-  matching %s AR address IPv6:[%s]\n", mx[i], buf);
+		    inet_ntop(AF_INET6, (void*) & usa.v6.sin6_addr, (char*)buf, sizeof(buf));
+		    printf("000-  matching %s AR address IPv6:[%s]\n", mxs[n].mx, buf);
 		  }
 #endif
 		  else
 		    printf("000- matching unknown %s AR address family address; AF=%d\n",
-			   mx[i], usa.v4.sin_family);
+			   mxs[n].mx, usa.v4.sin_family);
 		}
 #endif
 
-		i = matchmyaddress( &usa );
-		if (i == 1) {
+		j = validatev4address( &usa );
+		if (j) {
+		  /* Address is not valid, */
+		  mx_ip_is_invalid = 1;
+		}
+
+		j = matchmyaddress( &usa );
+		if (j == 1) {
 		  if (debug)
 		    printf("000-   AR ADDRESS MATCH!\n");
-		  for (i = 0; i < mxcount; ++i) if (mx[i]) free(mx[i]);
-		  return 1; /* Found a match! */
+		  mxs[n].islocal = 1;
+		  mxislocal = 1;
+		  /* Found a match! */
+		  goto ponder_mx_result;
+		} else if (j == 2) {
+		  if (debug)
+		    printf("000-   AR ADDRESS LOOPBACK MATCH!\n");
+		  mxs[n].islocal = 2;
+		  mxislocal = 2;
+		  /* Found a match! */
+		  goto ponder_mx_result;
 		} else
 		  if (debug)
-		    printf("000-   AR matchmyaddress() yields: %d\n", i);
+		    printf("000-   AR matchmyaddress() yields: %d\n", j);
 
 		break; /* Name matched, no need to spin more here.. */
 	      } /* Matched name! */
 	    } /* Name matching loop */
+	    continue;
 	  } /* type = T_A or T_AAAA */
 
-	  cp += n;
+	  /* All other cases.. */
 	  --arcount;
 	} /* Additional data collected! */
 
@@ -345,7 +470,7 @@ dnsmxlookup(host, depth, mxmode, qtype)
 
 	  memset(&req, 0, sizeof(req));
 
-	  switch(mxtype[n]) {
+	  switch(mxs[n].type) {
 	  case 0: /* no addresses seen! */
 	    req.ai_family   = 0; /* Both OK (IPv4/IPv6) */
 	    /* Definitely ask for it! */
@@ -377,26 +502,18 @@ dnsmxlookup(host, depth, mxmode, qtype)
 	  /* This resolves CNAME, it should not happen in case
 	     of MX server, though..    */
 #ifdef HAVE__GETADDRINFO_
-	  i = _getaddrinfo_(mx[n], "0", &req, &ai,
+	  rc = _getaddrinfo_(mxs[n].mx, "0", &req, &ai,
 			    debug ? stdout : NULL);
 #else
-	  i = getaddrinfo(mx[n], "0", &req, &ai);
+	  rc = getaddrinfo(mxs[n].mx, "0", &req, &ai);
 #endif
 	  if (debug)
-	    printf("000-  getaddrinfo('%s','0') -> r=%d, ai=%p\n",mx[n],i,ai);
+	    printf("000-  getaddrinfo('%s','0') -> r=%d, ai=%p\n",
+		   mxs[n].mx,rc,(void*)ai);
 	    
-	  if (i != 0)
+	  if (rc != 0)
 	    continue;		/* Well well.. spurious! */
 
-	  if (!mxmode) /* Accept if found ANYTHING! */ {
-	    if (debug) printf("000-  ... accepted!\n");
-	    freeaddrinfo(ai);
-	    for (i = 0; i < mxcount; ++i) if (mx[i]) free(mx[i]);
-
-	    return 1;
-	  }
-	  
-	    
 	  for ( ai2 = ai ; ai2 != NULL; ai2 = ai2->ai_next) {
 	    ++k;
 #if 1
@@ -406,26 +523,45 @@ dnsmxlookup(host, depth, mxmode, qtype)
 
 	      if (usa->v4.sin_family == AF_INET) {
 		inet_ntop(AF_INET, (void*) & usa->v4.sin_addr, buf, sizeof(buf));
-		printf("000-  matching %s address IPv4:[%s]\n", mx[n], buf);
+		printf("000-  matching %s address IPv4:[%s]\n",
+		       mxs[n].mx, buf);
 	      }
 #if defined(AF_INET6) && defined(INET6)
 	      else if (usa->v6.sin6_family == AF_INET6) {
 		inet_ntop(AF_INET6, (void*) & usa->v6.sin6_addr, buf, sizeof(buf));
-		printf("000-  matching %s address IPv6:[%s]\n", mx[n], buf);
+		printf("000-  matching %s address IPv6:[%s]\n",
+		       mxs[n].mx, buf);
 	      }
 #endif
 	      else
 		printf("000- matching %s unknown address family address; AF=%d\n",
-		       mx[n], usa->v4.sin_family);
+		       mxs[n].mx, usa->v4.sin_family);
 	    }
 #endif
-	    rc = matchmyaddress((Usockaddr *)ai2->ai_addr);
+
+	    rc = validatev4address( (Usockaddr *)ai2->ai_addr );
+	    if (rc) {
+	      /* Address is not valid, */
+	      mx_ip_is_invalid = 1;
+	    }
+
+	    rc = matchmyaddress( (Usockaddr *)ai2->ai_addr );
 	    if (rc == 1) {
 	      if (debug)
 		printf("000-   ADDRESS MATCH!\n");
 	      freeaddrinfo(ai);
-	      for (i = 0; i < mxcount; ++i) if (mx[i]) free(mx[i]);
-	      return 1; /* Found a match! */
+	      mxs[n].islocal = 1;
+	      mxislocal = 1;
+	      /* Found a match! */
+	      goto ponder_mx_result;
+	    } else if (rc == 2) {
+	      if (debug)
+		printf("000-   LOOPBACK ADDRESS MATCH!\n");
+	      freeaddrinfo(ai);
+	      mxs[n].islocal = 2;
+	      mxislocal = 2;
+	      /* Found a match! */
+	      goto ponder_mx_result;
 	    } else
 	      if (debug)
 		printf("000-   matchmyaddress() yields: %d\n", rc);
@@ -434,20 +570,51 @@ dnsmxlookup(host, depth, mxmode, qtype)
 	    printf("000-   No address match among %d address!\n", k);
 
 	  freeaddrinfo(ai);
+
+	  if (!mxmode) /* Accept if found ANYTHING! */ {
+	    int i;
+
+	    if (debug) printf("000-  ... accepted!\n");
+	    for (i = 0; i < mxcount; ++i)
+	      if (mxs[i].mx)
+		free(mxs[i].mx);
+
+	    if (mx_ip_is_invalid) return -EX_TEMPFAIL;
+
+	    return 1;
+	  }
+	} /* Thru all MXS[] ... */
+
+	/* No MX match found.. */
+	for (n = 0; n < mxcount; ++n) {
+	  if (mxs[n].mx)
+	    free(mxs[n].mx);
+	  mxs[n].mx = NULL;
 	}
 
-	for (i = 0; i < mxcount; ++i) if (mx[i]) free(mx[i]);
+	if (debug)
+	  printf("000-   saw_cname=%d  had_mx_record=%d  mxmode=%d  mxislocal=%d mx_ip_is_invalid=%d\n",
+		 saw_cname, had_mx_record, mxmode, mxislocal, mx_ip_is_invalid);
+
+	if (mx_ip_is_invalid) return -EX_TEMPFAIL;
 
 
 	/* Didn't find any, but saw CNAME ? Recurse with the real name */
 	if (saw_cname)
-	  return dnsmxlookup((void *)realname, depth+1, mxmode, qtype);
+	  return dnsmxlookup(state, (void *)realname, depth+1, mxmode, qtype);
 
 	if (had_mx_record && mxmode)
 	    return 2; /* We have SOME date, but no match on ourselves! */
 
+	if (!mxmode)
+	  return 1; /* Accept anything! */
+
+	return 0; /* No match, but had MXes.. */
+
 perhaps_address_record:
 	if (qtype == T_MX) {
+	  int i;
+
 	  /* No MX, perhaps A ? */
 	  memset(&req, 0, sizeof(req));
 	  req.ai_socktype = SOCK_STREAM;
@@ -456,7 +623,6 @@ perhaps_address_record:
 	  req.ai_family   = PF_INET;
 	  ai = NULL;
 
-
 	  /* This resolves CNAME, it should not happen in case
 	     of MX server, though..    */
 #ifdef HAVE__GETADDRINFO_
@@ -467,7 +633,7 @@ perhaps_address_record:
 	  i = getaddrinfo((const char*)host, "0", &req, &ai);
 #endif
 	  if (debug)
-	    printf("000-   getaddrinfo('%s','0') (PF_INET) -> r=%d (%s), ai=%p\n",host,i,gai_strerror(i),ai);
+	    printf("000-   getaddrinfo('%s','0') (PF_INET) -> r=%d (%s), ai=%p\n",host,i,gai_strerror(i),(void*)ai);
 
 #if defined(AF_INET6) && defined(INET6)
 	  if (use_ipv6) {
@@ -492,7 +658,7 @@ perhaps_address_record:
 	    n2 = getaddrinfo((const char *)host, "0", &req, &ai2);
 #endif
 	    if (debug)
-	      printf("000-   getaddrinfo('%s','0') (PF_INET6) -> r=%d (%s), ai=%p\n",host,n2,gai_strerror(n2),ai2);
+	      printf("000-   getaddrinfo('%s','0') (PF_INET6) -> r=%d (%s), ai=%p\n",host,n2,gai_strerror(n2),(void*)ai2);
 
 
 	    if (i != 0 && n2 == 0) {
@@ -511,25 +677,34 @@ perhaps_address_record:
 #endif
 
 	  if (i)
-	    return i;
+	    return -EX_TEMPFAIL; /* Bad lookup result -> no match */
 
+	  i = validatev4addresses( ai );
+	  if (i) {
+	    /* Address is not valid, */
+	    freeaddrinfo(ai);
+	    return -EX_TEMPFAIL;
+	  }
 
 	  i = matchmyaddresses(ai);
+
+	  freeaddrinfo(ai);
+
+	  if (i > 0)
+	    state->islocaldomain = 1;
 #if 1
 	  /* With this we can refuse to accept any message with
 	     source domain pointing back to loopback ! */
 	  if (i == 2) {
 	    /* Loopback ! */
-	    freeaddrinfo(ai);
-	    return 0;
+	    state->islocaldomain = 2;
+	    return 2;
 	  }
 #endif
 	  if (i == 0 && mxmode) {
-	    freeaddrinfo(ai);
 	    return 2; /* Didn't find our local address in client-MX-mode */
 	  }
 
-	  freeaddrinfo(ai);
 	  return 1; /* Found any address, or in client-MX-mode,
 		       a local address! */
 	}
@@ -538,6 +713,40 @@ perhaps_address_record:
 	  return 2; /* Not found, had no MX data either */
 	
 	return 0; /* Not found! */
+
+ ponder_mx_result:;
+	/* Ok, we have some MX match, lets see closer..
+	   If we have...
+	   - TODO: What shall we do with MX server sets ?
+	   - Can we set  state->islocaldomain   ??
+	*/
+#if 0
+	{
+	  int lowpref = 100000;
+
+	  for (n = 0; n < mxcount; ++n) {
+	    if ((mxs[n].pref < lowpref) && mxs[n].islocal)
+	      lowpref = mxs[n].pref;
+	  }
+
+	  for (n = 0; n < mxcount; ++n) {
+	    if (mxs[n].pref < lowpref)
+	      return 0;
+	  }
+
+	}
+#endif
+
+	for (n = 0; n < mxcount; ++n) {
+	  if (mxs[n].mx)
+	    free(mxs[n].mx);
+	}
+
+	/* If it was determined to be invalid... */
+	if (mx_ip_is_invalid) return -EX_TEMPFAIL;
+
+	/* Report as successfull match, didn't set  'state->islocaldomain' */
+	return 1;
 }
 
 
@@ -546,7 +755,8 @@ perhaps_address_record:
    For (retmode == '+'), and without MX, return 1.
  */
 
-int mx_client_verify(retmode, domain, alen)
+int mx_client_verify(state, retmode, domain, alen)
+     struct policystate *state;
      int retmode, alen;
      const char *domain;
 {
@@ -559,28 +769,29 @@ int mx_client_verify(retmode, domain, alen)
 	strncpy(hbuf, domain, alen);
 	hbuf[alen] = 0; /* Chop off the trailers from the name */
 
-	rc = dnsmxlookup(hbuf, 0, 1, T_MX);
+	rc = dnsmxlookup(state, hbuf, 0, 1, T_MX);
 
 	if (rc == 1) return 0; /* Found! */
 
 	if (rc == -EX_TEMPFAIL) {
 	  return -104;
 	}
+
+	if (rc == 2)
+	  return -3; /* Hard error, all the same */
+
 	if (retmode == '+') {
 	  if (rc == -EX_NOHOST ||
 	      rc == -EX_UNAVAILABLE)
 	    return -2; /* Definitely hard errors */
-	  if (rc == 2)
-	    return -103;
 	  return -102; /* Soft error */
 	}
 
-	if (rc == 2)
-	  return -3;
 	return -2;     /* Reject */
 }
 
-int sender_dns_verify(retmode, domain, alen)
+int sender_dns_verify(state, retmode, domain, alen)
+     struct policystate *state;
      int retmode, alen;
      const char *domain;
 {
@@ -593,7 +804,7 @@ int sender_dns_verify(retmode, domain, alen)
 	strncpy(hbuf, domain, alen);
 	hbuf[alen] = 0; /* Chop off the trailers from the name */
 
-	rc = dnsmxlookup(hbuf, 0, 0, T_MX);
+	rc = dnsmxlookup(state, hbuf, 0, 0, T_MX);
 
 	if (debug)
 	  printf("000- dnsmxlookup() did yield: %d, retmode='%c'\n",
@@ -604,30 +815,34 @@ int sender_dns_verify(retmode, domain, alen)
 	if (rc == -EX_TEMPFAIL) {
 	  return -104;
 	}
+
+	if (rc == 2)
+	  return -3; /* Hard error all the same */
+
 	if (retmode == '+') {
 	  if (rc == -EX_NOHOST      ||
 	      rc == -EX_UNAVAILABLE ||
+#ifdef EAI_NODATA
 	      rc == EAI_NODATA      ||
+#endif
 	      rc == EAI_NONAME)
 	    return -2; /* Definitely hard errors */
-	  if (rc == 2)
-	    return -103;
 	  return -102; /* Soft error */
 	}
 
-	if (rc == 2)
-	  return -3;
 	return -2;     /* Reject */
 }
 
-int client_dns_verify(retmode, domain, alen)
+int client_dns_verify(state, retmode, domain, alen)
+     struct policystate *state;
      int retmode, alen;
      const char *domain;
 {
-	return sender_dns_verify(retmode, domain, alen);
+	return sender_dns_verify(state,retmode, domain, alen);
 }
 
-int rbl_dns_test(ipaf, ipaddr, rbldomain, msgp)
+int rbl_dns_test(state, ipaf, ipaddr, rbldomain, msgp)
+     struct policystate *state;
      const int ipaf;
      const u_char *ipaddr;
      char *rbldomain;
@@ -665,9 +880,11 @@ int rbl_dns_test(ipaf, ipaddr, rbldomain, msgp)
 	     size...  */
 	  s = strchr(rbldomain, ':');
 	  if (s) *s = 0;
+#if 0 /* Remove the old hard-coded "+" macro ... */
 	  if (strcmp(rbldomain,"+") == 0)
 	    strncpy (suf, "rbl.maps.vix.com", hspc);
 	  else
+#endif
 	    strncpy (suf, rbldomain, hspc);
 	  suf[hspc] = '\0';
 
@@ -717,7 +934,7 @@ int rbl_dns_test(ipaf, ipaddr, rbldomain, msgp)
 	    if (debug)
 	      printf("000- looking up DNS TXT object: %s\n", hbuf);
 
-	    if (dnsmxlookup(hbuf, 0, 0, T_TXT) == 1) {
+	    if (dnsmxlookup(state, hbuf, 0, 0, T_TXT) == 1) {
 	      if (*msgp != NULL)
 		free(*msgp);
 	      *msgp = strdup(txt_buf);
@@ -731,7 +948,8 @@ int rbl_dns_test(ipaf, ipaddr, rbldomain, msgp)
 		    *s = ' ';
 		}
 	      }
-	      type(NULL,0,NULL,"Found DNS TXT object: %s\n", *msgp ? *msgp : "<nil>");
+	      type(NULL,0,NULL,"Found DNS TXT object: %s\n",
+		   (*msgp ? *msgp : "<nil>"));
 	    }
 	    return -1;
 	  }
