@@ -46,16 +46,8 @@ static const char *no_input = "No input";
 	(d?CHAR_DIGIT:0)|(e?CHAR_C:0)|(f?CHAR_X:0)|(g?CHAR_Q:0)| \
 	(h?CHAR_XDIGIT:0)|(i?CHAR_822ATM:0)|(j?CHAR_XCHAR:0)
 
-/* Could use 'unsigned short' here, but the Alpha machines dislike
-   anything smaller than int (32 bit) */
-static
-#if defined(__alpha)||defined(__alpha__)
- int
-#else
- short
-#endif
- char_array[] =
-{
+static const short int char_array[] = {
+
 		       /* /------------------------------ CHAR_ALPHA
 			  |  /--------------------------- CHAR_ALNUM
 			  |  |  /------------------------ CHAR_SPECL
@@ -101,7 +93,7 @@ static
 /*  31: `^_'   */ CHR_ENT(0, 0, 1, 0, 0, 1, 1, 0, 0, 0),
 /*  32: ` '    */ CHR_ENT(0, 0, 0, 0, 0, 1, 1, 0, 0, 0),
 /*  33: `!'    */ CHR_ENT(0, 0, 0, 0, 1, 1, 1, 0, 1, 1),
-/*  34: `"'    */ CHR_ENT(0, 0, 1, 0, 0, 1, 0, 0, 1, 1),
+/*  34: `"'    */ CHR_ENT(0, 0, 1, 0, 0, 1, 0, 0,-0, 1),
 /*  35: `#'    */ CHR_ENT(0, 0, 0, 0, 1, 1, 1, 0, 1, 1),
 /*  36: `$'    */ CHR_ENT(0, 0, 0, 0, 1, 1, 1, 0, 1, 1),
 /*  37: `%'    */ CHR_ENT(0, 0, 0, 0, 1, 1, 1, 0, 1, 1),
@@ -113,7 +105,7 @@ static
 /*  43: `+'    */ CHR_ENT(0, 0, 0, 0, 1, 1, 1, 0, 1, 0),
 /*  44: `,'    */ CHR_ENT(0, 0, 1, 0, 0, 1, 1, 0, 0, 1),
 /*  45: `-'    */ CHR_ENT(0, 0, 0, 0, 1, 1, 1, 0, 1, 1),
-/*  46: `.'    */ CHR_ENT(0, 0, 1, 0, 0, 1, 1, 0, 1, 1),
+/*  46: `.'    */ CHR_ENT(0, 0, 1, 0, 0, 1, 1, 0,-0, 1),
 /*  47: `/'    */ CHR_ENT(0, 0, 0, 0, 1, 1, 1, 0, 1, 1),
 /*  48: `0'    */ CHR_ENT(0, 1, 0, 1, 1, 1, 1, 1, 1, 1),
 /*  49: `1'    */ CHR_ENT(0, 1, 0, 1, 1, 1, 1, 1, 1, 1),
@@ -131,7 +123,7 @@ static
 /*  61: `='    */ CHR_ENT(0, 0, 0, 0, 1, 1, 1, 0, 1, 0),
 /*  62: `>'    */ CHR_ENT(0, 0, 1, 0, 0, 1, 1, 0, 0, 1),
 /*  63: `?'    */ CHR_ENT(0, 0, 0, 0, 1, 1, 1, 0, 1, 1),
-/*  64: `@'    */ CHR_ENT(0, 0, 1, 0, 0, 1, 1, 0, 1, 1),
+/*  64: `@'    */ CHR_ENT(0, 0, 1, 0, 0, 1, 1, 0,-0, 1),
 /*  65: `A'    */ CHR_ENT(1, 1, 0, 0, 1, 1, 1, 1, 1, 1),
 /*  66: `B'    */ CHR_ENT(1, 1, 0, 0, 1, 1, 1, 1, 1, 1),
 /*  67: `C'    */ CHR_ENT(1, 1, 0, 0, 1, 1, 1, 1, 1, 1),
@@ -334,55 +326,55 @@ static
 __MY_INLINE__ int is_821_alpha(chr)
 unsigned int chr;
 {
-    return (char_array[chr] & CHAR_ALPHA);
+    return (char_array[chr & 0xFF] & CHAR_ALPHA);
 }
 
 __MY_INLINE__ int is_821_alnum(chr)
 unsigned int chr;
 {
-    return (char_array[chr] & CHAR_ALNUM);
+    return (char_array[chr & 0xFF] & CHAR_ALNUM);
 }
 
 __MY_INLINE__ int is_821_specl(chr)
 unsigned int chr;
 {
-    return (char_array[chr] & CHAR_SPECL);
+    return (char_array[chr & 0xFF] & CHAR_SPECL);
 }
 
 __MY_INLINE__ int is_821_digit(chr)
 unsigned int chr;
 {
-    return (char_array[chr] & CHAR_DIGIT);
+    return (char_array[chr & 0xFF] & CHAR_DIGIT);
 }
 
 __MY_INLINE__ int is_821_xdigit(chr)
 unsigned int chr;
 {
-    return (char_array[chr] & CHAR_XDIGIT);
+    return (char_array[chr & 0xFF] & CHAR_XDIGIT);
 }
 
 __MY_INLINE__ int is_821_C(chr)
 unsigned int chr;
 {
-    return (char_array[chr] & CHAR_C);
+    return (char_array[chr & 0xFF] & CHAR_C);
 }
 
 __MY_INLINE__ int is_821_X(chr)
 unsigned int chr;
 {
-    return (char_array[chr] & CHAR_X);
+    return (char_array[chr & 0xFF] & CHAR_X);
 }
 
 __MY_INLINE__ int is_821_Q(chr)
 unsigned int chr;
 {
-    return (char_array[chr] & CHAR_Q);
+    return (char_array[chr & 0xFF] & CHAR_Q);
 }
 
 __MY_INLINE__ int is_822_atomchar(chr)
 unsigned int chr;
 {
-    return (char_array[chr] & CHAR_822ATM);
+    return (char_array[chr & 0xFF] & CHAR_822ATM);
 }
 
 /* ================================================================ */
@@ -853,8 +845,9 @@ int strict;
 	return s;		/* Pathological termination */
 
     q = rfc821_string(p, strict);
-    if (q == p)
-	return s;		/* Missing string */
+    if ((q == p) && strict && (*p == '.'))
+      return s;		/* Missing string */
+
     while (*q == '.') {		/* Well, intermediate dot.. */
 	p = q + 1;
 #if 0
@@ -865,13 +858,12 @@ int strict;
 	}
 #endif
 	q = rfc821_string(p, strict);
-	if (q == p) {
+	if (strict && (q == p) && (*p == '.')) {
 	  if (*q == '@') {
 	    rfc821_error = "Localpart must not end with unquoted dot!";
 	  }
 	  return s;		/* Missing string */
 	}
-	
     }
     return q;
 }
