@@ -2,7 +2,7 @@
  *  ZMailer smtpserver,  AUTH command things (RFC 2554, sort of);
  *  part of ZMailer.
  *
- *  by Matti Aarnio <mea@nic.funet.fi> 1999,2002,2003
+ *  by Matti Aarnio <mea@nic.funet.fi> 1999,2002,2003,2005
  *
  *  The basis of SASL[2] code is from Sendmail 8.12.3
  *
@@ -214,54 +214,92 @@ void smtp_auth(SS,buf,cp)
 
 	char c, co;
 	int i, rc;
-	char *uname;
+	char *uname = NULL;
 	long uid;
 	char *zpw;
 
-	if (SS->authuser != NULL) {
-	  type(SS, 503, m551, "Already authenticated, second attempt rejected!");
+	if (OCP->no_smtp_auth_on_25 &&
+	    (SS->with_protocol_set & WITH_SMTP)) {
+	  smtp_tarpit(SS); /* Double tarpit.. */
+	  smtp_tarpit(SS);
+	  type(SS, 503, m551, "Hi %s, AUTH NOT SUPPORTED AT PORT 25!", SS->rhostaddr);
+	  return;
+	}
+
+	rc = policytest(&SS->policystate, POLICY_AUTHFAIL,
+			uname, 0, SS->authuser);
+	if (rc < 0) {
+	  smtp_tarpit(SS);
+	  type(SS, 503, m551, "Hi %s, Too many authentication failures from your IP this hour..", SS->rhostaddr);
 	  return;
 	}
 
 	if (SS->state == Hello) {
-	  type(SS, 503, m551, "EHLO first, then - perhaps - AUTH!");
+	  policytest(&SS->policystate, POLICY_AUTHFAIL,
+		     uname, 1, SS->authuser);
+	  smtp_tarpit(SS);
+	  type(SS, 503, m551, "Hi %s, EHLO first, then - perhaps - AUTH!", SS->rhostaddr);
 	  return;
 	}
 	if (SS->state != MailOrHello && SS->state != Mail) {
-	  type(SS, 503, m551, "AUTH not allowed during MAIL transaction!");
+	  policytest(&SS->policystate, POLICY_AUTHFAIL,
+		     uname, 1, SS->authuser);
+	  smtp_tarpit(SS);
+	  type(SS, 503, m551, "Hi %s, AUTH not allowed during MAIL transaction!", SS->rhostaddr);
+	  return;
+	}
+	if (SS->authuser) {
+	  policytest(&SS->policystate, POLICY_AUTHFAIL,
+		     uname, 1, SS->authuser);
+	  smtp_tarpit(SS);
+	  type(SS, 503, m551, "Hello %s, already authenticated, second attempt rejected!",SS->rhostaddr);
 	  return;
 	}
 
 	if (*cp == ' ') ++cp;
-	if (!strict_protocol) while (*cp == ' ' || *cp == '\t') ++cp;
+	if (strict_protocol < 1)
+	  while (*cp == ' ' || *cp == '\t') ++cp;
 
 #ifdef HAVE_SASL2
-	if (!do_sasl)
+	if (!OCP->do_sasl)
 #endif
 	  {
 	    if (!CISTREQN(cp, "LOGIN", 5)) {
-	      type(SS, 504, m571, "Only 'AUTH LOGIN' supported.");
+	      policytest(&SS->policystate, POLICY_AUTHFAIL,
+			 uname, 1, SS->authuser);
+	      smtp_tarpit(SS);
+	      type(SS, 504, m571, "Hello %s, Only 'AUTH LOGIN' supported.", SS->rhostaddr);
 	      return;
 	    }
 
 #ifdef HAVE_OPENSSL
-	    if (!auth_login_without_tls && !SS->sslmode) {
+	    if (!OCP->auth_login_without_tls && !SS->sslmode) {
+	      policytest(&SS->policystate, POLICY_AUTHFAIL,
+			 uname, 1, SS->authuser);
+	      smtp_tarpit(SS);
 	      type(SS, 503, m571,
-		   "Plaintext password authentication must be run under SSL/TLS");
+		   "Hello %s, Plaintext password authentication must be run under SSL/TLS", SS->rhostaddr);
 	      return;
 	    }
 #endif /* - HAVE_OPENSSL */
 #ifndef HAVE_OPENSSL
-	    if (!auth_login_without_tls) {
+	    if (!OCP->auth_login_without_tls) {
+	      policytest(&SS->policystate, POLICY_AUTHFAIL,
+			 uname, 1, SS->authuser);
+	      smtp_tarpit(SS);
 	      type(SS, 503, m571,
-		   "Plaintext password authentication is not enabled in this system");
+		   "Hello %s, Plaintext password authentication is not enabled in this system",
+		   SS->rhostaddr);
 	      return;
 	    }
 #endif /* --HAVE_OPENSSL */
 
 	    cp += 5;
 	    if (*cp == ' ') ++cp;
-	    if (!strict_protocol) while (*cp == ' ' || *cp == '\t') ++cp;
+	    if (strict_protocol < 1)
+	      while (*cp == ' ' || *cp == '\t') ++cp;
+
+	    uname = NULL;
 	    
 	    if (*cp != 0) {
 
@@ -270,15 +308,25 @@ void smtp_auth(SS,buf,cp)
 	      bbuf[sizeof(bbuf)-1] = 0;
 	      if (debug)
 		type(SS, 0, NULL, "-> %s", bbuf);
-	      uname = strdup(bbuf);
+
 	      if (*ccp != 0) {
+		policytest(&SS->policystate, POLICY_AUTHFAIL,
+			   uname, 1, SS->authuser);
+		smtp_tarpit(SS);
 		type(SS, 501, m552, "unrecognized input/extra junk ??");
 		return;
 	      }
 
+	      uname = strdup(bbuf);
+
 	    } else {
 	      
-	      i = encodebase64string("Username:", 9, abuf, sizeof(abuf));
+	      if (!OCP->smtp_auth_username_prompt)
+		OCP->smtp_auth_username_prompt = "Username:";
+
+	      i = encodebase64string(OCP->smtp_auth_username_prompt,
+				     strlen(OCP->smtp_auth_username_prompt),
+				     abuf, sizeof(abuf));
 	      if (i >= sizeof(abuf)) i = sizeof(abuf)-1;
 	      abuf[i] = 0;
 	      type(SS, 334, NULL, "%s", abuf);
@@ -295,9 +343,14 @@ void smtp_auth(SS,buf,cp)
 		fprintf(logfp, "%s%04dr\t%s\n", logtag, (int)(now - logtagepoch), abuf);
 		fflush(logfp);
 	      }
-	      if (i == 0)	/* EOF ??? */
+	      if (i == 0) {	/* EOF ??? */
+		type(SS, 501, NULL, "Err... Seen EOF during AUTH ??");
 		return;
+	      }
 	      if (strcmp(abuf, "*") == 0) {
+		policytest(&SS->policystate, POLICY_AUTHFAIL,
+			   uname, 1, SS->authuser);
+		smtp_tarpit(SS);
 		type(SS, 501, NULL, "AUTH command cancelled");
 		return;
 	      }
@@ -308,7 +361,11 @@ void smtp_auth(SS,buf,cp)
 	      uname = strdup(bbuf);
 	    }
 
-	    i = encodebase64string("Password:", 9, abuf, sizeof(abuf));
+	    if (!OCP->smtp_auth_password_prompt)
+	      OCP->smtp_auth_password_prompt = "Password:";
+	    i = encodebase64string(OCP->smtp_auth_password_prompt,
+				   strlen(OCP->smtp_auth_password_prompt),
+				   abuf, sizeof(abuf));
 	    if (i >= sizeof(abuf)) i = sizeof(abuf)-1;
 	    abuf[i] = 0;
 	    type(SS, 334, NULL, "%s", abuf);
@@ -341,10 +398,14 @@ void smtp_auth(SS,buf,cp)
 #endif
 	    if (i == 0)	{ /* EOF ??? */
 	      if (uname) free(uname);
+	      type(SS, 501, NULL, "Err... Seen EOF during AUTH ??");
 	      return;
 	    }
 	    if (strcmp(abuf, "*") == 0) {
 	      if (uname) free(uname);
+	      policytest(&SS->policystate, POLICY_AUTHFAIL,
+			 uname, 1, SS->authuser);
+	      smtp_tarpit(SS);
 	      type(SS, 501, NULL, "AUTH command cancelled");
 	      return;
 	    }
@@ -354,20 +415,30 @@ void smtp_auth(SS,buf,cp)
 	    if (debug)
 	      type(SS, 0, NULL, "-> %s", bbuf);
 
-	    if (tls_loglevel > 3)
-	      type(NULL,0,NULL,"zpwmatch: user %s' password '%s'", uname, bbuf);
-	    else if (tls_loglevel > 0)
-	      type(NULL,0,NULL,"zpwmatch: user %s' (password: *not so easy*!)", uname);
+	    if (OCP->tls_loglevel > 3) {
+	      /* The TLS debugging dump does reveal arrived frame
+		 data content as is, so it tells also the password.. */
+	      type(NULL,0,NULL,"zpwmatch: user '%s' password '%s'", uname, bbuf);
+	    } else if (OCP->tls_loglevel > 0)
+	      type(NULL,0,NULL,"zpwmatch: user '%s' (password: *not so easy*!)", uname);
 	    
-	    if (smtpauth_via_pipe)
-	      zpw = pipezpwmatch(smtpauth_via_pipe, uname, bbuf, &uid);
+	    if (OCP->smtpauth_via_pipe)
+	      zpw = pipezpwmatch(OCP->smtpauth_via_pipe, uname, bbuf, &uid);
 	    else
 	      zpw = zpwmatch(uname, bbuf, &uid);
 
 	    if (zpw == NULL) {
 	      SS->authuser = uname;
 	      type(SS, 235, NULL, "Authentication successful.");
+	      SS->with_protocol_set |= WITH_AUTH;
+#ifdef DO_PERL_EMBED
+	      if (use_perlhook)
+		ZSMTP_hook_set_user(SS->authuser, "login");
+#endif
 	    } else {
+	      policytest(&SS->policystate, POLICY_AUTHFAIL,
+			 uname, 1, SS->authuser);
+	      smtp_tarpit(SS);
 	      type(SS, 535, NULL, "%s", zpw);
 	      if (uname) free(uname);
 	    }
@@ -391,6 +462,14 @@ void smtp_auth(SS,buf,cp)
 	  int result;
 	  char *auth_type;
 
+	  if (SS->authuser) {
+	    policytest(&SS->policystate, POLICY_AUTHFAIL,
+		       uname, 1, SS->authuser);
+	    smtp_tarpit(SS);
+	    type(SS, 503, m551, "Hello %s, already authenticated, second attempt rejected!",SS->rhostaddr);
+	    return;
+	  }
+
 	  /* make sure mechanism (p) is a valid string */
 	  for (q = (char*)cp; *q != '\0' && isascii(*q); q++) {
 	    if (isspace(*q)) {
@@ -406,6 +485,9 @@ void smtp_auth(SS,buf,cp)
 
 	  /* check whether mechanism is available */
 	  if (iteminlist(cp, SS->sasl.mechlist, " ") == NULL) {
+	    policytest(&SS->policystate, POLICY_AUTHFAIL,
+		       uname, 1, SS->authuser);
+	    smtp_tarpit(SS);
 	    type(SS, 503, "5.3.3", "AUTH mechanism %.32s not available", cp);
 	    return;
 	  }
@@ -416,9 +498,12 @@ void smtp_auth(SS,buf,cp)
 	    in = malloc(len);
 	    result = sasl_decode64(q, len-1, in, len, &inlen);
 	    if (result != SASL_OK) {
-	      type(SS, 501, "5.5.4", "cannot BASE64 decode '%s'", q);
 	      authenticating = SASL_NOT_AUTH;
 	      free(in);
+	      policytest(&SS->policystate, POLICY_AUTHFAIL,
+			 uname, 1, SS->authuser);
+	      smtp_tarpit(SS);
+	      type(SS, 501, "5.5.4", "cannot BASE64 decode '%s'", q);
 	      return;
 	    }
 	  } else {
@@ -435,7 +520,8 @@ void smtp_auth(SS,buf,cp)
 
 	  if (result != SASL_OK && result != SASL_CONTINUE) {
 
-	    type(SS, 500, "5.7.0", "authentication failed");
+	    policytest(&SS->policystate, POLICY_AUTHFAIL,
+		       uname, 1, SS->authuser);
 	    if (logfp){
 	      const char * e = sasl_errdetail(SS->sasl.conn);
 	      if (!e) e = "<-no-detail->";
@@ -444,6 +530,8 @@ void smtp_auth(SS,buf,cp)
 		      cp, sasl_errstring(result, NULL, NULL), result, e);
 	      fflush(logfp);
 	    }
+	    smtp_tarpit(SS);
+	    type(SS, 500, "5.7.0", "authentication failed");
 	    return;
 	  }
 
@@ -459,13 +547,16 @@ void smtp_auth(SS,buf,cp)
 	  result = sasl_encode64(out, outlen, out2, len, &out2len);
 
 	  if (result != SASL_OK) {
-	    type(SS, 454, "4.5.4", "Temporary authentication failure");
 	    if (logfp) {
 	      fprintf(logfp, "%s%d#\tAUTH encode64 error [%d for \"%s\"]\n",
 		      logtag, (int)(now - logtagepoch),
 		      result, out);
 	      fflush(logfp);
 	    }
+	    policytest(&SS->policystate, POLICY_AUTHFAIL,
+		       uname, 1, SS->authuser);
+	    smtp_tarpit(SS);
+	    type(SS, 454, "4.5.4", "Temporary authentication failure");
 	    /* start over? */
 	    authenticating = SASL_NOT_AUTH;
 	  } else {
@@ -473,6 +564,11 @@ void smtp_auth(SS,buf,cp)
 	    authenticating = SASL_PROC_AUTH;
 	  }
 
+
+
+
+	  /* Now we start spinning the SASL call state.. */
+
 	  while (authenticating == SASL_PROC_AUTH) {
 
 	    i = s_gets( SS, abuf, sizeof(abuf), &rc, &co, &c );
@@ -492,6 +588,9 @@ void smtp_auth(SS,buf,cp)
 
 	    if (abuf[0] == '\0' || i == 0) {
 	      authenticating = SASL_NOT_AUTH;
+	      policytest(&SS->policystate, POLICY_AUTHFAIL,
+			 uname, 1, SS->authuser);
+	      smtp_tarpit(SS);
 	      type(SS,  501, "5.5.2", "missing input");
 	      break;
 	    }
@@ -500,6 +599,9 @@ void smtp_auth(SS,buf,cp)
 	      authenticating = SASL_NOT_AUTH;
 	      
 	      /* rfc 2254 4. */
+	      policytest(&SS->policystate, POLICY_AUTHFAIL,
+			 uname, 1, SS->authuser);
+	      smtp_tarpit(SS);
 	      type(SS, 501, "5.0.0", "AUTH aborted");
 	      break;
 	    }
@@ -509,6 +611,9 @@ void smtp_auth(SS,buf,cp)
 	    if (result != SASL_OK) {
 	      authenticating = SASL_NOT_AUTH;
 	      /* rfc 2254 4. */
+	      policytest(&SS->policystate, POLICY_AUTHFAIL,
+			 uname, 1, SS->authuser);
+	      smtp_tarpit(SS);
 	      type(SS, 501, "5.5.4", "cannot decode AUTH parameter %s", abuf);
 	      continue;
 	    }
@@ -529,7 +634,14 @@ void smtp_auth(SS,buf,cp)
 	      result = sasl_getprop(SS->sasl.conn, SASL_USERNAME,
 				    (const void **)&SS->authuser); 
 	      /* XX: check result == SASL_OK ?? */
-	  
+	      SS->with_protocol_set |= WITH_AUTH;
+#ifdef DO_PERL_EMBED
+	      if (result == SASL_OK) {
+		if (use_perlhook)
+		  ZSMTP_hook_set_user(SS->authuser, "saslauth");
+	      }
+#endif
+
 # if 0
 	      /* get realm? */
 	      sasl_getprop(SS->sasl.conn, SASL_REALM, (const void **) &data);
@@ -592,6 +704,8 @@ void smtp_auth(SS,buf,cp)
 		/* correct code? XXX */
 		/* 454 Temp. authentication failure */
 		type(SS, 454, "4.5.4", "Internal error: unable to encode64");
+		policytest(&SS->policystate, POLICY_AUTHFAIL,
+			   uname, 1, SS->authuser);
 #if 0
 		if (LogLevel > 5)
 		  sm_syslog(LOG_WARNING, e->e_id,
@@ -614,6 +728,9 @@ void smtp_auth(SS,buf,cp)
 	    } else {
 
 	      /* not SASL_OK or SASL_CONT */
+	      policytest(&SS->policystate, POLICY_AUTHFAIL,
+			 uname, 1, SS->authuser);
+	      smtp_tarpit(SS);
 	      type(SS, 500, "5.7.0", "authentication failed");
 #if 0
 	      if (LogLevel > 9)
@@ -695,7 +812,7 @@ smtpauth_init(SS)
 #ifdef HAVE_SASL2
       int result;
 
-      if (do_sasl) {
+      if (OCP->do_sasl) {
 
 	SS->sasl.n_mechs = 0;
 
@@ -762,13 +879,18 @@ void
 smtpauth_ehloresponse(SS)
      SmtpState *SS;
 {
-	int result;
+
+	if (OCP->no_smtp_auth_on_25 &&
+	    (SS->with_protocol_set & WITH_SMTP)) return;
 
 #ifdef HAVE_SASL2
-	if (do_sasl) {
+	if (OCP->do_sasl) {
+
+	int result;
+
 	  if (SS->sasl.sasl_ok) {
 	    SS->sasl.ssp.security_flags = (SASLSecOpts & SASL_SEC_MAXIMUM);
-	    if (!(auth_login_without_tls || SS->sslmode)) {
+	    if (!(OCP->auth_login_without_tls || SS->sslmode)) {
 	      SS->sasl.ssp.security_flags |= SASL_SEC_NOPLAINTEXT;
 	    }
 	    SS->sasl.ssp.security_flags |= SASL_SEC_NOANONYMOUS;
@@ -834,8 +956,8 @@ smtpauth_ehloresponse(SS)
 	    type(NULL,0,NULL, "AUTH -- no mechlist!");
 	} else
 #endif
-	  if (auth_ok) {
-	    if (auth_login_without_tls || SS->sslmode) {
+	  if (OCP->auth_ok) {
+	    if (OCP->auth_login_without_tls || SS->sslmode) {
 	      type(SS, -250, NULL, "AUTH=LOGIN"); /* RFC 2554, NetScape/
 						     Sun Solstice/ ? */
 	      type(SS, -250, NULL, "AUTH LOGIN"); /* RFC 2554, M$ Exchange ? */
