@@ -5,7 +5,7 @@
  *  talking with us in order to figure out of there are too
  *  many parallel connections from same IP address out there..
  *
- *  Copyright Matti Aarnio <mea@nic.funet.fi> 1998-1999
+ *  Copyright Matti Aarnio <mea@nic.funet.fi> 1998-1999, 2003-2004
  *
  */
 
@@ -16,6 +16,7 @@ static int child_top   = 0;
 
 static struct {
   int pid;	/* PID of the working smtpserver (subprocess) */
+  int tag;	/* 0: smtp, 1: smtps, 2: submit, 3: lmtp, ... */
   time_t when;	/* When to next check on this child */
   Usockaddr addr; /* Address the connection comes from */
 } *childs = NULL;
@@ -24,9 +25,9 @@ static int child_poll_interval = 30;
 static time_t child_now = 0;
 
 
-int childsameip(addr, childcntp)
-Usockaddr *addr;
-int *childcntp;
+int childsameip(addr, socktag, childcntp)
+     Usockaddr *addr;
+     int socktag, *childcntp;
 {
     int i, cnt = 1; /* Ourself */
     int childcnt = 1; /* Ourself */
@@ -50,8 +51,8 @@ int *childcntp;
 	  continue;
 	}
       }
-      if (childs[i].pid != 0 &&
-	  /* PID non zero */
+      if (childs[i].pid != 0  /* PID non zero */ &&
+	  childs[i].tag == socktag /* Same type! */ &&
 	  addr->v4.sin_family == childs[i].addr.v4.sin_family) {
 	/* Same AddressFamily */
 	if ((addr->v4.sin_family == AF_INET &&
@@ -66,7 +67,8 @@ int *childcntp;
 	    )
 	  ++cnt;
       }
-      if (childs[i].pid != 0)
+      if (childs[i].tag == socktag /* Same type! */ &&
+	  childs[i].pid != 0)
 	++childcnt;
     }
 
@@ -74,12 +76,14 @@ int *childcntp;
     return cnt;
 }
 
-void childregister(cpid, addr)
-int cpid;
-Usockaddr *addr;
+void childregister(cpid, addr, socktag)
+     int cpid, socktag;
+     Usockaddr *addr;
 {
 	int i;
 
+	/* Called with SIGCHLD held ! */
+
 	if (kill(cpid, 0) < 0) {
 	  /* When there is no subprocess with this PID, DON'T
 	     register anything!  The subprocess is already
@@ -87,6 +91,13 @@ Usockaddr *addr;
 	  return;
 	}
 
+	/* Now count those processes.. */
+
+	MIBMtaEntry->ss.IncomingSMTPSERVERprocesses += 1;
+	MIBMtaEntry->ss.IncomingSMTPSERVERforks     += 1;
+
+	/* And do registering!         */
+
 	time(&child_now);
 
 	if (child_top == child_space) {
@@ -95,11 +106,14 @@ Usockaddr *addr;
 	  } else {
 	    child_space <<= 1;
 	  }
-	  if (childs == NULL) {
-	    childs = emalloc(child_space * sizeof(*childs));
-	  } else {
-	    childs = erealloc(childs, child_space * sizeof(*childs));
+	  /* 'childs' can be NULL here.. */
+	  childs = erealloc(childs, child_space * sizeof(*childs));
+	  /* Now it is NULL only in error.. and we are in deep trouble.. */
+	  if (!childs) {
+	    child_space = 0;
+	    return;
 	  }
+
 	  for (i = child_top; i < child_space; ++i)
 	    memset(&childs[i], 0, sizeof(childs[i]));
 	}
@@ -114,12 +128,13 @@ Usockaddr *addr;
 	    if (kill(childs[i].pid, 0) != 0) {
 	      /* No such process in there anymore !?
 		 We free this slot now! */
-	      memset(&childs[i], 0, sizeof(childs[i]));
+	      childreap(childs[i].pid);
 	    }
 	  }
 	  if (childs[i].pid == 0) { /* Free slot! */
-	    childs[i].pid = cpid;
+	    childs[i].pid  = cpid;
 	    childs[i].when = child_now + child_poll_interval;
+	    childs[i].tag  = socktag;
 	    if (addr->v4.sin_family == AF_INET) {
 	      childs[i].addr.v4.sin_family = AF_INET;
 	      memcpy(&childs[i].addr.v4.sin_addr, &addr->v4.sin_addr, 4);
@@ -138,6 +153,13 @@ Usockaddr *addr;
 	}
 }
 
+/* Started children call this to disable this tracking code */
+void disable_childreap()
+{
+	childs = NULL;
+	child_top = 0;
+}
+
 void childreap(cpid)
 int cpid;
 {
@@ -147,7 +169,29 @@ int cpid;
 
 	for (i = 0; i < child_top; ++i)
 	  if (childs[i].pid == cpid) {
+
+	    MIBMtaEntry->ss.IncomingSMTPSERVERprocesses -= 1;
+
+	    /* Drop the class-full counts; the main loop does
+	       setting of these gauges, we just speed their
+	       decrementing, just in case.  */
+
+	    switch (childs[i].tag) {
+	    case 0:
+	      MIBMtaEntry->ss.IncomingParallelSMTPconnects   -= 1;
+	      break;
+	    case 1:
+	      MIBMtaEntry->ss.IncomingParallelSMTPSconnects  -= 1;
+	      break;
+	    case 2:
+	      MIBMtaEntry->ss.IncomingParallelSUBMITconnects -= 1;
+	      break;
+	    default:
+	      break;
+	    }
+
 	    memset(&childs[i], 0, sizeof(childs[i]));
+
 	    break;
 	  }
 }
