@@ -4,7 +4,7 @@
  */
 /*
  *    Several extensive changes by Matti Aarnio <mea@nic.funet.fi>
- *      Copyright 1991-2003.
+ *      Copyright 1991-2007.
  */
 /*
  * Zmailer SMTP-server divided into bits
@@ -46,7 +46,7 @@ static int parsestatcode(ssp, statcodep)
      const char **statcodep;
 {
     int code = -1;
-    const char *ss = *ssp;
+    const unsigned char *ss = (const unsigned char *) *ssp;
     static char statcodebuf[6];
 
     *statcodep = NULL;
@@ -65,7 +65,7 @@ static int parsestatcode(ssp, statcodep)
       ss += 5;
     }
     SKIPSPACE(ss);
-    *ssp = ss;
+    *ssp = (const char *) ss;
     if (code < 200 || code > 599) code = 0;
     return code;
 }
@@ -83,13 +83,12 @@ const char *buf, *cp;
     char *fname;
     char taspid[30];
 
-
     MIBMtaEntry->ss.ReceivedMessagesSs  += 1;
     MIBMtaEntry->ss.ReceivedRecipientsSs += SS->ok_rcpt_count;
     MIBMtaEntry->ss.IncomingSMTP_DATA   += 1;
 
-    while (!strict_protocol && (*cp == ' ' || *cp == '\t')) ++cp;
-    if (strict_protocol && *cp != 0) {
+    while ((strict_protocol < 1) && (*cp == ' ' || *cp == '\t')) ++cp;
+    if ((strict_protocol > 0) && *cp != 0) {
 	MIBMtaEntry->ss.IncomingSMTP_DATA_bad += 1;
 	type(SS, 501, m554, "Extra junk after 'DATA' verb");
 	return 0;
@@ -115,8 +114,17 @@ const char *buf, *cp;
 	    break;
 	}
 	MIBMtaEntry->ss.IncomingSMTP_DATA_bad += 1;
-	type(SS, 503, m552, cp);
+	type(SS, 503, m552, "Hi %s, %s", SS->rhostaddr, cp);
 	typeflush(SS);
+	if (SS->mfp) {
+	  mail_abort(SS->mfp);
+	  SS->mfp = NULL;
+	}
+	if (SS->rcpt_count)
+	  policytest(&SS->policystate, POLICY_DATAABORT,
+		     NULL, SS->rcpt_count, NULL);
+	SS->rcpt_count = 0;
+
 	return 0;
     }
 
@@ -125,6 +133,9 @@ const char *buf, *cp;
 	typeflush(SS);
 	clearerr(SS->mfp);
 	mail_abort(SS->mfp);
+	policytest(&SS->policystate, POLICY_DATAABORT,
+		   NULL, SS->rcpt_count, NULL);
+
 	MIBMtaEntry->ss.IncomingSMTP_DATA_bad += 1;
 	SS->mfp = NULL;
 	reporterr(SS, tell, "message file error");
@@ -134,8 +145,11 @@ const char *buf, *cp;
 	type(SS, 550, "5.1.7", "No valid sender, rejecting all recipients");
 	typeflush(SS);
 	SS->state = MailOrHello;
-	if (SS->mfp)
+	if (SS->mfp) {
 	  mail_abort(SS->mfp);
+	  policytest(&SS->policystate, POLICY_DATAABORT,
+		     NULL, SS->rcpt_count, NULL);
+	}
 	MIBMtaEntry->ss.IncomingSMTP_DATA_bad += 1;
 	SS->mfp = NULL;
 	return 0;
@@ -150,17 +164,32 @@ const char *buf, *cp;
 	SS->mfp = NULL;
 	return 0;
     }
-    if ((SS->from_box != 0) && (SS->rcpt_count > MaxErrorRecipients)) {
+    if ((SS->from_box != 0) && (SS->rcpt_count > OCP->MaxErrorRecipients)) {
 	/* Too many recipients for a  "MAIL FROM:<>" */
 	type(SS, 550, "5.7.1", "SPAM trap -- too many recipients for an empty source address!");
 	typeflush(SS);
 	SS->state = MailOrHello;
 	mail_abort(SS->mfp);
+	policytest(&SS->policystate, POLICY_DATAABORT,
+		   NULL, SS->rcpt_count, NULL);
 	MIBMtaEntry->ss.IncomingSMTP_DATA_bad += 1;
 	SS->mfp = NULL;
 	return 0;
     }
-    SS->rcpt_count = 0;		/* now we can zero it.. */
+#ifdef DO_PERL_EMBED
+    if (use_perlhook) {
+      int rc;
+
+      if (ZSMTP_hook_univ(ZSMTP_HOOK_DATA, &SS->policystate, buf, cp-buf, &rc)) {
+	if (rc < 0) {
+	  /* 400-series response */
+	} else if (rc > 0) {
+	  /* 500-series response */
+	}
+      }
+    }
+#endif
+
     type(SS, 354, NULL, (char *) NULL);
     typeflush(SS);
     fputs("env-end\n", SS->mfp);
@@ -169,6 +198,17 @@ const char *buf, *cp;
       fprintf(SS->mfp, "X-Comment: RFC 2476 MSA function at %s logged sender identity as: %s\n", SS->myhostname, SS->authuser);
     }
 
+#ifdef Z_CHECK_SPF_DATA
+    if (OCP->spf_received && policyspfhdr(&SS->policystate)) {
+      fprintf(SS->mfp,"%s\n", policyspfhdr(&SS->policystate));
+    }
+#endif
+
+#ifdef DO_PERL_EMBED
+    if (use_perlhook && SS->policystate.ZSMTP_hook_hdr)
+      fprintf(SS->mfp, "%s\n", SS->policystate.ZSMTP_hook_hdr);
+#endif
+
     /* We set alarm()s inside the mvdata() */
     *msg = 0;
     filsiz = mvdata(SS, msg);
@@ -178,26 +218,46 @@ const char *buf, *cp;
     fname = mail_fname(SS->mfp);
     fstat(FILENO(SS->mfp), &stbuf);
 
-    taspoolid(taspid, stbuf.st_mtime, stbuf.st_ino);
+    taspoolid(taspid, stbuf.st_ino, stbuf.st_mtime,
+#ifdef HAVE_STRUCT_STAT_ST_ATIM_TV_NSEC
+	      stbuf.st_mtim.tv_nsec
+#else
+#ifdef HAVE_STRUCT_STAT_ST_ATIM___TV_NSEC
+	      stbuf.st_mtim.__tv_nsec
+#else
+#ifdef HAVE_STRUCT_STAT_ST_ATIMENSEC
+	      stbuf.st_mtimensec
+#else
+	      0
+#endif
+#endif
+#endif
+		);
+
     tell = stbuf.st_size,
 
-    report(SS, "Got '.'; tell=%ld", tell);
+    report(SS, "Got '.'; size=%ld", tell);
 
-    availspace = used_fd_statfs(FILENO(SS->mfp));
-    if (availspace >= 0)
-      MIBMtaEntry->sys.SpoolUsedSpace = availspace;
-    availspace = free_fd_statfs(FILENO(SS->mfp));
-    if (availspace < 0)
-	availspace = LONG_MAX / 1024;	/* Over 2G ? */
-    if (availspace >= 0)
-      MIBMtaEntry->sys.SpoolFreeSpace = availspace;
-    availspace -= minimum_availspace;
-    if (availspace > (LONG_MAX / 1024))
-      availspace = LONG_MAX / 1024;
-    availspace *= 1024;
+    {
+      long bavail, bused, iavail, iused;
+      if (0 == fd_statfs(FILENO(SS->mfp), &bavail, &bused, &iavail, &iused)) {
+	MIBMtaEntry->sys.SpoolUsedSpace = bused;
+	MIBMtaEntry->sys.SpoolFreeSpace = bavail;
+
+	availspace = bavail - minimum_availspace;
+	if (availspace > (LONG_MAX / 1024))
+	  availspace = LONG_MAX / 1024;
+	availspace *= 1024;
+	
+	MIBMtaEntry->sys.SpoolUsedFiles = iused;
+	MIBMtaEntry->sys.SpoolFreeFiles = iavail;
+      }
+    }
 
     if (*msg != 0) {
 	mail_abort(SS->mfp);
+	policytest(&SS->policystate, POLICY_DATAABORT,
+		   NULL, SS->rcpt_count, NULL);
 	MIBMtaEntry->ss.IncomingSMTP_DATA_bad += 1;
 	SS->mfp = NULL;
 	type(SS, 452, m430, "%s", msg);
@@ -208,8 +268,11 @@ const char *buf, *cp;
 	if (STYLE(SS->cfinfo,'D')) {
 	  /* Says: DON'T DISCARD -- aka DEBUG ERRORS! */
 	  mail_close_alternate(SS->mfp,"public",".DATA-EOF");
-	} else
+	} else {
 	  mail_abort(SS->mfp);
+	  policytest(&SS->policystate, POLICY_DATAABORT,
+		     NULL, SS->rcpt_count, NULL);
+	}
 	MIBMtaEntry->ss.IncomingSMTP_DATA_bad += 1;
 	SS->mfp = NULL;
 	reporterr(SS, tell, "premature EOF on DATA input");
@@ -223,10 +286,14 @@ const char *buf, *cp;
 	reporterr(SS, tell, ferror(SS->mfp) ? "write to spool file failed" : "system free storage under limit");
 	clearerr(SS->mfp);
 	mail_abort(SS->mfp);
+	policytest(&SS->policystate, POLICY_DATAABORT,
+		   NULL, SS->rcpt_count, NULL);
 	MIBMtaEntry->ss.IncomingSMTP_DATA_bad += 1;
 	SS->mfp = NULL;
     } else if (maxsize > 0 && filsiz > maxsize) {
 	mail_abort(SS->mfp);
+	policytest(&SS->policystate, POLICY_DATAABORT,
+		   NULL, SS->rcpt_count, NULL);
 	MIBMtaEntry->ss.IncomingSMTP_DATA_bad += 1;
 	SS->mfp = NULL;
 	type(SS, 552, "5.3.4", "Size of this message exceeds the fixed maximum size of  %ld  chars for received email ", maxsize);
@@ -248,9 +315,9 @@ const char *buf, *cp;
 	/* Lets see what the content-policy will tell now ? */
 
 	if (debug) typeflush(SS);
-	SS->policyresult = contentpolicy(policydb, &SS->policystate, fname);
+	SS->policyresult = contentpolicy(&SS->policystate, fname);
 
-	ss0 = ss  = policymsg(policydb, &SS->policystate);
+	ss0 = ss  = policymsg(&SS->policystate);
 
 	if (ss)
 	  type(NULL,0,NULL,
@@ -292,10 +359,17 @@ const char *buf, *cp;
 	  if (!statcode)  statcode = m571;
 	  if (!code)      code = 552;
 
-	  type(SS, -code, statcode, "Content-Policy msg: %s; %s", ss, taspid);
-	  for (j= 1; j <= sslinecnt; ++j)
-	    type(SS, -code, statcode, "msg: %s", sslines[j]);
-	  type(SS, code, statcode, "Content-Policy analysis rejected this message");
+	  if (sslinecnt == 0) {
+	    type(SS, code, statcode,
+		 "Content-Policy reject msg: %s; %s", ss, taspid);
+	  } else {
+	    type(SS, -code, statcode,
+		 "Content-Policy msg: %s; %s", ss, taspid);
+	    for (j= 1; j <= sslinecnt; ++j)
+	      type(SS, -code, statcode, "msg: %s", sslines[j]);
+	    type(SS, code, statcode,
+		 "Content-Policy analysis rejected this message");
+	  }
 
 	  if (lmtp_mode) for(i = 1; i < SS->ok_rcpt_count; ++i) {
 	    type(SS, -code, statcode, "Content-Policy msg: %s; %s", ss, taspid);
@@ -305,6 +379,8 @@ const char *buf, *cp;
 	  }
 
 	  mail_abort(SS->mfp);
+	  policytest(&SS->policystate, POLICY_DATAABORT,
+		     NULL, SS->rcpt_count, NULL);
 	  MIBMtaEntry->ss.IncomingSMTP_DATA_bad += 1;
 	  SS->mfp = NULL;
 	} else if (SS->policyresult > 0) {
@@ -329,10 +405,19 @@ const char *buf, *cp;
 	    if (!statcode)  statcode = "2.7.1";
 	    if (!code)      code = 250;
 
-	    type(SS, -code, statcode, "%s; %s", ss, taspid);
-	    for (j= 1; j <= sslinecnt; ++j)
-	      type(SS, -code, statcode, "%s", sslines[j]);
-	    type(SS, code, statcode, "Content-Policy accepted this message into freezer-%d; %s", SS->policyresult, taspid);
+	    if (sslinecnt == 0) {
+	      type(SS, code, statcode,
+		   "Content-Policy accept-into-freezer-%d msg: %s; %s",
+		   SS->policyresult, ss, taspid);
+	    } else {
+	      type(SS, -code, statcode,
+		   "Content-Policy msg: %s; %s", ss, taspid);
+	      for (j= 1; j <= sslinecnt; ++j)
+		type(SS, -code, statcode, "msg: %s", sslines[j]);
+	      type(SS, code, statcode,
+		   "Content-Policy accepted this message into freezer-%d; %s",
+		   SS->policyresult, taspid);
+	    }
 	    
 	    if (lmtp_mode) for(i = 1; i < SS->ok_rcpt_count; ++i) {
 	      type(SS, -code, statcode, "%s; %s", ss, taspid);
@@ -401,6 +486,9 @@ const char *buf, *cp;
 		       "%s: (%ldc) accepted from %s/%d", taspid, tell,
 		       SS->rhostname, SS->rport));
 		
+	    policytest(&SS->policystate, POLICY_DATAOK,
+		       NULL, SS->rcpt_count, NULL);
+
 	    MIBMtaEntry->ss.IncomingSMTP_DATA_ok    += 1;
 
 	    MIBMtaEntry->ss.TransmittedMessagesSs   += 1;
@@ -461,6 +549,15 @@ const char *buf, *cp;
 	if (msa_mode && SS->authuser != NULL ) {
 	  fprintf(SS->mfp, "X-Comment: RFC 2476 MSA function at %s logged sender identity as: %s\n", SS->myhostname, SS->authuser);
 	}
+#ifdef Z_CHECK_SPF_DATA
+	if (OCP->spf_received && policyspfhdr(&SS->policystate)) {
+	  fprintf(SS->mfp,"%s\n", policyspfhdr(&SS->policystate));
+	}
+#endif
+#ifdef DO_PERL_EMBED
+	if (use_perlhook && SS->policystate.ZSMTP_hook_hdr)
+	  fprintf(SS->mfp, "%s\n", SS->policystate.ZSMTP_hook_hdr);
+#endif
     }
     /* We set alarm()s inside the mvbdata() */
     *msg = 0;
@@ -475,7 +572,21 @@ const char *buf, *cp;
       fname = mail_fname(SS->mfp);
       fstat(FILENO(SS->mfp), &stbuf);
 
-      taspoolid(taspid, stbuf.st_mtime, stbuf.st_ino);
+      taspoolid(taspid, stbuf.st_ino, stbuf.st_mtime,
+#ifdef HAVE_STRUCT_STAT_ST_ATIM_TV_NSEC
+		stbuf.st_mtim.tv_nsec
+#else
+#ifdef HAVE_STRUCT_STAT_ST_ATIM___TV_NSEC
+		stbuf.st_mtim.__tv_nsec
+#else
+#ifdef HAVE_STRUCT_STAT_ST_ATIMENSEC
+		stbuf.st_mtimensec
+#else
+		0
+#endif
+#endif
+#endif
+		);
       tell = stbuf.st_size;
 
     } else {
@@ -486,7 +597,7 @@ const char *buf, *cp;
 
     }
 
-    report(SS, "BDAT %ld%s; tell=%ld", bdata_chunksize,
+    report(SS, "BDAT %ld%s; size=%ld", bdata_chunksize,
 	   bdata_last ? " LAST":"", tell);
 
     if (SS->state != BData) {
@@ -511,8 +622,11 @@ const char *buf, *cp;
 	  type(SS, 503, m552, cp);
 
 	typeflush(SS);
-	if (SS->mfp)
+	if (SS->mfp) {
 	    mail_abort(SS->mfp);
+	    policytest(&SS->policystate, POLICY_DATAABORT,
+		       NULL, SS->rcpt_count, NULL);
+	}
 	MIBMtaEntry->ss.IncomingSMTP_BDAT_bad += 1;
 	SS->mfp = NULL;
 	return 0;
@@ -529,13 +643,16 @@ const char *buf, *cp;
 
 	    typeflush(SS);
 	    SS->state = MailOrHello;
-	    if (SS->mfp)
+	    if (SS->mfp) {
 	      mail_abort(SS->mfp);
+	      policytest(&SS->policystate, POLICY_DATAABORT,
+			 NULL, SS->rcpt_count, NULL);
+	    }
 	    MIBMtaEntry->ss.IncomingSMTP_BDAT_bad += 1;
 	    SS->mfp = NULL;
 	    return 0;
 	}
-	if ((SS->from_box != 0) && (SS->rcpt_count > MaxErrorRecipients)) {
+	if ((SS->from_box != 0) && (SS->rcpt_count > OCP->MaxErrorRecipients)) {
 
 	  /* Too many recipients for a  "MAIL FROM:<>" */
 	  type(SS, 550, "5.7.1", "SPAM trap -- too many recipients for an empty source address!");
@@ -545,6 +662,8 @@ const char *buf, *cp;
 	  typeflush(SS);
 	  SS->state = MailOrHello;
 	  mail_abort(SS->mfp);
+	  policytest(&SS->policystate, POLICY_DATAABORT,
+		     NULL, SS->rcpt_count, NULL);
 	  MIBMtaEntry->ss.IncomingSMTP_BDAT_bad += 1;
 	  SS->mfp = NULL;
 	  return 0;
@@ -553,18 +672,21 @@ const char *buf, *cp;
 	SS->sender_ok = 0;
     }
 
-    availspace = used_fd_statfs(FILENO(SS->mfp));
-    if (availspace >= 0)
-      MIBMtaEntry->sys.SpoolUsedSpace = availspace;
-    availspace = free_fd_statfs(FILENO(SS->mfp));
-    if (availspace < 0)
-	availspace = LONG_MAX / 1024;	/* Over 2G ? */
-    if (availspace >= 0)
-      MIBMtaEntry->sys.SpoolFreeSpace = availspace;
-    availspace -= minimum_availspace;
-    if (availspace > (LONG_MAX / 1024))
-      availspace = LONG_MAX / 1024;
-    availspace *= 1024;
+    {
+      long bavail, bused, iavail, iused;
+      if (0 == fd_statfs(FILENO(SS->mfp), &bavail, &bused, &iavail, &iused)) {
+	MIBMtaEntry->sys.SpoolUsedSpace = bused;
+	MIBMtaEntry->sys.SpoolFreeSpace = bavail;
+
+	availspace = bavail - minimum_availspace;
+	if (availspace > (LONG_MAX / 1024))
+	  availspace = LONG_MAX / 1024;
+	availspace *= 1024;
+	
+	MIBMtaEntry->sys.SpoolUsedFiles = iused;
+	MIBMtaEntry->sys.SpoolFreeFiles = iavail;
+      }
+    }
 
     /* The common typeflush() is at the end... */
     if (SS->mfp == NULL) {
@@ -574,6 +696,8 @@ const char *buf, *cp;
 	MIBMtaEntry->ss.IncomingSMTP_BDAT_bad += 1;
     } else if (*msg != 0) {
 	mail_abort(SS->mfp);
+	policytest(&SS->policystate, POLICY_DATAABORT,
+		   NULL, SS->rcpt_count, NULL);
 	SS->mfp = NULL;
 	type(SS, 452, "%s", msg);
 	if (lmtp_mode && bdata_last) for(i = 1; i < SS->ok_rcpt_count; ++i)
@@ -584,8 +708,11 @@ const char *buf, *cp;
 	if (STYLE(SS->cfinfo,'D')) {
 	  /* Says: DON'T DISCARD -- aka DEBUG ERRORS! */
 	  mail_close_alternate(SS->mfp,"public",".BDAT-EOF");
-	} else
+	} else {
 	  mail_abort(SS->mfp);
+	  policytest(&SS->policystate, POLICY_DATAABORT,
+		     NULL, SS->rcpt_count, NULL);
+	}
 	MIBMtaEntry->ss.IncomingSMTP_BDAT_bad += 1;
 	SS->mfp = NULL;
 	reporterr(SS, tell, "premature EOF on BDAT input");
@@ -600,10 +727,14 @@ const char *buf, *cp;
 				    "system free storage under limit");
 	clearerr(SS->mfp);
 	mail_abort(SS->mfp);
+	policytest(&SS->policystate, POLICY_DATAABORT,
+		   NULL, SS->rcpt_count, NULL);
 	MIBMtaEntry->ss.IncomingSMTP_BDAT_bad += 1;
 	SS->mfp = NULL;
     } else if (maxsize > 0 && tell > maxsize) {
 	mail_abort(SS->mfp);
+	policytest(&SS->policystate, POLICY_DATAABORT,
+		   NULL, SS->rcpt_count, NULL);
 	MIBMtaEntry->ss.IncomingSMTP_BDAT_bad += 1;
 	SS->mfp = NULL;
 	type(SS, 552, "5.3.4", "Size of this message exceeds the fixed maximum size of  %ld  chars for received email ", maxsize);
@@ -624,8 +755,8 @@ const char *buf, *cp;
 	/* Lets see what the content-policy will tell now ? */
 
 	if (debug) typeflush(SS);
-	SS->policyresult = contentpolicy(policydb, &SS->policystate, fname);
-	ss0 = ss  = policymsg(policydb, &SS->policystate);
+	SS->policyresult = contentpolicy(&SS->policystate, fname);
+	ss0 = ss  = policymsg(&SS->policystate);
 
 	if (ss)
 	  type(NULL,0,NULL,
@@ -667,10 +798,17 @@ const char *buf, *cp;
 	  if (!statcode)  statcode = m571;
 	  if (!code)      code = 552;
 
-	  type(SS, -code, statcode, "Content-Policy msg: %s; %s", ss, taspid);
-	  for (j= 1; j <= sslinecnt; ++j)
-	    type(SS, -code, statcode, "msg: %s", sslines[j]);
-	  type(SS, code, statcode, "Content-Policy analysis rejected this message");
+	  if (sslinecnt == 0) {
+	    type(SS, code, statcode,
+		 "Content-Policy reject msg: %s; %s", ss, taspid);
+	  } else {
+	    type(SS, -code, statcode,
+		 "Content-Policy msg: %s; %s", ss, taspid);
+	    for (j= 1; j <= sslinecnt; ++j)
+	      type(SS, -code, statcode, "msg: %s", sslines[j]);
+	    type(SS, code, statcode,
+		 "Content-Policy analysis rejected this message");
+	  }
 
 	  if (lmtp_mode) for(i = 1; i < SS->ok_rcpt_count; ++i) {
 	    type(SS, -code, statcode, "Content-Policy msg: %s; %s", ss, taspid);
@@ -680,6 +818,8 @@ const char *buf, *cp;
 	  }
 
 	  mail_abort(SS->mfp);
+	  policytest(&SS->policystate, POLICY_DATAABORT,
+		     NULL, SS->rcpt_count, NULL);
 	  MIBMtaEntry->ss.IncomingSMTP_BDAT_bad += 1;
 	  SS->mfp = NULL;
 
@@ -706,10 +846,19 @@ const char *buf, *cp;
 	    if (!statcode)  statcode = "2.7.1";
 	    if (!code)      code = 250;
 
-	    type(SS, -code, statcode, "%s; %s", ss, taspid);
-	    for (j= 1; j <= sslinecnt; ++j)
-	      type(SS, -code, statcode, "%s", sslines[j]);
-	    type(SS, code, statcode, "Content-Policy accepted this message into freezer-%d; %s", SS->policyresult, taspid);
+	    if (sslinecnt == 0) {
+	      type(SS, code, statcode,
+		   "Content-Policy accept-into-freezer-%d msg: %s; %s",
+		   SS->policyresult, ss, taspid);
+	    } else {
+	      type(SS, -code, statcode,
+		   "Content-Policy msg: %s; %s", ss, taspid);
+	      for (j= 1; j <= sslinecnt; ++j)
+		type(SS, -code, statcode, "msg: %s", sslines[j]);
+	      type(SS, code, statcode,
+		   "Content-Policy accepted this message into freezer-%d; %s",
+		   SS->policyresult, taspid);
+	    }
 	    
 	    if (lmtp_mode) for(i = 1; i < SS->ok_rcpt_count; ++i) {
 	      type(SS, -code, statcode, "%s; %s", ss, taspid);
@@ -772,6 +921,9 @@ const char *buf, *cp;
 	  }
 #endif
 
+	  policytest(&SS->policystate, POLICY_DATAOK,
+		     NULL, SS->rcpt_count, NULL);
+
 
 	  MIBMtaEntry->ss.IncomingSMTP_BDAT_ok    += 1;
 
