@@ -1,7 +1,7 @@
 /*
  *  Globals of the ZMailer  smtp-server
  *
- *  Matti Aarnio <mea@nic.funet.fi> 1997-2001
+ *  Matti Aarnio <mea@nic.funet.fi> 1997-2005
  */
 
 
@@ -9,6 +9,8 @@
 #define SMTP_DATA_TIME_PER_LINE  600	/* 10 minutes of life.. */
 #define SMTP_REPLY_ALARM_IVAL    300	/*  5 minutes to write to socket.. */
 
+#define SUBSERVER_IDLE_TIMEOUT   600	/* 10 minutes of idle life.. */
+
 /*
  * The smtpserver connects to the router to ask it various questions, like,
  * is this a valid address?  What is the alias expansion of that? etc.
@@ -47,6 +49,9 @@
 #include <sys/stat.h>
 #include <sys/file.h>
 #include <sysexits.h>
+#ifdef HAVE_UNIDSTD_H
+#include <unistd.h>
+#endif
 #ifdef HAVE_FCNTL_H
 #include <fcntl.h>
 #endif
@@ -132,55 +137,17 @@ extern int wait();
 
 
 
-#ifdef HAVE_SYS_SELECT_H
-#include <sys/select.h>
-#endif
-
 #include <sys/time.h>
 
-#ifndef	NFDBITS
-/*
- * This stuff taken from the 4.3bsd /usr/include/sys/types.h, but on the
- * assumption we are dealing with pre-4.3bsd select().
- */
+#include "zmpoll.h"
 
-/* #error "FDSET macro susceptible" */
-
-typedef long	fd_mask;
-
-#ifndef	NBBY
-#define	NBBY	8
-#endif	/* NBBY */
-#define	NFDBITS		((sizeof fd_mask) * NBBY)
-
-/* SunOS 3.x and 4.x>2 BSD already defines this in /usr/include/sys/types.h */
-#ifdef	notdef
-typedef	struct fd_set { fd_mask	fds_bits[1]; } fd_set;
-#endif	/* notdef */
-
-#ifndef	_Z_FD_SET
-/* #warning "_Z_FD_SET[1]" */
-#define	_Z_FD_SET(n, p)   ((p)->fds_bits[0] |= (1 << (n)))
-#define	_Z_FD_CLR(n, p)   ((p)->fds_bits[0] &= ~(1 << (n)))
-#define	_Z_FD_ISSET(n, p) ((p)->fds_bits[0] & (1 << (n)))
-#define _Z_FD_ZERO(p)	  memset((char *)(p), 0, sizeof(*(p)))
-#endif	/* !FD_SET */
-#endif	/* !NFDBITS */
-
-#ifdef FD_SET
-/* #warning "_Z_FD_SET[2]" */
-#define _Z_FD_SET(sock,var) FD_SET(sock,&var)
-#define _Z_FD_CLR(sock,var) FD_CLR(sock,&var)
-#define _Z_FD_ZERO(var) FD_ZERO(&var)
-#define _Z_FD_ISSET(i,var) FD_ISSET(i,&var)
+#ifdef HAVE_SPF2_SPF_H
+#include <spf2/spf.h>
 #else
-/* #warning "_Z_FD_SET[3]" */
-#define _Z_FD_SET(sock,var) var |= (1 << sock)
-#define _Z_FD_CLR(sock,var) var &= ~(1 << sock)
-#define _Z_FD_ZERO(var) var = 0
-#define _Z_FD_ISSET(i,var) ((var & (1 << i)) != 0)
+#ifdef HAVE_SPF_ALT_SPF_H
+#include <spf_alt/spf.h>
+#endif
 #endif
-
 
 #include "policytest.h"
 
@@ -212,7 +179,7 @@ typedef enum {
     RecipientOrData, Data, Send, SendOrMail,
     SendAndMail, Reset, Verify, Expand, Help,
     NoOp, Quit, Turn, Tick, Verbose, DebugIdent,
-    Turnme, BData, DebugMode, Auth,
+    Turnme, BData, DebugMode, Auth, Report,
 #ifdef HAVE_OPENSSL
     StartTLS,
 #endif /* - HAVE_OPENSSL */
@@ -261,9 +228,9 @@ struct SmtpSASLState {
     char  *volatile auth_type;
     const char *mechlist;
     sasl_conn_t *conn;
-    volatile int sasl_ok;
-    volatile unsigned int n_auth;	/* count of AUTH commands */
-    volatile unsigned int n_mechs;
+    int sasl_ok;
+    unsigned int n_auth;	/* count of AUTH commands */
+    unsigned int n_mechs;
     unsigned int len;
     sasl_security_properties_t ssp;
 #if 0 /* Is in SASL-1, different/not in SASL-2 */
@@ -284,9 +251,18 @@ typedef struct SmtpState {
     long messagesize;		/* Reset at MAIL, add at BDATs / DATA */
     long sizeoptval;		/* "MAIL FROM:<xxx> SIZE=nnn" -value    */
     long sizeoptsum;
-    char myhostname[MAXHOSTNAMELEN + 1];
-    char rhostname[MAXHOSTNAMELEN + 1];
-    const char *with_protocol;	/* = WITH_SMTP */
+
+    int  with_protocol_set;	/* = WITH_SMTP */
+#define WITH_HELO		0x0001
+#define WITH_EHLO		0x0002
+#define WITH_SMTP		0x0004
+#define WITH_SUBMIT		0x0008
+#define WITH_SMTPS		0x0010
+#define WITH_TLS		0x0020
+#define WITH_AUTH		0x0040
+#define WITH_LMTP		0x0080
+#define WITH_BSMTP		0x0100
+
     const char *style;		/* = "ve" */
     Command state;		/* = Hello */
     int  VerboseCommand;
@@ -295,10 +271,24 @@ typedef struct SmtpState {
     int  policyresult, reject_net;
     int  postmasteronly;
     int  netconnected_flg;
-    int  tarpit;
-    int  tarpit_cval;		/* current tarpit value */
+    double  tarpit;
+    double  tarpit_cval;		/* current tarpit value */
+
+    char myhostname[MAXHOSTNAMELEN + 1];
+    char rhostname[MAXHOSTNAMELEN + 1];
+    int  rhostflags;
+#define RHOST_REVERSED_OK	0x0001
+#define RHOST_REVERSED_FAIL	0x0002
+#define RHOST_VERIFIED_OK	0x0010
+#define RHOST_VERIFIED_FAIL	0x0020
+#define RHOST_HELO_SYNTAX_OK	0x0100
+#define RHOST_HELO_SYNTAX_FAIL	0x0200
+#define RHOST_HELO_VERIFY_OK	0x1000
+#define RHOST_HELO_VERIFY_FAIL	0x2000
+
+    char rhostaddr[sizeof("[ipv6.ffff:ffff:ffff:ffff:ffff:ffff:255.255.255.255]") + 8];
     int  rport;
-    char ihostaddr[sizeof("[ipv6.ffff:ffff:ffff:ffff:ffff:ffff:255.255.255.255]") + 8];
+    int  lport;
     Usockaddr raddr;
     Usockaddr localsock;
 
@@ -324,7 +314,9 @@ typedef struct SmtpState {
     int  s_readout;
     int  s_status;
     int  s_readerrno;
-    char s_buffer[SMTPLINESIZE];
+    int  s_seen_eof;
+    char s_buffer[64*1024];	/* 64 kB */
+    int  s_buffer_size;		/* adjustable max size.. */
     int  s_ungetcbuf;
     int  s_seen_pipeline;
 
@@ -332,9 +324,12 @@ typedef struct SmtpState {
     int  rcpt_count;
     int  ok_rcpt_count;
     int  sender_ok;
+
     /* For BDAT -command */
     int  bdata_blocknum;
     int  mvbstate;
+
+    /* Who have we been authenticated as ? */
     char *authuser;
 
     char ident_username[MAXHOSTNAMELEN + MAXHOSTNAMELEN + 2];
@@ -353,15 +348,12 @@ typedef struct SmtpState {
     struct SmtpSASLState sasl;
 #endif
 
+    void *irouter_state;
 } SmtpState;
 
 
 #define STYLE(i,c)	(strchr(((i)==NULL ? style : (i)->flags), (c)) != NULL)
 
-#define	WITH_SMTP	"SMTP"
-#define	WITH_ESMTP	"ESMTP"
-#define	WITH_BSMTP	"BSMTP"
-
 #define HELPMAX 40
 extern char *helplines[];
 #define HDR220MAX 4
@@ -369,84 +361,155 @@ extern char *hdr220lines[];
 extern char logtag[];
 extern time_t logtagepoch, now;
 
+/* Spool related counters */
 extern long availspace;
 extern long minimum_availspace;
 extern long maxsize;
+
+
+/* Global parameters */
 extern int use_ipv6;
-extern int tarpit_initial;
-extern int tarpit_exponent;
-extern int tarpit_toplimit;
-extern int MaxErrorRecipients;
-extern int TcpRcvBufferSize;
-extern int TcpXmitBufferSize;
-extern int ListenQueueSize;
 extern int MaxSameIpSource;
 extern int MaxParallelConnections;
 extern int percent_accept;
 extern int smtp_syslog;
 extern int allow_source_route;
-extern int debugcmdok;
-extern int expncmdok;
-extern int vrfycmdok;
-extern int pipeliningok;
-extern int mime8bitok;
-extern int chunkingok;
-extern int enhancedstatusok;
-extern int multilinereplies;
-extern int dsn_ok;
-extern int auth_ok;
-extern int ehlo_ok;
-extern int etrn_ok;
-extern int starttls_ok;
-extern int ssmtp_listen;
-extern int msa_mode;
-extern int deliverby_ok;
-#define MAX_ETRN_CLUSTER_IDX 40
-typedef struct {
-  char *nodename; char *username; char *password;
-} etrn_cluster_ent;
-extern etrn_cluster_ent etrn_cluster[];
-extern const char *tls_cert_file, *tls_key_file, *tls_CAfile, *tls_CApath;
-extern const char *tls_dcert_file, *tls_dkey_file, *tls_dh1024_param;
-extern const char *tls_dh512_param;
-extern const char *tls_random_source;
-extern const char *tls_cipherlist;
-extern const char *SASL_Auth_Mechanisms;
-extern int tls_loglevel, tls_enforce_tls, tls_ccert_vd, tls_use_scache;
-extern int tls_ask_cert, tls_req_cert, tls_scache_timeout;
-extern int log_rcvd_whoson, log_rcvd_ident, log_rcvd_authuser;
-extern int log_rcvd_tls_mode, log_rcvd_tls_peer;
-extern int auth_login_without_tls;
-extern char *smtpauth_via_pipe;
+
 extern char *contentfilter;
 extern int debug_content_filter;
-extern int strict_protocol;
-extern int rcptlimitcnt;
+extern char *perlhookpath;
+extern int use_perlhook;
+
 extern int enable_router;
-extern int use_tcpwrapper;
+extern int enable_router_maxpar;
 extern int configuration_ok;
 extern int unknown_cmd_limit;
+
+extern int msa_mode;  /* ?? */
 extern int sum_sizeoption_value;
 extern int lmtp_mode;
-extern int do_sasl;
-extern int MaxSLBits;
-extern char *AuthMechanisms;
-extern int detect_incorrect_tls_use;
-extern int force_rcpt_notify_never;
 
-extern int bindaddr_set, bindport_set, testaddr_set;
+
+#define MAX_SMTPSERVER_CLUSTER_IDX 40
+typedef struct {
+  Usockaddr addr;
+  char *secret;
+} smtpserver_cluster_ent;
+extern smtpserver_cluster_ent smtpserver_cluster[];
+
+#define MAX_ETRN_CLUSTER_IDX       40
+typedef struct {  /* Talks with SCHEDULER process */
+  char *nodename;
+  char *username;
+  char *password;
+} etrn_cluster_ent;
+extern etrn_cluster_ent etrn_cluster[];
+
+typedef struct ConfigParams_ {
+  double tarpit_initial;
+  double tarpit_exponent;
+  double tarpit_toplimit;
+
+  int MaxErrorRecipients;
+  int TcpRcvBufferSize;
+  int TcpXmitBufferSize;
+  int ListenQueueSize;
+
+  struct policytest *policydb; /* group, can default */
+
+  int log_rcvd_whoson, log_rcvd_ident, log_rcvd_authuser;
+  int log_rcvd_tls_mode, log_rcvd_tls_peer;
+
+  int auth_ok;
+  int auth_login_without_tls;
+  int no_smtp_auth_on_25;
+  char *smtp_auth_username_prompt; /* group, can default */
+  char *smtp_auth_password_prompt; /* group, can default */
+  int auth_failrate;
+
+  char *smtpauth_via_pipe;		/* group, can default */
+  const char *SASL_Auth_Mechanisms;	/* group, can default */
+  int do_sasl;
+  int MaxSLBits;
+  char *AuthMechanisms;			/* group, can default */
+
+  int ehlo_ok;
+  int etrn_ok;
+  int starttls_ok;
+  int deliverby_ok;
+  int ssmtp_listen;
+  int submit_listen;
+  int debugcmdok;
+  int expncmdok;
+  int vrfycmdok;
+  int pipeliningok;
+  int mime8bitok;
+  int chunkingok;
+  int enhancedstatusok;
+  int multilinereplies;
+  int dsn_ok;
+
+	/* group, can default */
+  const char *tls_cert_file, *tls_key_file, *tls_CAfile, *tls_CApath;
+  const char *tls_dcert_file, *tls_dkey_file, *tls_dh1024_param;
+  const char *tls_dh512_param;
+  const char *tls_random_source;
+  const char *tls_cipherlist;
+  int tls_loglevel, tls_enforce_tls, tls_ccert_vd;
+  int tls_ask_cert, tls_req_cert;
+
+  int tls_scache_timeout, tls_use_scache;
+  char *tls_scache_name;
+
+
+  const char *contact_pointer_message; 	/* group, can default */
+
+  const char *reportauthfile;		/* group, can default */
+
+  int rcptlimitcnt;
+  int use_tcpwrapper;
+
+  int detect_incorrect_tls_use;
+  int force_rcpt_notify_never;
+
+  int use_spf, spf_received, spf_threshold, spf_whitelist_use_default;
+  char *spf_localpolicy;		/* group, can default */
+
+		/* group locals: */
+  int bindaddr_set;
+  Usockaddr *bindaddrs;
+  int       *bindaddrs_types;
+  int       *bindaddrs_ports;
+  int        bindaddrs_count;
+#define BINDADDR_ALL    0xFFFF
+#define BINDADDR_SMTP   0x0001
+#define BINDADDR_SMTPS  0x0002
+#define BINDADDR_SUBMIT 0x0004
+} ConfigParams;
+
+extern ConfigParams CPdefault;
+extern ConfigParams **CPpSet;
+extern int CPpSetSize;
+extern ConfigParams *CP;
+extern ConfigParams *OCP;
+
+extern int bindport_set;
 extern u_short   bindport;
-extern Usockaddr *bindaddrs;
-extern int        bindaddrs_count;
-extern Usockaddr testaddr;
+
+extern int strict_protocol;
+
+extern int testaddr_set;
+extern  Usockaddr testaddr;
+
 
 extern const char *progname;
-extern int debug, skeptical, checkhelo, ident_flag, verbose;
+extern int debug, skeptical, checkhelo, ident_flag, verbose, daemon_flg;
 
 extern const char *style;
 
 extern struct smtpconf *readcffile __((const char *fname));
 extern struct smtpconf *findcf __((const char *host));
+extern void ConfigParams_newgroup __((void));
 
 extern int loadavg_current __((void));
 
@@ -516,8 +579,7 @@ extern int X_settrrc;
 extern void header_from_mime __((char *, int *, int));
 #endif				/* USE_TRANSLATION */
 
-extern void killr __((SmtpState * SS, int rpid));
-extern void killcfilter __((SmtpState * SS, int rpid));
+extern void killcfilter __((SmtpState * SS));
 extern void typeflush __((SmtpState *));
 #if defined(HAVE_STDARG_H) && defined(HAVE_VPRINTF)
 extern void type __((SmtpState *, int code, const char *status, const char *fmt,...));
@@ -541,17 +603,21 @@ extern void runasrootuser __((void));
 extern int runastrusteduser __((void));
 extern char **environ;
 extern int kill __((pid_t, int));
+extern RETSIGTYPE sigchld  __((int sig));
+extern volatile int sawsigchld;
 
 extern const char *rfc822atom __((const char *str));
 extern const char *xtext_string __((const char *str));
 
-extern void s_setup  __((SmtpState * SS, int infd, int outfd));
 extern void s_ungetc __((SmtpState *SS, int ch));
 extern int s_feof __((SmtpState * SS));
+extern int s_seen_eof __((SmtpState * SS));
 extern int s_getc __((SmtpState * SS, int timeout_is_fatal));
 extern int s_hasinput __((SmtpState * SS));
 extern int s_gets __((SmtpState *SS, char *buf, int buflen, int *rcp, char *cop, char *cp));
 
+extern void zsleep __((int delay));
+
 extern int errno;
 extern int optind;
 extern char *optarg;
@@ -568,9 +634,10 @@ extern char *optarg;
 #define	putc	fputc
 #endif				/* lint */
 
-extern int  childsameip __((Usockaddr *addr, int *childcntp));
+extern int  childsameip __((Usockaddr *addr, int, int *childcntp));
 extern void childregister __((int cpid, Usockaddr *addr, int tag));
 extern void childreap   __((int cpid));
+extern void disable_childreap __((void));
 
 extern void smtp_helo   __((SmtpState * SS, const char *buf, const char *cp));
 extern int  smtp_mail   __((SmtpState * SS, const char *buf, const char *cp, int insecure));
@@ -586,6 +653,7 @@ extern void smtp_tarpit __((SmtpState * SS));
 extern void smtp_auth __((SmtpState * SS, const char *buf, const char *cp));
 extern void smtpauth_init __((SmtpState * SS));
 extern void smtpauth_ehloresponse __((SmtpState * SS));
+extern int  smtp_report __((SmtpState * SS, const char *buf, char *cp));
 
 #ifdef HAVE_OPENSSL
 extern int tls_start_servertls __((SmtpState *SS));
@@ -619,3 +687,128 @@ extern int decodebase64string __((const char *instr, int inlen, char *outstr, in
 extern int  fd_nonblockingmode __((int fd));
 extern int  fd_blockingmode __((int fd));
 extern void fd_restoremode __((int fd, int mode));
+
+/* subdaemons.c */
+extern int  subdaemons_init                   __((void));
+extern int  subdaemons_init_router            __((void));
+extern int  subdaemons_init_ratetracker       __((void));
+extern int  subdaemons_init_contentfilter     __((void));
+extern void subdaemons_kill_cluster_listeners __((void));
+
+struct fdgets_fdbuf {
+	int rdsize;
+	char rdbuf[100];
+};
+extern int fdgets __((char **bufp, int endi, int *buflenp, struct fdgets_fdbuf *fdp, int fd, int timeout));
+
+extern int  ratetracker_rdz_fd;
+extern int  ratetracker_server_pid;
+
+extern int  router_rdz_fd;
+extern int  router_server_pid;
+
+extern int  contentfilter_rdz_fd;
+extern int  contentfilter_server_pid;
+
+
+struct peerhead;           /* Forward declarator */
+struct subdaemon_handler;  /* Forward declarator */
+
+struct peerdata {
+	int  fd;
+	int  inlen;
+	int  outlen, outptr;
+	int  inpspace;
+	int  outspace;
+	int  in_job;		/* A job ready to send to server.. */
+	time_t when_in;		/* Arrival time */
+	char *inpbuf;		/* Grown to fit in an input line..
+				   The input protocol shall be SINGLE
+				   text line with '\n' at the end! */
+	char *outbuf;		/* Written out "synchronously" to the
+				   socket/pipe buffer space.
+				   About 200 chars should be enough.
+				   If not, adding automated buffer
+				   expansion codes is trivialish...  */
+	struct peerdata *prev, *next;	/* Job order.  Add to tail,
+					   process from head. */
+	struct peerhead *head;
+	struct subdaemon_handler *handler;
+	struct zmpollfd *pollfd;
+};
+
+struct peerhead {
+	struct peerdata		* head;
+	struct peerdata		* tail;
+	int			  queuecount;
+};
+
+struct subdaemon_state {
+  void *p; /* dummy */
+};
+
+struct subdaemon_handler {
+	int (*init)      __((struct subdaemon_state **statepp));
+	int (*input)     __((struct subdaemon_state *state, struct peerdata *));
+	int (*prepoll)   __((struct subdaemon_state *state, struct zmpollfd **fds, int *fdcount));
+	int (*postpoll)  __((struct subdaemon_state *state, struct zmpollfd *fds, int fdcount));
+	int (*shutdown)  __((struct subdaemon_state *state));
+        int (*killpeer)  __((struct subdaemon_state *state, struct peerdata *));
+	int (*reaper)    __((struct subdaemon_state *state));
+	int (*sigusr2)   __((struct subdaemon_state *state));
+	Vuint *reply_delay_G;	/* MIB variable pointer */
+	Vuint *reply_queue_G;   /* MIB variable pointer */
+};
+
+extern int subdaemon_send_to_peer __((struct peerdata *, const char *, int));
+
+
+extern struct subdaemon_handler subdaemon_handler_ratetracker;
+extern struct subdaemon_handler subdaemon_handler_contentfilter;
+
+extern void subdaemon_router        __((int fd));
+extern void subdaemon_ratetracker   __((int fd));
+extern void subdaemon_contentfilter __((int fd));
+
+/* subdaemon-rtr.c */
+extern char * call_subdaemon_rtr __((void **, const char*, const char *, int, int*));
+extern struct subdaemon_handler subdaemon_handler_router;
+
+/* subdaemon-trk.c */
+extern int call_subdaemon_trk __((void **statep,const char *cmd, char *retbuf, int retbuflen));
+extern int call_subdaemon_trk_getmore __((void *statep,char *retbuf, int retbuflen));
+extern int  smtp_report_ip __((SmtpState *SS, const char *ip));
+extern int  smtp_report_dump __((SmtpState *SS));
+extern int  call_rate_counter __((struct policystate *state, int incr, PolicyTest what, int *countp, int *countp2));
+
+
+
+
+/* contentpolicy.c */
+extern int contentpolicy __((struct policystate *ps, const char *fname));
+/* subdaemon-ctf.c */
+extern int contentfilter_maxctfs;
+extern char *contentfilter_proc __((void **, const char *fname));
+extern void  smtpcontentfilter_kill __((void *));
+
+extern int mx_client_verify  __((struct policystate *, int, const char *, int));
+extern int sender_dns_verify __((struct policystate *, int, const char *, int));
+extern int client_dns_verify __((struct policystate *, int, const char *, int));
+extern int rbl_dns_test __((struct policystate *, const int, const u_char *, char *, char **));
+
+/* smtphook.c */
+#ifdef DO_PERL_EMBED
+
+#define ZSMTP_HOOK_HELO        "ZSMTP::hook::helo"
+#define ZSMTP_HOOK_RSET        "ZSMTP::hook::rset"
+#define ZSMTP_HOOK_MAILFROM    "ZSMTP::hook::mailfrom"
+#define ZSMTP_HOOK_RCPTTO      "ZSMTP::hook::rcptto"
+#define ZSMTP_HOOK_DATA        "ZSMTP::hook::data"
+
+extern int  ZSMTP_hook_init          __((void));
+extern void ZSMTP_hook_atexit        __((void));
+extern void ZSMTP_hook_set_ipaddress __((const char *, int, const char *, const char *, int, const char *));
+extern void ZSMTP_hook_set_user      __((const char *, const char *));
+extern int  ZSMTP_hook_univ          __((const char *, struct policystate *, const char *, const int, int *));
+
+#endif
