@@ -4,7 +4,7 @@
  *
  *  Contains ALSO code for SMTP Transport Agent!
  *
- *  by Matti Aarnio <mea@nic.funet.fi> 1999, 2003
+ *  by Matti Aarnio <mea@nic.funet.fi> 1999, 2003-2006
  *
  *  Reusing TLS code for POSTFIX by:
  *     Lutz Jaenicke <Lutz.Jaenicke@aet.TU-Cottbus.DE>
@@ -16,26 +16,22 @@
 
 #ifdef HAVE_OPENSSL
 
-/*
- * We are saving sessions to disc, we want to make sure, that the lenght of
- * the filename is somehow limited. When saving client sessions, the hostname
- * is transformed to an MD5-hash, which is defined by RFC to be 16 bytes long.
- * The length of the actual session id is however not defined in RFC2246.
- * OpenSSL defines a SSL_MAX_SSL_SESSION_ID_LENGTH of 32, but nobody
- * guarantees, that a client might not try to resume a session with a longer
- * session id. So to make sure, we define an upper bound of 32.
- */
+#ifdef HAVE_DISTCACHE
+#include <distcache/dc_client.h>
+
+static DC_CTX *dc_ctx;
 
 static const char MAIL_TLS_SRVR_CACHE[] = "TLSsrvrcache";
 static const int id_maxlength = 32;	/* Max ID length in bytes */
 static char server_session_id_context[] = "ZMailer/TLS"; /* anything will do */
+#endif
 
 static int do_dump = 0;
 static int verify_depth = 1;
 static int verify_error = X509_V_OK;
 
-int tls_scache_timeout = 3600;
-int tls_use_scache = 0;
+#define SSL_SESSION_MAX_DER 10*1024
+
 
 /* We must keep some of info available */
 static const char hexcodes[] = "0123456789ABCDEF";
@@ -48,21 +44,12 @@ struct _randseed {
 } tls_randseed;
 
 
-static void
-mail_queue_path(buf, subdir, filename)
-     char *buf;
-     char *subdir;
-     char *filename;
-{
-  sprintf(buf, "%s/%s/%s", postoffice, subdir, filename);
-}
-
 void
 smtp_starttls(SS, buf, cp)
      SmtpState *SS;
      const char *buf, *cp;
 {
-    if (!starttls_ok) {
+    if (!OCP->starttls_ok) {
       /* Ok, then 'command not implemented' ... */
       type(SS, 502, m540, NULL);
       return;
@@ -77,7 +64,8 @@ smtp_starttls(SS, buf, cp)
       return;
     }
 
-    if (!strict_protocol) while (*cp == ' ' || *cp == '\t') ++cp;
+    if (strict_protocol < 1) /* Skip extra white-spaces */
+      while (*cp == ' ' || *cp == '\t') ++cp;
     if (*cp != 0) {
       type(SS, 501, m513, "Extra junk following 'STARTTLS' command!");
       MIBMtaEntry->ss.IncomingSMTP_STARTTLS_fail += 1;
@@ -89,6 +77,8 @@ smtp_starttls(SS, buf, cp)
     if (SS->mfp != NULL) {
       clearerr(SS->mfp);
       mail_abort(SS->mfp);
+      policytest(&SS->policystate, POLICY_DATAABORT,
+		 NULL, SS->rcpt_count, NULL);
       SS->mfp = NULL;
     }
 
@@ -100,11 +90,14 @@ smtp_starttls(SS, buf, cp)
       if (SS->mfp != NULL) {
 	clearerr(SS->mfp);
 	mail_abort(SS->mfp);
+	policytest(&SS->policystate, POLICY_DATAABORT,
+		   NULL, SS->rcpt_count, NULL);
 	SS->mfp = NULL;
       }
       MIBMtaEntry->ss.IncomingSMTP_STARTTLS_fail += 1;
       exit(2);
     }
+    SS->with_protocol_set |= WITH_TLS;
 }
 
 
@@ -139,8 +132,19 @@ int
 Z_pending(SS)
      SmtpState * SS;
 {
+    int rc;
+    static struct zmpollfd *pollfds = NULL;
+    int fdcount = 0;
+
     if (SS->sslmode)
       return SSL_pending(SS->TLS.ssl);
+
+    zmpoll_addfd( &pollfds, &fdcount, SS->inputfd, -1, NULL);
+
+    rc = zmpoll( pollfds, fdcount, 0 );
+
+    if (rc > 0) return 1;
+
     return 0;
 }
 
@@ -176,6 +180,23 @@ Z_SSL_flush(SS)
 }
 
 
+#ifdef HAVE_DISTCACHE
+
+
+static DC_CTX      *ssl_scache_dc_init     __((void));
+static SSL_SESSION *ssl_scache_dc_retrieve __((SSL *, unsigned char *, int));
+static int          ssl_scache_dc_store    __((SSL_SESSION *, unsigned char *, int, time_t));
+static void         ssl_scache_dc_remove   __((SSL_SESSION *, unsigned char *, int));
+
+
+/* we initialize this index at startup time
+ * and never write to it at request time,
+ * so this static is thread safe.
+ * also note that OpenSSL increments at static variable when
+ * SSL_get_ex_new_index() is called, so we _must_ do this at startup.
+ */
+static int TLScontext_index = -1;
+
 
 /*
  * Callback to retrieve a session from the external session cache.
@@ -184,61 +205,10 @@ static SSL_SESSION *get_session_cb(SSL *ssl, unsigned char *SessionID,
 				   int length, int *copy)
 {
     SSL_SESSION *session;
-    char *buf;
-    FILE *fp;
-    struct stat st;
-    char *idstring;
-    int n;
-    int uselength;
-    int verify_result;
 
-    if (length > id_maxlength)
-	uselength = id_maxlength;	/* Limit length of ID */
-    else
-	uselength = length;
+    session = ssl_scache_dc_retrieve(ssl, SessionID, length);
 
-    idstring = (char *)malloc(2 * uselength + 1);
-    if (!idstring) {
-	type(NULL,0,NULL, "could not allocate memory for IDstring");
-	return (NULL);
-    }
-
-    for(n=0 ; n < uselength ; n++)
-	sprintf(idstring + 2 * n, "%02X", SessionID[n]);
-    if (tls_loglevel >= 3)
-      type(NULL,0,NULL, "Trying to reload Session from disc: %s", idstring);
-
-    /*
-     * The constant "100" is taken from mail_queue.c and also used there.
-     * It must hold the name the postfix spool directory (if not chrooted)
-     * and the hash directory forest.
-     */
-    buf = malloc(100 + 2 * uselength + 1);
-    mail_queue_path(buf, MAIL_TLS_SRVR_CACHE, idstring);
-
-    /*
-     * Try to read the session from the file. If the file exists, but its
-     * mtime is so old, that the session has already expired, we dont
-     * waste time anymore, we rather delete the session file immediately.
-     */
-    session = NULL;
-    if (stat(buf, &st) == 0) {
-	if (st.st_mtime + tls_scache_timeout < time(NULL))
-            unlink(buf);
-	else if ((fp = fopen(buf, "r")) != 0) {
-	    if (fscanf(fp, "%d", &verify_result) <= 0)
-		verify_result = X509_V_ERR_APPLICATION_VERIFICATION;
-	    SSL_set_verify_result(ssl, verify_result);
-	    session = PEM_read_SSL_SESSION(fp, NULL, NULL, NULL);
-	    fclose(fp);
-	}
-    }
-
-    free(buf);
-    free(idstring);
-
-    if (session && (tls_loglevel >= 3))
-      type(NULL,0,NULL, "Successfully reloaded session from disc");
+    *copy = 0;
 
     return (session);
 }
@@ -249,66 +219,96 @@ static SSL_SESSION *get_session_cb(SSL *ssl, unsigned char *SessionID,
  */
 static int new_session_cb(SSL *ssl, SSL_SESSION *session)
 {
-    char *buf;
-    FILE *fp;
-    char *myname = "new_session_cb";
-    char *idstring;
-    int n;
-    int uselength;
-    int fd;
-    int success;
+    unsigned char *id;
+    unsigned int idlen;
+    int rc;
+    long timeout = OCP->tls_scache_timeout;
 
-    if (session->session_id_length > id_maxlength)
-	uselength = id_maxlength;	/* Limit length of ID */
-    else
-	uselength = session->session_id_length;
+    SSL_set_timeout(session, timeout);
+    id    = session->session_id;
+    idlen = session->session_id_length;
 
-    idstring = (char *)malloc(2 * uselength + 1);
-    if (!idstring) {
-      type(NULL,0,NULL, "could not allocate memory for IDstring");
-      return -1;
-    }
+    timeout += SSL_SESSION_get_time(session);
 
-    for(n=0 ; n < uselength ; n++)
-	sprintf(idstring + 2 * n, "%02X", session->session_id[n]);
+    rc = ssl_scache_dc_store(session,id,idlen,timeout);
 
-    if (tls_loglevel >= 3)
-      type(NULL,0,NULL, "Trying to save Session to disc: %s", idstring);
-
-    buf = malloc(100 + 2 * uselength + 1);
-    mail_queue_path(buf, MAIL_TLS_SRVR_CACHE, idstring);
-
-    /*
-     * Now open the session file in exclusive and create mode. If it
-     * already exists, we dont touch it and silently omit the save.
-     * We cannot use Wietses VSTREAM code here, as PEM_write uses
-     * Cs normal buffered library and we better dont mix.
-     * The return value of PEM_write_SSL_SESSION is nowhere documented,
-     * but from the source it seems to be something like the number
-     * of lines or bytes written. Anyway, success is positiv and
-     * failure is zero.
-     */
-    if ((fd = open(buf, O_WRONLY | O_CREAT | O_EXCL, 0600)) >= 0) {
-      if ((fp = fdopen(fd, "w")) == 0) {
-	type(NULL,0,NULL, "%s: could not fdopen %s: %s",
-	     myname, buf, strerror(errno));
-	return -1;
-      }
-      fprintf(fp, "%lu\n", (unsigned long)SSL_get_verify_result(ssl));
-      success = PEM_write_SSL_SESSION(fp, session);
-      fclose(fp);
-      if (success == 0)
-	unlink(buf);
-      else if (tls_loglevel >= 3)
-	type(NULL,0,NULL, "Successfully saved session to disc");
-    }
-
-    free(buf);
-    free(idstring);
-
-    return (0);
+    return rc;
 }
 
+/*
+ * Remove a session from the external cache
+ */
+static void remove_session_cb(SSL_CTX *ssl, SSL_SESSION *session)
+{
+    unsigned char *id;
+    unsigned int idlen;
+
+    id    = session->session_id;
+    idlen = session->session_id_length;
+
+    ssl_scache_dc_remove(session,id,idlen);
+}
+
+
+
+static void tls_scache_init(ssl_ctx)
+     SSL_CTX *ssl_ctx;
+{
+
+	/*
+	 * Initialize the DISTCACHE context.
+	 */
+
+	dc_ctx = ssl_scache_dc_init();
+	if (!dc_ctx) return; /* No can do.. */
+
+	/*
+	 * Initialize the session cache. We only want external caching to
+	 * synchronize between server sessions, so we set it to a minimum value
+	 * of 1. If the external cache is disabled, we won't cache at all.
+	 * The recall of old sessions "get" and save to disk of just created
+	 * sessions "new" is handled by the appropriate callback functions.
+	 *
+	 * We must not forget to set a session id context to identify to which
+	 * kind of server process the session was related. In our case, the
+	 * context is just the name of the patchkit: "Postfix/TLS".
+	 */
+
+	SSL_CTX_sess_set_cache_size(ssl_ctx, 1);
+	SSL_CTX_set_timeout(ssl_ctx, OCP->tls_scache_timeout);
+
+	SSL_CTX_set_session_id_context(ssl_ctx,
+				       (void*)&server_session_id_context,
+				       sizeof(server_session_id_context));
+
+	/*
+	 * The session cache is realized by distcache, if at all..
+	 */
+	if (OCP->tls_scache_name) {
+	    SSL_CTX_set_session_cache_mode(ssl_ctx,
+					   ( SSL_SESS_CACHE_SERVER |
+					     SSL_SESS_CACHE_NO_INTERNAL ));
+	    SSL_CTX_sess_set_get_cb(ssl_ctx,    get_session_cb);
+	    SSL_CTX_sess_set_new_cb(ssl_ctx,    new_session_cb);
+	    SSL_CTX_sess_set_remove_cb(ssl_ctx, remove_session_cb);
+	}
+	
+	/*
+	 * Finally create the global index to access TLScontext information
+	 * inside verify_callback.
+	 */
+	if (TLScontext_index < 0) {
+	  /* we _do_ need to call this twice */
+	  TLScontext_index = SSL_get_ex_new_index(0,
+						  "TLScontext ex_data index",
+						  NULL, NULL, NULL);
+	  TLScontext_index = SSL_get_ex_new_index(0,
+						  "TLScontext ex_data index",
+						  NULL, NULL, NULL);
+	}
+}
+#endif
+
 
 /* skeleton taken from OpenSSL crypto/err/err_prn.c */
 
@@ -390,7 +390,7 @@ tmp_rsa_cb(s, export, keylength)
     static RSA *rsa_tmp = NULL;
 
     if (rsa_tmp == NULL) {
-	if (tls_loglevel >= 2)
+	if (OCP->tls_loglevel >= 2)
 	    type(NULL,0,NULL,"Generating temp (%d bit) RSA key...", keylength);
 	rsa_tmp = RSA_generate_key(keylength, RSA_F4, NULL, NULL);
     }
@@ -450,7 +450,7 @@ verify_callback(ok, ctx)
     depth = X509_STORE_CTX_get_error_depth(ctx);
 
     X509_NAME_oneline(X509_get_subject_name(err_cert), buf, 256);
-    if (tls_loglevel >= 1)
+    if (OCP->tls_loglevel >= 1)
 	type(NULL,0,NULL,"Client cert verify depth=%d %s", depth, buf);
     if (!ok) {
 	type(NULL,0,NULL,"verify error:num=%d:%s", err,
@@ -477,7 +477,7 @@ verify_callback(ok, ctx)
 	type(NULL,0,NULL,"cert has expired");
 	break;
     }
-    if (tls_loglevel >= 1)
+    if (OCP->tls_loglevel >= 1)
 	type(NULL,0,NULL,"verify return:%d", ok);
     return (ok);
 }
@@ -504,7 +504,7 @@ apps_ssl_info_callback(s, where, ret)
 	str = "undefined";
 
     if (where & SSL_CB_LOOP) {
-	if (tls_loglevel >= 2)
+	if (OCP->tls_loglevel >= 2)
 	    type(NULL,0,NULL,"%s:%s", str, SSL_state_string_long(s));
     } else if (where & SSL_CB_ALERT) {
 	str = (where & SSL_CB_READ) ? "read" : "write";
@@ -611,6 +611,23 @@ bio_dump_cb(bio, cmd, argp, argi, argl, ret)
      long argl;
      long ret;
 {
+#if 0 /* NOT proper code!  must do in  bio->method->bread()  */
+    static int once = 1;
+
+    if (cmd == BIO_CB_READ && once) {
+      /* This callback is done ONCE per process lifetime ...
+	 ... thus we optimize a bit.  */
+      SmtpState *SS = (SmtpState *)BIO_get_callback_arg(bio);
+
+      once = 0;
+      if (SS->s_ungetcbuf >= 0) {
+	*(char*)argp = SS->s_ungetcbuf;
+	SS->s_ungetcbuf = -1;
+	return 1;
+      }
+    }
+#endif
+
     if (!do_dump)
 	return (ret);
 
@@ -837,7 +854,7 @@ static int tls_randseeder(const char *source)
   */
 
 static int tls_serverengine = 0;
-static SSL_CTX *ssl_ctx = NULL;
+static SSL_CTX * ssl_ctx;
 
 int
 tls_init_serverengine(verifydepth, askcert, requirecert)
@@ -857,7 +874,7 @@ tls_init_serverengine(verifydepth, askcert, requirecert)
 	if (tls_serverengine)
 	  return (0);				/* already running */
 
-	if (tls_loglevel >= 1)
+	if (OCP->tls_loglevel >= 1)
 	  type(NULL,0,NULL,"starting TLS engine");
 	
 	/*
@@ -906,7 +923,8 @@ tls_init_serverengine(verifydepth, askcert, requirecert)
 	while ( 1 ) {
 
 	  /* Parametrized version ? */
-	  if (tls_random_source && tls_randseeder(tls_random_source) >= 0)
+	  if (OCP->tls_random_source &&
+	      tls_randseeder(OCP->tls_random_source) >= 0)
 	    break;
 	  
 	  /* How about  /dev/urandom  ?  */
@@ -964,8 +982,8 @@ tls_init_serverengine(verifydepth, askcert, requirecert)
 	 * Set the list of ciphers, if explicitely given; otherwise the
 	 * (reasonable) default list is kept.
 	 */
-	if (tls_cipherlist) {
-	  if (SSL_CTX_set_cipher_list(ssl_ctx, tls_cipherlist) == 0) {
+	if (OCP->tls_cipherlist) {
+	  if (SSL_CTX_set_cipher_list(ssl_ctx, OCP->tls_cipherlist) == 0) {
 	    tls_print_errors();
 	    return (-1);
 	  }
@@ -988,14 +1006,14 @@ tls_init_serverengine(verifydepth, askcert, requirecert)
 	 * hand, the file is not really readable.
 	 */
 
-	if (!tls_CAfile || *tls_CAfile == 0)
+	if (!OCP->tls_CAfile || *OCP->tls_CAfile == 0)
 	  CAfile = NULL;
 	else
-	  CAfile = tls_CAfile;
-	if (!tls_CApath || *tls_CApath == 0)
+	  CAfile = OCP->tls_CAfile;
+	if (!OCP->tls_CApath || *OCP->tls_CApath == 0)
 	  CApath = NULL;
 	else
-	  CApath = tls_CApath;
+	  CApath = OCP->tls_CApath;
 	
 	/*
 	 * Now we load the certificate and key from the files and check,
@@ -1003,7 +1021,7 @@ tls_init_serverengine(verifydepth, askcert, requirecert)
 	 * set_cert_stuff().   We cannot run without.
 	 */
 
-	if (tls_ask_cert && (!CApath && !CAfile)) {
+	if (OCP->tls_ask_cert && (!CApath && !CAfile)) {
 	  type(NULL,0,NULL,"TLS engine: No CA certificate file/directory defined, and asking for client certs");
 	  return (-1);
 	}
@@ -1012,7 +1030,7 @@ tls_init_serverengine(verifydepth, askcert, requirecert)
 	    (!SSL_CTX_set_default_verify_paths(ssl_ctx))) {
 	  /* Consider this to be fatal ONLY if client
 	     certificates really are required ( = hardly ever) */
-	  if (tls_ask_cert && tls_req_cert) {
+	  if (OCP->tls_ask_cert && OCP->tls_req_cert) {
 	    type(NULL,0,NULL,"TLS engine: cannot load CA data");
 	    tls_print_errors();
 	    return (-1);
@@ -1035,24 +1053,24 @@ tls_init_serverengine(verifydepth, askcert, requirecert)
 	 * changed in the cipher setup.
 	 */
 
-	if (!tls_cert_file || *tls_cert_file == 0)
+	if (!OCP->tls_cert_file || *OCP->tls_cert_file == 0)
 	  s_cert_file = NULL;
 	else
-	  s_cert_file = tls_cert_file;
-	if (!tls_key_file  || *tls_key_file == 0)
+	  s_cert_file = OCP->tls_cert_file;
+	if (!OCP->tls_key_file  || *OCP->tls_key_file == 0)
 	  s_key_file = NULL;
 	else
-	  s_key_file = tls_key_file;
+	  s_key_file = OCP->tls_key_file;
 
-	if (!tls_dcert_file || *tls_dcert_file == 0)
+	if (!OCP->tls_dcert_file || *OCP->tls_dcert_file == 0)
 	  s_dcert_file = NULL;
 	else
-	  s_dcert_file = tls_dcert_file;
+	  s_dcert_file = OCP->tls_dcert_file;
 
-	if (!tls_dkey_file || *tls_dkey_file == 0)
+	if (!OCP->tls_dkey_file || *OCP->tls_dkey_file == 0)
 	  s_dkey_file = NULL;
 	else
-	  s_dkey_file = tls_dkey_file;
+	  s_dkey_file = OCP->tls_dkey_file;
 
 	if (s_cert_file) {
 	  if (!set_cert_stuff(ssl_ctx, s_cert_file, s_key_file)) {
@@ -1095,15 +1113,15 @@ tls_init_serverengine(verifydepth, askcert, requirecert)
 
 	SSL_CTX_set_tmp_dh_callback(ssl_ctx, tmp_dh_cb);
 	
-	if (tls_dh1024_param) {
-	  dh_1024 = load_dh_param(tls_dh1024_param);
+	if (OCP->tls_dh1024_param) {
+	  dh_1024 = load_dh_param(OCP->tls_dh1024_param);
 	  if (!dh_1024) {
 	    type(NULL,0,NULL,"TLS engine: could not load 1024bit DH parameters from given file; will use built-in default value");
 	    tls_print_errors();
 	  }
 	}
-	if (tls_dh512_param) {
-	  dh_512 = load_dh_param(tls_dh512_param);
+	if (OCP->tls_dh512_param) {
+	  dh_512 = load_dh_param(OCP->tls_dh512_param);
 	  if (!dh_512) {
 	    type(NULL,0,NULL,"TLS engine: could not load 512bit DH parameters from given file; will use builtin default value");
 	    tls_print_errors();
@@ -1153,67 +1171,10 @@ tls_init_serverengine(verifydepth, askcert, requirecert)
 	SSL_CTX_set_verify(ssl_ctx, verify_flags, verify_callback);
 	SSL_CTX_set_client_CA_list(ssl_ctx, SSL_load_client_CA_file(CAfile));
 	
-	/*
-	 * Initialize the session cache. We only want external caching to
-	 * synchronize between server sessions, so we set it to a minimum value
-	 * of 1. If the external cache is disabled, we won't cache at all.
-	 * The recall of old sessions "get" and save to disk of just created
-	 * sessions "new" is handled by the appropriate callback functions.
-	 *
-	 * We must not forget to set a session id context to identify to which
-	 * kind of server process the session was related. In our case, the
-	 * context is just the name of the patchkit: "Postfix/TLS".
-	 */
-
-	SSL_CTX_sess_set_cache_size(ssl_ctx, 1);
-	SSL_CTX_set_timeout(ssl_ctx, tls_scache_timeout);
-	{
-	  static char server_session_id_context[] = "ZMailer/Smtpserver/TLS"; /* anything will do */
-
-	  SSL_CTX_set_session_id_context(ssl_ctx,
-					 (void*)&server_session_id_context,
-					 sizeof(server_session_id_context));
-	}
 
 
-#if 0
-	/*
-	 * The session cache is realized by an external database file, that
-	 * must be opened before going to chroot jail. Since the session cache
-	 * data can become quite large, "[n]dbm" cannot be used as it has a
-	 * size limit that is by far too small.
-	 */
-	if (*var_smtpd_tls_scache_db) {
-	  /*
-	   * Insert a test against other dbms here, otherwise while writing
-	   * a session (content to large), we will receive a fatal error!
-	   */
-	  if (strncmp(var_smtpd_tls_scache_db, "sdbm:", 5))
-	    msg_warn("Only sdbm: type allowed for %s",
-		     var_smtpd_tls_scache_db);
-	  else
-	    scache_db = dict_open(var_smtpd_tls_scache_db, O_RDWR,
-				  ( DICT_FLAG_DUP_REPLACE | DICT_FLAG_LOCK |
-				    DICT_FLAG_SYNC_UPDATE ));
-	  if (scache_db) {
-	    SSL_CTX_set_session_cache_mode(ctx,
-					   ( SSL_SESS_CACHE_SERVER |
-					     SSL_SESS_CACHE_NO_AUTO_CLEAR ));
-	    SSL_CTX_sess_set_get_cb(ctx, get_session_cb);
-	    SSL_CTX_sess_set_new_cb(ctx, new_session_cb);
-	    SSL_CTX_sess_set_remove_cb(ctx, remove_session_cb);
-	  }
-	  else
-	    msg_warn("Could not open session cache %s",
-		     var_smtpd_tls_scache_db);
-	}
-	
-	/*
-	 * Finally create the global index to access TLScontext information
-	 * inside verify_callback.
-	 */
-	TLScontext_index = SSL_get_ex_new_index(0, "TLScontext ex_data index",
-						NULL, NULL, NULL);
+#ifdef HAVE_DISTCACHE
+	tls_scache_init(ssl_ctx);
 #endif
 
 	tls_serverengine = 1;
@@ -1288,6 +1249,38 @@ static void tls_reset(SMTPD_STATE *state)
 #endif
 
 
+static int z_rbio_bread __((BIO*, char *, int));
+static int
+z_rbio_bread(b, out, outl)
+	BIO *b;
+	char *out;
+	int outl;
+{
+	int ret=0;
+	SmtpState *SS = (SmtpState *)b->cb_arg;
+
+	/* type(NULL,0,NULL,"z_bio_bread(b, 0x%p, %d)", out, outl); */
+
+	if (out) {
+	  /* errno = 0; */
+	  if (SS && SS->s_ungetcbuf >= 0) {
+	    *out = SS->s_ungetcbuf;
+	    SS->s_ungetcbuf = -1;
+	    BIO_clear_retry_flags(b);
+	    /* BIO_set_retry_read(b); */
+	    return 1;
+	  }
+
+	  ret = read (b->num,out,outl);
+	  BIO_clear_retry_flags(b);
+	  if (ret <= 0) {
+	    if (BIO_sock_should_retry(ret))
+	      BIO_set_retry_read(b);
+	  }
+	}
+	return(ret);
+}
+
 int
 tls_start_servertls(SS)
      SmtpState *SS;
@@ -1297,7 +1290,12 @@ tls_start_servertls(SS)
     SSL_SESSION * session;
     SSL_CIPHER  * cipher;
     X509	* peer;
-    char	cbuf[4000];
+    char	  cbuf[ 4000 ];
+    const char  * kp;
+
+    BIO		*wbio, *rbio;
+    BIO_METHOD  *rbiomethod_old;
+    static BIO_METHOD  rbiomethod_new;
 
     /*
      * If necessary, setup a new SSL structure for a connection.
@@ -1365,6 +1363,21 @@ tls_start_servertls(SS)
     }
 
 
+    /*
+     * Now we do deep magic in rbio methods.
+     * We need to have our own low-level read routine
+     * so that we can do ungetc processing properly..
+     *
+     */
+    rbio = SSL_get_rbio(SS->TLS.ssl);
+    rbiomethod_old = rbio->method;
+    rbiomethod_new = *rbiomethod_old;
+    rbiomethod_new.bread = z_rbio_bread;
+    rbio->method = &rbiomethod_new;
+
+    BIO_set_callback_arg(rbio, SS);
+
+
 #if 0
     /*
      * Before really starting anything, try to seed the PRNG a little bit
@@ -1389,11 +1402,11 @@ tls_start_servertls(SS)
      * Well there is a BIO below the SSL routines that is automatically
      * created for us, so we can use it for debugging purposes.
      */
-    if (tls_loglevel >= 3)
-	BIO_set_callback(SSL_get_rbio(SS->TLS.ssl), bio_dump_cb);
+    /* if (OCP->tls_loglevel >= 3) */
+      BIO_set_callback(SSL_get_rbio(SS->TLS.ssl), bio_dump_cb);
 
     /* Dump the negotiation for loglevels 3 and 4*/
-    if (tls_loglevel >= 3)
+    if (OCP->tls_loglevel >= 3)
 	do_dump = 1;
 
     /*
@@ -1410,9 +1423,8 @@ tls_start_servertls(SS)
      */
     for (;;) {
 	int sslerr, rc, i;
-	BIO *wbio, *rbio;
-	fd_set rdset, wrset;
-	struct timeval tv;
+	static struct zmpollfd *fds = NULL;
+	int fdscount = 0;
 	int wantreadwrite = 0;
 
     ssl_accept_retry:;
@@ -1474,18 +1486,17 @@ tls_start_servertls(SS)
 	}
 
 	i = SSL_get_fd(SS->TLS.ssl);
-	_Z_FD_ZERO(wrset);
-	_Z_FD_ZERO(rdset);
+	fdscount = 0;
 
 	if (wantreadwrite < 0)
-	  _Z_FD_SET(i, rdset); /* READ WANTED */
+	  zmpoll_addfd(&fds, &fdscount, i, -1, NULL); /* WANT READ */
 	else if (wantreadwrite > 0)
-	  _Z_FD_SET(i, wrset); /* WRITE WANTED */
+	  zmpoll_addfd(&fds, &fdscount, -1, i, NULL); /* WANT WRITE */
+	else
+	  ; /* FIXME! What???  No reading, nor writing ??? */
 
-	tv.tv_sec = 300;
-	tv.tv_usec = 0;
 
-	rc = select(i+1, &rdset, &wrset, NULL, &tv);
+	rc = zmpoll( fds, fdscount, 300 * 1000 ); /* 300 seconds */
 	sslerr = errno;
 
 	if (rc == 0) {
@@ -1509,7 +1520,7 @@ tls_start_servertls(SS)
 
 
     /* Only loglevel==4 dumps everything */
-    if (tls_loglevel < 4)
+    if (OCP->tls_loglevel < 4)
 	do_dump = 0;
     /*
      * Lets see, whether a peer certificate is available and what is
@@ -1524,13 +1535,13 @@ tls_start_servertls(SS)
 
 	X509_NAME_oneline(X509_get_subject_name(peer),
 			  cbuf, sizeof(cbuf));
-	if (tls_loglevel >= 1)
+	if (OCP->tls_loglevel >= 1)
 	    type(NULL,0,NULL,"subject=%s", cbuf);
 	SS->TLS.peer_subject = strdup(cbuf);
 
 	X509_NAME_oneline(X509_get_issuer_name(peer),
 			  cbuf, sizeof(cbuf));
-	if (tls_loglevel >= 1)
+	if (OCP->tls_loglevel >= 1)
 	    type(NULL,0,NULL,"issuer=%s", cbuf);
 	SS->TLS.peer_issuer = strdup(cbuf);
 
@@ -1545,7 +1556,7 @@ tls_start_servertls(SS)
 	  cbuf[k] = 0;
 	  SS->TLS.peer_fingerprint = strdup(cbuf);
 
-	  if (tls_loglevel >= 1)
+	  if (OCP->tls_loglevel >= 1)
 	    type(NULL,0,NULL,"fingerprint=%s", SS->TLS.peer_fingerprint);
 	}
 
@@ -1557,7 +1568,7 @@ tls_start_servertls(SS)
 				  NID_commonName, cbuf, sizeof(cbuf));
  	SS->TLS.issuer_CN = strdup(cbuf);
 
- 	if (tls_loglevel >= 3)
+ 	if (OCP->tls_loglevel >= 3)
 	  type(NULL,0,NULL, "subject_CN=%s, issuer_CN=%s",
 	       SS->TLS.peer_CN ? SS->TLS.peer_CN : "",
 	       SS->TLS.issuer_CN ? SS->TLS.issuer_CN : "");
@@ -1568,9 +1579,21 @@ tls_start_servertls(SS)
     /*
      * Finally, collect information about protocol and cipher for logging
      */
-    SS->TLS.protocol = SSL_get_version(SS->TLS.ssl);
+    kp = SSL_get_version(SS->TLS.ssl);
+    if (kp)
+      SS->TLS.protocol = strdup(kp); /* This data belongs to SSL library,
+					make a copy of it for ourselves.
+					Darryl L. Miles */
+    else
+      SS->TLS.protocol = NULL;
     cipher = SSL_get_current_cipher(SS->TLS.ssl);
-    SS->TLS.cipher_name    = SSL_CIPHER_get_name(cipher);
+    kp    = SSL_CIPHER_get_name(cipher);
+    if (kp)
+      SS->TLS.cipher_name = strdup(kp); /* This data belongs to SSL library,
+					   make a copy of it for ourselves.
+					   Darryl L. Miles */
+    else
+      SS->TLS.cipher_name = NULL;
     SS->TLS.cipher_usebits = SSL_CIPHER_get_bits(cipher,
 						 &SS->TLS.cipher_algbits);
 
@@ -1596,8 +1619,10 @@ tls_start_servertls(SS)
 void
 Z_init __((void))
 {
-    if (starttls_ok)
-	tls_init_serverengine(tls_ccert_vd,tls_ask_cert,tls_req_cert);
+    if (OCP->starttls_ok)
+	tls_init_serverengine(OCP->tls_ccert_vd,
+			      OCP->tls_ask_cert,
+			      OCP->tls_req_cert);
 }
 
 void
@@ -1607,4 +1632,184 @@ Z_cleanup(SS)
     if (SS->sslmode)
 	tls_stop_servertls(SS, 0);
 }
+
+#ifdef HAVE_DISTCACHE
+/*                      _             _
+**  _ __ ___   ___   __| |    ___ ___| |  mod_ssl
+** | '_ ` _ \ / _ \ / _` |   / __/ __| |  Apache Interface to OpenSSL
+** | | | | | | (_) | (_| |   \__ \__ \ |  www.modssl.org
+** |_| |_| |_|\___/ \__,_|___|___/___/_|  ftp.modssl.org
+**                      |_____|
+**  ssl_scache_dc.c
+**  Distributed Session Cache (client support)
+*/
+
+/* ====================================================================
+ * THIS SOFTWARE IS PROVIDED BY GEOFF THORPE ``AS IS'' AND ANY
+ * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+ * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
+ * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL RALF S. ENGELSCHALL OR
+ * HIS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
+ * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+ * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
+ * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
+ * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
+ * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
+ * OF THE POSSIBILITY OF SUCH DAMAGE.
+ * ====================================================================
+ */
+
+/* Only build this code if it's enabled at configure-time. */
+
+#if !defined(DISTCACHE_CLIENT_API) || (DISTCACHE_CLIENT_API < 0x0001)
+#error "You must compile with a more recent version of the distcache-base package"
+#endif
+
+/*
+ * This cache implementation allows modssl to access 'distcache' servers (or
+ * proxies) to facilitate distributed session caching. It is based on code
+ * released as open source by Cryptographic Appliances Inc, and was developed by
+ * Geoff Thorpe, Steve Robb, and Chris Zimmerman.
+ */
+
+/*
+**
+** High-Level "handlers" as per ssl_scache.c
+**
+*/
+
+static void ssl_scache_dc_kill(void)
+{
+  if (dc_ctx)
+    DC_CTX_free(dc_ctx);
+  dc_ctx = NULL;
+}
+
+static DC_CTX *ssl_scache_dc_init()
+{
+    DC_CTX *ctx;
+
+    if (!OCP->tls_scache_name) return NULL;
+
+    /*
+     * Create a session context
+     */
+#if 0
+    /* If a "persistent connection" mode of operation is preferred, you *must*
+     * also use the PIDCHECK flag to ensure fork()'d processes don't interlace
+     * comms on the same connection as each other. */
+#define SESSION_CTX_FLAGS	SESSION_CTX_FLAG_PERSISTENT | \
+	    			SESSION_CTX_FLAG_PERSISTENT_PIDCHECK | \
+	    			SESSION_CTX_FLAG_PERSISTENT_RETRY | \
+	    			SESSION_CTX_FLAG_PERSISTENT_LATE
+#else
+    /* This mode of operation will open a temporary connection to the 'target'
+     * for each cache operation - this makes it safe against fork()
+     * automatically. This mode is preferred when running a local proxy (over
+     * unix domain sockets) because overhead is negligable and it reduces the
+     * performance/stability danger of file-descriptor bloatage. */
+#define SESSION_CTX_FLAGS	0
+#endif
+    ctx = DC_CTX_new(OCP->tls_scache_name, SESSION_CTX_FLAGS);
+    if(!ctx) {
+      type(NULL,0,NULL,"distributed scache failed to obtain context");
+      exit(1);
+    }
+    type(NULL,0,NULL, "distributed scache context initialised");
+
+    atexit(ssl_scache_dc_kill);
+
+    /* 
+     * Success .. we return the cache content to the caller
+     * :-)
+     */
+    return ctx;
+}
+
+
+static int ssl_scache_dc_store(pSession, id, idlen, timeout)
+     SSL_SESSION * pSession;
+     unsigned char *id;
+     int idlen;
+     time_t timeout;
+{
+    unsigned char der[SSL_SESSION_MAX_DER];
+    int der_len;
+    unsigned char *pder = der;
+    DC_CTX *ctx  =  dc_ctx;
+
+    if (!ctx) return FALSE;
+
+    /* Serialise the SSL_SESSION object */
+    der_len = i2d_SSL_SESSION(pSession, NULL);
+    if (der_len > SSL_SESSION_MAX_DER)
+        return FALSE;
+    i2d_SSL_SESSION(pSession, &pder);
+    /* !@#$%^ - why do we deal with *absolute* time anyway??? */
+    timeout -= time(NULL);
+    /* Send the serialised session to the distributed cache context */
+    if(!DC_CTX_add_session(ctx, id, idlen, der, der_len,
+			    (unsigned long)timeout * 1000)) {
+	/* ERROR INDICATION! */
+	type(NULL,0,NULL, "distributed scache 'add_session' failed");
+	return FALSE;
+    }
+    type(NULL,0,NULL, "distributed scache 'add_session' successful");
+    return TRUE;
+}
+
+static SSL_SESSION *ssl_scache_dc_retrieve(s, id, idlen)
+     SSL *s;
+     unsigned char *id;
+     int idlen;
+{
+    unsigned char der[SSL_SESSION_MAX_DER];
+    unsigned int der_len;
+    SSL_SESSION *pSession;
+    unsigned char *pder = der;
+    DC_CTX *ctx = dc_ctx;
+
+    if (!ctx) return FALSE;
+
+    /* Retrieve any corresponding session from the distributed cache context */
+    if(!DC_CTX_get_session(ctx, id, idlen, der, SSL_SESSION_MAX_DER,
+			    &der_len)) {
+	type(NULL,0,NULL,"distributed scache 'get_session' MISS");
+	return NULL;
+    }
+    if(der_len > SSL_SESSION_MAX_DER) {
+	type(NULL,0,NULL,"distributed scache 'get_session' OVERFLOW");
+	return NULL;
+    }
+    pSession = d2i_SSL_SESSION(NULL, &pder, der_len);
+    if(!pSession) {
+	type(NULL,0,NULL,"distributed scache 'get_session' CORRUPT");
+	return NULL;
+    }
+    type(NULL,0,NULL,"distributed scache 'get_session' HIT");
+    return pSession;
+}
+
+
+static void ssl_scache_dc_remove(s, id, idlen)
+     SSL_SESSION *s;
+     unsigned char *id;
+     int idlen;
+{
+    DC_CTX *ctx = dc_ctx;
+
+    if (!ctx) return;
+
+    /* Remove any corresponding session from the distributed cache context */
+    if(!DC_CTX_remove_session(ctx, id, idlen)) {
+	type(NULL,0,NULL, "distributed scache 'remove_session' MISS");
+    } else {
+	type(NULL,0,NULL, "distributed scache 'remove_session' HIT");
+    }
+}
+
+
+#endif
+
 #endif /* - HAVE_OPENSSL */
