@@ -65,6 +65,7 @@ static int run(rfp, wfp, path, argv, envp)
         break;
 
     case 0: /* Child */
+	/* Builtin presumption of: stdin==0, stdout==1, stderr==2 ... */
         if( pdo[0] != fileno(stdin)  ) (void) dup2( pdo[0], fileno(stdin)  );
         if( pdi[1] != fileno(stdout) ) (void) dup2( pdi[1], fileno(stdout) );
         if( pdi[1] != fileno(stderr) ) (void) dup2( pdi[1], fileno(stderr) );
@@ -138,11 +139,15 @@ pipeauth(cmd, msg, msgsize, uname, password)
     envp[envc++] = cp;
     envp[envc] = NULL;
     
+    wfp = rfp = NULL;
     pid = run( &rfp, &wfp, argv[0], argv, envp );
     if (pid > 1) {
         fprintf( wfp, "%s\n", password );
 	fflush(wfp);
         fclose( wfp );
+	wfp = NULL; /* Mark the pointer null too, so we won't
+		       double-close this futher below.
+		       "Darryl L. Miles" <darryl@netbauds.net> */
 	/* Following weird thing is because we have top-level
 	   child-death reaper code at the main part of this
 	   program... */
@@ -160,6 +165,8 @@ pipeauth(cmd, msg, msgsize, uname, password)
             fgets( msg, msgsize, rfp );
         }
     }
+    if (wfp) fclose(wfp);
+    if (rfp) fclose(rfp);
 
     return pipeauthchild_status;
 }
