@@ -27,18 +27,6 @@
 #define	PROGNAME	"errormail"
 #define	CHANNEL		"error"	/* the default channel name we deliver for */
 
-const char *dfltform[] = {
-		"From: The Post Office <postmaster>",
-		"Subject: email delivery error",
-		"Cc: The Postmaster <postmaster>",
-		"MIME-Version: 1.0",
-		"Precedence: junk", /* BSD sendmail */
-		"Content-Type: multipart/report; report-type=delivery-status;",
-		"",
-		"Processing your mail message caused the following errors:",
-		NULL
-};
-
 const char *progname;
 
 extern char *optarg;
@@ -61,6 +49,9 @@ int D_alloc = 0; /* For tmalloc() from libz.a ... */
 static void MIBcountCleanup __((void))
 {
 	MIBMtaEntry->taerrm.TaProcCountG -= 1;
+	if (MIBMtaEntry->taerrm.TaProcCountG > 99999U)
+	  MIBMtaEntry->taerrm.TaProcCountG = 0;
+
 }
 
 static void SHM_MIB_diag(rc)
@@ -120,6 +111,14 @@ main(argc, argv)
 	  SIGNAL_HANDLE(SIGQUIT, wantout);
 	SIGNAL_IGNORE(SIGPIPE);
 
+
+	/* This assignment MUST precede readzenv() ! */
+	progname = strrchr(argv[0], '/');
+	if (progname == NULL)
+	  progname = argv[0];
+	else
+	  ++progname;
+
 	if (getenv("ZCONFIG")) readzenv(getenv("ZCONFIG"));
 
 
@@ -130,11 +129,6 @@ main(argc, argv)
 
 	atexit(MIBcountCleanup);
 
-
-	if ((progname = strrchr(argv[0], '/')) == NULL)
-	  progname = argv[0];
-	else
-	  ++progname;
 	errflg = 0;
 	channel = CHANNEL;
 	while (1) {
@@ -287,7 +281,7 @@ process(dp)
 	struct ctldesc *dp;
 {
 	char buf[BUFSIZ];
-	const char **cpp, *mailshare, *mfpath;
+	const char *mailshare, *mfpath;
 	Sfio_t *mfp, *efp;
 	int n;
 	struct rcpt *rp;
@@ -295,8 +289,9 @@ process(dp)
 	char lastchar;
 	int reportcnt = 0;
 	struct stat stbuf;
-	long inum;
+	long inum, mtimems;
 	time_t mtime;
+	const char *fromuser;
 
 	if (fstat(dp->msgfd, &stbuf) != 0)
 	  abort(); /* This is a "CAN'T FAIL" case.. */
@@ -306,7 +301,10 @@ process(dp)
 	/* recipient host field is the error message file name in FORMSDIR */
 	/* recipient user field is the address causing the error */
 
-	if (dp->senders == NULL) {
+	rp = dp->recipients;
+
+	if ( dp->senders == NULL ||
+	     STREQ(rp->addr->link->channel,"error") ) {
 	  /*
 	   * If there was no error return address
 	   * it might be because this message was
@@ -334,8 +332,23 @@ process(dp)
 	  char *dom = mydomain(); /* transports/libta/buildbndry.c */
 	  struct stat stbuf;
 
+	  /* This is NOT spoolid, the mtime will change yet... */
 	  fstat(sffileno(mfp),&stbuf);
-	  taspoolid(boundarystr, stbuf.st_ctime, (long)stbuf.st_ino);
+	  taspoolid(boundarystr,  (long)stbuf.st_ino, stbuf.st_mtime,
+#ifdef HAVE_STRUCT_STAT_ST_ATIM_TV_NSEC
+		    stbuf.st_mtim.tv_nsec
+#else
+#ifdef HAVE_STRUCT_STAT_ST_ATIM___TV_NSEC
+		    stbuf.st_mtim.__tv_nsec
+#else
+#ifdef HAVE_STRUCT_STAT_ST_ATIMENSEC
+		    stbuf.st_mtimensec
+#else
+		    0
+#endif
+#endif
+#endif
+		    );
 	  strcat(boundarystr, "=_/errmail/");
 	  strcat(boundarystr, dom);
 	}
@@ -360,7 +373,8 @@ process(dp)
 
 	mfpath = emalloc(3 + strlen(mailshare) + strlen(FORMSDIR) +
 			 strlen(rp->addr->host));
-	sprintf((char*)mfpath, "%s/%s/%s", mailshare, FORMSDIR, rp->addr->host);
+	sprintf((char*)mfpath, "%s/%s/%s",
+		mailshare, FORMSDIR, rp->addr->host);
 
 	efp = sfopen(NULL, mfpath, "r");
 	if (efp != NULL) {
@@ -392,11 +406,13 @@ process(dp)
 	    } else {
 	      if (inhdr) {
 		inhdr = 0;
-		sfprintf(mfp,"MIME-Version: 1.0\n");
-		sfprintf(mfp,"Content-Type: multipart/report; report-type=delivery-status;\n");
-		sfprintf(mfp,"\tboundary=\"%s\"\n\n\n",boundarystr);
+		sfprintf(mfp, "MIME-Version: 1.0\n");
+		sfprintf(mfp, "Content-Type: multipart/report;\n");
+		sfprintf(mfp, "\treport-type=delivery-status;\n");
+		sfprintf(mfp, "\tboundary=\"%s\"\n\n\n",boundarystr);
 		sfprintf(mfp, "--%s\n", boundarystr);
-		sfprintf(mfp, "Content-Type: text/plain\n");
+		sfprintf(mfp, "Content-Transfer-Encoding: 7BIT\n");
+		sfprintf(mfp, "Content-Type: text/plain; charset=US-ASCII\n");
 	      }
 	      sfprintf(mfp, "%s", buf);
 	    }
@@ -405,13 +421,21 @@ process(dp)
 	} else {
 	  sfprintf(mfp,"to <postmaster>\n"); /* Default-form Cc: ... */
 	  sfprintf(mfp,"env-end\n");
-	  for (cpp = &dfltform[0]; *cpp != NULL; ++cpp)
-	    if (**cpp == '\0') {
-	      sfprintf(mfp, "\tboundary=\"%s\"\n\n\n", boundarystr);
-	      sfprintf(mfp, "--%s\n", boundarystr);
-	      sfprintf(mfp, "Content-Type: text/plain\n");
-	    } else
-	      sfprintf(mfp, "%s\n", *cpp);
+
+	  sfprintf(mfp, "From: The Post Office <postmaster>\n");
+	  sfprintf(mfp, "Subject: email delivery error\n");
+
+	  sfprintf(mfp, "To: <%s>\n",rp->addr->link->user);
+
+	  sfprintf(mfp, "Precedence: junk\n"); /* BSD sendmail */
+	  sfprintf(mfp,"MIME-Version: 1.0\n");
+	  sfprintf(mfp, "Content-Type: multipart/report;\n");
+	  sfprintf(mfp, "\treport-type=delivery-status;\n");
+	  sfprintf(mfp, "\tboundary=\"%s\"\n\n\n", boundarystr);
+	  sfprintf(mfp, "--%s\n", boundarystr);
+	  sfprintf(mfp, "Content-Transfer-Encoding: 7BIT\n");
+	  sfprintf(mfp, "Content-Type: text/plain; charset=US-ASCII\n");
+	  sfprintf(mfp, "\nProcessing your mail message caused the following errors:\n");
 	}
 	/* print out errors in standard format */
 	sfputc(mfp, '\n');
@@ -450,8 +474,16 @@ process(dp)
 	  decodeXtext(mfp,dp->envid);
 	  sfputc(mfp,'\n');
 	}
+
+	rp = dp->recipients;
+	fromuser = rp->addr->link->user;
+	if (*fromuser == 0 ||
+	    STREQ(rp->addr->link->channel, "error"))
+	  fromuser = "";
+
 	/* rfc822date() returns a string with trailing newline! */
 	sfprintf(mfp, "Arrival-Date: %s", rfc822date(&stbuf.st_ctime));
+	sfprintf(mfp, "Return-Path: <%.999s>\n", fromuser);
 	sfprintf(mfp, "\n");
 
 	for (rp = dp->recipients; rp != NULL; rp = rp->next) {
@@ -517,14 +549,14 @@ process(dp)
 	if (sferror(mfp)) {
 	  sfmail_abort(mfp);
 	  n = EX_IOERR;
-	} else if (_sfmail_close_(mfp, &inum, &mtime) == EOF)
+	} else if (_sfmail_close_(mfp, &inum, &mtime, &mtimems) == EOF)
 	  n = EX_IOERR;
 	else
 	  n = EX_OK;
 	{
 	  /* Ok, build response with proper "spoolid" */
 	  char taspid[30];
-	  taspoolid(taspid, mtime, inum);
+	  taspoolid(taspid, inum, mtime, mtimems);
 
 	  for (rp = dp->recipients; rp != NULL; rp = rp->next) {
 	    diagnostic(verboselog, rp, n, 0, taspid);
