@@ -1,7 +1,7 @@
 /*
  *	Copyright 1990 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
- *	Copyright 1992-2003 Matti Aarnio -- MIME processing et.al.
+ *	Copyright 1992-2005 Matti Aarnio -- MIME processing et.al.
  */
 
 #include "mailer.h"
@@ -15,6 +15,7 @@
 #include <sys/param.h>
 #include <sys/stat.h>
 #include <errno.h>
+#include <fcntl.h>
 
 #include "zmsignal.h"
 #include "zmalloc.h"
@@ -131,7 +132,7 @@ int	D_alloc = 0;
 extern char *optarg;
 extern int optind;
 extern void process __((struct ctldesc *));
-extern int  hold __((const char *, char **));
+extern int  hold __((struct ctldesc *, const char *, char **));
 extern char **environ;
 
 #ifndef strchr
@@ -144,10 +145,11 @@ extern char *strchr(), *strrchr();
 #endif	/* lint */
 
 
-
 static void MIBcountCleanup __((void))
 {
 	MIBMtaEntry->tahold.TaProcCountG -= 1;
+	if (MIBMtaEntry->tahold.TaProcCountG > 99999U)
+	  MIBMtaEntry->tahold.TaProcCountG = 0;
 }
 
 static void SHM_MIB_diag(rc)
@@ -214,6 +216,14 @@ main(argc, argv)
 	if (oldsig != SIG_IGN)
 	  SIGNAL_HANDLE(SIGHUP, wantout);
 
+
+	/* This assignment MUST precede readzenv() ! */
+	progname = strrchr(argv[0], '/');
+	if (progname == NULL)
+	  progname = argv[0];
+	else
+	  ++progname;
+
 	if (getenv("ZCONFIG")) readzenv(getenv("ZCONFIG"));
 
 
@@ -224,11 +234,6 @@ main(argc, argv)
 
 	atexit(MIBcountCleanup);
 
-
-	if ((progname = strrchr(argv[0], '/')) == NULL)
-	  progname = argv[0];
-	else
-	  ++progname;
 	errflg = 0;
 	channel = CHANNEL;
 	while (1) {
@@ -289,8 +294,27 @@ main(argc, argv)
 
 	  dp = ctlopen(filename, channel, host, &getout, NULL, NULL);
 	  if (dp != NULL) {
+
+	    if (verboselog) {
+	      fclose(verboselog);
+	      verboselog = NULL;
+	    }
+	    if (dp->verbose) {
+	      verboselog = fopen(dp->verbose,"a");
+	      if (verboselog) {
+		/* Buffering, and Close-On-Exec bit! */
+		setbuf(verboselog,NULL);
+		fcntl(FILENO(verboselog), F_SETFD, 1);
+	      }
+	    }
+
 	    process(dp);
 	    ctlclose(dp);
+
+	    if (verboselog) {
+	      fclose(verboselog);
+	      verboselog = NULL;
+	    }
 	  } else {
 	    printf("#resync %s\n",filename);
 	    fflush(stdout);
@@ -320,7 +344,7 @@ process(dp)
 	sawok = 0;
 	for (rp = dp->recipients; rp != NULL; rp = rp->next) {
 	  cp = rp->addr->user;
-	  rp->status = hold(rp->addr->host, (char**)&(rp->addr->user));
+	  rp->status = hold(dp, rp->addr->host, (char**)&(rp->addr->user));
 	  if (rp->status == EX_OK) {
 	    rp->addr->user = cp;
 	    sawok = 1;
@@ -457,39 +481,58 @@ process(dp)
  * string is canonicalized and parsed hierarchically.
  */
 
-extern int hold_ns	__(( const char* ));
-extern int hold_timeout	__(( const char* ));
-extern int hold_io	__(( const char* ));
-extern int hold_script	__(( const char* ));
-extern int hold_home	__(( const char* ));
+extern int hold_ns	__(( struct ctldesc *, const char* ));
+extern int hold_timeout	__(( struct ctldesc *, const char* ));
+extern int hold_io	__(( struct ctldesc *, const char* ));
+extern int hold_script	__(( struct ctldesc *, const char* ));
+extern int hold_home	__(( struct ctldesc *, const char* ));
+extern int hold_age	__(( struct ctldesc *, const char* ));
 
 struct holds_info {
 	const char	*type;
-	int	(*f) __(( const char * ));
+	int	(*f) __(( struct ctldesc *dp, const char * ));
 } holds[] = {
 	{	"ns",		hold_ns		},
 	{	"timeout",	hold_timeout	},
 	{	"io",		hold_io		},
 	{	"script",	hold_script	},
 	{	"home",		hold_home	},
+	{	"age",		hold_age	},
 	{	NULL,		NULL		},
 };
 
+/* Alloc this buffer to hold the 's' string, then copy
+   data there, and mutilate it there... */
+
+static char *holdparam;
 
 int
-hold(s, errmsgp)
+hold(dp, s, errmsgp)
+	struct ctldesc *dp;
 	const char *s;
 	char **errmsgp;
 {
 	char *cp, *colon;
 	struct holds_info *hip;
+#if 0 /* NO HOLD DATA CACHE */
 	static struct sptree *spt_hash = NULL;
 	struct spblk *spl;
-	int v;
 	spkey_t symid;
+#endif
+	int v;
+	int slen = strlen(s) + 1;
+
+	holdparam = realloc(holdparam, slen+2);
+	if (!holdparam) return EX_SOFTWARE;
+
+	memcpy(holdparam, s, slen);
+	s = holdparam;
+
+	if (verboselog)
+	  fprintf(verboselog, "hold(dp, \"%s\")\n", s);
 
 	colon = NULL;
-	for (cp = (char*)s; *cp != '\0'; ++cp) {
+	for (cp = holdparam; *cp != '\0'; ++cp) {
 	  unsigned char c = *cp;
 	  if (isascii(c) && isupper(c))
 	    *cp = tolower(c);
@@ -497,9 +540,15 @@ hold(s, errmsgp)
 	    colon = cp;
 	}
 
+	if (verboselog) {
+	  fprintf(verboselog, "hold(dp, \"%s\")\n", s);
+	}
+
+
 	if (colon == NULL)
 	  return EX_PROTOCOL;	/* invalid hold condition */
 
+#if 0  /* NO HOLD DATA CACHE! */
 	symid = symbol((void*)s);
 	if (spt_hash == NULL)
 	  spt_hash = sp_init();
@@ -507,6 +556,7 @@ hold(s, errmsgp)
 	  *errmsgp = (char *)spl->data;
 	  return spl->mark;
 	}
+#endif
 
 	*colon++ = '\0';
 	for (hip = &holds[0]; hip->type != NULL ; ++hip)
@@ -517,7 +567,11 @@ hold(s, errmsgp)
 
 	errormsg[0] = '\0';
 
-	if ((hip->f)(colon))
+	if (verboselog) {
+	  fprintf(verboselog, "Calling hold_%s('%s')\n", hip->type, colon);
+	}
+
+	if ((hip->f)(dp, colon))
 	  v = EX_OK;		/* resubmit the message address */
 	else
 	  v = EX_TEMPFAIL;	/* defer resubmission */
@@ -527,7 +581,10 @@ hold(s, errmsgp)
 	  strcpy(cp, errormsg);
 	} else
 	  cp = "deferred";
+#if 0 /* NO HOLD DATA CACHE! */
 	sp_install(symid, (const void *)cp, v, spt_hash);
+#endif
+
 	*errmsgp = cp;
 
 	return v;
@@ -564,7 +621,8 @@ struct qtypes {
 };
 
 int
-hold_ns(s)
+hold_ns(dp, s)
+	struct ctldesc *dp;
 	const char *s;
 {
 	struct qtypes *qtp;
@@ -618,7 +676,8 @@ struct qtypes {
 };
 
 int
-hold_ns(s)
+hold_ns(dp, s)
+	struct ctldesc *dp;
 	char *s;
 {
 	struct qtypes *qtp;
@@ -663,7 +722,8 @@ hold_ns(s)
 /* return true if the seconds-since-epoch in argument has passed */
 
 int
-hold_timeout(s)
+hold_timeout(dp, s)
+	struct ctldesc *dp;
 	const char *s;
 {
 	time_t now, then;
@@ -676,7 +736,8 @@ hold_timeout(s)
 /* return true if we should retry the I/O operation causing the error */
 
 int
-hold_io(s)
+hold_io(dp, s)
+	struct ctldesc *dp;
 	const char *s;
 {
 	return ranny(9) == 0;	/* 10% of the time */
@@ -686,7 +747,8 @@ hold_io(s)
    return exit status of "$MAILBIN/bin/command $user" command */
 
 int
-hold_script(command)
+hold_script(dp, command)
+	struct ctldesc *dp;
 	const char *command;
 {
 	int i, in[2];
@@ -803,8 +865,9 @@ hold_script(command)
    null, then we "succeed", to try to redeliver the mail. */
 
 int
-hold_home(user)
-const char *user;
+hold_home(dp, user)
+	struct ctldesc *dp;
+	const char *user;
 {
 	struct Zpasswd *pw;
 	struct stat st;
@@ -821,3 +884,33 @@ const char *user;
 	    S_ISDIR(st.st_mode)) return 1;
 	return 0;
 }
+
+
+
+/*
+ * A timeout until given age since message spool file creation
+ * Return TRUE when ready to release, 
+ *
+ */
+
+int
+hold_age(dp, age)
+	struct ctldesc *dp;
+	const char *age;
+{
+	time_t now, then;
+
+	time(&now);
+
+	then = parse_interval(age, NULL);
+
+	if (verboselog) {
+	  fprintf(verboselog,
+		  " hold_age('%s') -> interval=%ld,  msgmtime=%ld, now=%ld\n",
+		  age, then, dp->msgmtime, now);
+	}
+
+	then += dp->msgmtime;
+
+	return (now >= then);
+}
