@@ -4,7 +4,7 @@
  */
 
 /*
- *	Copyright 1994-2000 by Matti Aarnio
+ *	Copyright 1994-2003 by Matti Aarnio
  *
  * To really understand how headers (and their converted versions)
  * are processed you do need to draw a diagram.
@@ -80,11 +80,11 @@ static int mime_received_convert __((struct rcpt *rp, char *convertstr));
 
 static int strqcpy __((char **buf, int startpos, int *buflenp, const char *str));
 
-static int
+static int  /* FIXME: ??? LEAKS ?? */
 strqcpy(bufp, startpos, buflenp, str)
 	char **bufp;
 	const char *str;
-	int *buflenp;
+	int *buflenp, startpos;
 {
 	const char *s = str;
 	char *buf = *bufp;
@@ -99,7 +99,7 @@ strqcpy(bufp, startpos, buflenp, str)
 	  if (pos >= buflen) {
 	    *buflenp += 32;
 	    buflen = *buflenp -2;
-	    buf = realloc(buf, *buflenp + 5);
+	    buf = realloc(buf, buflen + 5);
 	  }
 
 	  if (!(('0' <= c && c <= '9') ||
@@ -125,7 +125,7 @@ strqcpy(bufp, startpos, buflenp, str)
 	if (pos+5 >= buflen) {
 	  *buflenp = pos + 32;
 	  buflen = *buflenp -2;
-	  buf = realloc(buf, *buflenp);
+	  buf = realloc(buf, buflen + 5);
 	}
 
 	buf[pos] = '"';
@@ -137,7 +137,7 @@ strqcpy(bufp, startpos, buflenp, str)
 	  if (pos >= buflen) {
 	    *buflenp = pos + 32;
 	    buflen = *buflenp - 4;
-	    buf = realloc(buf, *buflenp);
+	    buf = realloc(buf, buflen + 5);
 	  }
 
 	  if (c == '"' || c == '\\') {
@@ -153,7 +153,7 @@ strqcpy(bufp, startpos, buflenp, str)
 	if (pos+3 >= buflen) {
 	  *buflenp += pos+3;
 	  buflen = *buflenp;
-	  buf = realloc(buf, *buflenp);
+	  buf = realloc(buf, buflen +2);
 	}
 
 	buf[pos] = '"';
@@ -192,6 +192,7 @@ int cvtspace_copy(rp)
 
 	/* Allocate, and copy ! */
 
+	/* XX: FIXME: ?? LEAKS ? */
 	newcvt = (char **)malloc(sizeof(char *)*(hdrcnt+1));
 	if (newcvt != NULL) {
 
@@ -200,6 +201,7 @@ int cvtspace_copy(rp)
 
 	  while (*probe) { /* Copy over */
 	    int len = strlen(*probe)+1;
+	    /* XX: FIXME: ?? LEAKS ? */
 	    *ss = (char *)malloc(len);
 	    if (! *ss) return 0;
 	    memcpy(*ss,*probe,len);
@@ -610,12 +612,13 @@ static const char * skip_822linearcomments(p)
 	   defined LINEAR-WHITESPACES and COMMENTS */
 
 	const char *s = p;
+	char c;
 
 	while (*s) {
 
 	  /* To be exact we should scan over only SPC, TAB, CR and LF
 	     chars, but lets skip all controls just in case.. */
-	  while (0 <= *s && *s <= ' ') ++s; /* ASCII assumption! */
+	  while ((c = *s) && (0 <= c && c <= ' ')) ++s; /* ASCII assumption! */
 
 	  if (*s == '(') {
 	    s = skip_comment(s);
@@ -847,26 +850,24 @@ static const char *debugptrs[20];
 	    ct->name     = parval;
 	  } else {
 	    /* Unknown parameter.. */
-	    int unklen = strlen(parval)+5+strlen(paramname);
 	    int unkpos;
 	    char *unk;
-	    if (!ct->unknown) {
-	      ct->unknown = (char**)malloc(sizeof(char*)*2);
-	/* FIXME: malloc problem check ?? */
-	      ct->unknown[1] = NULL;
-	    } else {
-	      ct->unknown = (char**)realloc(ct->unknown,
-					    sizeof(char*)*(unknowncount+2));
-	/* FIXME: malloc problem check ?? */
-	    }
+	    int unklen = strlen(parval)+5+strlen(paramname);
+
+	    ct->unknown = (char**)realloc(ct->unknown,
+					  sizeof(char*)*(unknowncount+2));
+	    /* FIXME: malloc problem check ?? */
+
 	    unk = malloc(unklen);
-	/* FIXME: malloc problem check ?? */
+	    /* FIXME: malloc problem check ?? */
+
 	    sprintf(unk, "%s=", paramname);
 	    unkpos = strlen(unk);
+
 	    strqcpy(&unk, unkpos, &unklen, parval);
-	    ct->unknown[unknowncount] = unk;
-	    ct->unknown[unknowncount] = NULL;
-	    ++unknowncount;
+
+	    ct->unknown[unknowncount++] = unk;
+	    ct->unknown[unknowncount  ] = NULL;
 	    free (parval);
 	  }
 	  if (paramname)
@@ -875,6 +876,25 @@ static const char *debugptrs[20];
 	return ct;
 }
 
+void free_content_type(ct)
+     struct ct_data *ct;
+{
+	int i;
+
+	if (ct->basetype) free(ct->basetype);
+	if (ct->subtype)  free(ct->subtype);
+	if (ct->charset)  free(ct->charset);
+	if (ct->boundary) free(ct->boundary);
+	if (ct->name)     free(ct->name);
+	if (ct->unknown) {
+	  for (i = 0; ct->unknown[i]; ++i)
+	    free(ct->unknown[i]);
+	  free(ct->unknown);
+	}
+
+	free(ct);
+}
+
 struct cte_data *
 parse_content_encoding(cte_line)
      const char *cte_line;
@@ -906,6 +926,15 @@ parse_content_encoding(cte_line)
 }
 
 
+
+void free_content_encoding(cte)
+     struct cte_data *cte;
+{
+	if (cte->encoder) free(cte->encoder);
+	free(cte);
+}
+
+
 /*
  *  Check for  "Content-conversion: prohibited" -header, and return
  *  non-zero when found it. ---  eh, "-1", return '7' when QP-coding
@@ -1039,6 +1068,8 @@ downgrade_charset(rp, verboselog)
 	/* Delete the old one, and place there the new version.. */
 	output_content_type(rp,ct,CT);
 
+	if (ct) free_content_type(ct);
+
 	return 1;
 }
 
@@ -1062,7 +1093,8 @@ downgrade_headers(rp, convertmode, verboselog)
 	if (*(rp->newmsgheadercvt) != NULL)
 	  return 0; /* Already converted ! */
 
-	if (!cvtspace_copy(rp)) return 0; /* XX: auch! */
+	if (*(rp->newmsgheadercvt) == NULL)
+	  if (!cvtspace_copy(rp)) return 0; /* XX: auch! */
 
 	oldmsgheader = rp->newmsgheadercvt;
 
@@ -1204,10 +1236,273 @@ NULL };
 
 	}
 
+	if (ct) free_content_type(ct);
+
 	return 1; /* Non-zero for success! */
 }
 
 
+/* Returns the END of decoded string! */
+static char * decodeXtext __((const char *, char *, int));
+static char *
+decodeXtext(xtext, obuf, obuflen)
+	const char *xtext;
+	char *obuf;
+	int obuflen;
+{
+	char *s = obuf;
+	--obuflen;
+	for ( ; *xtext && obuflen > 0; --obuflen) {
+	  if (*xtext == '+') {
+	    int c = '?';
+	    sscanf(xtext+1,"%02X",&c);
+	    *s = c;
+	    if (*xtext) ++xtext;
+	    if (*xtext) ++xtext;
+	  } else {
+	    *s = *xtext;
+	  }
+	  ++s;
+	  ++xtext;
+	}
+	*s = '\0';
+	return s;
+}
+
+
+
+int
+header_received_for_clause(rp, rcptcnt, verboselog)
+     struct rcpt *rp;
+     int rcptcnt;
+     FILE *verboselog;
+{
+	int semicindex, receivedlen;
+	char *newreceived;
+	const char *newreceivedend;
+	const char *top_received, *sc;
+
+	char fchead[10], forclause[1024], fctail[20], *s;
+	int clauselen;
+	int col;
+
+	char **inhdr;
+
+	static int no_for_clause = -1;
+
+	if (no_for_clause < 0) {
+	  if (getzenv("NORECEIVEDFORCLAUSE"))
+	    no_for_clause = 1;
+	  else
+	    no_for_clause = 0;
+	}
+	if (no_for_clause) return 0; /* Forbidden.. */
+
+
+
+	if (*(rp->newmsgheadercvt) == NULL)
+	  if (!cvtspace_copy(rp)) return 0; /* XX: auch! */
+
+	inhdr = *(rp->newmsgheadercvt);
+
+
+	/* We have one advantage: The "Received:" header we
+	   want to fiddle with is the first one of them. */
+
+	if (!inhdr || !CISTREQN(*inhdr,"Received:",9)) {
+	  /* if (verboselog)fprintf(verboselog, "first *inhdr = '%s'\n",*inhdr ? *inhdr:"<NUL>"); */
+	  return 0; /* D'uh ??  Not 'Received:' ??? */
+	}
+
+
+
+	/* Begin at the indented line start.. */
+	/* strcpy(s, "\n\t"); s += 2; */
+
+	if (rp->orcpt) {
+
+	  strcpy(fchead, "(ORCPT");
+	  s = forclause;
+	  *s++ = '<';
+	  s = decodeXtext(rp->orcpt, s, 800);
+	  *s++ = '>';
+	  *s = 0;
+	  clauselen = s - forclause;
+
+	  /* ORCPT data tail, just ending closing parenthesis.. */
+	  if (rcptcnt > 2)
+	    sprintf(fctail, "+ %d others)", rcptcnt-1);
+	  else if (rcptcnt > 1)
+	    strcpy(fctail, "+ 1 other)");
+	  else {
+	    *fctail = 0;
+	    /* Close the comment on the ORCPT value! */
+	    strcat(s, ")");
+	    ++clauselen;
+	  }
+
+	} else { /* No  ORCPT  tail,  just 'for'-clause */
+
+	  strcpy(fchead, "for");
+	  sprintf(forclause, "<%.800s>", rp->addr->user);
+	  clauselen = strlen(forclause);
+
+	  if (rcptcnt > 2)
+	    sprintf(fctail, "(+ %d others)", rcptcnt-1);
+	  else if (rcptcnt > 1)
+	    strcpy(fctail, "(+ 1 other)");
+	  else
+	    *fctail = 0;
+	}
+
+
+
+	top_received = *inhdr;
+
+
+	/* Look for the LAST semicolon in this Received: header.. */
+
+	sc = strrchr(top_received, ';');
+	if (sc) {
+	  semicindex = sc - top_received;
+	} else {
+	  /* No semicolon at all, last NEWLINE then ? */
+	  semicindex = strlen(top_received);
+	  sc = top_received + semicindex;
+	  if (sc[-1] == '\n') { --sc; --semicindex; }
+	}
+
+	{
+	  /* Find out the column of the cut-point.. */
+	  const char *p = top_received;
+	  col = 0;
+	  for ( ; *p && (p <= sc) ; ++p) {
+	    const char c = *p;
+	    if ('\t' == c) col += 8; else ++col;
+	    if (('\n' == c) || ('\r' == c)) {
+	      col = 0;
+	    }
+	  }
+	}
+
+
+	receivedlen = strlen(top_received);
+	newreceived = realloc(rp->top_received,
+			      receivedlen + clauselen + 30); /* fchead[] +
+								fctail[] +
+								various
+								newlines.. */
+	rp->top_received = newreceived;
+	if (!newreceived) return 0; /* Failed malloc.. */
+
+	newreceivedend = newreceived + (receivedlen + clauselen + 30);
+
+
+	/* Begin.. */
+	s = newreceived;
+	memcpy(s, top_received, semicindex);
+	s += semicindex;
+
+	if ((col + 1 + strlen(fchead)) >= 78) {
+	  *s++ = '\n';
+	  *s++ = '\t';
+	  col = 8;
+	} else {
+	  *s++ = ' ';
+	  ++col;
+	}
+
+	/* For clause head */
+	{
+	  int fcheadlen = strlen(fchead);
+	  memcpy(s, fchead, fcheadlen +1);
+	  s   += fcheadlen;
+	  col += fcheadlen;
+	}
+
+	if ((col + 1 + clauselen) < 78) {
+	  *s++ = ' ';
+	  ++col;
+	} else {
+	  *s++ = '\n';
+	  *s++ = '\t';
+	  col = 8;
+	}
+	*s = 0; /* this is not absolutely required... */
+
+	/* For clause */
+	memcpy(s, forclause, clauselen+1);
+	s += clauselen;
+
+	col += clauselen;
+
+	/* Tail */
+
+	if (*fctail) {
+	  if ((col + 1 + strlen(fctail)) >= 78) {
+	    *s++ = '\n';
+	    *s++ = '\t';
+	    col = 8;
+	  } else {
+	    *s++ = ' ';
+	    ++col;
+	  }
+
+	  /* For clause tail */
+	  {
+	    int fctaillen = strlen(fctail);
+	    memcpy(s, fctail, fctaillen +1);
+	    s   += fctaillen;
+	    col += fctaillen;
+	  }
+	}
+
+
+	/* SEMIC + DATETIME tail */
+#if 1
+	if (verboselog) {
+	  fprintf(verboselog,"col=%d fc[] = '%s'(%d) ",
+		  col, forclause, clauselen);
+	  fprintf(verboselog,"sc[] = '%s' (%d)\n",
+		  sc, strlen(sc));
+	}
+#endif
+
+	if ( (strlen(sc) + col) < 78) {
+	  /* Shrink the tail a bit, unnecessary folding away. */
+	  *s++ = ';';
+	  *s++ = ' ';
+	} else {
+	  *s++ = ';';
+	  *s++ = '\n';
+	  *s++ = '\t';
+	}
+	*s = 0; /* this is not absolutely required... */
+
+	++sc; /* Skip the semicolon (or NEWLINE) .. and white-spaces */
+	while (*sc && (*sc == '\n' || *sc == '\r' || *sc == ' ' || *sc == '\t')) ++sc;
+	strcpy(s, sc);
+#if 0
+	if (verboselog) fprintf(verboselog,"sc[] = '%s'\n", sc);
+#endif
+	if (*sc == 0) strcat(s, "\n");
+
+	s += strlen(s);
+	if (s >= newreceivedend) abort();
+
+
+#if 1
+	if (verboselog) {
+	  fprintf(verboselog,"Rewriting 'Received:' headers.  ");
+	  fprintf(verboselog,"The new line is:\n%s\n",rp->top_received);
+	}
+#endif
+
+	return 1;
+}
+
+
+
 static int /* Return non-zero for success */
 mime_received_convert(rp, convertstr)
 	struct rcpt *rp;
@@ -1272,8 +1567,9 @@ qp_to_8bit(rp)
 	char **CTE, **CT;
 	struct ct_data *ct;
 
-	if (!cvtspace_copy(rp))
-	  return 0;	/* Failed to copy ! */
+	if (*(rp->newmsgheadercvt) == NULL)
+	  if (!cvtspace_copy(rp))
+	    return 0;	/* Failed to copy ! */
 
 	inhdr = *(rp->newmsgheadercvt);
 
@@ -1287,7 +1583,14 @@ qp_to_8bit(rp)
 	if (ct->basetype == NULL ||
 	    ct->subtype  == NULL ||
 	    !CISTREQ(ct->basetype,"text") ||
-	    !CISTREQ(ct->subtype,"plain") ) return 0; /* Not TEXT/PLAIN! */
+	    !CISTREQ(ct->subtype,"plain") ) {
+	  
+	  free_content_type(ct);
+
+	  return 0; /* Not TEXT/PLAIN! */
+	}
+
+	free_content_type(ct);
 
 	hdr = *CTE;
 
