@@ -1,7 +1,7 @@
 /*
  *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
- *	Copyright 1992-2003 Matti Aarnio -- MIME processing et.al.
+ *	Copyright 1992-2005 Matti Aarnio -- MIME processing et.al.
  */
 
 /* History:
@@ -43,6 +43,10 @@
 #endif
 #include <string.h>
 
+FILE *logfp = NULL;       /* needed very early.. */
+FILE *verboselog = NULL;
+
+
 #include "ta.h"
 
 #include "mail.h"
@@ -288,8 +292,6 @@ const char *defcharset;
 const char *progname;
 const char *channel;
 const char *logfile;
-FILE *logfp = NULL;
-FILE *verboselog = NULL;
 int   readalready = 0;		/* does buffer contain valid message data? */
 uid_t currenteuid;		/* the current euid */
 extern int nobody;		/* safe uid for file/program delivery */
@@ -395,6 +397,9 @@ static void zsfclose(fp)
 static void MIBcountCleanup __((void))
 {
 	MIBMtaEntry->tambox.TaProcCountG -= 1;
+	/* That is UNSIGNED, thus.. */
+	if (MIBMtaEntry->tambox.TaProcCountG > 99999U)
+	  MIBMtaEntry->tambox.TaProcCountG = 0;
 }
 
 static void SHM_MIB_diag(rc)
@@ -441,12 +446,21 @@ int sig;
 	/* Sigh, actually dummy routine.. */
 }
 
-static int free_spu __((void *));
+static int free_spu __((void *, struct spblk *));
 static int
-free_spu(p)
-     void *p;
+free_spu(p, spl)
+	void *p;
+	struct spblk *spl;
 {
-	free(p);
+	struct userhost *uhp = (struct userhost *) spl->data;
+	struct userhost *uhpn;
+
+	while (uhp) {
+	  uhpn = uhp->next;
+	  free(uhp);
+	  uhp = uhpn;
+	}
+
 	return 0;
 }
 
@@ -496,6 +510,13 @@ main(argc, argv)
 #endif
 
 
+	/* This assignment MUST precede readzenv() ! */
+	progname = strrchr(argv[0], '/');
+	if (progname == NULL)
+	  progname = argv[0];
+	else
+	  ++progname;
+
 	if (getenv("ZCONFIG")) readzenv(getenv("ZCONFIG"));
 
 	Z_SHM_MIB_Attach(1); /* we don't care if it succeeds or fails.. */
@@ -506,12 +527,6 @@ main(argc, argv)
 	atexit(MIBcountCleanup);
 
 
-	progname = strrchr(argv[0], '/');
-	if (progname == NULL)
-	  progname = argv[0];
-	else
-	  ++progname;
-
 	cs = getzenv("MAILBOX");
 	if (cs != NULL) {
 	  maildirs[0] = cs;
@@ -542,6 +557,8 @@ main(argc, argv)
 	    ++dirhashes;	/* For user "abcdefg" the mailbox path is:
 				   MAILBOX/a/b/abcdefg -- supported by
 				   qpopper, for example... */
+	    if (dirhashes > 10)
+	      dirhashes = 10;
 	    break;
 	  case 'P':		/* pjwhash32() from the last component of the
 				   mailbox file path (filename) is used to
@@ -549,9 +566,13 @@ main(argc, argv)
 				   (one or two) of modulo 26 ('A'..'Z') alike
 				   the scheduler does its directory hashes. */
 	    ++pjwhashes;
+	    if (pjwhashes > 10)
+	      pjwhashes = 10;
 	    break;
 	  case 'X':		/* Like pjwhash32() above, but with crc32() */
 	    ++crchashes;
+	    if (crchashes > 10)
+	      crchashes = 10;
 	    break;
 	  case 'F':
 	    if (STREQ(optarg,"edquot"))
@@ -721,7 +742,7 @@ main(argc, argv)
 #endif /* RBIFF */
 	  }
 #ifdef	HAVE_PROTOCOLS_RWHOD_H /* RBIFF */
-	  if (spt_users != NULL) {
+	  if (spt_users && (spt_users->eltscnt > 0)) {
 	    if (readrwho())
 	      for (nbp = biffs ; nbp != NULL; nbp = nbp->next) {
 		if (nbp->offset >= 0 && nbp->wantrbiff)
@@ -734,9 +755,8 @@ main(argc, argv)
 	    free(nbp->user);
 	    free(nbp);
 	  }
-	  sp_scan( (int(*)__((struct spblk*))) free_spu, NULL, spt_users);
-	  free((void*)spt_users);
-	  spt_users = NULL;
+	  sp_scan(free_spu, NULL, NULL, spt_users);
+	  symbol_null_db(NULL);
 #endif	/* HAVE_SOCKET */
 	}
 	exit(EX_OK);
@@ -1637,7 +1657,7 @@ void store_to_file(dp,rp,file,ismbox,usernam,st,uid,
      time_t starttime;
      const char *timestring;
 {
-	int fdmail;
+	int fdmail = -1;
 	struct stat s2;
 	Sfio_t *fp = NULL;
 	const char *mboxlocks = getzenv("MBOXLOCKS");
@@ -1654,12 +1674,87 @@ void store_to_file(dp,rp,file,ismbox,usernam,st,uid,
 	  filelocks = "";
 
 
+	alarm(180); /* Set an timed interrupt coming to us to break
+		       overlengthy file lock acquisition.. */
+
+	if (!S_ISREG(st->st_mode))
+	  /* don't lock non-files */;
+	else {
+
+	  int err;
+
+
+	  if (!ismbox)	/* Not mailbox, use file-locks */
+	    mboxlocks = filelocks;
+
+	  locks = mboxlocks;
+
+	  if (verboselog)
+	    fprintf(verboselog,"Locking sequence: '%s'\n",locks);
+
+	  err = 0;
+	  while (*locks != 0) {
+
+	    if (verboselog)
+	      fprintf(verboselog,"Calling lock: '%c' ..",*locks);
+
+	    switch(*locks) {
+	    case '"': /* ZENV variable may have quotes with it.. */
+	      break;
+	    case ':':
+	      err = -1; /* negative error is no REAL error */
+	      break;
+	    case '.':
+	      if (ismbox)
+		err = acquire_mboxlock(rp,file,uid);
+	      break;
+	    case 'N':
+	    case 'n':
+	      err = acquire_nfsmboxlock(rp,file);
+	      break;
+#if 0 /* L and F locks are ignored at this time! */
+	    case 'L':
+	    case 'l':
+	      err = acquire_lockflock(fdmail,rp,file);
+	      break;
+	    case 'F':
+	    case 'f':
+	      err = acquire_flocklock(fdmail,rp,file);
+	      break;
+#else
+	    case 'L': case 'l':
+	    case 'F': case 'f':
+	      break; /* IGNORE */
+#endif
+	    default:
+	      err = 2; /* BAD INPUT */
+	      break;
+	    }
+	    if (verboselog)
+	      fprintf(verboselog," err=%d\n", err);
+
+	    if (err) break;
+	    ++locks; /* Advance on success only! */
+	  }
+
+	  /* Turn off the alarm */
+	  alarm(0);
+
+	  if (err > 0) {
+	    goto lock_release;
+	  }
+	}
+
+	/* We have locks that must be acquired before... */
+
+
 	if (verboselog)
 	  fprintf(verboselog,
 		  "To open a file with euid=%d egid=%d ismbox=%d file='%s'\n",
 		  (int)geteuid(), (int)getegid(), ismbox, file);
 
 	fdmail = open(file, O_RDWR|O_APPEND);
+
 	if (fdmail < 0) {
 	  char fmtbuf[512];
 	  int saverrno = errno;
@@ -1716,6 +1811,8 @@ void store_to_file(dp,rp,file,ismbox,usernam,st,uid,
 	alarm(180); /* Set an timed interrupt coming to us to break
 		       overlengthy file lock acquisition.. */
 
+	/*  Now we acquire locks that are at the mailbox file itself.. */
+
 	if (!S_ISREG(st->st_mode))
 	  /* don't lock non-files */;
 	else {
@@ -1734,80 +1831,45 @@ void store_to_file(dp,rp,file,ismbox,usernam,st,uid,
 	  err = 0;
 	  while (*locks != 0) {
 	    switch(*locks) {
-	      case '"': /* ZENV variable may have quotes with it.. */
-		break;
-	      case ':':
-		err = -1; /* negative error is no REAL error */
-		break;
-	      case '.':
-		if (ismbox)
-		  err = acquire_mboxlock(rp,file,uid);
-		break;
-	      case 'N':
-	      case 'n':
-		err = acquire_nfsmboxlock(rp,file);
-		break;
-	      case 'L':
-	      case 'l':
-		err = acquire_lockflock(fdmail,rp,file);
-		break;
-	      case 'F':
-	      case 'f':
-		err = acquire_flocklock(fdmail,rp,file);
-		break;
-	      default:
-		err = 1;
-		break;
+	    case '"': /* ZENV variable may have quotes with it.. */
+	      break;
+	    case ':':
+	      err = -1; /* negative error is no REAL error */
+	      break;
+#if 0 /* Ignore dot and N locks */
+	    case '.':
+	      if (ismbox)
+		err = acquire_mboxlock(rp,file,uid);
+	      break;
+	    case 'N':
+	    case 'n':
+	      err = acquire_nfsmboxlock(rp,file);
+	      break;
+#else
+	    case '.':
+	    case 'N':
+	    case 'n':
+	      break; /* IGNORE */
+#endif
+
+	    case 'L':
+	    case 'l':
+	      err = acquire_lockflock(fdmail,rp,file);
+	      break;
+	    case 'F':
+	    case 'f':
+	      err = acquire_flocklock(fdmail,rp,file);
+	      break;
+	    default:
+	      err = 2; /* BAD INPUT */
+	      break;
 	    }
 	    if (err) break;
 	    ++locks; /* Advance on success only! */
 	  }
 
 	  if (err > 0) {
-	    --locks; /* Don't try to revert the failed (= last) lock! */
-	    while (locks >= mboxlocks) {
-	      switch (*locks) {
-		case '"':
-		  break;
-		case '.':
-#ifdef	HAVE_MAILLOCK
-		  if (havemaillock && ismbox)
-		    mailunlock();
-		  havemaillock = 0;
-#endif	/* HAVE_MAILLOCK */
-#ifdef	HAVE_DOTLOCK
-		  if (ismbox)
-		    dotunlock(file);
-		  havedotlock = 0;
-#endif /*HAVE_DOTLOCK*/
-		  break;
-	        case 'N':
-		case 'n':
-#ifdef	USE_NFSMBOX
-		  unlock(file);
-#endif	/* USE_NFSMBOX */
-		  break;
-	        case 'L':
-		case 'l':
-#if defined(HAVE_LOCKF) && defined(F_LOCK) /* If one, also the other */
-		  lseek(fdmail,(off_t)0,SEEK_SET);
-		  lockf(fdmail, F_ULOCK, 0);
-#endif	/* HAVE_LOCKF */
-		  break;
-	        case 'F':
-		case 'f':
-#ifdef HAVE_FLOCK
-		  flock(fdmail, LOCK_UN);
-#endif
-		  break;
-		default:
-		  break;
-	      }
-	      --locks;
-	    }
-	    close(fdmail);
-	    setrootuid(rp);
-	    return;
+	    goto lock_release;
 	  }
 	}
 
@@ -1823,6 +1885,10 @@ void store_to_file(dp,rp,file,ismbox,usernam,st,uid,
 #endif	/* BIFF || RBIFF */
 
 	fp = putmail(dp, rp, fdmail, "a+", timestring, file, uid);
+
+
+	/* Successes and failures: do lock releases here! */
+ lock_release:
 	
 	if (S_ISREG(st->st_mode)) {
 
@@ -1879,7 +1945,8 @@ void store_to_file(dp,rp,file,ismbox,usernam,st,uid,
 	setrootuid(rp);
 	time(&endtime);
 
-	close(fdmail);
+	if (fdmail >= 0) close(fdmail);
+
 	if (fp != NULL) { /* Dummy marker! */
 	  notary_setxdelay((int)(endtime-starttime));
 	  notaryreport(rp->addr->user, s_delivered,
@@ -2170,8 +2237,8 @@ putmail(dp, rp, fdmail, fdopmode, timestring, file, uid)
 	      rc = EX_UNAVAILABLE; /* Quota-exceeded is instant permanent reject ? */
 
 	      notaryreport(file, notaryacct(rc, s_delivered),
-			   "5.2.2 Mailbox Quota Exceeded",
-			   "x-local; 550 (Mailbox Quota Exceeded)");
+			   "5.2.2 Mailbox Quota Exceeded / Mailbox Full",
+			   "x-local; 550 (Mailbox Quota Exceeded / Mailbox Full)");
 	    } else {
 	      notaryreport(file, notaryacct(rc, s_delivered),
 			   "4.2.0 mailbox write failure",
@@ -2234,7 +2301,7 @@ putmail(dp, rp, fdmail, fdopmode, timestring, file, uid)
 	   * The mtime will be set to now by the following write() calls.
 	   */
 	  err = errno;
-	  sfseek(fp, (Sfoff_t)-2LL, SEEK_END);
+	  sfseek(fp, (Sfoff_t)-2, SEEK_END);
 	  len = sfread(fp, buf, 2);
 	  sfseek(fp, (Sfoff_t)0,    SEEK_END);
 	  /* to end of file, again */
@@ -2520,7 +2587,7 @@ program(dp, rp, cmdbuf, user, timestring, uid)
 	
 	  env[envi++] = cp;
 	  strcpy(cp, "MSGSPOOLID="); cp += 11;
-	  taspoolid(cp, rp->desc->msgmtime, rp->desc->msginonumber);
+	  strcat(cp, rp->desc->taspoolid);
 	  cp += strlen(cp) + 1;
 	  if (cp > cpe) break;
 
@@ -2750,52 +2817,39 @@ static void mkhashpath(s, uname)
      char *s;
      const char *uname;
 {
-	extern long pjwhash32 __((const char *));
-	extern long crc32     __((const char *));
+	if (pjwhashes || crchashes) {
+	  int i, h = 0, hlim = 0;
+	  int hashes[10]; /* Hard-coded max of 10 levels */
 
-	if (pjwhashes) {
-	  int h = pjwhash32(uname);
-	  switch (pjwhashes) {
-	  case 1:
-	    h %= 26;
-	    sprintf(s,"%c/", ('A' + h));
-	    break;
-	  default:
-	    h %= (26*26);
-	    sprintf(s,"%c/%c/", ('A' + (h / 26)), ('A' + (h % 26)));
-	    break;
+	  if (pjwhashes) {
+	    h = pjwhash32(uname);
+	    hlim = pjwhashes;
 	  }
-	}
-	if (crchashes) {
-	  int h = crc32(uname);
-	  switch (crchashes) {
-	  case 1:
-	    h %= 26;
-	    sprintf(s,"%c/", ('A' + h));
-	    break;
-	  default:
-	    h %= (26*26);
-	    sprintf(s,"%c/%c/", ('A' + (h / 26)), ('A' + (h % 26)));
-	    break;
+	  if (crchashes) {
+	    h = crc32(uname);
+	    hlim = crchashes;
 	  }
-	}
-	if (dirhashes) {
-	  switch (dirhashes) {
-	  case 1:
-	    sprintf(s,"%c/",uname[0]);
+	  for (i = 0; i < hlim; ++i) {
+	    hashes[i] = h % 26;
+	    h = h / 26;
+	  }
+	
+	  for (i = hlim-1; i >= 0; --i) {
+	    sprintf(s, "%c/", 'A' + hashes[i]);
 	    s += 2;
-	    break;
-	  case 2:
-	    if (uname[1])
-	      sprintf(s,"%c/%c/",uname[0],uname[1]);
-	    else /* Err.... One char userid ?? TROUBLE TIME! */
-	      sprintf(s,"%c/%c/",uname[0],uname[0]);
-	    s += 4;
-	    break;
-	  default:
-	    break;
 	  }
 	}
+
+	if (dirhashes) {
+	  const char *p = uname;
+	  int i;
+	  for (i = 0; i < dirhashes; ++i) {
+	    if (!*p) break;
+	    *s++ = *p++;
+	    *s++ = '/';
+	  }
+	  *s = 0;
+	}
 	strcat(s, uname);
 }
 
@@ -3362,8 +3416,9 @@ readrwho()
 	struct userhost *uhp;
 	struct spblk *spl;
 	
-	if (spt_users == NULL)
+	if (spt_users == NULL || spt_users->eltscnt == 0)
 	  return 0;
+
 	now = time(NULL);
 	if (chdir(RWHODIR) || (dirp = opendir(".")) == NULL ) {
 	  /* BE SILENT -- failing RBIFF is no fault! */
@@ -3424,7 +3479,8 @@ rbiff(nbp)
 	spkey_t symid;
 
 	symid = symbol((void*)(nbp->user));
-	if ((spl = sp_lookup(symid, spt_users)) == NULL)
+	spl = sp_lookup(symid, spt_users);
+	if (!spl)
 	  return;
 	for (uhp = (struct userhost *)spl->data; uhp != NULL; uhp = uhp->next)
 	  biff(uhp->hostname, nbp->user, nbp->offset);
@@ -3766,7 +3822,21 @@ return_receipt (dp, retrecptaddr, uidstr)
 	  struct stat stbuf;
 
 	  fstat(sffileno(mfp),&stbuf);
-	  taspoolid(boundarystr, stbuf.st_ctime, (long)stbuf.st_ino);
+	  taspoolid(boundarystr, (long)stbuf.st_ino, stbuf.st_ctime,
+#ifdef HAVE_STRUCT_STAT_ST_ATIM_TV_NSEC
+		  stbuf.st_mtim.tv_nsec
+#else
+#ifdef HAVE_STRUCT_STAT_ST_ATIM___TV_NSEC
+		  stbuf.st_mtim.__tv_nsec
+#else
+#ifdef HAVE_STRUCT_STAT_ST_ATIMENSEC
+		  stbuf.st_mtimensec
+#else
+		  0
+#endif
+#endif
+#endif
+		    );
 	  strcat(boundarystr, "=_/return-receipt/");
 	  strcat(boundarystr, dom);
 	}
