@@ -227,7 +227,7 @@ appendlet(SS, dp, convertmode, CT)
 		return EX_IOERR;
 	      }
 	      sprintf(SS->remotemsg,
-		      "smtp; 500 (msgbuffer write IO-error[1]! [%s] DATA %d/%d [%d%%])",
+		      "smtp; 500 (msgbuffer write IO-error[3]! [%s] DATA %d/%d [%d%%])",
 		      strerror(errno),
 		      SS->hsize, SS->msize, (SS->hsize*100+SS->msize/2)/SS->msize);
 	      return EX_IOERR;
@@ -274,7 +274,7 @@ appendlet(SS, dp, convertmode, CT)
 	      /* We NEVER get timeouts here.. We get anything else.. */
 	      if (rc != i) {
 		sprintf(SS->remotemsg,
-			"500 (msgbuffer write IO-error[2]! [%s] DATA %d/%d [%d%%])",
+			"500 (msgbuffer write IO-error[4]! [%s] DATA %d/%d [%d%%])",
 			strerror(errno),
 			SS->hsize, SS->msize, (SS->hsize*100+SS->msize/2)/SS->msize);
 		return EX_IOERR;
@@ -297,7 +297,7 @@ appendlet(SS, dp, convertmode, CT)
 	      return EX_IOERR;
 	    }
 	    sprintf(SS->remotemsg,
-		    "500 (msgbuffer write IO-error[3]! [%s] DATA %d/%d [%d%%])",
+		    "500 (msgbuffer write IO-error[5]! [%s] DATA %d/%d [%d%%])",
 		    strerror(errno),
 		    SS->hsize, SS->msize, (SS->hsize*100+SS->msize/2)/SS->msize);
 	    if (bufferfull > 1) readalready = 0;
@@ -329,7 +329,7 @@ ssputc(SS, ch, fp)
 {
   if (SS->chunkbuf == NULL) {
     if (sferror(fp)) return EOF;
-    if (sfputc(fp, ch) < 0) return EOF;
+    if (sfputc(fp, ch) < 0 || sferror(fp)) return EOF;
     return 0;
   }
   if (SS->chunksize >= CHUNK_MAX_SIZE) {
@@ -395,7 +395,7 @@ writebuf(SS, buf, len)
 
 	  if (state && c != '\n') {
 	    state = 0;
-	    if (c == '.' && !SS->chunking) {
+	    if (c == '.' && !SS->doing_chunking) {
 	      if (ssputc(SS, c, fp) == EOF || ssputc(SS, c, fp) == EOF) {
 		time(&endtime);
 		notary_setxdelay((int)(endtime-starttime));
@@ -562,7 +562,7 @@ writemimeline(SS, buf, len, convertmode)
 	    }
 	  } /* .... end convertmode	*/
 
-	  if (column == 0 && c == '.' && !SS->chunking) {
+	  if (column == 0 && c == '.' && !SS->doing_chunking) {
 	    if (ssputc(SS, c, fp) == EOF) {
 	      time(&endtime);
 	      notary_setxdelay((int)(endtime-starttime));
