@@ -1,4 +1,4 @@
-/* Copyright 1993-2002 - Matti Aarnio
+/* Copyright 1993-2006 - Matti Aarnio
 
    The way the Zmailer uses DBM entries is by using strings with
    their terminating NULL as keys, and as data..  Thus the length
@@ -146,7 +146,7 @@ static int store_db(dbf, typ, overwritemode, linenum, t, tlen, s, slen)
      void *dbf;
      const int typ, linenum;
      const void *s, *t;
-     const int slen, tlen;
+     const int slen, tlen, overwritemode;
 {
 #ifdef HAVE_NDBM
 	DBM *ndbmfile = dbf;
@@ -910,7 +910,7 @@ char *argv[];
 	err = db_create(&dbfile, NULL, 0);
 	if (! err)
 	  err = dbfile->open(dbfile,
-#if (DB_VERSION_MAJOR == 4) && (DB_VERSION_MINOR == 1)
+#if (DB_VERSION_MAJOR == 4) && (DB_VERSION_MINOR >= 1)
 			     NULL, /* TXN id was added at SleepyDB 4.1 */
 #endif
 			     dbasename, NULL, DB_BTREE,
@@ -928,7 +928,7 @@ char *argv[];
 	err = db_create(&dbfile, NULL, 0);
 	if (! err)
 	  err = dbfile->open(dbfile,
-#if (DB_VERSION_MAJOR == 4) && (DB_VERSION_MINOR == 1)
+#if (DB_VERSION_MAJOR == 4) && (DB_VERSION_MINOR >= 1)
 			     NULL, /* TXN id was added at SleepyDB 4.1 */
 #endif
 			     dbasename, NULL, DB_HASH,
