@@ -1,14 +1,12 @@
 /*
 	Libc fill-in for ZMailer using IPv6 API
-	by Matti Aarnio <mea@nic.funet.fi> 1997, 2000
+	by Matti Aarnio <mea@nic.funet.fi> 1997, 2001
 
 	The original Craig Metz code is deeply Linux specific,
 	this adaptation tries to be way more generic..
 */
 
-#include "config.h"
-#ifndef HAVE_GAI_STRERROR
-
+#include "../config.h"
 #include <stdio.h>
 #include <sys/types.h>
 #include <netdb.h>
@@ -20,6 +18,8 @@
 # define const
 #endif
 
+#ifndef HAVE_GAI_STRERROR
+
 /*
 %%% copyright-cmetz-97
 This software is Copyright 1997 by Craig Metz, All Rights Reserved.
@@ -29,7 +29,7 @@ you didn't get a copy, you may request one from <license@inner.net>.
 
 */
 
-const char *gai_strerror(errnum)
+char *gai_strerror(errnum)
 int errnum;
 {
   static char buffer[24];
@@ -44,8 +44,10 @@ int errnum;
       return "temporary failure in name resolution";
     case EAI_FAIL:
       return "non-recoverable failure in name resolution";
+#ifdef EAI_NODATA
     case EAI_NODATA:
       return "no address associated with name";
+#endif
     case EAI_FAMILY:
       return "ai_family not supported";
     case EAI_SOCKTYPE:
