@@ -2,10 +2,10 @@
  *  vacation -- originally BSD vacation by Eric Allman,
  *
  *  Adapted to ZMailer by Rayan Zachariassen, and further
- *  modified by Matti Aarnio over years 1988(?) thru 1998.
+ *  modified by Matti Aarnio over years 1988(?) thru 2004
  */
 
-#include "hostenv.h"
+#include "mailer.h"
 
 #include <stdio.h>
 #include <ctype.h>
@@ -26,30 +26,24 @@
 /* #include <tzfile.h> */  /* Not needed ? */
 #include <errno.h>
 #include <sysexits.h>
-#ifdef	HAVE_NDBM_H
+#include <ctype.h>
+
+/* Database format preferrence order:
+   NDBM, GDBM, SleepyCat4, SleepyCat3, SleepyCat2, BSD DB 1
+*/
+
+#ifdef	HAVE_NDBM
 #include <ndbm.h>
 #include <fcntl.h>
 #else
-#ifdef  HAVE_GDBM_H
+#ifdef  HAVE_GDBM
 #include <gdbm.h>
 #include <fcntl.h>
 #else
-#if defined(HAVE_DB_H)||defined(HAVE_DB1_DB_H)||defined(HAVE_DB2_DB_H)
-#if defined(HAVE_DB_185_H) && !defined(HAVE_DB_OPEN2)
-# include <db_185.h>
-#else
-#ifdef HAVE_DB2_DB_H
-# include <db2/db.h>
-#else
-#ifdef HAVE_DB_H
-# include <db.h>
-#else
-#ifdef HAVE_DB1_DB_H
-# include <db1/db.h>
-#endif
-#endif
-#endif
-#endif
+#if defined(HAVE_DB_H)     || defined(HAVE_DB1_DB_H) || \
+    defined(HAVE_DB2_DB_H) || defined(HAVE_DB3_DB_H) || \
+    defined(HAVE_DB4_DB_H)
+#include "sleepycatdb.h"
 #else
 :error:error:error "To compile, VACATION needs ndbm.h, gdbm.h, or db.h; none found!"
 #endif
@@ -118,15 +112,15 @@ extern void   usage __((void));
 
 typedef struct alias {
 	struct alias *next;
-	char *name;
+	const char *name;
 } ALIAS;
 ALIAS *names = NULL;
 
-#ifdef	HAVE_NDBM_H
+#ifdef	HAVE_NDBM
 DBM *db;
 #define DBT datum
 #else
-#ifdef HAVE_GDBM_H
+#ifdef HAVE_GDBM
 GDBM_FILE db;
 #define DBT datum
 #else /* HAVE_DB_H */
@@ -142,8 +136,7 @@ char from[MAXLINE];
 char *subject_str = NULL;	/* Glob subject from input */
 int dblog = 1;
 
-extern void purge_input __((void));
-extern struct passwd *getpwnam();
+extern void purge_input __((const char *));
 extern int optind, opterr;
 extern char *optarg;
 extern FILE *freopen(), *tmpfile();
@@ -151,16 +144,41 @@ extern char *getenv();
 #ifndef HAVE_STDLIB_H
 extern char *malloc();
 #endif
-extern void usrerr __((char *));
-extern void syserr __((char *));
-extern void readheaders __((void));
+extern void usrerr __((const char *));
+extern void syserr __((const char *));
+extern void readheaders __((const char *));
 extern char *strerror __((int));
 static int recent __((void));
 static int junkmail __((void));
-static int nsearch __((char *name, char *str));
+static int nsearch __((const char *name, const char *str));
 static void sendmessage __((const char *msgf, const char *myname));
+static void str_to_lower __((char *));
 
 const char *progname;
+const char *zenv_vinterval;
+const char *orcpt;
+const char *inrcpt;
+const char *msgspoolid = "S00xx";
+
+
+static void vacation_exit_handler()
+{
+	if (dblog && db) {
+#ifdef	HAVE_NDBM
+	  dbm_close(db);
+#else
+#ifdef HAVE_GDBM
+	  gdbm_close(db);
+#else
+#ifdef HAVE_DB_CLOSE2
+	  db->close(db, 0);
+#else
+	  db->close(db);
+#endif
+#endif
+#endif
+	}
+}
 
 int
 main(argc, argv)
@@ -168,14 +186,29 @@ main(argc, argv)
      char *argv[];
 {
 	register char *p;
-	struct passwd *pw;
+	struct Zpasswd *pw;
 	ALIAS *cur;
 	time_t interval;
 	char *msgfile = NULL;
 	int ch, iflag, ret;
 
+	/* ZMailer thingies.. */
+	char *zenv = getenv("ZCONFIG");
+
 	progname = argv[0];
 
+	if (zenv  ) readzenv(zenv);
+	zenv_vinterval = getzenv("VACATIONINTERVAL");
+
+	zopenlog("vacation", LOG_PID, LOG_MAIL);
+
+	atexit(vacation_exit_handler);
+
+	orcpt  = getenv("ORCPT");
+	inrcpt = getenv("INRCPT");
+	if (getenv("MSGSPOOLID"))
+	  msgspoolid = getenv("MSGSPOOLID");
+
 	/* process arguments */
 	opterr = iflag = 0;
 	interval = -1;
@@ -186,6 +219,7 @@ main(argc, argv)
 	    if (!cur)
 	      break;
 	    cur->name = optarg;
+	    str_to_lower((char*)(cur->name));
 	    cur->next = names;
 	    names = cur;
 	    break;
@@ -202,9 +236,12 @@ main(argc, argv)
 	  case 't':
 	  case 'r':
 	    if (isdigit(*optarg)) {
-	      interval = atol(optarg) * (24*60*60);
-	      if (interval < 0)
+	      const char *rest;
+	      interval = parse_interval(optarg,&rest);
+	      if (interval < 0 || *rest)
 		usage();
+	      if (interval < 30)
+		interval *= (60*60*24); /* Old "days" multiplier */
 	    }
 	    else
 	      interval = INT_MAX;
@@ -232,60 +269,71 @@ main(argc, argv)
 	if (argc != 1) {
 	  if (!iflag)
 	    usage();
-	  pw = getpwuid(getuid());
+	  pw = zgetpwuid(getuid());
 	  if (!pw) {
 	    fprintf(stderr, "vacation: no such user uid %ld.\n", (long)getuid());
+	    zsyslog((LOG_NOTICE, "%s: vacation: no such user uid: %ld", msgspoolid, (long)getuid()));
 	    exit(EX_NOUSER);
 	  }
-	} else if (!(pw = getpwnam(*argv))) {
+	} else if (!(pw = zgetpwnam(*argv))) {
 	  fprintf(stderr, "vacation: no such user %s.\n", *argv);
+	  zsyslog((LOG_NOTICE, "%s: vacation: no such user: '%s'", msgspoolid, *argv));
 	  exit(EX_NOUSER);
 	}
 	if (chdir(pw->pw_dir)) {
 	  fprintf(stderr, "vacation: no such directory %s.\n", pw->pw_dir);
+	  zsyslog((LOG_NOTICE, "%s: vacation: no such directory '%s'", msgspoolid, pw->pw_dir));
 	  exit(EX_NOUSER);
 	}
 
-	/* verify recipient argument */
-#ifdef ZMAILER
-	if (argc == 0) {
-	  p = getenv("USER");
-	  if (p == NULL) {
-	    usrerr("Zmailer error: USER environment variable not set");
-	    exit(EX_USAGE+102);
-	  }
-	}
-#endif /* ZMAILER */
-
-#ifdef	HAVE_NDBM_H
-	if (dblog)
+	if (dblog) {
+#ifdef	HAVE_NDBM
 	  db = dbm_open(VDB, O_RDWR | (iflag ? O_TRUNC|O_CREAT : 0),
 			S_IRUSR|S_IWUSR);
 #else	/* !NDBM */
-#ifdef HAVE_GDBM_H
-	if (dblog)
+#ifdef HAVE_GDBM
 	  db = gdbm_open(VDB ".pag" /* Catenates these strings */, 8192,
 			 iflag ? GDBM_NEWDB : GDBM_WRITER,
 			 S_IRUSR|S_IWUSR, NULL );
 #else
-	db = NULL;
-#ifdef HAVE_DB_OPEN2
-	if (dblog)
+#ifdef HAVE_DB_CREATE
+	  ret = db_create(&db, NULL, 0);
+#ifndef DB_UPGRADE
+#define DB_UPGRADE 0
+#endif
+	  if (ret == 0)
+	    ret = db->open(db,
+#if (DB_VERSION_MAJOR == 4) && (DB_VERSION_MINOR >= 1)
+			    NULL, /* TXN id was added at SleepyDB 4.1 */
+#endif
+			   VDB ".db", NULL, DB_BTREE,
+			   DB_CREATE|DB_UPGRADE, S_IRUSR|S_IWUSR);
+	  if (ret) {
+#ifdef HAVE_DB_CLOSE2
+	    db->close(db, 0);
+#else
+	    db->close(db);
+#endif
+	    db = NULL;
+	  }
+#elif defined(HAVE_DB_OPEN2)
 	  db_open(VDB ".db", DB_BTREE, DB_CREATE, S_IRUSR|S_IWUSR,
 		  NULL, NULL, &db);
 #else
-	if (dblog)
 	  db = dbopen(VDB ".db", iflag ? (O_RDWR|O_CREAT) : O_RDWR,
 		      S_IRUSR|S_IWUSR, DB_BTREE, NULL);
 #endif
 #endif
 #endif
-
+	}
 	ret = EX_OK;
 
 	if (dblog && !db) {
+	  int e = errno;
 	  fprintf(stderr, "vacation: %s.* database file(s): %s\n", 
-		  VDB, strerror(errno));
+		  VDB, strerror(e));
+	  zsyslog((LOG_NOTICE, "%s: vacation: %s.* database file(s): %s",
+		   msgspoolid, VDB, strerror(e)));
 	  exit(EX_CANTCREAT);
 	}
 
@@ -298,38 +346,47 @@ main(argc, argv)
 	  if (!cur) {
 	    ret = EX_TEMPFAIL;
 	  } else {
-	    cur->name = pw->pw_name;
+	    const char *myname = pw->pw_name;
+
+	    cur->name = strdup(pw->pw_name);
+	    str_to_lower((char*)(cur->name));
 	    cur->next = names;
 	    names = cur;
 
+	    if (orcpt) {
+	      const char *s = strchr(orcpt, ';');
+	      if (s)
+		myname = s+1;
+	      else
+		myname = orcpt;
+	    }
+	    if (inrcpt) {
+	      const char *s = strchr(inrcpt, ';');
+	      if (s)
+		myname = s+1;
+	      else
+		myname = inrcpt;
+	    }
+
 	    /* read message from standard input (just from line) */
-	    readheaders();
-	    purge_input();
+	    readheaders(myname);
+	    purge_input(myname);
+
 	    if (!recent()) {
 	      setreply();
-	      sendmessage(msgfile,pw->pw_name);
+	      sendmessage(msgfile,myname);
+
+zsyslog((LOG_DEBUG, "%s: vacation: sent message from '%s' to '%s'",
+	 msgspoolid, myname, from));
+	    } else {
+
+zsyslog((LOG_DEBUG, "%s: vacation: from '%s' to '%s' is too recent.",
+	 msgspoolid, from, myname));
+ 
 	    }
 	  }
 	}
 
-#ifdef	HAVE_NDBM_H
-	if (dblog)
-	  dbm_close(db);
-#else
-#ifdef HAVE_GDBM_H
-	if (dblog)
-	  gdbm_close(db);
-#else
-#ifdef HAVE_DB_OPEN2
-	if (dblog)
-	  db->close(db, 0);
-#else
-	if (dblog)
-	  db->close(db);
-#endif
-#endif
-#endif
-
 	exit(ret);
 }
 /*
@@ -374,7 +431,8 @@ sendmessage(msgf, myname)
 	fprintf(mf, "from %s\n",myname);
 	fprintf(mf, "to %s\n", from);
 	fprintf(mf, "env-end\n");
-	fprintf(mf,"To: %s\n", from);
+	fprintf(mf, "From: %s\n", myname);
+	fprintf(mf, "To: %s\n", from);
 	while (!feof(f) && !ferror(f)) {
 	  if (fgets(linebuf,sizeof(linebuf),f) == NULL) break;
 	  if ((s = strchr(linebuf,'$')) != NULL) {
@@ -412,9 +470,10 @@ sendmessage(msgf, myname)
 
 void
 usrerr(msg)
-	char *msg;
+	const char *msg;
 {
 	fprintf(stderr, "vacation: %s\n",msg);
+	zsyslog((LOG_NOTICE, "%s: vacation: usererr: %s", msgspoolid, msg));
 }
 /*
 **  SYSERR -- print system error
@@ -433,9 +492,10 @@ usrerr(msg)
 /* VARARGS 1 */
 void
 syserr(msg)
-	char *msg;
+	const char *msg;
 {
 	fprintf(stderr, "vacation: %s\n", msg);
+	zsyslog((LOG_NOTICE, "%s: vacation: syserr: %s", msgspoolid, msg));
 	exit(EX_USAGE+103);
 }
 /*
@@ -461,7 +521,7 @@ newstr(s)
 	if (p == NULL)
 	{
 		syserr("newstr: cannot alloc memory");
-		exit(EX_OSERR);
+		exit(EX_TEMPFAIL);
 	}
 	strcpy(p, s);
 	return p;
@@ -471,7 +531,8 @@ newstr(s)
  *	read mail headers
  */
 void
-readheaders()
+readheaders(myname)
+     const char *myname;
 {
 	register ALIAS *cur;
 	register char *p;
@@ -494,7 +555,9 @@ readheaders()
 		strcpy(from,buf);
 		has_from = 1;
 		if (junkmail()) {
-			purge_input();
+			purge_input(myname);
+zsyslog((LOG_DEBUG, "%s: vacation: considering this message to '%s' to be from JUNK source", msgspoolid, myname));
+
 			exit(EX_OK);
 		}
 	}
@@ -513,7 +576,8 @@ readheaders()
 				if (p != NULL)
 					*p = '\0';
 				if (junkmail()) {
-					purge_input();
+					purge_input(myname);
+zsyslog((LOG_DEBUG, "%s: vacation: considering this message to '%s' to be from JUNK source", msgspoolid, myname));
 					exit(EX_OK);
 				}
 			}
@@ -530,7 +594,8 @@ readheaders()
 				break;
 			if (!strncasecmp(p, "junk", 4) ||
 			    !strncasecmp(p, "bulk", 4)) {
-				purge_input();
+				purge_input(myname);
+zsyslog((LOG_DEBUG, "%s: vacation: this message to '%s' has Precedence: '%s'", msgspoolid, myname, p));
 				exit(EX_OK);
 			}
 			break;
@@ -560,15 +625,15 @@ readheaders()
 findme:			for (cur = names; !tome && cur; cur = cur->next)
 				tome += nsearch(cur->name, buf);
 		} /* switch() */
-	if (!tome) {
-		purge_input();
-		exit(EX_OK);
-	}
 	if (!*from) {
-	  zopenlog("vacation", LOG_PID, LOG_MAIL);
-	  zsyslog((LOG_NOTICE, "vacation: no initial \"From\" line.\n"));
+	  zsyslog((LOG_NOTICE, "%s: vacation: no initial \"From\" line.", msgspoolid));
 	  exit(EX_USAGE+105);
 	}
+	if (!tome) {
+		purge_input(myname);
+		zsyslog((LOG_DEBUG, "%s: vacation: not found to be for me (%s)", msgspoolid, myname));
+		exit(EX_OK);
+	}
 }
 
 /*
@@ -577,13 +642,14 @@ findme:			for (cur = names; !tome && cur; cur = cur->next)
  */
 static int
 nsearch(name, str)
-	register char *name, *str;
+	register const char *name, *str;
 {
 	register int len;
 
-	for (len = strlen(name); *str; ++str)
-		if (*str == *name && !strncasecmp(name, str, len))
+	for (len = strlen(name); *str; ++str) {
+		if (tolower(*str) == *name && !strncasecmp(name, str, len))
 			return(1);
+	}
 	return(0);
 }
 
@@ -611,6 +677,9 @@ junkmail()
 	register int len;
 	register char *p;
 
+	if (strcmp(from, "<>") == 0)
+		return(1);
+
 	/*
 	 * This is mildly amusing, and I'm not positive it's right; trying
 	 * to find the "real" name of the sender, assuming that addresses
@@ -642,7 +711,8 @@ junkmail()
  *
  */
 void
-purge_input()
+purge_input(myname)
+     const char *myname;
 {
 	char buf[256];
 	int read_rc;
@@ -666,7 +736,7 @@ recent()
 	DBT key, data;
 	time_t then, next;
 
-	if (!dblog) return 0;
+	if (!dblog) return 0;  /* No DB -> always recent! */
 
 	memset(&key, 0, sizeof(key));
 	memset(&data, 0, sizeof(data));
@@ -674,10 +744,10 @@ recent()
 	/* get interval time */
 	key.dptr = VIT;
 	key.dsize = sizeof(VIT);
-#ifdef HAVE_NDBM_H
+#ifdef HAVE_NDBM
 	data = dbm_fetch(db, key);
 #else
-#ifdef HAVE_GDBM_H
+#ifdef HAVE_GDBM
 	data = gdbm_fetch(db, key);
 #else
 #ifdef DB_INIT_TXN
@@ -689,10 +759,14 @@ recent()
 #endif
 #endif
 #endif
-	if (data.dptr == NULL)
-		next = (60*60*24*7); /* One week */
-	else
-		memcpy(&next, data.dptr, sizeof(next));
+	if (data.dptr == NULL) {
+	    next = 0;
+	    if (zenv_vinterval)
+		next = parse_interval(zenv_vinterval, NULL);
+	    if (next < 30)
+		next = (60*60*24*1); /* One DAY (like is documented) */
+	} else
+	    memcpy(&next, data.dptr, sizeof(next));
 
 	memset(&key, 0, sizeof(key));
 	memset(&data, 0, sizeof(data));
@@ -700,10 +774,10 @@ recent()
 	/* get record for this address */
 	key.dptr = from;
 	key.dsize = strlen(from);
-#ifdef HAVE_NDBM_H
+#ifdef HAVE_NDBM
 	data = dbm_fetch(db, key);
 #else
-#ifdef HAVE_GDBM_H
+#ifdef HAVE_GDBM
 	data = gdbm_fetch(db, key);
 #else
 #ifdef DB_INIT_TXN
@@ -717,7 +791,7 @@ recent()
 #endif
 	if (data.dptr) {
 		memcpy(&then, data.dptr, sizeof(then));
-		if (next == INT_MAX || then + next > time(NULL))
+		if (next == INT_MAX || ((then + next) > time(NULL)))
 			return(1);
 	}
 	return(0);
@@ -744,10 +818,10 @@ setinterval(interval)
 	data.dptr = (void*)&interval;
 	data.dsize = sizeof(interval);
 
-#ifdef HAVE_NDBM_H
+#ifdef HAVE_NDBM
 	dbm_store(db, key, data, DBM_REPLACE);
 #else
-#ifdef HAVE_GDBM_H
+#ifdef HAVE_GDBM
 	gdbm_store(db, key, data, GDBM_REPLACE);
 #else
 #ifdef DB_INIT_TXN
@@ -781,10 +855,10 @@ setreply()
 	data.dptr = (void*)&now;
 	data.dsize = sizeof(now);
 
-#ifdef HAVE_NDBM_H
+#ifdef HAVE_NDBM
 	dbm_store(db, key, data, DBM_REPLACE);
 #else
-#ifdef HAVE_GDBM_H
+#ifdef HAVE_GDBM
 	gdbm_store(db, key, data, GDBM_REPLACE);
 #else
 #ifdef DB_INIT_TXN
@@ -802,3 +876,11 @@ usage()
 	fprintf(stderr,"vacation: [-i] [-d] [-a alias] [-m msgfile] [-r interval] {login | 'start' | 'stop' }\n");
 	exit(EX_USAGE);
 }
+
+static void
+str_to_lower(str)
+	char *str;
+{
+	for (; *str; ++str)
+	  *str = tolower(*str);
+}
